/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import priorseditor.engine.data.FeatureDataset;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.protocol.Protocol;

/**
 *
 * @author kjetikl
 */
public class AddSequencesTask extends ExecutableTask {
    private static final int FORMAT_MANUAL_ENTRY=1;
    private static final int FORMAT_SEQUENCES=2;
    
    private PriorsEditorEngine engine;
    private int format=0;
    private Object[][] sequenceData=null;
    private Sequence[] sequences=null;

    /**
     * Creates a new AddSequencesTask object instantiated with the given data
     * originating from manual entry of sequence and gene data. In the data object
     * the number of rows (1st dimension) is the number of sequences while the
     * columns (2nd dimension) specifies
     *      1) Sequence name (String)
     *      2) Organism (Integer) NCBI taxonomy ID
     *      3) Genome build
     *      4) Chromosome (String)
     *      5) Sequence start (Integer)
     *      6) Sequence end (Integer)
     *      7) Gene name (String)
     *      8) Gene start (Integer)
     *      9) Gene end (Integer)
     *      10) Gene orientation (Integer)
     * 
     * NOTE: This method is kept for historic purposes and is deprecated. 
     * It is better to use: AddSequencesTask(PriorsEditorEngine engine, Sequence[] sequences) 
     * 
     * @param engine
     * @param data
     */
    @Deprecated
    public AddSequencesTask(PriorsEditorEngine engine, Object[][] sequenceSpecificationData) {
        super("Add Sequences");
        this.engine=engine;
        this.format=FORMAT_MANUAL_ENTRY;
        this.sequenceData=sequenceSpecificationData;
    }
    
    
    /**
     * Creates a new AddSequencesTask object instantiated with the given data
     * @param engine
     * @param data
     */    
    public AddSequencesTask(PriorsEditorEngine engine, Sequence[] sequences) {
        super("Add Sequences");
        this.engine=engine;
        this.format=FORMAT_SEQUENCES;
        this.sequences=sequences;
    }
    
    
    /**
     * Executes the operation associated with this task
     * Implements the Runnable interface
     */
    @Override
    public void run() throws InterruptedException, Exception {
        //System.err.println("Executing AddSequencesTask");//debug();
        boolean done=false;
        setStatus(RUNNING);  
        setProgress(5);
        if (engine.hasDataItemsOfType(FeatureDataset.class)) {
            setStatus(ERROR);
            throw new ExecutionError("New sequences can not be added after feature data has been loaded");            
        }          
        if (undoMonitor!=null) undoMonitor.register();
        while (!done){
          try {      
            execute();
            done=true;
          } catch (InterruptedException e) { // task aborted by the user
              if (undoMonitor!=null) undoMonitor.deregister(false);
              setStatus(ABORTED);
              throw e;
          } catch (Exception e) { // other errors
              if (shouldRetry(e)) {
                  setRetrying(true);
              } else {
                  if (undoMonitor!=null) undoMonitor.deregister(false);
                  setStatus(ERROR);
                  throw e;
              }
          }            
        }
        setProgress(100);
        if (undoMonitor!=null) undoMonitor.deregister(true);        
        setStatus(DONE);
        setStatusMessage(null);
    }

    
    private boolean execute() throws Exception {
        int numberOfSequences=(format==FORMAT_SEQUENCES)?sequences.length:sequenceData.length;
        for (int i=0;i<numberOfSequences;i++) { // for each sequence 
            String sequenceName=(format==FORMAT_SEQUENCES)?((Sequence)sequences[i]).getSequenceName():(String)sequenceData[i][0];
            Sequence sequence=(format==FORMAT_SEQUENCES)?sequences[i]:getSequenceFromData(i);
            if (sequence==null) throw new ExecutionError("Unable to instantiate sequence '"+sequenceName+"'");
            setStatusMessage("Adding sequence '"+sequenceName+"' ("+(i+1)+"/"+numberOfSequences+")");
            setProgress(i+1, numberOfSequences);
            Thread.yield();
            checkExecutionLock(); // checks to see if this task should suspend execution
            if (Thread.interrupted() || getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
            engine.storeDataItem(sequence);
        }   
        return true;
    }
    
    /** DEPRECATED */
    private Sequence getSequenceFromData(int index) {
        return new Sequence((String)sequenceData[index][0], ((Integer)sequenceData[index][1]).intValue(), (String)sequenceData[index][2], (String)sequenceData[index][3], ((Integer)sequenceData[index][4]).intValue(), ((Integer)sequenceData[index][5]).intValue(), (String)sequenceData[index][6], ((Integer)sequenceData[index][7]), ((Integer)sequenceData[index][8]), ((Integer)sequenceData[index][9]).intValue());
    }
    

    @Override
    public String getCommandString(Protocol protocol) {
        return null;
    }    
    
    public int getNumberofSequencesToAdd() {
        if (format==FORMAT_SEQUENCES) {
            if (sequences==null) return 0;
            else return sequences.length;
        } else if (format==FORMAT_MANUAL_ENTRY) {
            if (sequenceData==null) return 0;
            else return sequenceData.length;
        }
        return 0;
    }
    
}
