/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import priorseditor.engine.data.DataSegment;
import priorseditor.engine.data.FeatureSequenceData;


/**
 * This class represents a cache for storing downloaded Feature data locally so
 * that download from internet source is not necessary every time if the data has
 * been used before
 * 
 * @author kjetikl
 */
public class Cache {
    private PriorsEditorEngine engine;
    private String cacheDirectory;
    
    public Cache(PriorsEditorEngine engine, String cacheDirectory) {
        this.engine=engine;
        this.cacheDirectory=cacheDirectory;
         // create cache directory if it does not exist
        File dir=new File(cacheDirectory);
        if (!dir.exists()) {
            try {
                if (!dir.mkdirs()) throw new IOException("Unable to create local data cache directory in '"+cacheDirectory+"'");
            } catch (Exception e) {engine.logMessage("SYSTEM ERROR: Data cache error: "+e.getMessage());}
        } 
        //System.err.println("Installed new cache in: "+cacheDirectory);
    }
    
    /** 
     * Stores a dataobject in cache
     * @return TRUE if the operation was successful 
     */
    public boolean saveData(String trackName, int organism, String genomebuild, FeatureSequenceData data) {
        String chromosome=data.getChromosome();
        if (chromosome.equals("?")) return false; // do not store this data. We don't know where it is from
        int start=data.getRegionStart();
        int end=data.getRegionEnd();
        Object datavalue=null;
        ArrayList<String> toBeDeleted=new ArrayList<String>();
        ArrayList<String> overlaps=lookupRegion(trackName, organism, genomebuild, chromosome, start, end); // get cached data that overlaps with new data
        if (overlaps!=null) { // Overlaps existing files. Adjust start/end-positions of segment to save
            for (String filename:overlaps) {
                int[] pos=getStartAndEndFromFilename(filename);
                if (pos==null) return false;
                if (pos[0]<=start && pos[1]>=end) {
                    //System.err.println("Segment is completely cached already!");
                    return true; // new data is completely cached already. No need to store again
                }
                if (pos[0]>=start && pos[1]<=end) { // cached segment lies fully within new data => Delete this smaller file and store the new one instead
                    //System.err.println("Deleting smaller covered region!");
                    toBeDeleted.add(new String(cacheDirectory+File.separator+filename));
                } else if (pos[0]<start && pos[1]<end) { // cached segment partially overlaps at left end
                    start=pos[1]+1; // set new start to position after cached end
                    //System.err.println("Left flank partially covered. Adjusting start to position="+start);
                } else if (pos[0]>start && pos[1]>end) { // cached segment partially overlaps at right end
                    end=pos[0]-1; // set new end to position before cached start
                    //System.err.println("Right flank partially covered. Adjusting end to position="+end);
                } else {
                    System.err.println("WARNING: Reached 'unreachable' else-clause in cache.saveData(): start="+start+", end="+end+", cachedstart="+pos[0]+", cachedend="+pos[1]);
                }
            }
        }
        datavalue=data.getValueInGenomicInterval(start, end);
        if (datavalue!=null) {
           DataSegment cacheddata=new DataSegment(trackName, organism, genomebuild, chromosome, start, end, datavalue);
           String filename=trackName+"_"+organism+"_"+genomebuild+"_"+chromosome+"_"+start+"_"+end;
           filename=cacheDirectory+File.separator+filename;
           if (toBeDeleted.contains(filename)) return true; // The new trimmed region is already stored
           for (String deletefilename:toBeDeleted) { // delete files that lie within the new larger region
               File oldfile=new File(deletefilename);
               //System.err.println("Deleting smaller file: "+deletefilename);
               if (!oldfile.delete()) return false; // unable to delete old file. Abort caching to keep consistency intact
           }
           //System.err.println("Saving new cache data to: "+filename);
           return saveObjectToFile(filename, cacheddata);
        }
        else return false;
    } 
    
    /** Extracts start and end positions from a given cachefilename*/
    private int[] getStartAndEndFromFilename(String filename) {
        int[] positions=new int[]{-1,-1};
        try {
            String[] elements=filename.split("_");
            int start=Integer.parseInt(elements[elements.length-2]); // second last element
            int end=Integer.parseInt(elements[elements.length-1]); // last element
            positions[0]=start;
            positions[1]=end;                         
       } catch (NumberFormatException ne) {engine.logMessage("SYSTEM ERROR: Unable to parse expected numeric value: '"+ne.getMessage()+"'"); return null;}       
       return positions; 
    }
   
    
    /** Saves an object to file */
    private boolean saveObjectToFile(String filename,Object object) {
        ObjectOutputStream outputStream = null;
        try { 
            outputStream = new ObjectOutputStream(new FileOutputStream(filename));
            outputStream.writeObject(object);
            outputStream.close();                
        } catch (Exception e) {
            engine.logMessage("SYSTEM ERROR: Unable to store data in cache: "+e.getMessage());
            return false;
        } finally {
            try {if (outputStream!=null) outputStream.close();} catch (Exception x) {engine.logMessage("SYSTEM ERROR: Unable to close ObjectOutputStream in cache: "+x.getMessage());}
        }          
       return true;        
    }
    
    /** Loads an object from a file */
    private Object loadObjectFromFile(String filename) {
        ObjectInputStream inputStream = null;
        try { 
            inputStream = new ObjectInputStream(new FileInputStream(filename));
            Object value=inputStream.readObject();
            inputStream.close();  
            return value;
        } catch (InvalidClassException e) {
            // this can happen if a data class has been updated in a never version compared to 
            // the version of the data object stored in the cache
            try {if (inputStream!=null) inputStream.close();} catch (Exception x){}
            File file=new File(filename);
            file.delete();
            return null;
        } catch (Exception e) {
            engine.logMessage("SYSTEM ERROR: Unable to load data from cache: "+e.getMessage());
            return null;
        } finally {
            try {if (inputStream!=null) inputStream.close();} catch (Exception x) {engine.logMessage("SYSTEM ERROR: Unable to close ObjectInputStream in cache: "+x.getMessage());}
        }                 
    }

    
    /** 
     * Tries to retrieve cached data that overlaps a specified genomic segment of a dataTrack
     * Returns null if no data from this dataTrack/organism/segment was found in cache.
     * If some relevant data was found, the method will return a list of DataSegment objects 
     * that each represent a contiguous subsegment of the segment. The subsegments are 
     * non-overlapping and are returned in sorted order. (Note that the segments might exceed
     * the bounds of the supplied FeatureSequenceData object)    
     */
    public ArrayList<DataSegment> loadCachedSegments(FeatureSequenceData data, String trackName, int organism, String genomebuild) {
        String chromosome=data.getChromosome(); 
        int start=data.getRegionStart();
        int end=data.getRegionEnd();
        ArrayList<DataSegment> cachedData=new ArrayList<DataSegment>();
        ArrayList<String> overlaps=lookupRegion(trackName, organism, genomebuild, chromosome, start, end);
        if (overlaps!=null){     
            for (String filename:overlaps) {
                  filename=cacheDirectory+File.separator+filename;
                  Object saved=loadObjectFromFile(filename);
                  if (saved!=null && saved instanceof DataSegment) cachedData.add((DataSegment)saved);                        
            }
            Collections.sort(cachedData);
        }
        return cachedData;        
    }
   
    /** 
     * Returns a list of filenames that stores data originating from the given 
     * DataTrack and genomic segment, or null if there is no such data available
     */
    private ArrayList<String> lookupRegion(String trackName, int organism, String genomebuild, String chromosome, int start, int end) {
        final String fileprefix=trackName+"_"+organism+"_"+genomebuild+"_"+chromosome+"_";
        ArrayList<String> inside=new ArrayList<String>();
        File dir=new File(cacheDirectory);
        String[] list=dir.list(new FilenameFilter() {
            @Override
            public boolean accept(File dir, String name) {
                return name.startsWith(fileprefix);
            }
        });
        if (list.length==0) return null;
        try {
            for(String filename:list) { // filename = trackname_organism_genmebuild_chromosome_start_end
                // System.err.print("Checking '"+fileprefix+"' against cached file: "+filename);
                String[] elements=filename.split("_");
                int segmentstart=Integer.parseInt(elements[elements.length-2]); // second last element
                int segmentend=Integer.parseInt(elements[elements.length-1]); // last element
                if (start>segmentend || end<segmentstart) {
                    continue; // segment outside
                } else {
                    inside.add(filename); // add filename to list
                }
            } 
            return inside;                                
       } catch (NumberFormatException ne) {engine.logMessage("SYSTEM ERROR: Unable to parse expected numeric value: '"+ne.getMessage()+"'"); return null;}       
  }    
    
    
    
    
    /** 
     * Clears all the contents in the cache
     * @return TRUE if all cached files were successfully deleted, else FALSE
     */
    public boolean clearCache() {
        File directory=new File(cacheDirectory);
        File[] files=directory.listFiles();
        boolean alldeleted=true;
        for (File file:files) {
            if (!file.delete()) {
               engine.logMessage("SYSTEM ERROR: Unable to delete cache-file: "+file.getAbsolutePath());
               alldeleted=false;       
            }
        }
        return alldeleted;
    }
}
