/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.util.HashMap;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamResult;
import priorseditor.engine.datatracks.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import priorseditor.engine.data.RegionDataset;
import priorseditor.engine.data.NumericDataset;
import priorseditor.engine.data.DNASequenceDataset;

/**
 * The DataConfiguration class contains information about the currently
 * registered servers and available data tracks.
 * It contains methods to load and save a configuration from an XML-file
 * and to return lists (or rather maps) of available tracks and servers. 
 * 
 * @author kjetikl
 */
public class DataConfiguration implements Cloneable {
    private HashMap<String,Server> servers; // the key is the serveraddress
    private HashMap<String,DataTrack> availableTracks; // the key is the datatrack name    
    
    public DataConfiguration() {}
    
    public DataConfiguration(HashMap<String,DataTrack> availableTracks, HashMap<String,Server> servers) {
        this.availableTracks=availableTracks;
        this.servers=servers;
    }
    
    /** Returns a datastructure containing the available datatracks (with datasources) */
    public HashMap<String,DataTrack> getAvailableTracks() {
        return availableTracks;
    }
    
    /** Returns a datastructure containing information about particular servers */    
    public HashMap<String,Server> getServers() {
        return servers;
    }
    
    public void setAvailableTracks(HashMap<String,DataTrack> tracks) {
        availableTracks=tracks;
    }
    
    public void setServers(HashMap<String,Server> servers) {
        this.servers=servers;
    }
       
    private Document getXMLrepresentation() throws ParserConfigurationException {
         DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
         Document document = builder.newDocument();
         Element root=document.createElement("root");
         for (Server server:servers.values()) {
             org.w3c.dom.Element serverelement=server.getXMLrepresentation(document);
             root.appendChild(serverelement);             
         }
         for (DataTrack track:availableTracks.values()) {
             org.w3c.dom.Element trackelement=track.getXMLrepresentation(document);
             root.appendChild(trackelement);
         }
         document.appendChild(root);
         return document;
    }
    
    /**
     * Saves this DataConfiguration in XML format to the given file
     */
    public void saveConfigurationToFile(File configurationfile) throws Exception{
        Document document=getXMLrepresentation();
        TransformerFactory factory=TransformerFactory.newInstance();
        factory.setAttribute("indent-number", new Integer(3));
        Transformer transformer=factory.newTransformer();
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        DOMSource source=new DOMSource(document); 
        StreamResult result=new StreamResult(new OutputStreamWriter(new FileOutputStream(configurationfile),"UTF-8"));
        transformer.transform(source, result);
    }

    
    public void loadConfigurationFromFile(File configurationfile) throws SystemError {
        try {
            FileInputStream inputstream=new FileInputStream(configurationfile);
            loadConfigurationFromStream(inputstream); 
        }
        catch (FileNotFoundException fnf) {throw new SystemError("File not found error: "+configurationfile.getAbsolutePath());}
        catch (SystemError e) {throw e;}          
    }
    
    /**
     * Initializes an new DataConfiguration object based on descriptions in the given XML settings file
     */
    public void loadConfigurationFromStream(InputStream inputstream) throws SystemError {
        HashMap<String, DataTrack> newTracks=new HashMap<String, DataTrack>();
        HashMap<String, Server> newServers=new HashMap<String, Server>();
        try {
         DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
         Document doc = builder.parse(inputstream);
         NodeList nodes = doc.getElementsByTagName("Server");
         for (int i=0;i<nodes.getLength();i++) {
            int serverdelay=0;
            Element server = (Element) nodes.item(i);
            String serveraddress=server.getAttribute("address");
            String delayString=server.getAttribute("delay");
            if (delayString!=null && !delayString.isEmpty()) {
               try {serverdelay=Integer.parseInt(delayString);} catch(Exception e){ throw new SystemError("Unable to parse server delay attribute for server="+serveraddress+". Expected integer value, got '"+delayString+"'");} 
            }
            Server newserver=new Server(serveraddress, serverdelay);
            NodeList mirrornodes = server.getElementsByTagName("Mirror");
            for (int j=0;j<mirrornodes.getLength();j++) {
               Element mirror = (Element) mirrornodes.item(j);
               String mirroraddress=mirror.getTextContent();
               newserver.addMirror(mirroraddress);
            }
            newServers.put(serveraddress,newserver);
         }                  
         nodes = doc.getElementsByTagName("DataTrack");
         for (int i=0;i<nodes.getLength();i++) {
            Element datatrack = (Element) nodes.item(i);
            String datatrackName=datatrack.getAttribute("name");
            String datatypeName=datatrack.getAttribute("type");
            String sourceSite=datatrack.getAttribute("source");
            String description=datatrack.getAttribute("description");
            Class datatype;
                 if (datatypeName.equalsIgnoreCase(DataTrack.NUMERIC_DATA) || datatypeName.equalsIgnoreCase("wiggle")) datatype=NumericDataset.class; // 'wiggle' is a legacy name for numeric data
            else if (datatypeName.equalsIgnoreCase(DataTrack.REGION_DATA)) datatype=RegionDataset.class;
            else if (datatypeName.equalsIgnoreCase(DataTrack.SEQUENCE_DATA)) datatype=DNASequenceDataset.class;
            else throw new SystemError("Unrecognized data type '"+datatypeName+"' for DataTrack="+datatrackName);
            DataTrack track=new DataTrack(datatrackName, datatype, sourceSite, description);
            if (newTracks.containsKey(datatrackName)) throw new SystemError("Duplicate entry for datatrack='"+datatrackName+"' in 'datatracks.txt'");
            newTracks.put(datatrackName,track);
            NodeList datasources = datatrack.getElementsByTagName("DataSource");
            for (int j=0;j<datasources.getLength();j++) {
                Element datasourcenode = (Element) datasources.item(j);
                String organismString=datasourcenode.getAttribute("organism");
                int organism=0;
                try {
                    organism=Integer.parseInt(organismString);
                } catch (NumberFormatException nfe) {throw new SystemError("Organism specification must be an NCBI Taxonomy ID (integer) for DataSource["+(j+1)+"] for DataTrack="+datatrackName+". Got value="+organismString);}                
                String build=datasourcenode.getAttribute("build");               
                Element protocol = (Element)datasourcenode.getElementsByTagName("Protocol").item(0);
                Element dataformat = (Element)datasourcenode.getElementsByTagName("dataformat").item(0);
                String baseURL=null;
                String featureName=null;
                String dataformatName=null;
                String protocoltype=null;
                String parameterstring=null;
                if (protocol==null) throw new SystemError("Missing Protocol speficification for DataSource["+(j+1)+"] for DataTrack="+datatrackName);
                else {
                    protocoltype=protocol.getAttribute("type");
                    if (protocoltype==null) throw new SystemError("Missing Protocol Type attribute for DataSource["+(j+1)+"] for DataTrack="+datatrackName);
                    Element url = (Element)protocol.getElementsByTagName("URL").item(0);
                    if (url!=null) baseURL=url.getTextContent();
                    Element feature = (Element)protocol.getElementsByTagName("Feature").item(0);
                    if (feature!=null) {featureName=feature.getTextContent();}
                    Element parameters = (Element)protocol.getElementsByTagName("Parameter").item(0);
                    if (parameters!=null) parameterstring=parameters.getTextContent();
                }
                ParameterSettings dataformatsettings=null;
                if (dataformat!=null) {
                    dataformatName=dataformat.getAttribute("name");
                    NodeList settingslist = dataformat.getElementsByTagName("setting");
                    if (settingslist.getLength()>0) dataformatsettings=new ParameterSettings();
                    for (int m=0;m<settingslist.getLength();m++) {                                      
                        Element settingnode = (Element) settingslist.item(m);
                        String settingname=settingnode.getAttribute("name"); 
                        String settingvalue=settingnode.getTextContent();
                        dataformatsettings.setParameter(settingname, settingvalue);
                    }
                }
                DataSource datasource=null;
                if (protocoltype.equals(DataSource.HTTP_GET)) {
                    if (baseURL==null) throw new SystemError("Missing base URL tag for DataSource["+(j+1)+"] for DataTrack="+datatrackName);
                    datasource=new DataSource_http_GET(track, organism, build, baseURL, dataformatName, parameterstring);
                    datasource.setDataFormatSettings(dataformatsettings);
                } else if (protocoltype.equals(DataSource.DAS_SERVER)) {
                    if (baseURL==null) throw new SystemError("Missing base URL tag for DataSource["+(j+1)+"] for DataTrack="+datatrackName);
                    datasource=new DataSource_DAS(track, organism, build, baseURL, dataformatName, featureName);
                    datasource.setDataFormatSettings(dataformatsettings);                   
                }
                if (datasource!=null) {
                   track.addDataSource(datasource);
                }
                // System.err.println("     DataSource: "+organism+" ["+build+"]["+serverdelay+"] '"+featureName+"'  {"+dataformatName+"}  URL=> "+baseURL);                
            }
         }
       } // -- end try
       catch (Exception e) {
          throw new SystemError("["+e.getClass().toString()+"] "+e.getMessage());
       }
       this.availableTracks=newTracks;
       this.servers=newServers;
    }    
    
    
    public void loadDefaultConfiguration() throws SystemError {
         loadConfigurationFromStream(this.getClass().getResourceAsStream("/priorseditor/engine/datatracks/DataTracks.xml"));
    }    
    
    @SuppressWarnings("unchecked")
    @Override
    public DataConfiguration clone() {
        DataConfiguration copy=new DataConfiguration((HashMap<String,DataTrack>)availableTracks.clone(), (HashMap<String,Server>)servers.clone());
        return copy;
    }      
}
