/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Set;
import priorseditor.engine.data.*;
import priorseditor.engine.dataformat.DataFormat;
import priorseditor.engine.datatracks.*;
import priorseditor.engine.protocol.ParseError;


/**
 *
 * @author kjetikl
 */
public class DataLoader {
    private PriorsEditorEngine engine;
    private HashMap<String,Server> servers; // the key is the serveraddress
    private HashMap<String,DataTrack> availableTracks; // the key is the datatrack name
    private DataConfiguration dataconfiguration;
    private Cache cache=null;
    private boolean useCache=true;
    
    public DataLoader(PriorsEditorEngine engine) {
        this.engine=engine;      
        servers=new HashMap<String, Server>();
        availableTracks=new HashMap<String, DataTrack>();
        dataconfiguration=new DataConfiguration(availableTracks, servers);
        DataSource.setDataLoader(this); // the DataLoader property is static in DataSource
    }
    
    /** Returns a reference to the file that contains the XML formatted file of available datatracks
     * (and other settings) which should be read into a DataConfiguration object
     */
    public File getDataConfigurationFile() {
        String filename=engine.getWorkDirectory()+File.separator+"DataTracks.xml";
        return new File(filename);
    }
    
    /** Specifies a cache directory to use for this dataloader */
    public void setCacheDirectory(String cacheDirectory) {
        cache=new Cache(engine, cacheDirectory);
    }
    
    public PriorsEditorEngine getEngine() {
        return engine;
    }
    
    /** Specifies whether this DataLoader should make use of a local disc cache to reduce connections to external DataSources */
    public void setUseCache(boolean flag) {
        useCache=flag;
    }
    
    /** Clears the installed Cache for this DataLoader
     * @return TRUE if the cached was successfully clear (or if no cached was installed) or
     * FALSE if the installed cached could not be successfully cleared
     */
    public boolean clearCache() {
        if (cache!=null) return cache.clearCache();
        else return true;
    }
    
    /** 
     * This method should be called whenever data is loaded from a specific server
     * It will store a timestamp for when the server was last accessed
     */
    public void notifyDataSourceAccess(String serveraddress) {
        if (serveraddress!=null) {
            Server server=servers.get(serveraddress);
            if (server!=null) {server.setTimeOfLastAccess(new Date());}
        }
    }
    
    /** 
     * This method will return the number of milliseconds one should wait before accessing the 
     * given server again. The number is based on the time of last access (as registered by 
     * the last call to notifyDataSourceAccess(serveraddress)) and the "delay" property for this server
     */
    public int getWaitingPeriodForServer(String serveraddress) {
        if (serveraddress==null) return 0;
        Server server=servers.get(serveraddress);
        if (server==null) return 0;    
        Date lastDate=server.getTimeOfLastAccess();
        if (lastDate==null) return 0;
        long lasttime=lastDate.getTime();
        long timeout=lasttime+server.getServerDelay();
        long now=new Date().getTime();
        if (now>timeout) return 0; 
        else return (int)(timeout-now);
    }    
    
    /**
     * Returns an array containing all servers that have specific settings
     * @return
     */
    public ArrayList<Server> getAvailableServersAsList() {
        ArrayList<Server> serverlist=new ArrayList<Server>(servers.size());
        Set<String> keys=servers.keySet();
        for (String key:keys) {
            serverlist.add(servers.get(key));
        }
        return serverlist;
    }
    
    /**
     * Returns an array containing all datatracks that are availble to the system
     * (according to a list of predefined datasets)
     * @return
     */
    public DataTrack[] getAvailableDatatracks() {
        DataTrack[] tracks=new DataTrack[availableTracks.size()];
        Set<String> keys=availableTracks.keySet();
        int i=0;
        for (String key:keys) {
            tracks[i]=availableTracks.get(key);
            i++;
        }
        return tracks;
    }
    /**
     * Returns an array containing all datatracks that are availble to the system
     * (according to a list of predefined datasets)
     * @return
     */
    public ArrayList<DataTrack> getAvailableDatatracksAsList() {
        ArrayList<DataTrack> tracks=new ArrayList<DataTrack>(availableTracks.size());
        Set<String> keys=availableTracks.keySet();
        for (String key:keys) {
            tracks.add(availableTracks.get(key));
        }
        return tracks;
    }
    
    /**
     * Returns an array containing all datatracks of the given type that are 
     * availble to the system (according to a list of predefined datasets)
     * @return
     */
    public DataTrack[] getAvailableDatatracks(Class datatype) {
        int size=0;
        Set<String> keys=availableTracks.keySet();
        for (String key:keys) {
            if (availableTracks.get(key).getDataType()==datatype) size++;
        }
        DataTrack[] tracks=new DataTrack[size];
        int i=0;
        for (String key:keys) {
            DataTrack track=availableTracks.get(key);
            if (track.getDataType()==datatype) {tracks[i]=track;i++;}
        }
        return tracks;
    }
    
    
    
    /** Replaces the current tracks database with the given configuration */
    public void installConfiguration(DataConfiguration newconfig) {
        dataconfiguration=newconfig;
        servers=newconfig.getServers();
        availableTracks=newconfig.getAvailableTracks();
    }

    /** Returns the current DataConfiguration (not a copy!) */
    public DataConfiguration getCurrentDataConfiguration() {
        return dataconfiguration;
    }
    
    /**
     * Loads the selected Data Track from available Data Sources and returns a FeatureDataset
     * If caching is used, the method first examines if any parts of the data
     * can be found in the cache. Data which is not available from cache is then
     * loaded from primary source(s)
     * @param track
     * @return
     * @throws ExecutionError
     */
    public FeatureDataset loadDataTrack(String datatrackName, ExecutableTask task) throws Exception {
        DataTrack datatrack=availableTracks.get(datatrackName);
        if (datatrack==null) throw new ExecutionError("Unknown DataTrack="+datatrackName);
        SequenceCollection allSequences=engine.getDefaultSequenceCollection();
        ArrayList<Sequence>list=allSequences.getAllSequences(engine);
        FeatureDataset dataset=null;
             if (datatrack.getDataType()==DNASequenceDataset.class) dataset=new DNASequenceDataset(datatrackName);
        else if (datatrack.getDataType()==NumericDataset.class) dataset=new NumericDataset(datatrackName);
        else if (datatrack.getDataType()==RegionDataset.class) dataset=new RegionDataset(datatrackName);
        else throw new ExecutionError("Unknown DataTrack type="+datatrack.getDataType());
        int counter=0;
        int size=list.size();
        for (Sequence sequence:list) { // for each sequence
            counter++;
            if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) { throw new InterruptedException();}           
            task.setStatusMessage("Loading "+datatrackName+" for "+sequence.getName()+" ("+counter+"/"+size+")");
            engine.logMessage("["+counter+"/"+size+"]  Loading "+datatrackName+" for "+sequence.getName());
            task.setProgress(counter, size);
            int organism=sequence.getOrganism();
            String genomebuild=sequence.getGenomeBuild();
            FeatureSequenceData sequencedata=null;
                 if (datatrack.getDataType()==DNASequenceDataset.class) sequencedata=new DNASequenceData(sequence,'N');
            else if (datatrack.getDataType()==NumericDataset.class) sequencedata=new NumericSequenceData(sequence,0);
            else if (datatrack.getDataType()==RegionDataset.class) sequencedata=new RegionSequenceData(sequence);
            ArrayList<DataSegment> segments=null;
            // Load available segments from cache
            if (useCache && cache!=null) segments=cache.loadCachedSegments(sequencedata, datatrackName, organism, genomebuild);
            else segments=new ArrayList<DataSegment>();   
            // fill in with new empty segments
            //engine.logMessage("Got "+segments.size()+" cached segments for "+sequence.getName());
            if (segments.size()>0) {
                String chromosome=sequence.getChromosome();
                ArrayList<DataSegment> inbetween=new ArrayList<DataSegment>();
                int firststart=segments.get(0).getSegmentStart();
                if (firststart>sequence.getRegionStart()) inbetween.add(new DataSegment(datatrackName, organism, genomebuild, chromosome, sequence.getRegionStart(), firststart-1, null));
                int lastend=segments.get(segments.size()-1).getSegmentEnd();
                if (lastend<sequence.getRegionEnd()) inbetween.add(new DataSegment(datatrackName, organism, genomebuild, chromosome, lastend+1, sequence.getRegionEnd(), null));
                for (int i=0;i+1<segments.size();i++) {
                    DataSegment first=segments.get(i);
                    DataSegment second=segments.get(i+1);
                    if (second.getSegmentStart()-first.getSegmentEnd()>1) inbetween.add(new DataSegment(datatrackName, organism, genomebuild, chromosome, first.getSegmentEnd()+1, second.getSegmentStart()-1, null));
                }     
                segments.addAll(inbetween);
                Collections.sort(segments);
            } else { // No cached segments. Add single empty segment that spans the whole sequence
                segments.add(new DataSegment(datatrackName, organism, genomebuild, sequence.getChromosome(), sequence.getRegionStart(), sequence.getRegionEnd(), null));
            }
            // Prune segments list
            optimizeSegments(segments);
            // Load each of the missing segments anew
            for (DataSegment segment:segments) {
                try { // I added this try-catch block so that dataloading errors occurring in single sequences would not ruin for other sequences that works fine 
                    if (segment.isEmpty()) segment=obtainDataSegment(segment,datatrack,organism,genomebuild,counter,size,sequence.getName(),task);
                    if (segment.isEmpty()) throw new ExecutionError("No data returned for DataTrack="+datatrackName+" for sequence="+sequence.getName()+". This could be correct or information could be missing for this region");            
                    segment.copyData(sequencedata); // transfer data from segment into the sequencedata object
                } catch (InterruptedException e) {
                    throw e;
                } catch (Exception e) {
                    throw e;
                    //engine.errorMessage(e.getMessage(), 0);
                }
            }
            if (useCache && cache!=null) cache.saveData(datatrackName,organism, genomebuild, sequencedata); // save 
            dataset.addSequence(sequencedata);
        }
        return dataset;
    }
    
    
    /** 
     * Optimizes a list of DataSegments that spans a genomic region.
     * Some of the segments might represent data loaded from cache while others
     * just contain empty placeholder for which data must be obtained from 
     * original DataSources. This method prunes the list by possibly deleting
     * cached segments and merging flanking empty segments into single larger
     * segments in order to reduce the number of DataSource connections that
     * has to be made (while possibly increasing the amount of data that has to
     * be loaded anew, since it is generally considered to be more efficient to 
     * load a larger amount of data in fewer attempts rather than to load many 
     * small segments one at a time). Note that the argument ArrayList might be
     * changed as a result of this optimization. Note also that the optimization
     * might not be "optimal" in the classic sense of the word.  
     * The present implementation simply finds the first "empty" segment (which
     * must be loaded from original DataSource) and the last emtpy segment in 
     * the list and (provided that they are not the same) merge these two and
     * all segments in between into a single continuous empty segment
     */
    private void optimizeSegments(ArrayList<DataSegment> segments) {
        int first=-1; int last=-1;
        for (int i=0;i<segments.size();i++) {
            DataSegment segment=segments.get(i);
            if (segment.containsData()) continue;
            else {
                if (first<0) first=i;
                else last=i;
            }
        }
        if (first>0 && last>0) { // two non-identical empty segments have been found
            int endcoordinate=segments.get(last).getSegmentEnd();
            segments.get(first).setSegmentEnd(endcoordinate);
            int len=last-first;
            for (int j=0;j<len;j++) {segments.remove(first+1);} // remove "same" coordinate j times 
        }
    }
    
    /** Obtains (uncached) DataTrack data for a segment using any one of the available applicable sources for that DataTrack */
    private DataSegment obtainDataSegment(DataSegment segment, DataTrack datatrack, int organism, String genomebuild, int counter, int size, String sequenceName, ExecutableTask task) throws Exception {
        //engine.logMessage("Loading '"+datatrack+"' from DataSource in "+segment.getChromosome()+":"+segment.getSegmentStart()+"-"+segment.getSegmentEnd()); 
        DataSource[] sources=datatrack.getDataSources(organism,genomebuild);
            String datatrackName=datatrack.getName();
            if (sources==null || sources.length==0) throw new ExecutionError("Unsupported organism/genomebuild '"+Organism.getCommonName(organism)+":"+genomebuild+"' (sequence "+sequenceName+") for DataTrack="+datatrackName);            
            boolean ok=false;
            int i=0;
            while(!ok && i<sources.length) { // try to obtain sequence data from different sources (mirrors)            
                if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) { throw new InterruptedException();}           
                DataSource source=sources[i];
                if (i>0) engine.logMessage("["+counter+"/"+size+"]  Loading "+datatrackName+" for "+sequenceName+":  trying mirror site at "+source.getServerAddress());
                try {
                    segment=source.loadDataSegment(segment, task);
                    ok=true; // datatrack has been loaded for this sequence. Proceed with next sequence
                } catch (ParseError e) { // data was loaded but could not be parsed
                    engine.logMessage("["+counter+"/"+size+"]  Loading "+datatrackName+" for "+sequenceName+":  ERROR: Unable to parse data file: "+e.getMessage());
                    if (i==sources.length-1) {throw new ExecutionError("Unable to parse data file:\n"+e.getMessage(), e);}
                    // just try next mirror
                } catch (SocketTimeoutException e) {
                    engine.logMessage("["+counter+"/"+size+"]  Loading "+datatrackName+" for "+sequenceName+":  ERROR: Network timeout!");
                    if (i==sources.length-1) {throw new ExecutionError("Network timeout");}
                    // just try next mirror
                } catch (InterruptedException e) {
                    throw e; // user aborted
                } catch (Exception e) {
                    engine.logMessage("["+counter+"/"+size+"]  Loading "+datatrackName+" for "+sequenceName+":  ERROR: "+e.getClass().getSimpleName()+":"+e.getMessage());
                    if (i==sources.length-1) {throw new ExecutionError(e.getMessage(), e);}
                    // just try next mirror
                } 
                i++; // try next mirror
            }
            return segment;
    }
    
    /** This method should be called before DataLoader is used 
     *  It will install a DataConfiguration based on an XML-file description of available data tracks
     *   
     */
    public void setup() throws SystemError {
        engine.logMessage("Importing Data Tracks");
        File dataTracksFile=getDataConfigurationFile();
        if (!dataTracksFile.exists()) revertSettings();
        DataConfiguration config=new DataConfiguration();
        config.loadConfigurationFromFile(dataTracksFile);
        availableTracks=config.getAvailableTracks();
        servers=config.getServers();
        dataconfiguration=config;
    }
    
        
    
    /** 
     * Copies the 'immutable' DataTracks.xml that came with the installation into the working directory 
     * where it can be manipulated by the user. 
     */
    private void revertSettings() {
        BufferedReader inputStream=null;
        PrintWriter outputStream=null; 
        File outfile=getDataConfigurationFile();
        try {
            inputStream=new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("/priorseditor/engine/datatracks/DataTracks.xml")));
            outputStream=new PrintWriter(outfile);
            String line;
            while((line=inputStream.readLine())!=null) {outputStream.println(line);}
        } catch (IOException e) { 
            engine.errorMessage("An error occurred when reverting datatrack settings: ["+e.getClass().toString()+"] "+e.getMessage(),0);
        } finally {
            try {
                if (inputStream!=null) inputStream.close();
                if (outputStream!=null) outputStream.close();
            } catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader or PrintWriter in DataLoader.revertSettings(): "+ioe.getMessage());}
        }       
    }
    
    
    /**
     * The type of data to return is determined by the 'target' data object.
     * If the target object is NULL. the datatype String is used instead to determine
     * what kind of data object to return
     *
     */
    public Data loadData(File file, String datatype, Data target, DataFormat format, ParameterSettings settings, ExecutableTask task) throws Exception {
        if (!file.exists()) throw new ExecutionError("File does not exist: '"+file.getAbsolutePath()+"'");
            if (!file.canRead()) throw new ExecutionError("Unable to read file: '"+file.getAbsolutePath()+"'");
            if (target==null) {
                if (datatype==null || datatype.isEmpty()) throw new ExecutionError("Missing data type specification");
                if (datatype.equals(NumericDataset.getType())) {
                    target=new NumericDataset("temporary");
                } else if (datatype.equals(RegionDataset.getType())) {
                    target=new RegionDataset("temporary");
                } else if (datatype.equals(DNASequenceDataset.getType())) {
                    target=new DNASequenceDataset("temporary");
                } else if (datatype.equals(MotifCollection.getType())) {
                    target=new MotifCollection("temporary");
                } else if (datatype.equals(BackgroundModel.getType())) {
                    target=new BackgroundModel("temporary");
                } else if (datatype.equals(SequenceCollection.getType())) {
                    target=new SequenceCollection("temporary");
                } else throw new ExecutionError("Unknown or unapplicable data type: "+datatype);
            }
            if (!format.canParseInput(target)) {
               throw new ExecutionError("Data format '"+format.getName()+" can not be used to parse "+datatype);
            }              
            try {
                InputStream inputStream=new FileInputStream(file);
                target=format.parseInput(inputStream, target, settings, task);
                inputStream.close();// maybe this is redundant
            }           
            catch (InterruptedException ie) {throw ie;}
            catch (Exception e) {
                String exceptionText=e.getClass().getSimpleName();
                if (exceptionText.contains("ParseError")) exceptionText="";
                else exceptionText+=":";            
                throw new ExecutionError(exceptionText+e.getMessage(), e); 
            }
            return target;
    }
    
}
