/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashSet;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;
import priorseditor.engine.protocol.Protocol;
import priorseditor.gui.UndoMonitor;
/**
 *
 * @author kjetikl
 */
public abstract class ExecutableTask extends AbstractUndoableEdit {
    public static final String PENDING="Pending"; // Task has not yet been started
    public static final String RUNNING="Running"; // Task has been started and is currently running
    public static final String DONE="Done";       // Task has completed execution
    public static final String ABORTED="Aborted"; // Task has been aborted
    public static final String WAITING="Waiting"; // Task is running but must wait for an indeterminate time for another process to finish before continuing
    public static final String ERROR="Error";     // Task has ended because of an error
    public static final String STATUS="status";
    public static final String PROGRESS="progress";
    public static final String STATUS_MESSAGE="statusMessage";
    private int progress=0;
    private String statusMessage="";
    private String status=PENDING;    
    private HashSet<PropertyChangeListener> propertyChangeListeners=new HashSet<PropertyChangeListener>();
    private String taskName="task";
    private int queueNumber=0; 
    private Object executionLock=null; //
    private PriorsEditorClient client=null;
    private int lineNumber=0;
    private boolean retrying=false;
    protected UndoMonitor undoMonitor=null;
    private UndoableEdit protocolEdit=null;
    
    /**
     * 
     * @param name
     */
    public ExecutableTask(String name) {
        taskName=name;
    }
    
    public void setTaskName(String name) {
        taskName=name;
    }    
    
    /** 
     * This method is called automatically by some Executor when it is time to 
     * execute the task.
     * Subclasses should override this method to do the work real work
     */
    public abstract void run() throws InterruptedException, Exception;
                          
    
    /**
     * Provides an object which can be used as an execution lock to temporarily suspend execution 
     * of a task. Running tasks should regularly check this lock by calling the method
     * {@code checkExecutionLock}. If this lock has been requested by some outside object, the execution
     * of the task (and its thread) will block until the lock is released 
     * 
     * @param lock
     */
    public void setExecutionLock(Object lock) {
        executionLock=lock;
    }
    
    /**
     * Returns the executionLock for this object
     */
    public Object getExecutionLock() {
        return executionLock;
    }
    
    /**
     * Running tasks should call this method on a regular basis to check whether some other thread has
     * requested the task executing thread to suspend temporarily.
     */
    public void checkExecutionLock() {
        if (executionLock==null) return;
        synchronized(executionLock) {}; // we do not need to do anything, really. This command will block and wait for the lock to be released (unless it is already free)
    }
    
    /** 
     * Returns an integer value in the range 0 to 100 that represents the progress 
     * made so far to complete this OperationTask 
     * @return A value between 0 and 100
     */
    public int getProgress() {
        return progress;
    }
    
    /** 
     * Sets the progress value for this operation task to a number between 0 and 100
     * 
     * @param A value between 0 and 100
     */
    public void setProgress(int i) {
        int oldvalue=progress;
        if (i>100) progress=100;
        else if (i<0) progress=0;
        else progress=i;
        firePropertyChange(PROGRESS, oldvalue, progress);
    }
    
    /** 
     * Sets the progress value for this operation according to the relative number
     * of "subtasks" that have been completed so far
     * 
     * @param i The number of subtasks currently completed
     * @param n The total number of subtasks that must be completed
     */
    public void setProgress(int i, int n) {
        int oldvalue=progress;
        if (i>n) progress=100;
        else if (i<0) progress=0;
        else progress=(int)((i*1.0/n)*100);
        firePropertyChange(PROGRESS, oldvalue, progress);
    }
    
    
    /**
     * Returns TRUE if the task was attempted executed before but was aborted due
     * to some error or exception and is now (by user request) attempted executed
     * again. If a task is "retrying" some of the actions performed the last time
     * might still be valid and need not be repeated from scratch.
     * @return
     */
    public boolean isRetrying() {
        return retrying;
    }
    
    /**
     * Sets the "retrying" flag for this task
     * @return
     */
    protected void setRetrying (boolean isretrying) {
        retrying=isretrying;
    }
    
    /** Returns a status message for this operation execution 
     * @return a status message
     */
    public String getStatusMessage() {
        return statusMessage;
    }
    
    /** Sets a new status message for this operation execution
     * @param newstatus The new status message
     */
    public void setStatusMessage(String newstatusMessage) {
        String oldstatusMessage=statusMessage;
        statusMessage=newstatusMessage;
        firePropertyChange(STATUS_MESSAGE, oldstatusMessage, statusMessage);
    }
    
    /**
     * Sets a new status for this OperationTask
     * The class ExecutableTask defines static string constants that can be 
     * used to specify the status 
     * @param status
     */
     public void setStatus(String status) {
        String oldstatus=this.status;
        this.status=status;
        firePropertyChange(STATUS, oldstatus, status);
     }

    /**
     * Returns the status for this OperationTask
     * The class ExecutableTask defines static integer constants for different statu
     * @param status
     */
     public String getStatus() {
       return status;
     }

     /** 
      * This function should return true if an executing task should attempt to
      * redo tasks that have failed because of unforseen errors (for instance 
      * network timeouts) or false if the error should be propagated by throwing
      * an exception. The default implementation first tries to ask an installed
      * PriorsEditorClient or returns false if no PriorsEditorClient is installed.
      * 
      */
     public boolean shouldRetry(Exception e) {
         if (client==null) return false;
         else return client.shouldRetry(this, e);
     }
     
     /** 
      * Sets a PriorsEditorClient to use for this task. PriorsEditorClient decides whether
      * tasks should attempt to retry executions which have failed due to unforseen
      * errors (such as network timeouts). If set to null the default behaviour 
      * would be to never retry execution and instead propagate the error by
      * throwing an Exception
      */
     public void setPriorsEditorClient(PriorsEditorClient client) {
         this.client=client;
     }
     
     /** 
      * Returns the PriorsEditorClient which installed for this task
      */
     public PriorsEditorClient getPriorsEditorClient() {
         return client;
     }
     
     
    /** Registers a new propertyChangeListener which will receive notifications of important updates */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeListeners.add(listener);
    }
    
    /** Removes a previously registered propertyChangeListener */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeListeners.remove(listener);
    }
    
    /** Notifies interested listeners of a change in a property value */
    private void firePropertyChange(String name, Object oldvalue, Object newvalue) {
        PropertyChangeEvent event=new PropertyChangeEvent(this, name, oldvalue, newvalue);
        for (PropertyChangeListener listener:propertyChangeListeners) {
            listener.propertyChange(event);
        }
    }    

     /**
     * Provides an undoMonitor to be used by this Task to trace information about 
     * updates that needs to be undone. Without such a monitor the changes made by 
     * this task cannot be undone or redone (unless monitored at a higher level)  
     * @param flag
     */
    public void setUndoMonitor(UndoMonitor monitor) {
        undoMonitor=monitor;
    }   
    
    @Override
    public boolean canRedo() {
        return (super.canRedo() && undoMonitor!=null && undoMonitor.canRedo());
    }
    
    @Override
    public boolean canUndo() {
        return (super.canUndo() && undoMonitor!=null && undoMonitor.canUndo());
    }
    
   
    @Override
    public void undo() throws CannotUndoException {
        super.undo();
        if (undoMonitor==null || !undoMonitor.canUndo()) throw new CannotUndoException();
        undoMonitor.undo();
        if (protocolEdit!=null) {
            //System.err.println("Undoing protocol edit:"+protocolEdit.getPresentationName());
            protocolEdit.undo();
        }
    }
    
    @Override
    public void redo() throws CannotRedoException {
        super.redo();
        if (undoMonitor==null || !undoMonitor.canRedo()) throw new CannotRedoException();
        undoMonitor.redo();
        if (protocolEdit!=null) protocolEdit.redo();
    }

    
    public void setProtocolEdit(UndoableEdit edit) {
        protocolEdit=edit;
    }  
    @Override
    public void die() {
        super.die();
        if (undoMonitor!=null) undoMonitor.die();
    }
    
    @Override
    public String getPresentationName() {
        if (status.equals(ABORTED) || status.equals(ERROR)) return "aborted "+taskName;
        else return taskName;
    }
    
    /**
     * Sets a queue number for this task
     * @param number
     */
    public void setQueueNumber(int number) {
        queueNumber=number;
    }
    /**
     * Returns a queueNumber for this task. The number should represents the 
     * order assigned to the task when it first arrived in a queue, not necessarily
     * its current placement.
     */
    public int getQueueNumber() {
        return queueNumber;
    }

        /**
     * Returns the line number for this operation if it was executed as part of a script
     * (or 0 if the operation is not associated with a script)
     */
    public int getLineNumber() {
        return lineNumber;
    }
    
    /**
     * Sets a line number for this operation 
     */
    public void setLineNumber(int line) {
        lineNumber=line;
    }

    
   /**
     * Returns a textual representation of this task's command (with parameters) to be included in protocol scripts
     */    
    public abstract String getCommandString(Protocol protocol);
}
