/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;
import priorseditor.engine.data.Organism;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.protocol.ParseError;

/**
 * This class contains methods for obtaining gene names and transcription start 
 * and end sites given one or more Gene Identifiers in various formats
 * (ENSEMBL, Entrez, UniGene etc.)
 *  
 * @author kjetikl
 */
public class GeneIDResolver {
    private HashMap<Integer,String> databaseMap=new HashMap<Integer,String>();
    private HashMap<String,String> idFormatMap=new HashMap<String,String>();
    private HashMap<String,String> buildMap=new HashMap<String,String>(); // contains references to older biomarts for older builds
    private PriorsEditorEngine engine;
    private String cachefilename="GeneIDmappingCache.ser";
    private HashMap<String,ArrayList<GeneIDmapping>> cache=null;
    private String cacheDirectory=null;
    private boolean useCache=true;    
    private boolean cacheUpdated=false;
    private String biomartCurrent="http://www.biomart.org/biomart/martservice";
            
     public GeneIDResolver( PriorsEditorEngine engine) {
        databaseMap.put(Organism.HUMAN, "hsapiens_gene_ensembl");
        databaseMap.put(Organism.MOUSE, "mmusculus_gene_ensembl");
        databaseMap.put(Organism.RAT, "rnorvegicus_gene_ensembl");
        databaseMap.put(Organism.FRUITFLY, "dmelanogaster_gene_ensembl");
        databaseMap.put(Organism.CHICKEN, "ggallus_gene_ensembl");
        databaseMap.put(Organism.COW, "btaurus_gene_ensembl");
        databaseMap.put(Organism.YEAST, "scerevisiae_gene_ensembl");        
        databaseMap.put(Organism.FROG, "xtropicalis_gene_ensembl");        
        databaseMap.put(Organism.ZEBRAFISH, "drerio_gene_ensembl"); 
        idFormatMap.put("ensembl gene", "ensembl_gene_id");
        idFormatMap.put("ensembl transcript", "ensembl_transcript_id");
        idFormatMap.put("entrez gene", "entrezgene");
        idFormatMap.put("embl", "embl");
        idFormatMap.put("hgnc symbol", "hgnc_symbol");
        idFormatMap.put("unigene id", "unigene");
        idFormatMap.put("ccds id", "ccds");
        idFormatMap.put("refseq dna", "refseq_dna");
        buildMap.put("hg18","http://nov2008.archive.ensembl.org/biomart/martservice");        
        buildMap.put("mm8","http://aug2007.archive.ensembl.org/biomart/martservice");        
        buildMap.put("mm7","http://apr2006.archive.ensembl.org/biomart/martservice");        
        buildMap.put("danRer5","http://nov2008.archive.ensembl.org/biomart/martservice");        
        this.engine=engine;  
    }       

    /** Specifies a cache directory to use for this GeneIDResolver */
    public void setCacheDirectory(String cacheDir) {
        cacheDirectory=cacheDir;
    }
    
    /** Specifies whether this GeneIDResolver should make use of a local disc cache to store resolved Gene ID mappings */
    public void setUseCache(boolean flag) {
        useCache=flag;
    }
    
    /** Clears the installed Cache for this GeneIDResolver
     * @return TRUE if the cached was successfully clear (or if no cached was installed) or
     * FALSE if the installed cached could not be successfully cleared
     */
    public boolean clearCache() {
        try {
          String fullpath=cacheDirectory+File.separator+cachefilename;
          File file=new File(fullpath);
          if (file.exists()) return file.delete();
        } catch (Exception e) {return false;}
        cache=new HashMap<String, ArrayList<GeneIDmapping>>();
        return true;
    }
    
    /** */
    @SuppressWarnings("unchecked")
    private void installCache() {
        String fullpath=cacheDirectory+File.separator+cachefilename;
        File file=new File(fullpath);
        if (file.exists()) cache=(HashMap<String,ArrayList<GeneIDmapping>>)getObjectFromFile(fullpath);
        if (cache==null) cache=new HashMap<String, ArrayList<GeneIDmapping>>();      
    }
    
    private void storeCacheToDisc() {
        String fullpath=cacheDirectory+File.separator+cachefilename;
        //System.err.println("Storing cache to disc with "+cache.size()+" entries");
        if (storeObjectInFile(fullpath,cache)) cacheUpdated=false;      
    }
    
    private void storeInCache(GeneIDmapping mapping,String format, int organism, String build) {
        String key=mapping.geneID+"\t"+format+"\t"+organism+"\t"+build;
        ArrayList<GeneIDmapping> list=cache.get(key);
        if (list==null) {
            list = new ArrayList<GeneIDmapping>();
        } 
        if (!listContainsMapping(list, mapping)) list.add(mapping);  
        cache.put(key, list);
        cacheUpdated=true;
    }
    
    /** Returns a list of organisms that this class can resolve Gene IDs for 
     * The numbers returned are NCBI taxonomy IDs. Common and latin names
     * corresponding to these IDs are obtained from the Organism class
     */
    public int[] getSupportedOrganisms() {
        return new int[]{Organism.HUMAN,Organism.MOUSE,Organism.RAT,Organism.FRUITFLY,Organism.CHICKEN,Organism.COW,Organism.YEAST,Organism.FROG,Organism.ZEBRAFISH};
    }
    
    /** Returns true if the specified organism is supported by this GeneIDResolver */
    public boolean isOrganismSupported(int organism) {
        for (int supported:getSupportedOrganisms()) {
            if (organism==supported) return true;
        }
        return false;        
    }
    
    /** Returns a list of ID formats that this class can resolve Gene IDs for */
    public String[] getSupportedIDFormats() {
        return new String[]{"Ensembl Gene","Ensembl Transcript","Entrez Gene","EMBL","HGNC symbol","UniGene ID","CCDS ID","RefSeq DNA"};
    }
    
    /** Returns true if the specified ID format is supported by this GeneIDResolver */
    public boolean isIDFormatSupported(String format) {
        for (String supported:getSupportedIDFormats()) {
            if (format.equalsIgnoreCase(supported)) return true;
        }
        return false;
    }
    
    /** Returns true if the specified genome build is supported by this GeneIDResolver */    
    public boolean isGenomeBuildSupported(String genomebuild) {
        return Organism.isGenomeBuildSupported(genomebuild);
    }
    /** 
     * This method tries to resolve a list of gene identifiers in the given format.
     * The format and organism must be stated explicitly for each identifier
     * and need not be the same for all identifiers
     */
    public ArrayList<GeneIDmapping> resolveIDs(ArrayList<GeneIdentifier> list) throws Exception {        
        ArrayList<GeneIDmapping> result=new ArrayList<GeneIDmapping>(list.size());
            HashMap<String,ArrayList<String>> subgroups=new HashMap<String,ArrayList<String>>();
            for (GeneIdentifier geneID:list) {
                String key=geneID.format+"\t"+geneID.organism+"\t"+geneID.build;
                if (useCache) {
                    if (cache==null) installCache();
                    ArrayList<GeneIDmapping> cacheresults=resolveIDFromCache(geneID.identifier,geneID.format,geneID.organism, geneID.build);
                    //System.err.println("Got "+cacheresults.size()+" hits in cache");
                    if (cacheresults.size()>0) {result.addAll(cacheresults); continue;}
                }
                if (!subgroups.containsKey(key)) subgroups.put(key,new ArrayList<String>());
                ArrayList<String> uniformlist=subgroups.get(key);
                uniformlist.add(geneID.identifier); // these uniform lists contains IDs from the same organism/build and in the same format
            }
            Set<String> keys=subgroups.keySet();
            for (String key:keys) {
                String[] elements=key.split("\t");
                String format=elements[0];
                String organismString=elements[1];
                String build=elements[2];
                int organism=0;
                try {organism=Integer.parseInt(organismString);} catch (NumberFormatException e) {System.err.println("SYSTEM ERROR: Unable to parse expected numeric value in GeneIDResolver.resolveIDs(): "+e.getMessage());}
                ArrayList<String> sublist=subgroups.get(key);
                ArrayList<GeneIDmapping> partial=resolveIDlist(sublist,format,organism,build);
                result.addAll(partial);
                if (useCache) {
                    for (GeneIDmapping mapping:partial) storeInCache(mapping,format,organism,build);
                }
            }
            if (useCache && cacheUpdated) storeCacheToDisc();
            return result;
    }
    
    /** Returns true if the list already contains a similar mapping */
    private boolean listContainsMapping(ArrayList<GeneIDmapping> list, GeneIDmapping target) {
        for (GeneIDmapping mapping:list) {
            if (mapping.chromosome.equals(target.chromosome) && mapping.geneID.equals(target.geneID) && mapping.geneName.equals(target.geneName) && mapping.TSS==target.TSS && mapping.TES==target.TES && mapping.strand==target.strand) return true;
        }
        return false;
    }
    
    /** Resolves a list of identifiers in a single format from a single organism */
    private ArrayList<GeneIDmapping> resolveIDlist(ArrayList<String> list, String format, int organism, String build) throws Exception {
        boolean useTranscript=false;
        if (format.equalsIgnoreCase("Ensembl Transcript")) useTranscript=true;
        String database=databaseMap.get(organism);      
        if (database==null) throw new ExecutionError("Unsupported organism '"+Organism.getCommonName(organism)+"' in Gene ID Resolver");
        String idformat=idFormatMap.get(format.toLowerCase());
        if (idformat==null )  throw new ExecutionError("Unsupported format '"+format+"' in Gene ID Resolver");
        String xml=getXMLQueryString(idformat, database, list, useTranscript);
        if (!Organism.isGenomeBuildSupported(build)) throw new ExecutionError("Unsupported genome build '"+build+"' in Gene ID Resolver");
        String biomartURL=buildMap.get(build);
        if (biomartURL==null) biomartURL=biomartCurrent;
        //String urlString=biomartURL+"?query="+xml;
        //URL url=new URL(urlString); 
        //ArrayList<String> page=getPageUsingHttpGET(url,engine.getNetworkTimeout());
        URL url=new URL(biomartURL); 
        ArrayList<String> page=getPageUsingHttpPost(url,xml,engine.getNetworkTimeout());
        return parseResults(page);        
    }
    
    /** Resolves a single identifier by looking up in the cache */    
    private ArrayList<GeneIDmapping> resolveIDFromCache(String identifier, String format, int organism, String build) throws Exception {
        ArrayList<GeneIDmapping> result=null;
        if (cache==null) return new ArrayList<GeneIDmapping>();
        String key=identifier+"\t"+format+"\t"+organism+"\t"+build;
        result=cache.get(key);
        if (result!=null) return result;
        else return new ArrayList<GeneIDmapping>();
    }
    
    /** 
     * Parses the gene ID mapping table obtained from BioMart, and returns a list of GeneIDmappings
     * which might possibly be incomplete and contain duplicate entries (?)
     */    
    private ArrayList<GeneIDmapping> parseResults(ArrayList<String> table) throws ParseError {
        int numproperties=6;
        ArrayList<GeneIDmapping> result=new ArrayList<GeneIDmapping>(table.size());
        for (int i=0;i<table.size();i++) {
            String line=table.get(i);
            String[] elements=line.split("\t");
            //System.err.println(line);
            if (elements.length==1) {
                engine.logMessage("Unable to parse result: "+line);
                throw new ParseError("Unable to parse result: "+line); // no tab separation probably means an error has occurred!
            }
            if (elements.length!=numproperties) throw new ParseError("Expected "+numproperties+" columns in  gene ID mapping table, got "+elements.length+": "+line);
            try {
                GeneIDmapping newmapping=null;
                int start=Integer.parseInt(elements[3]);
                int end=Integer.parseInt(elements[4]);
                int strand=Integer.parseInt(elements[5]);
                if (strand==Sequence.DIRECT) newmapping=new GeneIDmapping(elements[0], elements[1], elements[2], start, end, strand);
                else newmapping=new GeneIDmapping(elements[0], elements[1], elements[2], end, start, strand);
                if (!listContainsMapping(result,newmapping)) result.add(newmapping);
            } catch (NumberFormatException e) { throw new ParseError("Unable to parse expected numeric value: "+e.getLocalizedMessage()+" in gene ID mapping table");}
        }        
        return result;
    }

    /** Retrieves a web page given an URL address using an http POST-request */    
    private ArrayList<String> getPageUsingHttpPost(URL url, String message, int timeout) throws Exception {    
        ArrayList<String> document=new ArrayList<String>();
        InputStream inputStream = null;
        BufferedReader dataReader = null;
        PrintWriter writer = null;
        HttpURLConnection connection=(HttpURLConnection)url.openConnection();
        connection.setConnectTimeout(timeout);

        connection.setDoOutput(true);
        connection.setDoInput(true);
        connection.setUseCaches(false);
        connection.setAllowUserInteraction(false);
        connection.setRequestMethod("POST");
        //connection.setRequestProperty("Content-type", "text/xml; charset=UTF-8");  
        try {
            message="query="+message;
            writer = new PrintWriter(connection.getOutputStream());
            writer.println(message);
            writer.flush();
        } catch (Exception e) {throw e;}
        finally {
            if (dataReader!=null) dataReader.close();
            if (inputStream!=null) inputStream.close();
        }  
        
        inputStream=connection.getInputStream();
        dataReader = new BufferedReader(new InputStreamReader(inputStream));
        String line;
        try {
           while ((line=dataReader.readLine())!=null){
              document.add(line);
           }
        } catch (Exception e) {throw e;}
        finally {
            if (dataReader!=null) dataReader.close();
            if (inputStream!=null) inputStream.close();
        }    
        return document;
    }

    
    /** Retrieves a web page given an URL address using an http GET-request */
    private ArrayList<String> getPageUsingHttpGET(URL url, int timeout) throws Exception {
        ArrayList<String> document=new ArrayList<String>();
        InputStream inputStream = null;
        BufferedReader dataReader = null;
        URLConnection connection=url.openConnection();
        connection.setConnectTimeout(timeout);
        inputStream=connection.getInputStream();
        dataReader = new BufferedReader(new InputStreamReader(inputStream));
        String line;
        try {
           while ((line=dataReader.readLine())!=null){
              document.add(line);
           }
        } catch (Exception e) {throw e;}
        finally {
            if (dataReader!=null) dataReader.close();
            if (inputStream!=null) inputStream.close();
        }    
        return document;
    }
    
    
    
    /** Builds the XML query which is send to the BioMart web service*/
    private String getXMLQueryString(String IDformat, String database, ArrayList<String> IDs, boolean transcript) {
        String idListAsString="";
        for (int i=0;i<IDs.size()-1;i++) {
            idListAsString+=IDs.get(i)+",";
        }
        idListAsString+=IDs.get(IDs.size()-1);
        String xml="<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
        xml+="<!DOCTYPE Query>";
        xml+="<Query  virtualSchemaName = \"default\" formatter = \"TSV\" header = \"0\" uniqueRows = \"0\" count = \"\" datasetConfigVersion = \"0.6\" >";			
            xml+="<Dataset name = \""+database+"\" interface = \"default\" >";
		xml+="<Filter name = \""+IDformat+"\" value = \""+idListAsString+"\"/>";
		xml+="<Attribute name = \""+IDformat+"\" />";
		xml+="<Attribute name = \"external_gene_id\" />";
                xml+="<Attribute name = \"chromosome_name\" />";
		if (transcript) {
		   // xml+="<Attribute name = \"ensembl_gene_id\" />";
		   xml+="<Attribute name = \"transcript_start\" />";
		   xml+="<Attribute name = \"transcript_end\" />";                    
                } else {
		   // xml+="<Attribute name = \"ensembl_transcript_id\" />";
                   xml+="<Attribute name = \"start_position\" />";
		   xml+="<Attribute name = \"end_position\" />";                    
                }
		xml+="<Attribute name = \"strand\" />";
	   xml+="</Dataset>";
       xml+="</Query>";
       xml=xml.replaceAll(" ", "%20");
       return xml;
    }

    /** Stores a serialized version of an object to the specified file */
    private boolean storeObjectInFile(String filename, Object value) {
        ObjectOutputStream stream=null;
        try {
             stream = new ObjectOutputStream(new FileOutputStream(filename));
             stream.writeObject(value);
             stream.close();         
        } catch (Exception e) {
                System.err.println("SYSTEM ERROR: GeneIDResolver.storeObjectInFile : "+e.getClass().toString()+"  : "+e.getMessage());
            return false;
        } finally {try {if (stream!=null) stream.close();} catch (Exception x){}}
        return true;
    }
    
    /** Fetches an object that has been serialized on file*/
    private Object getObjectFromFile(String filename) {
        ObjectInputStream stream=null;
        try {
             stream=new ObjectInputStream(new FileInputStream(filename));
             Object value=stream.readObject();
             stream.close();
             return value;        
        } catch (Exception e) {
                System.err.println("SYSTEM ERROR: GeneIDResolver.getObjectFromFile :  "+e.getClass().toString()+"  : "+e.getMessage());
                return null;
        } finally {try {if (stream!=null) stream.close();} catch (Exception x){}}

    }
    
    

}