/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;

/**
 * This class represents a framework for easy construction of graphs
 *
 * @author Kjetil
 */
public class Graph {
    public static final int NONE=0;
    public static final int SOLID=1;
    public static final int DASHED=2;
    public static final int DOTTED=3;
    public static final int BOX=0;
    public static final int CROSS=1;

    private Graphics2D g;
    private float minX=0;
    private float maxX=0;
    private float minY=0;
    private float maxY=0;

    private int graphwidth=0; // the width of the graph in pixels
    private int graphheight=0; // the height of the graph in pixels

    private int translateX=0; // a translation offset for the top-left corner of the graph-box (axes)
    private int translateY=0; // a translation offset for the top-left corner of the graph-box (axes)
    public static Color lightGray=new Color(230,230,230);

    public Graph(Graphics2D graphics, float minX, float maxX, float minY, float maxY, int graphwidth, int graphheight, int translateX, int translateY) {
        g=graphics;
        this.minX=minX;
        this.maxX=maxX;
        this.minY=minY;
        this.maxY=maxY;
        this.graphwidth=graphwidth;
        this.graphheight=graphheight;
        this.translateX=translateX;
        this.translateY=translateY;
    }

    /**
     * Sets new limits for this Graph. Note that an existing graph must be completely redrawn in order to update to the new limits
     * @param minX
     * @param maxX
     * @param minY
     * @param maxY
     */
    public void setLimits(float minX, float maxX, float minY, float maxY) {
        this.minX=minX;
        this.maxX=maxX;
        this.minY=minY;
        this.maxY=maxY;
    }


    /**
     * Draws an axes system (X and Y axis) for this graph
     * @param axestype A number specifying the layout of the axes.
     *        BOX: the axes (x and y axes with ticks) will be drawn "outside" the box like an L (to the left and at the bottom)
     *        CROSS: the axes (x and y axes with ticks) be drawn at origo like a cross
     * @param gridlines A number specifying the type of gridlines to use. NONE, SOLID, DASHED or DOTTED
     * @param boundingBox Draw a box around the whole graph
     * @param xAxisIsPercentages A flag to indicate whether the ticks on the X axis represent percentages
     * @param yAxisIsPercentages A flag to indicate whether the ticks on the Y axis represent percentages
     *
     */
    public void drawAxes(int axeslayout, int gridlines, boolean boundingBox) {
        drawAxes(axeslayout, gridlines, boundingBox, true, false, false);
    }

    /**
     * Draws an axes system (X and Y axis) for this graph
     * @param axestype A number specifying the layout of the axes.
     *        BOX: the axes (x and y axes with ticks) will be drawn "outside" the box like an L (to the left and at the bottom)
     *        CROSS: the axes (x and y axes with ticks) be drawn at origo like a cross
     * @param gridlines A number specifying the type of gridlines to use. NONE, SOLID, DASHED or DOTTED
     * @param boundingBox Draw a box around the whole graph
     * @param drawTicks Decides if tick marks should be drawn or not
     * @param xAxisIsPercentages A flag to indicate whether the ticks on the X axis represent percentages
     * @param yAxisIsPercentages A flag to indicate whether the ticks on the Y axis represent percentages
     *
     */
    public void drawAxes(int axeslayout, int gridlines, boolean boundingBox,  boolean drawTicks, boolean xAxisIsPercentages, boolean yAxisIsPercentages) {
         Stroke gridlinestroke=null;
         Stroke defaultStroke=g.getStroke();
         if (gridlines==DASHED) {
             gridlinestroke=new BasicStroke(1f,BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1f, new float[]{2f,3f}, 0f);
         } else if (gridlines==DOTTED) {
             gridlinestroke=new BasicStroke(1f,BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1f, new float[]{2f}, 0f);
         } else if (gridlines==SOLID) {
             gridlinestroke=defaultStroke;
         }
         if (gridlinestroke!=null) {
             g.setStroke(gridlinestroke);
             g.setColor(lightGray);
             drawHorizontalGridLines();
             drawVerticalGridLines();
             g.setStroke(defaultStroke);
         }
         g.setColor(Color.BLACK);
         if (drawTicks) {
             int xAxisPos=(axeslayout==CROSS)?getYforValue(0):translateY+graphheight;
             int yAxisPos=(axeslayout==CROSS)?getXforValue(0):translateX;
             drawXaxisWithTicks(xAxisPos,(axeslayout==CROSS), xAxisIsPercentages);
             drawYaxisWithTicks(yAxisPos,(axeslayout==CROSS), yAxisIsPercentages);
         }
         if (boundingBox) {
             g.drawRect(translateX, translateY, graphwidth, graphheight);
         }
    }


    /**
     * Draws an X-axis with sensible ticks within the range minX to maxX
     * @param translateY The Y-offset where the axis-line should be drawn
     * @param skipZero If TRUE then no label will be drawn for 0
     */
    public void drawXaxisWithTicks(int positionY, boolean skipZero, boolean formatAsPercentages) {
        // X-axis ticks
        Object aliasing=g.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        float range=(maxX-minX);
        double tickscaleX=ceilToNearestTenth(range);
        if (range/tickscaleX>50) tickscaleX=roundToNearestX(tickscaleX*5, tickscaleX); // check if there are too many ticks
        else if (range/tickscaleX>25) tickscaleX=roundToNearestX(tickscaleX*3, tickscaleX); // check if there are too many ticks
        else if (range/tickscaleX>15) tickscaleX=roundToNearestX(tickscaleX*2, tickscaleX); // check if there are too many ticks
        double tickAnchorX=0;
        if ((minX<0 && maxX<0) || (minX>0 && maxX>0)) {
            // 0 is not included in the range. Find some other anchor
            tickAnchorX=ceilToNearestX(minX,tickscaleX);
        }
        int tickAnchorXcoordinate=(int)((tickAnchorX-minX)/range*graphwidth);
        g.drawLine(tickAnchorXcoordinate+translateX, positionY, tickAnchorXcoordinate+translateX, positionY+5);
        if (tickAnchorX!=0 || !skipZero) drawAlignedNumber(tickAnchorX,tickscaleX,tickAnchorXcoordinate+translateX,positionY+7,0.5f,1f,formatAsPercentages);
        double lefttick=tickAnchorX-tickscaleX;
        int lefttickXcoordinate=(int)((lefttick-minX)/range*graphwidth);
        //System.err.println("TickScaleX="+tickscaleX+",  lefttick="+lefttick+",   xcoord="+lefttickXcoordinate);
        while (lefttickXcoordinate>=0) {
           g.drawLine(lefttickXcoordinate+translateX, positionY, lefttickXcoordinate+translateX, positionY+5);
           drawAlignedNumber(lefttick,tickscaleX,lefttickXcoordinate+translateX,positionY+7,0.5f,1f,formatAsPercentages);
           lefttick-=tickscaleX;
           lefttickXcoordinate=(int)((lefttick-minX)/range*graphwidth);
        }
        double righttick=tickAnchorX+tickscaleX;
        int righttickXcoordinate=(int)((righttick-minX)/range*graphwidth);
        while (righttickXcoordinate<=graphwidth) {
           g.drawLine(righttickXcoordinate+translateX, positionY, righttickXcoordinate+translateX, positionY+5);
           drawAlignedNumber(righttick,tickscaleX,righttickXcoordinate+translateX,positionY+7,0.5f,1f,formatAsPercentages);
           righttick+=tickscaleX;
           righttickXcoordinate=(int)((righttick-minX)/range*graphwidth);
        }
        g.drawLine(translateX,positionY,translateX+graphwidth,positionY);
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, aliasing);
    }


    /**
     * Draws vertical gridlines in the current color and stroke
     */
    public void drawVerticalGridLines() {
        // X-axis ticks
        float range=(maxX-minX);
        double tickscaleX=ceilToNearestTenth(range);
        if (range/tickscaleX>50) tickscaleX=roundToNearestX(tickscaleX*5, tickscaleX); // check if there are too many ticks
        else if (range/tickscaleX>25) tickscaleX=roundToNearestX(tickscaleX*3, tickscaleX); // check if there are too many ticks
        else if (range/tickscaleX>15) tickscaleX=roundToNearestX(tickscaleX*2, tickscaleX); // check if there are too many ticks
        double tickAnchorX=0;
        if ((minX<0 && maxX<0) || (minX>0 && maxX>0)) {
            // 0 is not included in the range. Find some other anchor
            tickAnchorX=ceilToNearestX(minX,tickscaleX);
        }
        int tickAnchorXcoordinate=(int)((tickAnchorX-minX)/range*graphwidth);
        g.drawLine(tickAnchorXcoordinate+translateX, translateY, tickAnchorXcoordinate+translateX, translateY+graphheight);
        double lefttick=tickAnchorX-tickscaleX;
        int lefttickXcoordinate=(int)((lefttick-minX)/range*graphwidth);
        //System.err.println("TickScaleX="+tickscaleX+",  lefttick="+lefttick+",   xcoord="+lefttickXcoordinate);
        while (lefttickXcoordinate>=0) {
           g.drawLine(lefttickXcoordinate+translateX, translateY, lefttickXcoordinate+translateX, translateY+graphheight);
           lefttick-=tickscaleX;
           lefttickXcoordinate=(int)((lefttick-minX)/range*graphwidth);
        }
        double righttick=tickAnchorX+tickscaleX;
        int righttickXcoordinate=(int)((righttick-minX)/range*graphwidth);
        while (righttickXcoordinate<=graphwidth) {
           g.drawLine(righttickXcoordinate+translateX, translateY, righttickXcoordinate+translateX, translateY+graphheight);
           righttick+=tickscaleX;
           righttickXcoordinate=(int)((righttick-minX)/range*graphwidth);
        }
    }

    /**
     * Draws a curve that goes through the list of points specified as (x,y)-value
     * the values are graph-values (not pixel coordinates)
     */
    public void drawCurve(float[][] points) {
        if (points==null || points.length<2) return;
        GeneralPath graph=new GeneralPath();
        graph.moveTo(getXforValue(points[0][0]),getYforValue(points[0][1]));
        for (int i=1;i<points.length;i++) {
            graph.lineTo(getXforValue(points[i][0]),getYforValue(points[i][1]));
        }
        g.draw(graph);
    }

    /**
     * Draws a curve that goes through the list of points specified
     * the values are graph-values (not pixel coordinates)
     */
    public void drawCurve(float[] Xpoints, float[] Ypoints) {
        if (Xpoints==null || Xpoints==null ||  Xpoints.length!=Ypoints.length || Xpoints.length<2) return;
        GeneralPath graph=new GeneralPath();
        graph.moveTo(getXforValue(Xpoints[0]),getYforValue(Ypoints[0]));
        for (int i=1;i<Xpoints.length;i++) {
            graph.lineTo(getXforValue(Xpoints[i]),getYforValue(Ypoints[i]));
        }
        g.draw(graph);
    }

    /**
     * Draws a histogram based on the bin-counts provided. 
     * The bins must span the whole value range and the count values
     * must lie within the Y-range of the graph (with 0 as minimum value)
     * @param counts
     */
    public void drawHistogram(float[] counts) {
        float binwidth=graphwidth/counts.length;
        Color color=g.getColor();
        Polygon histogram=new Polygon();
        histogram.addPoint(0, graphheight);
        for (int i=0;i<counts.length;i++) {
            float binvalue=counts[i];
            int stackheight=(int)Math.ceil((binvalue/maxY)*graphheight);
            histogram.addPoint((int)(i*binwidth), graphheight-stackheight);
            histogram.addPoint((int)((i+1)*binwidth), graphheight-stackheight);
        }
        histogram.addPoint((int)(counts.length*binwidth), graphheight);
        histogram.translate(translateX, translateY);
        g.setColor(new Color(color.getRed(),color.getGreen(),color.getBlue(),80));
        g.fillPolygon(histogram);
        g.setColor(color);
        g.drawPolygon(histogram);
    }



    /**
     * Returns the graphics X coordinate corresponding to a given value within the graph
     * @param value the requested value in the graph
     * @return
     */
    public int getXforValue(float value) {
        return (int)((value-minX)/(maxX-minX)*graphwidth)+translateX;
    }

    /**
     * Returns the graphics Y coordinate corresponding to a given value within the graph
     * @param value the requested value in the graph
     * @return
     */
    public int getYforValue(float value) {
        return graphheight-(int)((value-minY)/(maxY-minY)*graphheight)+translateY;
    }
    /**
     * Draws a Y-axis with sensible ticks within the range minY to maxY
     * @param positionX The X-offset where the axis-line should be drawn
     * @param skipZero If TRUE then no label will be drawn for 0
     */
    public void drawYaxisWithTicks(int positionX, boolean skipZero, boolean formatAsPercentages) {
        // X-axis ticks
        Object aliasing=g.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        float range=(maxY-minY);
        double tickscaleY=ceilToNearestTenth(range);
        if (range/tickscaleY>50) tickscaleY=roundToNearestX(tickscaleY*5, tickscaleY); // check if there are too many ticks
        else if (range/tickscaleY>25) tickscaleY=roundToNearestX(tickscaleY*3, tickscaleY); // check if there are too many ticks
        else if (range/tickscaleY>15) tickscaleY=roundToNearestX(tickscaleY*2, tickscaleY); // check if there are too many ticks
        double tickAnchorY=0;
        if ((minY<0 && maxY<0) || (minY>0 && maxY>0)) {
            // 0 is not included in the range. Find some other anchor
            tickAnchorY=ceilToNearestX(minY,tickscaleY);
        }
        int tickAnchorYcoordinate=(int)((tickAnchorY-minY)/range*graphheight);
        g.drawLine(positionX-5, graphheight-tickAnchorYcoordinate+translateY, positionX, graphheight-tickAnchorYcoordinate+translateY);
        if (tickAnchorY!=0 || !skipZero) drawAlignedNumber(tickAnchorY,tickscaleY,positionX-7, graphheight-tickAnchorYcoordinate+translateY,1f,0.3f,formatAsPercentages);
        double lefttick=tickAnchorY-tickscaleY;
        int lefttickYcoordinate=(int)((lefttick-minY)/range*graphheight);
        //System.err.println("TickScaleX="+tickscaleX+",  lefttick="+lefttick+",   xcoord="+lefttickXcoordinate);
        while (lefttickYcoordinate>=0) {
           g.drawLine(positionX-5, graphheight-lefttickYcoordinate+translateY, positionX, graphheight-lefttickYcoordinate+translateY);
           drawAlignedNumber(lefttick,tickscaleY,positionX-7,graphheight-lefttickYcoordinate+translateY,1f,0.3f,formatAsPercentages);
           lefttick-=tickscaleY;
           lefttickYcoordinate=(int)((lefttick-minY)/range*graphheight);
        }
        double righttick=tickAnchorY+tickscaleY;
        int righttickYcoordinate=(int)((righttick-minY)/range*graphheight);
        while (righttickYcoordinate<=graphheight) {
           g.drawLine(positionX-5, graphheight-righttickYcoordinate+translateY, positionX, graphheight-righttickYcoordinate+translateY);
           drawAlignedNumber(righttick,tickscaleY,positionX-7,graphheight-righttickYcoordinate+translateY,1f,0.3f,formatAsPercentages);
           righttick+=tickscaleY;
           righttickYcoordinate=(int)((righttick-minY)/range*graphheight);
        }
        g.drawLine(positionX,translateY,positionX,translateY+graphheight);
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, aliasing);
    }

    /**
     * Draws vertical gridlines in the current color and stroke
     */
    public void drawHorizontalGridLines() {
        // X-axis ticks
        float range=(maxY-minY);
        double tickscaleY=ceilToNearestTenth(range);
        if (range/tickscaleY>50) tickscaleY=roundToNearestX(tickscaleY*5, tickscaleY); // check if there are too many ticks
        else if (range/tickscaleY>25) tickscaleY=roundToNearestX(tickscaleY*3, tickscaleY); // check if there are too many ticks
        else if (range/tickscaleY>15) tickscaleY=roundToNearestX(tickscaleY*2, tickscaleY); // check if there are too many ticks
        double tickAnchorY=0;
        if ((minY<0 && maxY<0) || (minY>0 && maxY>0)) {
            // 0 is not included in the range. Find some other anchor
            tickAnchorY=ceilToNearestX(minY,tickscaleY);
        }
        int tickAnchorYcoordinate=(int)((tickAnchorY-minY)/range*graphheight);
        g.drawLine(translateX, graphheight-tickAnchorYcoordinate+translateY, translateX+graphwidth, graphheight-tickAnchorYcoordinate+translateY);
        double lefttick=tickAnchorY-tickscaleY;
        int lefttickYcoordinate=(int)((lefttick-minY)/range*graphheight);
        //System.err.println("TickScaleX="+tickscaleX+",  lefttick="+lefttick+",   xcoord="+lefttickXcoordinate);
        while (lefttickYcoordinate>=0) {
           g.drawLine(translateX, graphheight-lefttickYcoordinate+translateY, translateX+graphwidth, graphheight-lefttickYcoordinate+translateY);
           lefttick-=tickscaleY;
           lefttickYcoordinate=(int)((lefttick-minY)/range*graphheight);
        }
        double righttick=tickAnchorY+tickscaleY;
        int righttickYcoordinate=(int)((righttick-minY)/range*graphheight);
        while (righttickYcoordinate<=graphheight) {
           g.drawLine(translateX, graphheight-righttickYcoordinate+translateY, translateX+graphwidth, graphheight-righttickYcoordinate+translateY);
           righttick+=tickscaleY;
           righttickYcoordinate=(int)((righttick-minY)/range*graphheight);
        }
    }

    /**
     * Draws a text string at the given pixel-coordinate in the graph
     * @param g
     * @param text
     * @param x
     * @param y
     * @param alignX
     * @param alignY
     */
    public void drawAlignedString(String text, int x, int y, float alignX, float alignY) {
        int stringlength=g.getFontMetrics().stringWidth(text);
        int ascent=g.getFontMetrics().getAscent();
        g.drawString(text,  x-stringlength*alignX, y+ascent*alignY);
    }

    /**
     * Draws a vertically oriented text string at the given pixel-coordinate in the graph
     * @param text
     * @param x
     * @param y
     * @param up If true the text will be read upwards (with head tilted to the left). If false the text will be read downwards (with head tilted to the right)
     */
    public void drawVerticalString(String text, int x, int y, boolean up) {
        //int stringlength=g.getFontMetrics().stringWidth(text);
        int ascent=g.getFontMetrics().getAscent();
        AffineTransform oldtransform=g.getTransform();
        if (up) g.rotate(3*Math.PI/2,x,y);
        else g.rotate(Math.PI/2,x,y);
        g.drawString(text,  x, y+(int)((float)ascent/2f)-2); //
        g.translate(x, y);
        g.setTransform(oldtransform);
    }

    /** */
    public void drawPieChart(float[] values, Color[] colors, boolean drawOutLine) {
        float sum=0;
        for (float v:values) sum+=v;
        int start=0;
        for (int i=0;i<values.length;i++) {
            int angle=(int)Math.round((values[i]/sum)*360);
            //float percentage=(values/sum)*2*Math.PI;
            g.setColor(colors[i]);
            g.fillArc(translateX, translateY, graphwidth, graphheight, start, angle);
            start+=angle;
        }
        if (drawOutLine) {
            g.setColor(Color.BLACK);
            g.drawOval(translateX, translateY, graphwidth, graphheight);
        }
    }

    /**
     * Draws a number at the given pixel-coordinate in the graph
     * @param number
     * @param tickscale
     * @param x
     * @param y
     * @param alignX
     * @param alignY
     * @param formatAsPercentage
     */
    public void drawAlignedNumber(double number, double tickscale, int x, int y, float alignX, float alignY, boolean formatAsPercentage) {
        number=roundToNearestX(number,tickscale);
        String tickscaleasstring=""+tickscale;
        int significantdecimals=tickscaleasstring.indexOf(".");
        if (significantdecimals<0) significantdecimals=0;
        else significantdecimals=(tickscaleasstring.length()-1)-significantdecimals;
        String text=(formatAsPercentage)?(""+(number*100)):(""+number);
        int decimalpointpos=text.indexOf('.');
        if (number==0) text="0";
        else if (tickscale>1) { // remove decimals for numbers if tickscale is larger than 1 (these should not have significant decimals anyway)
            if (decimalpointpos>=0) text=text.substring(0,decimalpointpos);
        }
        else if (text.indexOf('.')>=0) { // tickscale<=1
            int pos=decimalpointpos+significantdecimals;
            if (pos<text.length()) text=text.substring(0,pos+1);
        }
        if (formatAsPercentage) text=text+"%";
        int stringlength=g.getFontMetrics().stringWidth(text);
        int ascent=g.getFontMetrics().getAscent();
        g.drawString(text,  x-stringlength*alignX, y+ascent*alignY);
    }


    private static float ceilToNearestX(float value, float tick) {
        int div=(int)(value/tick);
        return (div+1)*tick;
    }

    private static double ceilToNearestX(double value, double tick) {
        int div=(int)(value/tick);
        return (div+1)*tick;
    }

    private static double roundToNearestX(double value, double tick) {
        //System.err.print("Rounding "+value+"  to   ");
        int div=(int)(value/tick);
        double nearest=div*tick;
        if (Math.abs(value-div*tick)>Math.abs(value-(div+1)*tick)) nearest=(div+1)*tick;
        //System.err.println(""+nearest);
        return nearest;
    }

    private static double ceilToNearestTenth(double value) {
        double tick=1E-5;
        while (tick*1E2<=value) {
            tick*=1E1;
        }
        return tick;
    }
}
