/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.util.concurrent.BlockingQueue;
import priorseditor.gui.PriorsEditorGUI;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.HashSet;
import java.util.concurrent.LinkedBlockingQueue;
import priorseditor.gui.VisualizationSettingsListener;



/**
 * This class schedules the execution of tasks that need to be sequentially executed
 * in the order they were submitted.
 *
 * @author kjetikl
 */
public class GuiScheduler implements PropertyChangeListener {
    private final BlockingQueue<ExecutableTask> queue = new LinkedBlockingQueue<ExecutableTask>();; // tasks are added to this queue in the order submitted
    private ScheduledExecutor executor;
    private PriorsEditorEngine engine;
    private PriorsEditorGUI gui;
    private HashSet<PropertyChangeListener> propertyChangeListeners;
    private int queueNumber=0;
    private Object executionLock=new Object();



    /**
     *
     */
    public GuiScheduler(PriorsEditorGUI gui) {
       this.gui=gui;
       this.engine=gui.getEngine();
       propertyChangeListeners = new HashSet<PropertyChangeListener>();
    }



    /**
     * Submits a task to be scheduled for execution as soon as possible
     * @return
     */
    public synchronized void submit(ExecutableTask task) {
        //engine.logMessage("Submitting task for execution in GuiScheduler");
        if (executor==null || !executor.isAlive()) {
            executor=new ScheduledExecutor(this);
            executor.start(); // just start it right away. It will block anyway until tasks are submitted
        }
        if (queueNumber==0) {
            queueNumber=1;
            task.setQueueNumber(1);
        } else {
          queueNumber++;
          task.setQueueNumber(queueNumber);
        }
        queue.offer(task);
    }


    /**
     * Aborts the specified task. If the task is still pending it will simply
     * be removed from the queue without further ado. If the task is currently being
     * executed the Executor will be asked to shut it down
     * @param task
     * @return true if the task was currently executing and had to be aborted or
     * false otherwise
     */
    public boolean abort(ExecutableTask task) {
         //gui.debugMessage("aborting task in GuiScheduler");
         synchronized(queue) {
            if (executor==null) return false;
            ExecutableTask currentTask=executor.getCurrentTask();
            if (currentTask==null) return false;
            if (task==currentTask) {
                executor.interrupt();
                try {
                    executor.join(); // wait until the executor thread finishes
                } catch (Exception e) {}
                return true;
            } else {
                queue.remove(task);
                return false;
            }
        }
    }

    /** Aborts execution of all tasks */
    public void abort() {
       if (executor!=null) executor.abort();
       try {
            executor.join(); // wait until the executor thread finishes
       } catch (Exception e) {}
    }

    /** Cancels the current execution if an error/exception has made it impossible to cancel
     *  executions the normal way (i.e. using abort() ).
     *  This method is called by the GUI's default UncaughtExceptionHandler in order to clean
     *  up any executing tasks that might be "stuck"
     */
    public void abortEmergency() {
       if (executor!=null) executor.abortEmergency();
       try {
            executor.join(); // wait until the executor thread finishes
       } catch (Exception e) {}
       executor=null;
    }

    /**
     * Returns a lock object that can be used to temporarily suspend execution of tasks
     * by synchronizing on the lock. Note that the running task might not block right away
     * (tasks are only required to check this lock sporadically).
     */
    public Object getExecutionLock() {
        return executionLock;
    }

    /**
     * Returns the number of tasks in queue still waiting to be executed
     * (not counting the one currently executing)
     * @param waitForIt
     */
    public int pendingTasks() {
        return queue.size();
    }

        /** Returns true if the execution thread is not currently executing tasks */
        public boolean isIdle() {
            return (executor==null || executor.isIdle());
        }

    /**
     * Returns the queue number for the task currently being executed
     * The queue number for tasks increases every time a task is added to a
     * non-empty queue but resets when the queue empties
     * @returns The queue number of the task currently executing (1+) or 0 if there is no such task
     */
    public int getCurrentTaskQueueNumber() {
        if (executor==null) return 0;
        ExecutableTask ct=executor.getCurrentTask();
        if (ct!=null) return ct.getQueueNumber();
        else return 0;
    }
    /**
     * Returns the queue number for the last task added to the "current queue"
     * The queue number for tasks increases every time a task is added to a
     * non-empty queue but resets when the queue empties
     */
    public int getLastQueueNumber() {
        return queueNumber;
    }


    /**
     * This method tries to shutdown the currently running execution thread
     * Pending tasks will be flushed
     * @param waitForIt If this is set to true the method will wait until the
     * Executor thread is finished aborting. If false the method will return
     * right away.
     */
    public void shutdown(boolean waitForIt) {
        if (executor==null) return;
        executor.shutdown();
        if (waitForIt)
        try {
            executor.join(); // wait until the executor thread finishes
        } catch (Exception e) {}
    }



    /** Registers a new propertyChangeListener which will receive notifications of important updates */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeListeners.add(listener);
    }

    /** Removes a previously registered propertyChangeListener */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeListeners.remove(listener);
    }

    // The scheduler registers itself as PropertyChangeListener with ScheduledTasks to receive notifications about progress
    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        // just propagate the event for now. Maybe I will add more functionality later
        // I should probably check here if errors have occurred?
        for (PropertyChangeListener listener:propertyChangeListeners) listener.propertyChange(evt);
    }


    /**
     * The following private class represents the worker thread that
     * executes tasks. All it does is to remove scheduled tasks from
     * the queue and execute them. When the queue is empty the ScheduledExecutor
     * blocks until additional tasks are added, in which case the thread wakes
     * up and continues its work.
     */
    private class ScheduledExecutor extends Thread {
        private GuiScheduler scheduler;
        private boolean exitThread=false;
        ExecutableTask task=null; // the currently executing task

        public ScheduledExecutor(GuiScheduler scheduler) {
            super();
            this.scheduler=scheduler;
            setDaemon(true);
        }

        /** Cancels the currently used execution thread. (new ones can be created later !) */
        public void shutdown() {
            exitThread=true;
            this.interrupt();
        }

        /** Cancels the currently used execution thread. (new ones can be created later !) */
        public void abort() {
            this.interrupt(); //
        }

        /** Cancels the current execution if an error/exception has made it impossible to cancel
         *  executions the normal way (i.e. using abort() ).
         */
        public void abortEmergency() {
            if (task!=null) {
                PriorsEditorClient arbiter=task.getPriorsEditorClient();
                if ((!(task instanceof ProtocolTask) || arbiter.shouldRollback()) && gui.getUndoManager().canUndo()) {
                   gui.getUndoManager().undo(); // always rollback "atomic" tasks (they have probably not been executed properly anyway)
                   gui.getUndoManager().pop(); // remove the undo (after it is undone) so that it can't be redone
                }
                task.setStatus(ExecutableTask.ERROR);  // this will notify listeners, i.e. The scheduler
                task.setPriorsEditorClient(null);
                task.setExecutionLock(null);
                task.removePropertyChangeListener(scheduler);
                gui.updateUndoRedoStates();
            }
            task=null;
            flushQueue();
            engine.removeTemporaryDataItems();
            gui.reportExecutionEnded();
        }


        /** Returns true if the execution thread is not currently executing tasks */
        public boolean isIdle() {
            return (queue.size()==0 && task==null);
        }

        public ExecutableTask getCurrentTask() {
            return task;
        }

        /** Retrieves tasks from the queue one by one (blocking if the queue is empty) and executes them */
        @Override
        public void run() {
            boolean shouldAbort=false;
            while(!shouldAbort && !exitThread) { // infinite loop?
                if (queue.size()==0) queueNumber=0;
                try {
                    task=queue.take(); // this will block the thread until a new task is available in the queue
                    gui.reportExecutionStarted();
                    task.addPropertyChangeListener(scheduler);
                    task.setExecutionLock(executionLock);
                    task.setPriorsEditorClient(gui);
                    if (task instanceof AddSequencesTask) { // turn off visualization updates to avoid too many redundant updates
                         gui.getVisualizationSettings().enableVisualizationSettingsNotifications(false); // disable
                    }
                    task.run(); // <-- THIS IS IT. HERE WE GO !!!
                    task.setPriorsEditorClient(null);
                    task.setExecutionLock(null);
                    task.removePropertyChangeListener(scheduler);
                    if (task instanceof ProtocolTask) gui.logMessage("Protocol execution finished");
                    else if (task instanceof AddSequencesTask) {
                        gui.getVisualizationSettings().enableVisualizationSettingsNotifications(true); // reenable
                        gui.getVisualizationSettings().notifyListenersOfSequenceOrderRearrangement(VisualizationSettingsListener.REORDERED, engine.getDefaultSequenceCollection());
                    }
                    engine.removeTemporaryDataItems();
                    task=null; // clear before retrieving next task
                } catch (NullPointerException e) { // this should not happen
                    gui.getVisualizationSettings().enableVisualizationSettingsNotifications(true); // reenable
                    gui.debugMessage("SYSTEM ERROR: Caught unexpected NullPointerException");
                    if (task!=null) {
                        task.setStatus(ExecutableTask.ABORTED); // this will notify listeners, i.e. The scheduler
                        task.setExecutionLock(null);
                        task.removePropertyChangeListener(scheduler);
                        gui.updateUndoRedoStates();
                    }
                    shouldAbort=true;
                    task=null;
                    flushQueue(); // flush queue
                    throw e; // throws the nullpointer Exception to be handled by the defaultExceptionHandler
                    //e.printStackTrace(System.err);
                    //break;
                } catch (InterruptedException e) { // Execution aborted interactively by the user
                    gui.getVisualizationSettings().enableVisualizationSettingsNotifications(true); // reenable
                    if (task instanceof ProtocolTask) gui.logMessage("Protocol script execution aborted");
                    else gui.logMessage("Execution aborted");
                    if (task!=null) {
                        task.setStatusMessage("Aborted!");
                        task.setStatus(ExecutableTask.ABORTED); // this will notify listeners, i.e. The scheduler
                        task.setExecutionLock(null);
                        task.removePropertyChangeListener(scheduler);
                        gui.updateUndoRedoStates();
                    }
                    shouldAbort=true;
                    task=null;
                    flushQueue(); // flush queue
                    //break;

                } catch (Exception e) { // Execution aborted because of unforseen errors
                    gui.getVisualizationSettings().enableVisualizationSettingsNotifications(true); // reenable
                    if (task!=null) {
                        PriorsEditorClient arbiter=task.getPriorsEditorClient();
                        if ((!(task instanceof ProtocolTask) || arbiter.shouldRollback()) && gui.getUndoManager().canUndo()) {
                           gui.getUndoManager().undo(); // always rollback "atomic" tasks (they have probably not been executed properly anyway)
                           gui.getUndoManager().pop(); // remove the undo (after it is undone) so that it can't be redone
                        }
                        task.setStatus(ExecutableTask.ERROR);  // this will notify listeners, i.e. The scheduler
                        task.setPriorsEditorClient(null);
                        task.setExecutionLock(null);
                        task.removePropertyChangeListener(scheduler);
                        gui.updateUndoRedoStates();
                    }
                    shouldAbort=true;
                    task=null;
                    flushQueue();
                    //e.printStackTrace(System.err); // only for debug. Remove later!!
                    //break;
                } finally  {
                    gui.getVisualizationSettings().enableVisualizationSettingsNotifications(true); // reenable
                    engine.removeTemporaryDataItems();
                    gui.reportExecutionEnded();
                }
            }
            flushQueue(); // flushes the remaining scheduled tasks upon abortion/errors.
            //gui.debugMessage("ScheduledExecutor thread finished");
            task=null;
            queueNumber=0;
        } // --- end run() ---


        private void flushQueue() {
            queue.clear();
            queueNumber=0;
        }




    } // ******** end class ScheduledExecutor **********



}
