/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import priorseditor.engine.operations.*;
import priorseditor.engine.data.Data;
import java.util.HashMap;
import java.util.Set;

import priorseditor.engine.protocol.Protocol;

/**
 * Whereas the Operation classes represents singleton objects that only hold the
 * code necessary to perform different operations and contain no field variables
 * to hold instance specific data or parameters, the OperationTasks objects are
 * used to store the values for all parameters needed to perform an operation on
 * a specified set of data objects.   
 * 
 * OperationTasks also contain functions to obtain and update information about
 * the progress and status of an operation execution.
 * 
 * @author kjetikl
 */
public class OperationTask extends ExecutableTask {
    public static final String SOURCE="source";
    public static final String SOURCE_NAME="sourceName";
    public static final String TARGET="target";
    public static final String TARGET_NAME="targetName";
    public static final String OPERATION="operation";
    public static final String OPERATION_NAME="operationName";
    public static final String ENGINE="engine";
    public static final String SEQUENCE_COLLECTION_NAME="sequenceCollectionName"; // this field should hold the NAME (String) of the SequenceCollection
    public static final String SEQUENCE_COLLECTION="sequenceCollection"; // this field should hold the SequenceCollection itself
    private static final String RESERVED_NAMES="_RESERVED_NAMES";

    private HashMap<String,Object> storage;

   
    
    public OperationTask(String taskName) {
        super(taskName);
        storage=new HashMap<String,Object>();
    }
   
    /**
     * Sets a specific parameter to be used for execution of an Operation
     * @param key
     * @param value
     */
    public void setParameter(String key, Object value) {
        storage.put(key, value);
    } 

    /**
     * Gets a specific parameter to be used for execution of an Operation
     * @param key
     */    
    public Object getParameter(String key) {
        return storage.get(key);
    }
    
    public String[] getParameters() {
        Set<String> keys=storage.keySet();
        String[] res=new String[keys.size()];
        int i=0;
        for (String k:keys) {res[i]=k;i++;}
        return res;
        
    }
    
    /**
     * Returns the Operation object associated with this task
     */    
    public Operation getOperation() {
       Operation operation=(Operation)getParameter(OPERATION);
       if (operation!=null) return operation;
       else {
            String name=(String)getParameter(OPERATION_NAME);
            if (name==null) return null;
            PriorsEditorEngine engine=(PriorsEditorEngine)getParameter(ENGINE);
            if (engine!=null) return engine.getOperation(name);
       }
       return null;
    }
    /**
     * Returns the name of the Operation object associated with this task
     */    
    public String getOperationName() {
       String name=(String)getParameter(OPERATION_NAME);
       if (name!=null) return name;
       else {
          Operation operation=(Operation)getParameter(OPERATION);
          if (operation!=null) return operation.getName();
       }
       return null;
    }
    
    
    /**
     * Returns a reference to the engine (if set)
     */    
    public PriorsEditorEngine getEngine() {
        PriorsEditorEngine engine=(PriorsEditorEngine)getParameter(ENGINE);
        if (engine==null) {
          Operation operation=(Operation)getParameter(OPERATION);
          if (operation!=null) {
              engine=operation.getEngine();
              setParameter(ENGINE, engine);
          }      
        }
        return engine;
    }

    
    /**
     * Returns the Target dataobject associated with this task (if it exists)
     */    
    public Data getTargetData() {
       Data data=(Data)getParameter(TARGET);
       return data;
    }  
    
    /**
     * Returns the name of the Target dataobject associated with this task
     */    
    public String getTargetDataName () {
       return (String)getParameter(TARGET_NAME);
    }
    
    /**
     * Returns the Source dataobject associated with this task
     */    
    public Data getSourceData() {
       Data data=(Data)getParameter(SOURCE);
       return data;
    }
    
    /**
     * Returns the name of the Source dataobject associated with this task
     */    
    public String getSourceDataName() {
       return (String)getParameter(SOURCE_NAME);
    }
    
    /**
     * Returns a textual representation of this command (with parameters) to be included in protocol scripts
     */   
    @Override
    public String getCommandString(Protocol protocol) {
       if (protocol==null) return null;
       return protocol.getCommandString(this);
    }

    /**
     * This method can be used to 'reserve' names for new data objects that will be created by this task
     * The names of "default target(s)" will automatically be marked as reserved so there is no need to explicitly reserve these
     * but other names might be necessary to reserve (for instance the names of motif collections for motif discovery operations)
     */
    public void reserveDataName(String name) {
        String reserved=(String)getParameter(RESERVED_NAMES);
        if (reserved==null) reserved=name; else reserved=reserved+","+name;
        setParameter(RESERVED_NAMES, reserved);
    }

    /** Returns a comma-separated list of names that this task wishes to reserve for its output data */
    public String getReserveDataNames() {
        String reserved=(String)getParameter(RESERVED_NAMES);
        if (reserved==null) return getTargetDataName();
        else return getTargetDataName()+","+reserved;
    }
        
    /**
     * Executes the operation associated with this task
     * Implements the Runnable interface
     */
    @Override
    public void run() throws InterruptedException, Exception {
        //System.err.println("Executing operation");debug();
        boolean done=false;
        Operation operation=getOperation();  
        if (operation==null) return;
        setProgress(5);
        setStatus(RUNNING);  
        if (undoMonitor!=null) undoMonitor.register();
        while (!done){
          try {      
            operation.execute(this);
            done=true;
          } catch (InterruptedException e) { // task aborted by the user
              if (undoMonitor!=null) undoMonitor.deregister(false);
              setStatus(ABORTED);
              throw e;
          } catch (Exception e) { // other errors
              if (shouldRetry(e)) {
                  setRetrying(true);
              } else {
                  if (undoMonitor!=null) undoMonitor.deregister(false);
                  setStatus(ERROR);
                  throw e;
              }
          }            
        }
        setProgress(100);
        if (undoMonitor!=null) undoMonitor.deregister(true);
        setStatus(DONE);
        setStatusMessage(null);
    }

    
    public void debug() {
        System.err.println("DEBUG: OperationTask="+getOperationName());
        for (String paramName:storage.keySet()) {
            Object value=getParameter(paramName);
            if (value==null) System.err.println("    > "+paramName+" is null");
            else System.err.println("    > "+paramName+"=>"+value.toString());
        }
    }
}
