/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.util.HashMap;
import java.util.Set;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.NumericVariable;

/**
 * This class implements a set of values for parameters
 * The values are initially set as strings representing the value in question
 * (for instance the string "10.3" to represent the floating point number 10.3 
 * or the string "X" which refers to the name of a Numerical Variable)
 * 
 * When the method resolve() is called, the values for all parameters are replaced
 * with values of their proper type after parsing the initial strings and, 
 * if necessary, resolving any references to variables.
 * 
 * @author kjetikl
 */
public class ParameterSettings implements Cloneable {
    private HashMap<String,Object> storage;  // String is the name of the parameter, Object is the value
    
    public ParameterSettings() {
        storage=new HashMap<String,Object>();
    } 
    
    /**
     * Sets the value of the specified parameter
     * @param key
     * @param value
     */
    public void setParameter(String parameterName, Object value) {
        storage.put(parameterName, value);
    } 

    /**
     * Gets the value of the specified parameter as a String
     * @param key
     */    
    public String getParameterAsString(String parameterName, Parameter[] parameterFormats) {
        Object value=storage.get(parameterName);
        if (value==null) value=getDefaultValueForParameter(parameterName, parameterFormats);
        if (value==null) return null;
        else if (value instanceof String && ((String)value).isEmpty()) return null; // ?
        else return value.toString();
    }  
    
    /**
     * Gets the resolved value of the specified parameter. If there is no explicit value
     * set for this parameter, the parameters default value (if any) is returned
     * @param key
     */    
    public Object getResolvedParameter(String parameterName, Parameter[] parameterFormats, PriorsEditorEngine engine) throws ExecutionError {
        Object value=storage.get(parameterName);
        if (value==null) return getDefaultValueForParameter(parameterName, parameterFormats);
        if (!(value instanceof String)) return value; // already "resolved"
        Class type=getTypeForParameter(parameterName, parameterFormats);
        String valueAsString=(String)value;
        if (type==null) {
            return null;
        } else if (type.equals(String.class)) {
            return (String)value;
        } else if (type.equals(Integer.class)) {
            try {
                int intval=Integer.parseInt(valueAsString);
                return new Integer(intval);
            } catch (NumberFormatException ne) {
                Data dataitem=engine.getDataItem(valueAsString);
                if (dataitem instanceof NumericVariable) return new Integer((int)((NumericVariable)dataitem).getValue().floatValue());
                else throw new ExecutionError("Value for parameter '"+parameterName+"' is not a valid integer or Numeric Variable");
            }
        } else if (type.equals(Float.class)) {
            try {
                float floatval=Float.parseFloat(valueAsString);
                return new Float(floatval);
            } catch (NumberFormatException ne) {
                Data dataitem=engine.getDataItem(valueAsString);
                if (dataitem instanceof NumericVariable) return new Float(((NumericVariable)dataitem).getValue().floatValue());
                else throw new ExecutionError("Value for parameter '"+parameterName+"' is not a valid number or Numeric Variable");
            }            
        } else if (type.equals(Boolean.class)) {
                if (valueAsString.equalsIgnoreCase("TRUE") || valueAsString.equalsIgnoreCase("YES")) return Boolean.TRUE;
                else return Boolean.FALSE;
                // I haven't made a boolean Data type yet...
        } else  { // some data-type probably
            if (valueAsString.isEmpty()) return null;
            Data dataitem=engine.getDataItem(valueAsString);
            if (type.isInstance(dataitem)) return dataitem;
            else throw new ExecutionError("Value for parameter '"+parameterName+"' is not of correct type");
        } 
    }   

    /**
     * Returns an object representing the default value for the parameter with the given name,
     * or null if no there is no such parameter
     * @param parameterName
     * @param parameterFormats
     * @return
     */
    public Object getDefaultValueForParameter(String parameterName, Parameter[] parameterFormats) {
        Object def=null;
        if (parameterFormats==null) return null;
        for(Parameter item:parameterFormats) {
            if (item.getName().equals(parameterName)) {def=item.getDefaultValue();break;}
        }
        return def;
    }  
    
    /**
     * Returns an object representing the type for the parameter with the given name,
     * or null if no there is no such parameter. The object should either be of type "Class"
     * or "Class[]"
     * @param parameterName
     * @param parameterFormats
     * @return
     */
    public Class getTypeForParameter(String parameterName, Parameter[] parameterFormats) {
        Class type=null;
        if (parameterFormats==null) return null;
        for(Parameter item:parameterFormats) {
            if (item.getName().equals(parameterName)) {type=item.getType();break;}
        }
        return type;
    }    


    /**
     * Returns a list with the names of all (specified) parameters
     * @return
     */
    public String[] getParameters() {
        Set<String>keys=storage.keySet();
        String[]list=new String[keys.size()];
        return storage.keySet().toArray(list);        
    }
    

    /** Clears all parameter key,value-pairs stored in this object */
    public void clear() {
        storage.clear();
    }

    /** Dumps the contents of this settings object to STDERR*/
    public void debug() {
        Set<String>keys=storage.keySet();
        for (String key:keys) {
            System.err.println("  "+key+" => "+storage.get(key));
        }
    }


    @Override
    public Object clone()  {
        ParameterSettings newclone=new ParameterSettings();
        Set<String>keys=storage.keySet();
        for (String key:keys) {
            newclone.storage.put(key, storage.get(key));
        }
        return newclone;
    }


      
}
