/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Set;
import org.jdesktop.application.Application;
import priorseditor.engine.data.BackgroundModel;
import priorseditor.engine.data.DNASequenceData;
import priorseditor.engine.data.DNASequenceDataset;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.MotifCollection;
import priorseditor.engine.data.NumericVariable;
import priorseditor.engine.data.Organism;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.data.SequenceCollection;
import priorseditor.engine.data.StringCollection;
import priorseditor.engine.data.StringVariable;
import priorseditor.engine.dataformat.DataFormat;
import priorseditor.engine.dataformat.DataFormat_FASTA;
import priorseditor.engine.dataformat.DataFormat_INCLUSive_Background;
import priorseditor.engine.operations.Operation;
import priorseditor.engine.operations.Operation_new;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.protocol.StandardProtocol;
import priorseditor.gui.PriorsEditorApp;

/**
 * This class implements the command-line interface (CLI) version of PriorsEditor 
 * 
 * @author kjetikl
 */
public class PriorsEditor implements PriorsEditorClient {
    
    private static int MANUAL_ENTRY_FIELDS_COUNT=8;
    private static int GENE_ID_FIELDS_COUNT=6;
    
    
    private PriorsEditorEngine engine;
    private String protocolFilename;
    private String sequencesFilename;
    private BufferedWriter logger=null;
    private boolean silentmode=false; // in silent mode the user will not be prompted for values. Default values will be used in stead.
    
    public PriorsEditor(PriorsEditorEngine engine) {
        this.engine=engine;
        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() { shutdown(); }
        });         
    }
    
    public void setProtocolFilename(String protocol) {
        this.protocolFilename=protocol;
    }
    
    public void setSequencesFilename(String sequencesfilename) {
        this.sequencesFilename=sequencesfilename;
    }
    
    /** Specifies a BufferedWriter that should log all messages passed to the client */
    public void setLogger(BufferedWriter logger) {
        this.logger=logger;
    }
    
    
    @Override
    public boolean handleUncaughtException(Throwable e) {
        if (e instanceof OutOfMemoryError) {
            errorMessage("SYSTEM ERROR: Out of memory", 0);
        } else {
          errorMessage("SYSTEM ERROR: Fatal uncaught exception '"+e.getClass().getSimpleName()+"': "+e.getMessage(), 0); 
          engine.reportError(e);
        }
        return false;
    }

    @Override
    public void initializeClient(PriorsEditorEngine engine) {
        // emtpy interface implementation
    }

    @Override
    public Data promptValue(Data item, String message) throws ExecutionError {
        if (item==null) throw new ExecutionError("Unknown datatype for prompt");
        Data copy=item.clone(); // this is necessary
             if (item instanceof SequenceCollection) return promptValueSequenceCollection((SequenceCollection)copy, message);
        else if (item instanceof MotifCollection) return promptValueMotifCollection((MotifCollection)copy, message);
        else if (item instanceof BackgroundModel) return promptValueBackgroundModel((BackgroundModel)copy, message);
        else if (item instanceof NumericVariable) return promptValueNumericVariable((NumericVariable)copy, message);
        else if (item instanceof StringVariable) return promptValueStringVariable((StringVariable)copy, message);
        else if (item instanceof StringCollection) return promptValueStringCollection((StringCollection)copy, message);
        else throw new ExecutionError("Unable to prompt for values for data objects of type: "+item.getTypeDescription());
    }

    
    private Data promptValueSequenceCollection(SequenceCollection item, String message) {
        if (message==null || message.isEmpty()) System.out.println("\nPlease select sequences for collection '"+item.getName()+"': ");
        else System.out.println("\n"+message+": ");
        item.clearAll(engine);
        ArrayList<Data> sequences=engine.getAllDataItemsOfType(Sequence.class);
        for (Data sequence:sequences) {
            System.out.print("Include sequence '"+sequence.getName()+"' [Y/n]: ");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) item.addSequence((Sequence)sequence);
            else if (userInput.startsWith("y") || userInput.startsWith("Y")) item.addSequence((Sequence)sequence);
        }
        return item;
    }
    
    private Data promptValueMotifCollection(MotifCollection item, String message) throws ExecutionError {
        System.out.println("");
        String dataname=item.getName();
        Set<String> predef=engine.getPredefinedMotifCollections();
        String[] collectionNames=new String[predef.size()];
        collectionNames=predef.toArray(collectionNames);        
        for (int i=0;i<collectionNames.length;i++) {
            String name=collectionNames[i];
            int size=engine.getSizeForMotifCollection(name);
            System.out.println("["+(i+1)+"] "+name+" ("+size+" motifs)");
        }
        if (message==null || message.isEmpty()) System.out.println("\nPlease select motifs for collection '"+item.getName()+"': ");
        else System.out.println("\n"+message+": ");  
        System.out.println("Select a predefined collection by typing the number,");
        System.out.println("enter the filename of a motifs collection file");
        System.out.println("or press Enter to use default ("+item.getTypeDescription()+")");  
        boolean ok=false;
        while(!ok) {
            System.out.print("Select motifs>");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) return item;                     
            else {
                try { // did the user select a predefined model (i.e. type a number) ?
                    int selection=Integer.parseInt(userInput);
                    if (selection>0 && selection<=collectionNames.length) {
                        String collectionName=collectionNames[selection-1];
                        item=getPredefinedMotifCollection(item.getName(),collectionName);
                        ok=true;
                    }
                } catch (NumberFormatException e) { // the user typed a text (i.e. a filename)
                    File file=new File(userInput.trim());
                    if (!file.exists()) System.out.println("File not found! Please specify another motif collection");
                    else {
                        try {
                            item=getMotifCollectionFromFile(dataname,file);
                            ok=true;
                        } catch (Exception ex) {
                            System.out.println(ex.getMessage());
                        }
                    }                    
                }
            }
        }        
        return item;    
    }
    
    private Data promptValueBackgroundModel(BackgroundModel item, String message) throws ExecutionError {
        System.out.println("");
        String dataname=item.getName();
        Object[][] predef=BackgroundModel.predefinedModels;
        for (int i=0;i<predef.length;i++) {
            String name=(String)predef[i][0];
            int order=(Integer)predef[i][1];
            int organism=(Integer)predef[i][2];
            String organismName=Organism.getCombinedName(organism);
            if (organismName.startsWith("NCBI")) organismName="UNKNOWN";
            System.out.println("["+(i+1)+"] "+name+" (order="+order+", "+organismName+")");
        }
        if (message==null || message.isEmpty()) System.out.println("\nPlease select a background model for '"+item.getName()+"': ");
        else System.out.println("\n"+message+": ");  
        System.out.println("Select a predefined model by typing the number,");
        System.out.println("enter the filename of a background model file");
        System.out.println("or press Enter to use default ("+item.getTypeDescription()+")");  
        boolean ok=false;
        while(!ok) {
            System.out.print("Select background>");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) return item;                     
            else {
                try { // did the user select a predefined model (i.e. type a number) ?
                    int selection=Integer.parseInt(userInput);
                    if (selection>0 && selection<=predef.length) {
                        String modelname=(String)predef[selection-1][0];
                        item=getPredefinedBackgroundModel(item.getName(),modelname,BackgroundModel.getFilenameForModel(modelname));
                        ok=true;
                    }
                } catch (NumberFormatException e) { // the user typed a text (i.e. a filename)
                    File file=new File(userInput.trim());
                    if (!file.exists()) System.out.println("File not found! Please specify another background");
                    else {
                        try {
                            item=getBackgroundModelFromFile(dataname,file);
                            ok=true;
                        } catch (Exception ex) {
                            System.out.println(ex.getMessage());
                        }
                    }                    
                }
            }
        }        
        return item;        
    }
        
    private Data promptValueStringVariable(StringVariable item, String message) {
        //String defaultValue=item.getValue();
        if (message==null || message.isEmpty()) System.out.print("\nPlease provide a String value for '"+item.getName()+"' : ");
        else System.out.print("\n"+message+": ");
        String userInput=System.console().readLine();
        if (userInput!=null && !userInput.trim().isEmpty()) item.setValue(userInput.trim());        
        return item;       
    }
    
    private Data promptValueStringCollection(StringCollection item, String message) {
        item.clearAll();
        if (message==null || message.isEmpty()) System.out.println("\nPlease provide values for String Collection'"+item.getName()+"': ");
        else System.out.println("\n"+message+": ");
        boolean finished=false;
        while (!finished) {
           System.out.print("Enter string: ");
           String userInput=System.console().readLine();
           if (userInput==null || userInput.trim().isEmpty()) finished=true;
           else item.addString(userInput.trim());
        }
        return item;        
    }
    
    private Data promptValueNumericVariable(NumericVariable item, String message) {    
        float value=item.getValue();
        boolean ok=false;
        System.out.println("");
        while (!ok) {
            if (message==null || message.isEmpty()) System.out.print("Please provide a numeric value for '"+item.getName()+"' (default="+value+") : ");
            else System.out.print(message+": ");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) ok=true; // use default in this case
            else try {
                value=Float.parseFloat(userInput);
                ok=true;
            } catch (NumberFormatException e) {}       
        }   
        item.setValue(value);
        return item;
    }
    
    /** Returns a predefined Background model based on specifications */
    private BackgroundModel getPredefinedBackgroundModel(String targetName, String modelName, String filename) throws ExecutionError {
            if (filename==null) throw new ExecutionError("Unknown Background Model: "+modelName);
            BufferedReader inputStream=null;
            ArrayList<String> input=new ArrayList<String>();
            try {
                inputStream=new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("/priorseditor/engine/resources/"+filename)));
                String line;
                while((line=inputStream.readLine())!=null) {input.add(line);}
            } catch (IOException e) { 
                throw new ExecutionError("An error occurred when loading predefined Background model: ["+e.getClass().getSimpleName()+"] "+e.getMessage(),0);
            } finally {
                try {
                    if (inputStream!=null) inputStream.close();
                } catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader: "+ioe.getMessage());}
            }           
            DataFormat_INCLUSive_Background format=new DataFormat_INCLUSive_Background();   
            BackgroundModel data=new BackgroundModel(targetName);
            try {data=(BackgroundModel)format.parseInput(input, data,null, null);}
            catch (Exception e) {throw new ExecutionError(e.getMessage());} 
            data.setPredefinedModel(modelName);
            return data;        
    }    

    /** Returns a Background loaded from file */
    private BackgroundModel getBackgroundModelFromFile(String targetName, File file) throws ExecutionError {
            BufferedReader inputStream=null;
            ArrayList<String> input=new ArrayList<String>();
            try {
                inputStream=new BufferedReader(new FileReader(file));
                String line;
                while((line=inputStream.readLine())!=null) {input.add(line);}
            } catch (IOException e) { 
                throw new ExecutionError("An error occurred when loading Background model: ["+e.getClass().getSimpleName()+"] "+e.getMessage(),0);
            } finally {
                try {
                    if (inputStream!=null) inputStream.close();
                } catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader: "+ioe.getMessage());}
            }  
            if (input.isEmpty()) throw new ExecutionError("Background file is empty");
            String formatName=BackgroundModel.determineDataFormatFromHeader(input.get(0));
            if (formatName==null) throw new ExecutionError("Unable to determine format of backgroundmodel file");
            DataFormat format=engine.getDataFormat(formatName);  
            BackgroundModel data=new BackgroundModel(targetName);
            try {data=(BackgroundModel)format.parseInput(input, data,null, null);}
            catch (Exception e) {throw new ExecutionError(e.getMessage());} 
            return data;   
    }      

    /** Returns a predefined Motif Collection based on specifications */
    private MotifCollection getPredefinedMotifCollection(String targetName, String collectionName) throws ExecutionError {
            String filename=engine.getFilenameForMotifCollection(collectionName);
            if (filename==null) throw new ExecutionError("Unknown motif collection: "+collectionName);
            BufferedReader inputStream=null;
            ArrayList<String> input=new ArrayList<String>();
            try {
                inputStream=new BufferedReader(new InputStreamReader(new FileInputStream(filename)));
                String line;
                while((line=inputStream.readLine())!=null) {input.add(line);}
            } catch (IOException e) { 
                throw new ExecutionError("An error occurred when loading predefined motif collection: ["+e.getClass().getSimpleName()+"] "+e.getMessage(),0);
            } finally {
                try {
                    if (inputStream!=null) inputStream.close();
                } catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader: "+ioe.getMessage());}
            }           
            String dataformatname="PriorsEditorMotif";
            DataFormat format=engine.getDataFormat(dataformatname);
            if (format==null) throw new ExecutionError("Unknown data format '"+dataformatname+"'");
            MotifCollection data=new MotifCollection(targetName);
            try {data=(MotifCollection)format.parseInput(input, data,null, null);}
            catch (Exception e) {throw new ExecutionError(e.getMessage());} 
            data.setPredefinedCollectionName(collectionName);
            return data;        
    }    

    /** Returns a Background loaded from file */
    private MotifCollection getMotifCollectionFromFile(String targetName, File file) throws ExecutionError {
            BufferedReader inputStream=null;
            ArrayList<String> input=new ArrayList<String>();
            try {
                inputStream=new BufferedReader(new FileReader(file));
                String line;
                while((line=inputStream.readLine())!=null) {input.add(line);}
            } catch (IOException e) { 
                throw new ExecutionError("An error occurred when loading motif collection: ["+e.getClass().getSimpleName()+"] "+e.getMessage(),0);
            } finally {
                try {
                    if (inputStream!=null) inputStream.close();
                } catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader: "+ioe.getMessage());}
            }  
            if (input.isEmpty()) throw new ExecutionError("Motif collection file is empty");
            String formatName=MotifCollection.determineDataFormatFromHeader(input.get(0));
            if (formatName==null) throw new ExecutionError("Unable to determine format of motif collection file");
            DataFormat format=engine.getDataFormat(formatName);  
            MotifCollection data=new MotifCollection(targetName);
            try {data=(MotifCollection)format.parseInput(input, data, null, null);}
            catch (Exception e) {throw new ExecutionError(e.getMessage());} 
            return data;   
    }    
    
    
    
private Sequence[] parseSequencesFromFastaFile(String filename) { 
    DataFormat dataformat=engine.getDataFormat("FASTA");
    DataFormat_FASTA fasta=(DataFormat_FASTA)dataformat;
    DNASequenceDataset dataset=null;
    try {
       dataset=loadSequences(fasta, filename, null);
    } catch (Exception e) {
           errorMessage("An error occurred while attempting to read FASTA file\n"+e.getClass().getSimpleName()+" : " +e.getMessage(),0);
           System.exit(1);        
    }
    if (dataset==null) {
           errorMessage("No dataset returned",0);
           System.exit(1);         
    }  
    int size=dataset.getSize();
    Sequence[] sequences=new Sequence[size];
    for (int i=0;i<size;i++) {
        DNASequenceData seq=(DNASequenceData)dataset.getSequenceByIndex(i);
        sequences[i]=new Sequence(seq.getSequenceName(), 0, "", seq.getChromosome(),seq.getRegionStart(), seq.getRegionEnd(), seq.getGeneName(), seq.getTSS(), seq.getTES(), seq.getStrandOrientation());      
        if (seq.getTemporaryOrganism()!=null) sequences[i].setOrganism(seq.getTemporaryOrganism());
        if (seq.getTemporaryBuild()!=null) sequences[i].setGenomeBuild(seq.getTemporaryBuild());
    }
    return sequences;    
}    


private DNASequenceDataset loadSequences(DataFormat_FASTA fasta,String filename, ParameterSettings settings) throws Exception {
            ArrayList<String>input=new ArrayList<String>();
            File file=new File(filename);
            if (!file.exists()) throw new ExecutionError("File '"+filename+"' does not exist");
            if (!file.canRead()) throw new ExecutionError("Unable to read file '"+filename+"'");
            BufferedReader inputStream=null;
            try {
                inputStream=new BufferedReader(new FileReader(file));
                String line;
                int counter=0;
                while((line=inputStream.readLine())!=null) {
                    if (counter%50==0 && Thread.interrupted()) { throw new InterruptedException();}
                    counter++;
                    input.add(line);
                }
            } 
            catch (InterruptedException ie) {throw ie;}
            catch (Exception ioe) { 
                throw new ExecutionError(ioe.getMessage(), ioe);
            } finally {
                try {if (inputStream!=null) inputStream.close();} catch (IOException ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing BufferedReader: "+ioe.getMessage());}
            }                
            try {
                DNASequenceDataset result=fasta.parseNewSequenceDatasetFromFASTA(input, "DNA", settings);
                return result;
            }           
            catch (InterruptedException ie) {throw ie;}
            catch (Exception e) {throw new ExecutionError(e.getClass().getSimpleName()+":"+e.getMessage(), e); }               
}    
    
    
    @Override
    public boolean shouldRetry(ExecutableTask task, Exception e) {
        System.out.println("ERROR encountered: "+e.getClass().getSimpleName()+" : "+e.getMessage());
        //e.printStackTrace(System.err);
        boolean ok=false;
        while (!ok) {
            System.out.print("Retry? [y|n]:");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) continue;
            //System.err.println("Value='"+userInput+"'");
                 if (userInput.equalsIgnoreCase("YES") || userInput.equalsIgnoreCase("Y")) return true;
            else if (userInput.equalsIgnoreCase("NO") || userInput.equalsIgnoreCase("N")) return false;            
        }
        return false;
    }

    @Override
    public boolean shouldRollback() {
        return false; // not really necessary to roll back
    }

    @Override
    public void shutdown() {
        // --- For debugging ---
//          ArrayList<Data> list=engine.getAllDataItemsOfType(Data.class);
//          for (Data data:list) {
//              System.err.println("------------ "+data.getName()+" --------------");
//              System.err.println(data.output());
//              System.err.println("============================");              
//          }
          if (logger!=null) {
              try {
                  logger.close();
                  logger=null;
              } catch (IOException ioe) {System.err.println("SYSTEM ERROR (non-fatal): An error occurred when closing BufferedReader: "+ioe.getMessage());}
          }
//          File workdir=new File(engine.getWorkDirectory());
          engine.shutdown();
    }

    @Override
    public void errorMessage(String msg, int errortype) {
        if (logger!=null) {
            try {logger.append(msg+"\n");} catch (IOException e) {System.err.println("Unable to write error message to file: "+e.getMessage());}
        }
        else System.err.println(msg);
    }

    @Override
    public void logMessage(String msg) {
        if (logger!=null) {
            try {logger.append(msg+"\n");} catch (IOException e) {System.err.println("Unable to write log message to file: "+e.getMessage());}
        }
        else System.err.println(msg);
    }

    @Override
    public void statusMessage(String msg) {
       if (!msg.startsWith("Exec")) return; // only output execution messages
        if (logger!=null) {
            try {logger.append(msg+"\n");} catch (IOException e) {System.err.println("Unable to write log message to file: "+e.getMessage());}
        }
        else System.err.println(msg);       
    }

    
    /** This method does the main body of work */
    public void run() {
        PriorsEditorApp application=Application.getInstance(PriorsEditorApp.class);
        String workDir=application.getContext().getLocalStorage().getDirectory().getAbsolutePath(); 
        // System.err.println("Workdir="+workDir);
        engine.setWorkDirectory(workDir);
        engine.addMessageListener(this);
        engine.initialize();   
        engine.setClient(this);  
        logMessage("Reading sequences file");
        Sequence[] sequences=null;
        boolean sequenceFromFasta=isFASTAfile(sequencesFilename);
        if (sequenceFromFasta) {
            sequences=parseSequencesFromFastaFile(sequencesFilename);  
        } else {
            sequences=parseSequencesFile(sequencesFilename);            
        }
        logMessage("Reading protocol file");
        StandardProtocol protocol=readProtocolFromFile(protocolFilename);
        try {
            AddSequencesTask addSequencesTask=new AddSequencesTask(engine, sequences);
            addSequencesTask.setPriorsEditorClient(this);                    
            addSequencesTask.run(); 
            if (sequenceFromFasta) {
                DataFormat fasta=engine.getDataFormat("FASTA");
                Operation operation=engine.getOperation("new");      
                OperationTask newDNATrackTask=new OperationTask("new "+DNASequenceDataset.getType());
                newDNATrackTask.setParameter(OperationTask.OPERATION, operation);
                newDNATrackTask.setParameter(OperationTask.OPERATION_NAME, operation.getName());
                newDNATrackTask.setParameter(OperationTask.TARGET_NAME, "DNA");
                newDNATrackTask.setParameter(OperationTask.SOURCE_NAME, "DNA");
                newDNATrackTask.setParameter(Operation_new.DATA_TYPE, DNASequenceDataset.getType());    
                newDNATrackTask.setParameter(Operation_new.PARAMETERS,Operation_new.FILE_PREFIX); // the FILE_PREFIX part of the parameter is necessary for proper recognition by Operation_new
                newDNATrackTask.setParameter(Operation_new.FILENAME,sequencesFilename);
                newDNATrackTask.setParameter(Operation_new.DATA_FORMAT,fasta.getName());
                newDNATrackTask.setParameter(Operation_new.DATA_FORMAT_SETTINGS,null);  
                engine.executeTask(newDNATrackTask);                  
            }
            engine.executeProtocol(protocol);
        } catch (ParseError p) {
            errorMessage("Parse Error: "+p.getMessage(),0);
        } catch (ExecutionError e) {
            errorMessage("Execution Error: "+e.getMessage(),0);
        } catch (Exception e) {
            errorMessage("SYSTEM ERROR ("+e.getClass().getSimpleName()+") : "+e.getMessage(),0);
        }
        // now save all "OutputData" object to files before ending the program
        saveOutputData();
        shutdown();
    }
    
    
    /** returns true if the file with the given name is believed to be FASTA formatted
     */
    private boolean isFASTAfile(String filename) {
        BufferedReader inputStream=null;
        try {
            File file=new File(filename);
            inputStream=new BufferedReader(new FileReader(file));
            String line;
            if ((line=inputStream.readLine())!=null) { // first line
                if (line.startsWith(">")) return true;
                else return false;
            } return false;
        } catch (IOException e) { 
             errorMessage("SYSTEM ERROR: While parsing sequences file '"+filename+"': "+e.getMessage(), 0);
             System.exit(1);
        } finally {
            try {if (inputStream!=null) inputStream.close();} catch (IOException ioe) {System.err.println("SYSTEM ERROR (non-fatal): An error occurred when closing BufferedReader: "+ioe.getMessage());}
        }
        return false;    
    }
    
    private boolean saveOutputData() {
        ArrayList<Data> outputlist=engine.getAllDataItemsOfType(OutputData.class);
        for (Data data:outputlist) {
            OutputData output=(OutputData)data;
            String dataFormatName=output.getDataFormat();
            String suffix="txt";
            if (dataFormatName!=null && !dataFormatName.equals("*")) {
                DataFormat format=engine.getDataFormat(dataFormatName);
                if (format!=null) suffix=format.getSuffix();
            }
            String filename=output.getName()+"."+suffix;
            try {
                output.saveToFile(filename);
            } catch (IOException io) {
                errorMessage("ERROR: Unable to save file '"+filename+"' ("+io.getMessage()+")", 0);
                return false;
            }
        } // end output all 
        return true;
    }
    
    private StandardProtocol readProtocolFromFile(String filename) {
        BufferedReader inputStream=null;
        String text="";
        try {
            File file=new File(filename);
            inputStream=new BufferedReader(new FileReader(file));
            String line;
            while((line=inputStream.readLine())!=null) {text+=line+"\n";}
        } catch (IOException e) { 
             errorMessage("SYSTEM ERROR: While parsing protocol file '"+filename+"': "+e.getMessage(), 0);
             System.exit(1);
        } finally {
            try {if (inputStream!=null) inputStream.close();} catch (IOException ioe) {System.err.println("SYSTEM ERROR (non-fatal): An error occurred when closing BufferedReader: "+ioe.getMessage());}
        }
        StandardProtocol protocol=new StandardProtocol(engine,text); 
        return protocol;
    }
    
    
    /**
     * Method used to start PriorsEditor with command-line interface
     * @param args
     */
    public static void main (String[] args) {
        String protocolFilename=null;
        String sequencesFilename=null;
        BufferedWriter logger=null;
        for (int i=0;i<args.length;i++) {
            if (!args[i].startsWith("-")) reportErrorAndAbort("Options must start with '-'");    
            else if (args[i].equals("-p") || args[i].equals("-protocol")) {
                protocolFilename=getOptionValue(args, i);
                i++; // skip ahead            
            }
            else if (args[i].equals("-s") || args[i].equals("-sequences")) {
                sequencesFilename=getOptionValue(args, i);
                i++; // skip ahead            
            }
            else if (args[i].equals("-l") || args[i].equals("-log")) {
                String logfilename=getOptionValue(args, i);
                try {
                    File file=new File(logfilename);
                    logger=new BufferedWriter(new FileWriter(file));
                } catch (IOException e) {
                    reportErrorAndAbort(e.getMessage());
                }
                i++; // skip ahead
            }
            else if (args[i].equals("-h")) {showUsage(System.out);}
            else {
                reportErrorAndAbort("Unknown option: "+args[i]);
            }            
        }
        if (protocolFilename==null)  reportErrorAndAbort("No protocol file specified");
        if (sequencesFilename==null) reportErrorAndAbort("No sequences file specified");
        PriorsEditor priorsEditor=new PriorsEditor(new PriorsEditorEngine());
        priorsEditor.setProtocolFilename(protocolFilename);
        priorsEditor.setSequencesFilename(sequencesFilename);
        priorsEditor.setLogger(logger);
        priorsEditor.run();
    }
     
    
    private static void reportErrorAndAbort(String msg) {
        System.err.println(msg);
        System.exit(1);
    }
    
    private static String getOptionValue(String[] args, int optionpos) {
        if (optionpos+1>=args.length || args[optionpos+1].startsWith("-")) {
            reportErrorAndAbort("Missing value for option '"+args[optionpos]+"'");
            return null;
        }
        else return args[optionpos+1];
    }
    
    /**
     * Parses the specified sequence-file and returns an array of corresponding Sequence objects
     * for sequences found
     * @param filename
     * @return
     */
    private Sequence[] parseSequencesFile(String filename) {
        ArrayList<Sequence> sequences=new ArrayList<Sequence>();
        ArrayList<String[]> list=readSequencesFile(filename);
        ArrayList<GeneIdentifier> geneIDs=new ArrayList<GeneIdentifier>();
        // locate all Gene ID entries, parse these and add GeneIdentifier object to list to be resolved
        for (int i=0;i<list.size();i++) {
            String[] entry=list.get(i);
            if (entry.length==GENE_ID_FIELDS_COUNT) { // line specifies a sequence using Gene ID
                geneIDs.add(processGeneIDLine(entry));
            }
        }
        ArrayList<GeneIDmapping> resolved=resolveGenesFromList(geneIDs);
        if (resolved==null) resolved=new ArrayList<GeneIDmapping>();
        // Process full list again, this time resolved Gene ID info should be available
        for (int i=0;i<list.size();i++) {
            String[] entry=list.get(i);
            if (entry.length==GENE_ID_FIELDS_COUNT) { // line specifies a sequence using Gene ID
                Object[] result=processGeneIDLineInFull(entry);
                GeneIdentifier geneid=(GeneIdentifier)result[0];                
                int upstream=(Integer)result[1]; 
                int downstream=(Integer)result[2]; 
                String anchor=(String)result[3]; 
                ArrayList<Sequence> choice=resolveSequenceForGeneID(geneid,resolved,upstream,downstream,anchor);                
                sequences.addAll(choice);
            } else if (entry.length==MANUAL_ENTRY_FIELDS_COUNT){
                 Sequence seq=processManualEntryLine(entry);
                 if (seq!=null) {
                     if (seq.getSize()>engine.getMaxSequenceLength()) errorMessage("Warning: Size of sequence '"+seq.getName()+"' exceeds preset maximum ("+engine.getMaxSequenceLength()+" bp). Sequence skipped",0);
                     else sequences.add(seq);
                 } 
            } else {
                errorMessage("Wrong number of arguments in sequence file",0);
            }
        }     
        if (sequences.isEmpty()) {
           errorMessage("Execution Error: No sequences to analyse",0);
           System.exit(1);              
        }
        Sequence[] result=new Sequence[sequences.size()];
        for (int i=0;i<result.length;i++) result[i]=sequences.get(i);
        return result;
    }
    
    
    /** 
     * Reads a file containing sequence specifications. Each line should specify one sequence
     * in a comma or tab-delimited format using either MANUAL_ENTRY_FIELDS_COUNT fields (for manual entries) or GENE_ID_FIELDS_COUNT fields (for gene ID refs)
     * @return an ArrayList with String arrays containing relevant info. The String-arrays have lengths GENE_ID_FIELDS_COUNT or MANUAL_ENTRY_FIELDS_COUNT
     */
    private ArrayList<String[]> readSequencesFile(String filename) {
        ArrayList<String[]> results=new ArrayList<String[]>();
        BufferedReader inputStream=null;
        try {
            File file=new File(filename);
            inputStream=new BufferedReader(new FileReader(file));
            String originalLine;
            while((originalLine=inputStream.readLine())!=null) {
                String line=originalLine.trim();
                if (!(line.isEmpty() || line.startsWith("#"))) {
                    line=line.replace("\t", ",");
                    String[] fields=line.split("\\s*,+\\s*");
                    if (fields.length==GENE_ID_FIELDS_COUNT || fields.length==MANUAL_ENTRY_FIELDS_COUNT) {
                       results.add(fields);
                    } else {
                       errorMessage("Parse Error in '"+filename+"': number of fields on one line must be either "+GENE_ID_FIELDS_COUNT+" or "+MANUAL_ENTRY_FIELDS_COUNT+" (found "+fields.length+")\n=>"+originalLine,0);
                       System.exit(1);                          
                    }
                }
            }
        } catch (IOException e) { 
             errorMessage("SYSTEM ERROR: While parsing sequences file '"+filename+"': "+e.getMessage(), 0);
             System.exit(1);
        } finally {
            try {if (inputStream!=null) inputStream.close();} catch (IOException ioe) {System.err.println("SYSTEM ERROR (non-fatal): An error occurred when closing BufferedReader: "+ioe.getMessage());}
        }
        return results;      
    }
    
    
    /** Processes a line with MANUAL_ENTRY_FIELDS_COUNT String entries and returns a Sequence Object */
    private Sequence processManualEntryLine(String[] fields)  {
        String geneName=fields[0];
        String genomeBuild=fields[1];
        int organism=Organism.getOrganismForGenomeBuild(genomeBuild);
        String chromosome=fields[2];
        int start=0;
        int end=0;
        Integer TSS=null;
        Integer TES=null;
        int orientation=Sequence.DIRECT;
        try {start=Integer.parseInt(fields[3]);} 
        catch (NumberFormatException ne) {
            errorMessage("Parse Error: Unable to parse expected numeric value for 'start' (value='"+fields[3]+"')",0);
            System.exit(1);
        }        
        try {end=Integer.parseInt(fields[4]);} 
        catch (NumberFormatException ne) {
            errorMessage("Parse Error: Unable to parse expected numeric value for 'end' (value='"+fields[4]+"')",0);
            System.exit(1);
        }        
        if (!fields[5].isEmpty() && !fields[5].equalsIgnoreCase("NULL")) {
            try {
                int tss=Integer.parseInt(fields[5]);
                TSS=new Integer(tss);
            }
            catch (NumberFormatException ne) {
                errorMessage("Parse Error: Unable to parse expected numeric value for 'TSS' (value='"+fields[5]+"')",0);
                System.exit(1);
            }        
        }
        if (!fields[6].isEmpty() && !fields[6].equalsIgnoreCase("NULL")) {
            try {
                int tes=Integer.parseInt(fields[6]);
                TES=new Integer(tes);
            }
            catch (NumberFormatException ne) {
                errorMessage("Parse Error: Unable to parse expected numeric value for 'TES' (value='"+fields[6]+"')",0);
                System.exit(1);
            }        
        }   
        if (fields[7].equalsIgnoreCase("DIRECT") || fields[7].equals("1") || fields[7].startsWith("+")) orientation=Sequence.DIRECT;
        else if (fields[7].equalsIgnoreCase("REVERSE") || fields[7].startsWith("-")) orientation=Sequence.REVERSE;
        else {
            errorMessage("Parse Error: Unable to parse value for orientation (value='"+fields[7]+"')",0);
            System.exit(1);            
        }        
        return new Sequence(geneName, organism, genomeBuild, chromosome, start, end, geneName, TSS, TES, orientation);
    }

    
    /** Processes a line with GENE_ID_FIELDS_COUNT String entries and returns a GeneIdentifier */
    private GeneIdentifier processGeneIDLine(String[] fields)  { 
        String identifier=fields[0];
        String idFormat=fields[1];
        String build=fields[2];
        int organism=Organism.getOrganismForGenomeBuild(build);       
        return new GeneIdentifier(identifier, idFormat, organism, build);
    }
    
    /** 
     * Processes a line with GENE_ID_FIELDS_COUNT String entries and returns an object array
     * containing 
     * [0] a GeneIdentifier,
     * [1] A 'from' (upstream) specification 
     * [2] A 'to' (downstream) specification
     * [3] An anchor (eg. "TSS" or "TES"
     * 
     */
    private Object[] processGeneIDLineInFull(String[] fields)  {  
        Object[] result=new Object[]{null,null,null,null};
        String identifier=fields[0];
        String idFormat=fields[1];
        String build=fields[2];
        int organism=Organism.getOrganismForGenomeBuild(build);       
        result[0]=new GeneIdentifier(identifier, idFormat, organism, build);
        try {
            int from=Integer.parseInt(fields[3]);
            result[1]=new Integer(from);
        } catch (NumberFormatException ne) {
            errorMessage("Parse Error: Unable to parse expected numeric range-value for 'from' (value='"+fields[3]+"')",0);
            System.exit(1);
        }
        try {
            int to=Integer.parseInt(fields[4]);
            result[2]=new Integer(to);
        } catch (NumberFormatException ne) {
            errorMessage("Parse Error: Unable to parse expected numeric range-value for 'to' (value='"+fields[4]+"')",0);
            System.exit(1);
        }
        result[3]=fields[5];
        return result;
    }
    

    /** Resolves a list of GeneIdentifiers */
    private ArrayList<GeneIDmapping> resolveGenesFromList(ArrayList<GeneIdentifier> idlist) {
        GeneIDResolver idResolver=engine.getGeneIDResolver();
        ArrayList<GeneIDmapping> resolvedList=null;
        try {
            resolvedList=idResolver.resolveIDs(idlist);
        } catch (Exception e) { 
            errorMessage("An error occurred while resolving gene IDs: "+e.getClass().getSimpleName()+": "+e.getMessage(), 0);            
        }  
        return resolvedList;
    }
    
    
    /** 
     * Based on a GeneID specified by the user and a list of previously resolved gene IDs (among which the first argument geneid should hopefully be present)
     * the method returns a list of Sequence objects corresponding to the selected gene ID.
     * Note that when resolving Gene IDs, several different hits could be returned. 
     * The user is prompted to select the correct one (or several) and the ones the user selects are turned into Sequence objects
     * and returned as a list
     */
    private ArrayList<Sequence> resolveSequenceForGeneID(GeneIdentifier geneid, ArrayList<GeneIDmapping> resolvedList, int upstream, int downstream, String anchor) {
           ArrayList<Sequence> sequences=new ArrayList<Sequence>();
           ArrayList<GeneIDmapping> listForGene=getEntriesForID(resolvedList,geneid.identifier);
           if (listForGene.size()==0) {
               logMessage("Unable to find information about "+Organism.getCommonName(geneid.organism).toLowerCase()+" "+geneid.format+" identifier: "+geneid.identifier);
           } else if (listForGene.size()>1) {
               if (!silentmode) listForGene=selectCorrectMappings(listForGene,geneid.identifier); // prune list based on user selections
           } 
           for (GeneIDmapping mapping:listForGene) { // add all those mappings that are left for this ID
               Sequence sequence=new Sequence(mapping.geneID, new Integer(geneid.organism), geneid.build, mapping.chromosome, 0, 0, mapping.geneID, mapping.TSS, mapping.TES, mapping.strand);
               fillInStartAndEndPositions(sequence, upstream, downstream, anchor);
               if (sequence.getSize()>engine.getMaxSequenceLength()) {
                   errorMessage("Warning: Size of sequence '"+sequence.getName()+"' exceeds preset maximum ("+engine.getMaxSequenceLength()+" bp). Sequence skipped",0);
               } else sequences.add(sequence);
           }        
           return sequences;
    }        
        
      

    
    
        
    
    /** Goes through a list of GeneIDmapping and returns only those entries that correspond to the given gene id */
    private ArrayList<GeneIDmapping> getEntriesForID(ArrayList<GeneIDmapping> list, String id) {
        ArrayList<GeneIDmapping> result=new ArrayList<GeneIDmapping>();
        for (GeneIDmapping entry:list) {
            if (entry.geneID.equals(id)) result.add(entry);
        }
        return result;
    }
    
/** Fills in upstream and downstream coordinates based on user selections and gene orientation */
    private void fillInStartAndEndPositions(Sequence sequence, int upstream, int downstream, String anchor) {
        if (upstream>0) upstream--; // to account for direct transition from -1 to +1 at TSS
        if (downstream>0) downstream--; // to account for direct transition from -1 to +1 at TSS
        int tss=sequence.getTSS();
        int tes=sequence.getTES();
        if (anchor.equalsIgnoreCase("Transcription Start Site") || anchor.equalsIgnoreCase("TSS")) {
            if (sequence.getStrandOrientation()==Sequence.DIRECT) {
               sequence.setRegionStart(tss+upstream);
               sequence.setRegionEnd(tss+downstream);           
            } else { // Reverse Strand
               sequence.setRegionStart(tss-downstream);
               sequence.setRegionEnd(tss-upstream);                
            }
        } else if (anchor.equalsIgnoreCase("Transcription End Site") || anchor.equalsIgnoreCase("TES")) {
            if (sequence.getStrandOrientation()==Sequence.DIRECT) {
               sequence.setRegionStart(tes+upstream);
               sequence.setRegionEnd(tes+downstream);           
            } else { // Reverse Strand
               sequence.setRegionStart(tes-downstream);
               sequence.setRegionEnd(tes-upstream);                
            }        
        } else {
            errorMessage("Unsupported anchor site: "+anchor,0);
            System.exit(0);
        }
    }      
    
    /** This method is used whenever multiple mappings are returned for a single gene ID
     * (I am not sure if this is possible but I will account for it anyway). A dialog is
     * displayed so that the user can select the mapping wanted (or even several mappings, or none)
     */
    private ArrayList<GeneIDmapping> selectCorrectMappings(ArrayList<GeneIDmapping> list, String id) {
        ArrayList<GeneIDmapping> result=new ArrayList<GeneIDmapping>();
        System.out.println("Found several matches for identifier '"+id+"'. Please select the correct one (or comma-separated list)");
        for (int i=0;i<list.size();i++) {
            GeneIDmapping mapping = list.get(i);
            int start=(mapping.TSS<mapping.TES)?mapping.TSS:mapping.TES;
            int end  =(mapping.TSS<mapping.TES)?mapping.TES:mapping.TSS;
            String string=id+"_"+(i+1);
            string=padToLength(string, 20, false);
            string+=mapping.geneName;
            string=padToLength(string, 30, false);
            String chromosome="chr"+mapping.chromosome+":"+start+"-"+end;
            chromosome=padToLength(chromosome, 30, true);
            string+=chromosome+"       ";
            string+=((mapping.strand==Sequence.DIRECT)?"Direct          ":"Reverse          ");              
            System.out.println("["+(i+1)+"] "+string);
        }
        boolean ok=false;
        ArrayList<Integer> indices=new ArrayList<Integer>();
        while (!ok) {
            System.out.print("Enter selection: ");
            String userInput=System.console().readLine();
            if (userInput==null || userInput.trim().isEmpty()) continue;
            String[] elements=userInput.split("\\s*,\\s*");     
            for (String element:elements) {
                indices.clear();
                try {
                    Integer number=Integer.parseInt(element);
                    if (number<0 || number>list.size()) {
                        System.out.println("Selected number out of range: '"+number+"'");
                        break;                        
                    } else indices.add(number);
                } catch (NumberFormatException e) {
                    System.out.println("Unable to parse number: '"+element+"'");
                    break;
                }
                ok=true;
            }
        }
        for (Integer index:indices) {
            try {
                GeneIDmapping mapping=list.get(index);
                mapping.geneID=mapping.geneID+"_"+(index+1);
                result.add(mapping);
            } catch (NumberFormatException e) {}
        }        
        return result;
    }    
    
    
    private String padToLength(String string, int size, boolean front) {
        while (string.length()<size) {
            if (front) string=" "+string;
            else string+=" ";
        }
        return string;
    }
    
    /** Prints usage information */
    private static void showUsage(PrintStream stream) {
        stream.println("Usage:");
        stream.println("java -cp PriorsEditor.jar priorseditor.engine.PriorsEditor -p <protocolfile> -s <sequencesfile> [optional arguments]");
        stream.println("");
        stream.println("Recognized arguments:");
        stream.println("      -p | -protocol <protocolfile>    The protocol to be executed");
        stream.println("      -s | -sequences <sequencesfile>  A file containing either sequences");
        stream.println("                                       in FASTA-format or a specification");
        stream.println("                                       of which sequences to use");
        stream.println("      -l | -log <logfile>              log execution messages");
        stream.println("      -h                               print this message");
        
    }
}
