/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.management.ManagementFactory;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import priorseditor.engine.data.*;
import priorseditor.engine.operations.*;
import java.util.*;
import java.util.prefs.Preferences;
import priorseditor.engine.dataformat.*;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.protocol.Protocol;
import priorseditor.external.ExternalProgram;



/**
 *
 * @author kjetikl
 */
public class PriorsEditorEngine implements MessageListener, DataListener, Thread.UncaughtExceptionHandler {
 
    public static final String PREFERENCES_USE_CACHE_FEATUREDATA="useFeatureDataCache";
    public static final String PREFERENCES_USE_CACHE_GENE_ID_MAPPING="useGeneIDMappingCache";    
    public static final String PREFERENCES_NETWORK_TIMEOUT="networkTimeout";    
    public static final String MAX_SEQUENCE_LENGTH="maxSequenceLength";    
    
    private DataStorage storage;
    private HashSet<MessageListener> messagelisteners=new HashSet<MessageListener>();
    private HashSet<DataListener> datalisteners=new HashSet<DataListener>();
    private HashMap<String,Operation> operationslist=new HashMap<String,Operation>();
    private HashMap<String,DataFormat> dataformatslist=new HashMap<String,DataFormat>();
    private HashMap<Class,String> defaultdataformats=new HashMap<Class,String>();
    private HashMap<String,Object[]> predefinedMotifCollections=new HashMap<String,Object[]>();// (String)Object[0]=filename, (Integer)Object[1]=collection size
    private String defaultSequenceCollectionName="AllSequences";
    private String defaultOutputStreamName="Output";
    private Object executionLock=new Object(); // this is used to synchronize execution threads with other threads
    private String[] operations=new String[]{"combine_numeric","combine_regions","convert","count","decrease","distance","divide","execute","extend","filter","increase","interpolate","mask","merge","motifDiscovery","motifScanning","new","normalize","output","physical","prompt","rescore","score","search","set","smooth","statistic","weight"};
    private PriorsEditorClient client=null;
    private DataLoader dataLoader;
    private GeneIDResolver geneIDResolver;
    private HashMap<String,ExternalProgram> externalPrograms=new HashMap<String,ExternalProgram>();
    private HashSet<String> reservedwords=new HashSet<String>(); // list of Protocol-specific words that can not be used as variable names
    private Preferences preferences;
    private int networkTimeoutDelay=25000; // milliseconds until the DataLoader should call a network timeout and proceed to the next mirror in the list (or throw Exception if there are no more mirrors)
    private String workDirectoryPath="";
    private int maxSequenceLength=10000;
            
    public PriorsEditorEngine() {
        storage=new DataStorage();
        SequenceCollection defaultSequenceCollection=new SequenceCollection(defaultSequenceCollectionName);
        try {storeDataItem(defaultSequenceCollection);} catch (Exception e) {}
        preferences = Preferences.userNodeForPackage(this.getClass());
        Thread.setDefaultUncaughtExceptionHandler(this);  
    }
    
    /**
     * Initializes the engine and sets up necessary resources
     */
    public void initialize() {
        logMessage("Initializing PriorsEditorEngine");
        dataLoader=new DataLoader(this);  
        geneIDResolver=new GeneIDResolver(this);
        try {dataLoader.setup();}
        catch (Exception e) {errorMessage(e.getMessage(), 0);}     
        dataLoader.setUseCache(preferences.getBoolean(PREFERENCES_USE_CACHE_FEATUREDATA, false));
        geneIDResolver.setUseCache(preferences.getBoolean(PREFERENCES_USE_CACHE_GENE_ID_MAPPING, false));
        networkTimeoutDelay=preferences.getInt(PREFERENCES_NETWORK_TIMEOUT, networkTimeoutDelay);
        setDataCacheDirectory(getWorkDirectory()+java.io.File.separator+"cache");
        setGeneIDCacheDirectory(getWorkDirectory());
        if (!setupTempDirectory()) System.exit(1);      
        importOperations(); // import known operations
        importDataFormats(); // import known data formats     
        importPredefinedMotifCollections(); // import information about predefined motif collections           
        importExternalPrograms(); // import external programs
        registerReservedWord("Visualization"); // these words are reserved to avoid problems in the GUI
        registerReservedWord("Protocol");
        registerReservedWord("within");
        Preferences enginePreferences=Preferences.userNodeForPackage(PriorsEditorEngine.class);        
        maxSequenceLength=enginePreferences.getInt(PriorsEditorEngine.MAX_SEQUENCE_LENGTH, 10000);          
    }

    /**
     * This method executes a protocol script and is called by the command-line client.
     * Note that the graphical user interface has its own protocol execution class called GuiScheduler
     * and does not use this method
     * @param protocol
     * @throws priorseditor.engine.ExecutionError
     */
    public void executeProtocol(Protocol protocol) throws ParseError,ExecutionError {
        ProtocolTask task = protocol.parse();
        task.setPriorsEditorClient(client);
        try {      
            logMessage("Executing protocol");
            task.run(); 
            logMessage("Execution finished");
        } catch (ExecutionError e) { 
            task.setStatus(ExecutableTask.ERROR);  // this will notify listeners  
            throw e;
        } catch (NullPointerException e) { // this should not happen
            errorMessage("SYSTEM ERROR: Caught unexpected NullPointerException",0);                   
            task.setStatus(ExecutableTask.ABORTED); // this will notify listeners                    
        } catch (InterruptedException e) { // Execution aborted interactively by the user
            //gui.debugMessage("ABORT!!!   ScheduledExecutor thread interrupted"); 
            logMessage("Protocol script execution aborted");
            task.setStatusMessage("Aborted!");
            task.setStatus(ExecutableTask.ABORTED); // this will notify listeners              
        } catch (Exception e) { // Execution aborted because of unforseen errors 
            task.setStatus(ExecutableTask.ERROR);  // this will notify listeners
            throw new ExecutionError(e.getClass().getSimpleName()+":"+e.getMessage());
        }           
        task.setPriorsEditorClient(null);
    }
    
    /**
     * This method executes an OperationTask and is called by the command-line client
     * Note that the graphical user interface has its own execution class called GuiScheduler
     * and does not use this method
     * @param task
     * @throws priorseditor.engine.ExecutionError
     */
    public void executeTask(OperationTask task) throws ExecutionError {
        task.setPriorsEditorClient(client);
        try {      
            task.run(); 
        } catch (ExecutionError e) { 
            task.setStatus(ExecutableTask.ERROR);  // this will notify listeners  
            throw e;
        } catch (NullPointerException e) { // this should not happen
            errorMessage("SYSTEM ERROR: Caught unexpected NullPointerException",0);                   
            task.setStatus(ExecutableTask.ABORTED); // this will notify listeners                    
        } catch (InterruptedException e) { // Execution aborted interactively by the user
            //gui.debugMessage("ABORT!!!   ScheduledExecutor thread interrupted"); 
            logMessage("Protocol script execution aborted");
            task.setStatusMessage("Aborted!");
            task.setStatus(ExecutableTask.ABORTED); // this will notify listeners              
        } catch (Exception e) { // Execution aborted because of unforseen errors 
            task.setStatus(ExecutableTask.ERROR);  // this will notify listeners
            throw new ExecutionError(e.getClass().getSimpleName()+":"+e.getMessage());
        }           
        task.setPriorsEditorClient(null);
    }
    
    
    
    /**
     * Sets a client to use for this session
     * It will also call "initializeClient" in the client
     * @param client
     */
    public void setClient(PriorsEditorClient client) {
        this.client=client;
        addMessageListener(client);
        client.initializeClient(this);
    }
    
    /**
     * Returns a reference to the current client
     * @param client
     */
    public PriorsEditorClient getClient() {
        return client;
    }
    
    /** Sets a maximum sequence length permitted by the engine */
    public void setMaxSequenceLength(int newmax) {
         maxSequenceLength=newmax;
    }

    /** Returns the maximum sequence length permitted by the engine */
    public int getMaxSequenceLength() {
        return maxSequenceLength;
    }
    
    /** 
     * Returns a string denoting the path of the work directory for this PriorsEditor installation 
     * The work directory contains such information as GUI-settings, Datatrack configuration settings,
     * data caches, undo caches etc.
     */
    public String getWorkDirectory() {
        return workDirectoryPath;
    }
    
    /** 
     * Sets the path of the work directory to use for this PriorsEditor installation 
     * The work directory contains such information as GUI-settings, Datatrack configuration settings,
     * data caches, undo caches etc.
     */
    public void setWorkDirectory(String workDirectoryPath) {




        File file=new File(workDirectoryPath);
        boolean success=true;
        if (!file.exists()) success=file.mkdirs();
        if (!success) errorMessage("Unable to create work-directory: "+workDirectoryPath,0);
        this.workDirectoryPath=workDirectoryPath;
    }
    
    /** Creates a temp directory for this PriorsEditor session */
    private boolean setupTempDirectory() {
       File dir=new File(getTempDirectory());
       if (!dir.mkdirs()) {logMessage("Unable to create temp directory on local disc => "+getTempDirectory()); return false;}
       dir.deleteOnExit(); // tells the VM to delete the temp directory automatically on exit
       return true;
    }
    
    /** 
     * Returns a string denoting the path of the temp directory for this PriorsEditor session.
     * The temp directory is located beneath the work directory for the PriorsEditor installation
     * but is different for each PriorsEditor session. Work directories are emptied and deleted 
     * when the process exits.
     */
    public String getTempDirectory() {
        return workDirectoryPath+File.separator+ManagementFactory.getRuntimeMXBean().getName();
    }

    /** 
     * Creates and returns a reference to a temporary file (with unique filename).
     * Temp files are automatically deleted when the PriorsEditor session is over
     * @return A File object which holds the pathname for the temp-file, or NULL if the file could not be created
     */
    public File createTempFile() {
        return createTempFile(".tmp");
    }
    
    /** 
     * Creates and returns a reference to a temporary file (with unique filename).
     * Temp files are automatically deleted when the PriorsEditor session is over
     * @return A File object which holds the pathname for the temp-file, or NULL if the file could not be created
     */
    public File createTempFile(String suffix) {
        File temp=null;
        try {
            temp=File.createTempFile("PriorsEditor", suffix, new File(getTempDirectory()));
            temp.deleteOnExit();
        } catch (Exception e) {errorMessage("Unable to create temp file", 0);}
        return temp;
    }
    
 
    
    /**
     * Returns a data item with the specified name that has been registered with the engine
     * Registered data items are usually datasets or scalar variables. Individual sequences 
     * within datasets should not be stored individually but rather as part of the datasets 
     * Thus to obtain a single sequence, use method to obtain the dataset, then use methods
     * in the dataset to obtain the sequence.
     * @param key
     */
    public Data getDataItem(String key) {
        return storage.getDataItem(key);
    }
    
    /**
     * Registers a new dataitem with the engine or replaces an old Data object of the same name
     * If a Data item of the same name already exists in storage, the previous Data object will be removed
     * from storage before the new Data item is added
     *
     * @param dataitem The new dataitem to be added 
     */
    public void storeDataItem(Data dataitem) throws ExecutionError {
        String namecheck=checkNameValidity(dataitem.getName(),false);
        if (!(dataitem instanceof Sequence) && namecheck!=null) throw new ExecutionError("Invalid data name '"+dataitem.getName()+"': "+namecheck);
        Data olditem=storage.getDataItem(dataitem.getName()); // is there a previous Data object with the same name?
        if (olditem!=null && dataitem.getClass()!=olditem.getClass()) {
           if (olditem.getClass()==Sequence.class) throw new ExecutionError("Cannot store new data object with name '"+dataitem.getName()+"' since a Sequence with that name already exists");
           else if (olditem.getClass()==Motif.class) throw new ExecutionError("Cannot store new data object with name '"+dataitem.getName()+"' since a Motif with that name already exists");
           else if (olditem.getClass()==OutputData.class) throw new ExecutionError("Cannot store new data object with name '"+dataitem.getName()+"' since an Output object with that name already exists");
        }
        if (olditem!=null) removeDataItem(olditem.getName());
        dataitem.addDataListener(this);
        if (dataitem instanceof MotifCollection) {((MotifCollection)dataitem).initializeFromPayload(this);}
        storage.storeDataItem(dataitem);
        for (DataListener listener:datalisteners) listener.dataAdded(dataitem); // Notify all datalisteners that a new dataitem has been added
        if (dataitem instanceof Sequence) {
           SequenceCollection defaultSC=getDefaultSequenceCollection();
           for (DataListener listener:datalisteners) listener.dataUpdate(defaultSC,null); // Notify all datalisteners that the DefaultSequenceCollection has been updated
           defaultSC.addSequence((Sequence)dataitem);
        }

    }


    /**
     * Updates a data item already registered with the engine by providing an argument Dataitem whose
     * values will be imported into the existing Data object.
     * If the two objects are of incompatible types a ClassCastException will be thrown
     * If two items are compatible registered DataListeners will be notified through the callback
     * interface dataUpdate(old, new) before the actual data import occurs. If no previous Data
     * items with the same name exists, the method will add the Data item as a new object by passing
     * on the same argument to storeDataItem(dataitem)
     * exec
     * NOTE: The procedure might fail if the dataitem provided is already stored in the engine
     * (i.e. if one attempts to update the dataitem itself using this method).
     * This applies for instance to Collection objects where the target (stored in the engine)
     * is cleared of all entries before copying the entries from object provided as argument.
     * Thus, if the target and source are the same object (not just having the same names), the data
     * in the target (and hence the source) will be cleared and there will be nothing left to
     * import into the target.
     *
     * @param dataitem The new dataitem to be added (or an object holding values to be imported in an old data item)
     */
    public void updateDataItem(Data dataitem) throws ExecutionError, ClassCastException {
        Data olditem=storage.getDataItem(dataitem.getName()); // is there a previous Data object with the same name?
        if (olditem==null) {
            storeDataItem(dataitem);
        } else {
            if (dataitem.getClass()!=olditem.getClass()) { // incompatible types?
                if (olditem.getClass()==Sequence.class || dataitem.getClass()==Sequence.class) throw new ClassCastException("Illegal type change from/to Sequence");
                else if (olditem.getClass()==Motif.class || dataitem.getClass()==Motif.class) throw new ClassCastException("Illegal type change from/to Motif");
                else if (olditem.getClass()==OutputData.class || dataitem.getClass()==OutputData.class) throw new ClassCastException("Illegal type change from/to Output object");
                for (DataListener listener:datalisteners) listener.dataUpdate(olditem,dataitem); // Notify all datalisteners that an update is about to occur
                storeDataItem(dataitem);
            } else { //
                for (DataListener listener:datalisteners) listener.dataUpdate(olditem,dataitem); // Notify all datalisteners that an update is about to occur
                if (dataitem instanceof MotifCollection) {((MotifCollection)dataitem).initializeFromPayload(this);}
                olditem.importData(dataitem);
                olditem.notifyListenersOfDataUpdate();
            }
        }
    }
   
    
    
    /**
     * Removes a specified dataitem from storage and notifies any listeners.
     * @param key The name of the dataitem
     * @return the dataitem that was removed or null if it was not found
     */
    public Data removeDataItem(String key) {
        Data dataitem=storage.removeDataItem(key);
        if (dataitem==null) return null;
        dataitem.removeDataListener(this);
        // notify all datalisteners      
        if (dataitem instanceof Sequence) getDefaultSequenceCollection().removeSequence((Sequence)dataitem);
        // this should be removed from all collections!!!
        for (DataListener listener:datalisteners) {
            listener.dataRemoved(dataitem);
        }  
        return dataitem;
    }
    
    /**
     * Renames the specified data item. The rename() method in individual Data items
     * should not be called directly, rather Data items should be renamed with this function
     * so that the hash-keys in the DataStorage are updated accordingly.
     * Listeners are not notified of the name change by the Engine but perhaps from some other source 
     * @param oldname 
     * @param newname 
     * @return the dataitem that was renamed or null if it was not found
     */
    public Data renameDataItem(String oldname, String newname) {
        if (oldname.equals(defaultSequenceCollectionName)) {defaultSequenceCollectionName=newname;}
        Data dataitem=storage.renameDataItem(oldname,newname);
        if (dataitem==null) return null;
        dataitem.rename(newname);
        return dataitem;
    }
        
    /** 
     * Returns true if the specified dataitem is designated as a temporary data item
     * Currently temporary data items are those whose names begin with an underscore character
     */
    public static boolean isTemporary(Data item) {
        if (item!=null && item.getName().startsWith("_")) return true;
        else return false;
    }
    
    /** 
     * Returns true if there exists a data item with the specified name and of the specified type
     * if the type parameter is null, the function will return true if a data item with the given
     * name exists irrespective of the data item's class type.
     */
    public boolean dataExists(String name, Class type) {
        Data item=storage.getDataItem(name);
        if (type==null) return (item!=null);
        if (item==null || !(type.isInstance(item))) return false;
        return true;
    }
    
    
    
    /**
     * Removes all temporary data items currently registered with the engine
     */
    public void removeTemporaryDataItems() {
        ArrayList<Data> list=storage.getAllDataItemsOfType(Data.class);
        ArrayList<String> names=new ArrayList<String>();
        for (Data item:list) {
            if (isTemporary(item)) names.add(item.getName());
        }
        for (String name:names) {
            removeDataItem(name);
        }               
    }
    
    
    /**
     * Removes every data item from storage (except the default Sequence Collection)
     */
    public void clearAllData() {
        ArrayList<String> names=new ArrayList<String>();
        ArrayList<Data> list=storage.getAllDataItemsOfType(Data.class);
        for (Data item:list) { // First add all non-Sequence data items
            if (!(item instanceof Sequence || item instanceof Motif) && item!=getDefaultSequenceCollection()) names.add(item.getName()); 
        }
        list=storage.getAllDataItemsOfType(Motif.class);
        for (Data item:list) { 
            names.add(item.getName());
        }          
        list=storage.getAllDataItemsOfType(Sequence.class);
        for (Data item:list) { // Sequences  must be deleted last because other data objects might contain references to them
            names.add(item.getName());
        }      
        for (String name:names) {
            removeDataItem(name);
        }          
    }
    
   /**
     * Removes every data item from storage of type Motif, MotifCollection and Module
     */
    public void clearAllMotifData() {
        ArrayList<String> names=new ArrayList<String>();
        ArrayList<Data> list=storage.getAllDataItemsOfType(MotifCollection.class);
        for (Data item:list) {
            names.add(item.getName()); 
        }
        list=storage.getAllDataItemsOfType(Motif.class);
        for (Data item:list) { 
            names.add(item.getName());
        }            
        for (String name:names) {
            removeDataItem(name);
        }          
    }
    
    
    
    
    
    
    /**
     * Removes all Sequences, FeatureData Sequence Collections and Partitions from storage (except the default Sequence Collection)
     */
    public void clearAllSequences() {
        ArrayList<String> names=new ArrayList<String>();
        ArrayList<Data> list=storage.getAllDataItemsOfType(FeatureDataset.class);
        for (Data item:list) { // First add all non-Sequence data items
            names.add(item.getName()); 
        }   
        list=storage.getAllDataItemsOfType(SequenceCollection.class);
        for (Data item:list) { // Next add sequence collections (note that the default sequence collection should never be 'removed')
            if (item!=getDefaultSequenceCollection()) names.add(item.getName()); 
        }    
        list=storage.getAllDataItemsOfType(Sequence.class);
        for (Data item:list) { // Sequences  must be deleted last because other data object might contain references to them
            names.add(item.getName());
        }      
        for (String name:names) {
            removeDataItem(name);
        }          
    }
    
    /**
     * Returns a new name that can be used for data objects. 
     * The name consists of the given prefix followed by an assigned number which
     * together makes the name unique. For instance, if the engine has registered 
     * a set of motifs with names MM0001,MM0002,MM0003,MM0004,MM0005 a call to
     * getNextAvailableDataName("MM",4) will return "MM0006". The name returned 
     * is the lowest available name, so if there exists motifs MM0001,MM0002,MM0004
     * and MM0005 (but not MM0003) the method will return MM0003 as the suggested name.
     * @param prefix 
     * @param digits the number of digits to use to format the number after the prefix 
     *        (If the assigned number is shorter than the number of digits it will be preceeded by 0's)
     *        If this argument is 0 then the number of digits will vary depending on the assigned number
     * @return
     */
    public String getNextAvailableDataName(String prefix, int digits) {
        int counter=1;      
        boolean done=false;
        while (!done) {
            String name=""+counter;
            while (name.length()<digits) name="0"+name;
            name=prefix+name;
            if (!dataExists(name, null)) return name;
            else counter++;
        }
        return "x";
    }
    /** 
     * This method works the same way as getNextAvailableDataName() but returns 
     * a list of available names
     */
    public String[] getNextAvailableDataNames(String prefix, int digits, int numberofnames) {
        int counter=0;      
        String[] list=new String[numberofnames];
        for (int i=0;i<numberofnames;i++) {
           while(1<2) { // the loop will exit with a break
              String name=""+(++counter);
              while (name.length()<digits) name="0"+name;
              name=prefix+name;
              if (!dataExists(name, null)) {list[i]=name;break;}
           }
        }
        return list;
    }
    
    /**
     * Returns an ArrayList containing all Data of the specified classtype
     * currently available in storage
     * 
     * @param classtype the class of data to return (subclass of Data)
     * @return An ArrayList of Data objects
     */
    public ArrayList<Data> getAllDataItemsOfType(Class classtype) {
        return storage.getAllDataItemsOfType(classtype);
    }
    
    /**
     * Returns TRUE if there is currently registered any Data items of the specified classtype
     * 
     * @param classtype the class of data (subclass of Data)
     * @return 
     */
    public boolean hasDataItemsOfType(Class classtype) {
        return storage.hasDataItemsOfType(classtype);
    }
    
    /**
     * Returns the number of currently registered Data items of the specified classtype
     * 
     * @param classtype the class of data (subclass of Data)
     * @return 
     */
    public int countDataItemsOfType(Class classtype) {
        return storage.countDataItemsOfType(classtype);
    }
    
    /**
     * Returns the default SequenceCollection which contains all sequences registered
     */
    public SequenceCollection getDefaultSequenceCollection() {
        return (SequenceCollection)getDataItem(defaultSequenceCollectionName);
    }
    
    /**
     * Returns the name of the default SequenceCollection which contains all sequences registered
     */
    public String getDefaultSequenceCollectionName() {
        return defaultSequenceCollectionName;
    }
    
    /**
     * Returns the default DataOutput object to use for output if no other destination is specified
     */
    public OutputData getDefaultOutputObject() {
        OutputData defaultout=(OutputData)getDataItem(defaultOutputStreamName);
        if (defaultout==null) defaultout=new OutputData(defaultOutputStreamName);           
        return defaultout;
    }
    
    /**
     * Returns the name of the default DataOutput object to use for output if no other destination is specified
     */
    public String getDefaultOutputObjectName() {
        return defaultOutputStreamName;
    }
    
    
    
    
    /**
     * Returns a list of all operations known to the engine.
     * These operations are implemented and available to be used in protocol scripts.
     * 
     * @return An ArrayList of Operation objects
     */
    public ArrayList<Operation> getAllOperations() {
        ArrayList<Operation> list=new ArrayList<Operation>();
        Set<String> keys = operationslist.keySet();
        for (String key:keys) {
            list.add(operationslist.get(key));
        }
        return list;
    }
    
    /**
     * Returns an instance of the operation with the specified name
     * 
     * @param name The name (command) of the operation
     * @return an Operation instance 
     */
    public Operation getOperation(String name) {
        return operationslist.get(name);
    }
  
    
    /**
     * Returns a list of all data formats known to the engine.
     * 
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getAllDataFormats() {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            list.add(dataformatslist.get(key));
        }
        return list;
    }
    
    /**
     * Returns a list of all data format object that can format output for the given data object
     * @param data A data item used as template
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getDataOutputFormats(Data data) {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            DataFormat formatter=dataformatslist.get(key);
            if (formatter.canFormatOutput(data)) list.add(formatter);
        }
        return list;
    }
    
    /**
     * Returns a list of all data format object that can format output for data objects of the given class
     * @param data A data item used as template
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getDataOutputFormats(Class dataclass) {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            DataFormat formatter=dataformatslist.get(key);
            if (formatter.canFormatOutput(dataclass)) list.add(formatter);
        }
        return list;
    }
    
     /**
     * Returns a list of all data format object that can parse input for the given data object
     * @param data A data item used as template
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getDataInputFormats(Data data) {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            DataFormat formatter=dataformatslist.get(key);
            if (formatter.canParseInput(data)) list.add(formatter);
        }
        return list;
    }   
    
    /**
     * Returns a list of all data format object that can parse input for data objects of the given class
     * @param data A data item used as template
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getDataInputFormats(Class dataclass) {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            DataFormat formatter=dataformatslist.get(key);
            if (formatter.canParseInput(dataclass)) list.add(formatter);
        }
        return list;
    }
    
    /**
     * Returns a list of all data format objects that can parse input for some kind of Feature Data
     * @return An ArrayList of Data Format objects
     */
    public ArrayList<DataFormat> getFeatureDataInputFormats() {
        ArrayList<DataFormat> list=new ArrayList<DataFormat>();
        Set<String> keys = dataformatslist.keySet();
        for (String key:keys) {
            DataFormat formatter=dataformatslist.get(key);
            if (formatter.canParseInput(DNASequenceData.class) || formatter.canParseInput(NumericSequenceData.class) || formatter.canParseInput(RegionSequenceData.class)) list.add(formatter);
        }
        return list;
    }
    
    
    /**
     * Returns an instance of the data format with the specified name
     * 
     * @param name The name (command) of the data format
     * @return an DataFormat instance 
     */
    public DataFormat getDataFormat(String name) {
        return dataformatslist.get(name);
    }
      
    /**
     * Sets the name of the default data format for Data items of the given class
     * 
     */
    public void setDefaultDataFormat(Class dataclass, String dataformatname) {
         defaultdataformats.put(dataclass, dataformatname);
    }
    
    
    /**
     * Returns an instance of the default data format for Data items of the given type
     * 
     * @param data The data 
     * @return a DataFormat instance which is registered as default for this data type 
     */
    public DataFormat getDefaultDataFormat(Data data) {
        String name=defaultdataformats.get(data.getClass());
        return getDataFormat(name);
    }
    
    /**
     * Returns an instance of the default data format for Data items of the given type
     * 
     * @param data The data 
     * @return anDataFormat instance which is registered as default for this data type 
     */
    public DataFormat getDefaultDataFormat(Class dataclass) {
        String name=defaultdataformats.get(dataclass);
        return getDataFormat(name);
    }
      
    /**
     * Returns the DataLoader used by the system. 
     * @return
     */
    public DataLoader getDataLoader() {
        return dataLoader;
    }
    /**
     * Returns the Gene ID Resolver used by the system. 
     * @return
     */
    public GeneIDResolver getGeneIDResolver() {
        return geneIDResolver;
    }
    
    /** Returns the number of milliseconds to wait for network connection before calling timeout */
    public int getNetworkTimeout() {
        return networkTimeoutDelay;
    }
    
    /** Specifies the number of milliseconds to wait for network connection before calling timeout */
    public void setNetworkTimeout(int milliseconds) {
        networkTimeoutDelay=milliseconds;
    }
    
    /** Specifies a directory location for the data cache */
    public void setDataCacheDirectory(String cacheDirectory) {
        if (dataLoader!=null) dataLoader.setCacheDirectory(cacheDirectory);
    }
    
    /** Specifies a cache file for the gene id mapping cache */
    public void setGeneIDCacheDirectory(String cacheDirectory) {
        if (geneIDResolver!=null) geneIDResolver.setCacheDirectory(cacheDirectory);
    }
    
    
    /** Returns a list of names for available motif discovery algorithms */
    public String[] getAvailableMotifDiscoveryAlgorithms() {
        ArrayList<String> list=new ArrayList<String>();
        for (ExternalProgram program:externalPrograms.values()) {
            if (program.getProgramClass().equalsIgnoreCase("MotifDiscovery")) list.add(program.getName());
        }
        String[] result=new String[list.size()];
        result=list.toArray(result);
        Arrays.sort(result);
        return result;
    }
    
    /** Returns a list of names for available motif scanning algorithms */
    public String[] getAvailableMotifScanningAlgorithms() {
        ArrayList<String> list=new ArrayList<String>();
        for (ExternalProgram program:externalPrograms.values()) {
            if (program.getProgramClass().equalsIgnoreCase("MotifScanning")) list.add(program.getName());
        }
        String[] result=new String[list.size()];
        result=list.toArray(result);
        Arrays.sort(result);
        return result;
    }

    /** Returns a list of names for available module discovery algorithms */
    public String[] getAvailableModuleDiscoveryAlgorithms() {
        ArrayList<String> list=new ArrayList<String>();
        for (ExternalProgram program:externalPrograms.values()) {
            if (program.getProgramClass().equalsIgnoreCase("ModuleDiscovery")) list.add(program.getName());
        }
        String[] result=new String[list.size()];
        result=list.toArray(result);
        Arrays.sort(result);
        return result;
    }

    /** Returns a list of names for available module scanning algorithms */
    public String[] getAvailableModuleScanningAlgorithms() {
        ArrayList<String> list=new ArrayList<String>();
        for (ExternalProgram program:externalPrograms.values()) {
            if (program.getProgramClass().equalsIgnoreCase("ModuleScanning")) list.add(program.getName());
        }
        String[] result=new String[list.size()];
        result=list.toArray(result);
        Arrays.sort(result);
        return result;
    }


    /** Returns a list of names for available external programs that are NOT motif scanning or discovery programs */
    public String[] getOtherExternalPrograms() {
        ArrayList<String> list=new ArrayList<String>();
        for (ExternalProgram program:externalPrograms.values()) {
            String programClass=program.getProgramClass();
            if (!programClass.equalsIgnoreCase("MotifScanning") && !programClass.equalsIgnoreCase("MotifDiscovery") && !programClass.equalsIgnoreCase("ModuleScanning") && !programClass.equalsIgnoreCase("ModuleDiscovery")) list.add(program.getName());
        }
        String[] result=new String[list.size()];
        result=list.toArray(result);
        Arrays.sort(result);
        return result;
    }  
    
    /** Returns a registered external program with the given name (or null if not found) */
    public ExternalProgram getExternalProgram(String programname) {
        return externalPrograms.get(programname);
    }

    
    /** Returns a collection containing all External Programs registered with the engine */
    public Collection<ExternalProgram> getAllExternalPrograms() {
        return externalPrograms.values();
    }
      
    
    /** Returns true if the argument data object is the default sequence collection */
    public boolean isDefaultSequenceCollection(Data data) {
        return (data!=null && data==getDataItem(defaultSequenceCollectionName));
    }
    
    /** 
     * Checks if the given word is marked as 'reserved', which means that it
     * can not be used as the name of a datatrack or variable in a protocol script
     */
    public boolean isReservedWord(String word) {
        return reservedwords.contains(word);
    }
    
    /** 
     * Marks the given word as 'reserved', which means that it
     * can not be used as the name of a datatrack or variable in a protocol script
     */
    public void registerReservedWord(String word) {
        reservedwords.add(word);
    }
    
    /**
     * Returns TRUE if there are currently no Data items in storage (except the DefaultSequenceCollection)
     * @return
     */
    public boolean isDataStorageEmpty() {
        return (storage.getSize()<=1);
    }
    
    /** Given the class of a Data object (e.g. MotifCollection.class) the method will 
     *  return a String representation of the data type (e.g. "Motif Collection").
     *  If the class is not recognized as a Data type the method will return the String "unknown type"
     */
    public String getTypeNameForDataClass(Class dataclass) {
             if (dataclass==BackgroundModel.class) return BackgroundModel.getType();
        else if (dataclass==DNASequenceDataset.class) return DNASequenceData.getType();
        else if (dataclass==NumericDataset.class) return NumericDataset.getType();
        else if (dataclass==RegionDataset.class) return RegionDataset.getType();
        else if (dataclass==DNASequenceData.class) return DNASequenceData.getType();
        else if (dataclass==NumericSequenceData.class) return NumericSequenceData.getType();
        else if (dataclass==RegionSequenceData.class) return RegionSequenceData.getType();
        else if (dataclass==Motif.class) return Motif.getType();
        else if (dataclass==MotifCollection.class) return MotifCollection.getType();
        else if (dataclass==NumericConstant.class) return NumericConstant.getType();
        else if (dataclass==NumericVariable.class) return NumericVariable.getType();
        else if (dataclass==NumericMapVariable.class) return NumericMapVariable.getType();
        else if (dataclass==OutputData.class) return OutputData.getType();
        else if (dataclass==Sequence.class) return Sequence.getType();
        else if (dataclass==SequenceCollection.class) return SequenceCollection.getType();
        else if (dataclass==StringVariable.class) return StringVariable.getType();
        else if (dataclass==StringCollection.class) return StringCollection.getType();
        else return "unknown type";
    }

    /** Given a string representing a data type (e.g. "Motif Collection") the method will 
     *  return the corresponding class (e.g. MotifCollection.class) or NULL if the data type is not known
     */
    public Class getDataClassForTypeName(String typeName) {
             if (typeName==null) return null;
        else if (typeName.equalsIgnoreCase(BackgroundModel.getType())) return BackgroundModel.class;
        else if (typeName.equalsIgnoreCase(DNASequenceDataset.getType())) return DNASequenceDataset.class;
        else if (typeName.equalsIgnoreCase(NumericDataset.getType())) return NumericDataset.class;
        else if (typeName.equalsIgnoreCase(RegionDataset.getType())) return RegionDataset.class;
        else if (typeName.equalsIgnoreCase(DNASequenceData.getType())) return DNASequenceData.class;
        else if (typeName.equalsIgnoreCase(NumericSequenceData.getType())) return NumericSequenceData.class;
        else if (typeName.equalsIgnoreCase(RegionSequenceData.getType())) return RegionSequenceData.class;
        else if (typeName.equalsIgnoreCase(Motif.getType())) return Motif.class;
        else if (typeName.equalsIgnoreCase(MotifCollection.getType())) return MotifCollection.class;
        else if (typeName.equalsIgnoreCase(NumericVariable.getType())) return NumericVariable.class;
        else if (typeName.equalsIgnoreCase(NumericMapVariable.getType())) return NumericMapVariable.class;
        else if (typeName.equalsIgnoreCase(OutputData.getType())) return OutputData.class;
        else if (typeName.equalsIgnoreCase(Sequence.getType())) return Sequence.class;
        else if (typeName.equalsIgnoreCase(SequenceCollection.getType())) return SequenceCollection.class;
        else if (typeName.equalsIgnoreCase(StringVariable.getType())) return StringVariable.class;
        else if (typeName.equalsIgnoreCase(StringCollection.getType())) return StringCollection.class;
        else return null;
    }    
    
    /** 
     * Checks if the given name can be used as the name for a NEW datatrack or variable
     * The method checks whether the name is syntactically correct, if it is a reserved
     * word or if it is already in use by another data object.
     * If the name can be used the method returns NULL, otherwise it returns a string
     * containing the reason why the name can not be used
     * @param name The name to be checked (this should not contain flanking whitepace)
     * @param checkInUse If this flag is set to TRUE an error message will be reported if
     * the name is already used by another data object. If FALSE then this check will be
     * be performed
     */    
    public String checkNameValidity(String name, boolean checkInUse) {
            if (name.isEmpty()) return "The name is empty";        
       else if (name.matches(".*[^a-zA-Z_0-9].*")) return "The name contains illegal characters";        
       else if (!name.matches("^[a-zA-Z_].*")) return "The name must start with a letter";        
       else if (isReservedWord(name)) return "This word is reserved and can not be used";
       else if (checkInUse && getDataItem(name)!=null) return "The name entered is already in use";     
       else return null; // Name is OK
    }    

    /**
     * Checks if the given name can be used as the name for a sequence
     * The method checks whether the name is syntactically correct, if it is a reserved
     * word or if it is already in use by another data object.
     * If the name can be used the method returns NULL, otherwise it returns a string
     * containing the reason why the name can not be used
     * @param name The name to be checked (this should not contain flanking whitepace)
     * @param checkInUse If this flag is set to TRUE an error message will be reported if
     * the name is already used by another data object. If FALSE then this check will be
     * be performed
     */
    public String checkSequenceNameValidity(String name, boolean checkInUse) {
            if (name.isEmpty()) return "The name is empty";
       else if (name.matches(".*[^a-zA-Z_0-9].*")) return "The name contains illegal characters";
       else if (isReservedWord(name)) return "This word is reserved and can not be used";
       else if (checkInUse && getDataItem(name)!=null) return "The name entered is already in use";
       else return null; // Name is OK
    }

    // *********************************    MESSAGE LISTENING    ***************************************

        
    /**
     * Registers a new message listener with the engine
     *  
     * @param listener The object that wants to receive messages from the engine
     */
    public void addMessageListener(MessageListener listener) {
        messagelisteners.add(listener);
    }
    
     /**
     * Removes a previously registered message listener
     *  
     * @param listener The object that wants to receive messages from the engine
     */   
    public void removeMessageListener(MessageListener listener) {
        messagelisteners.remove(listener);
    }
    
    
    /**
     * Takes an incoming error message and distributes it to all registered listeners
     * @param msg The error message itself
     * @param errortype An errorcode specifying the type of error
     */
    @Override    
    public void errorMessage(String msg, int errortype) {
        for (MessageListener listener:messagelisteners) {
            listener.errorMessage(msg, errortype);
        }
    }
    
    /**
     * Takes an incoming log message and distributes it to all registered listeners
     * @param msg The log message itself
     */
    @Override    
    public void logMessage(String msg) {
        if (messagelisteners.isEmpty()) System.err.println(msg);
        for (MessageListener listener:messagelisteners) {
            listener.logMessage(msg);
        }   
    }
    

    
    /**
     * Takes an incoming status message and distributes it to all registered listeners
     * @param msg The status message itself
     */
    @Override
    public void statusMessage(String msg) {
        for (MessageListener listener:messagelisteners) {
            listener.statusMessage(msg);
        } 
    }
    
    // *********************************    DATA LISTENING    ***************************************
    
    /**
     * Registers a new data listener with the engine
     * Although interested parties can register as DataListeners directly with Data objects
     * it is preferable that they rather register as listener with the Engine. 
     * The engine will automatically track any changes to all Data Items stored (being itself
     * a registered listener with all Data items) and will propagate important messages to
     * all listeners that have registered with the Engine.
     *  
     * @param listener The object that wants to receive data event messages from the engine
     */
    public void addDataListener(DataListener listener) {
        datalisteners.add(listener);
    }
    
     /**
     * Removes a previously registered data listener
     *  
     * @param listener The object that wants to receive data event  messages from the engine
     */   
    public void removeDataListener(DataListener listener) {
        datalisteners.remove(listener);
    }
    
    @Override
    public void dataAdded(Data data) {
        for (DataListener listener:datalisteners) {
            listener.dataAdded(data);
        }
    }
    @Override
    public void dataAddedToSet(Data parent,Data child) {
        for (DataListener listener:datalisteners) {
            listener.dataAddedToSet(parent,child);
        }
    }
    @Override
    public void dataRemoved(Data data) {
        for (DataListener listener:datalisteners) {
            listener.dataRemoved(data);
        }
    }
    @Override
    public void dataRemovedFromSet(Data parent,Data child) {
        for (DataListener listener:datalisteners) {
            listener.dataRemovedFromSet(parent,child);
        }
    }
       
    /**
     * Propagates notification of Data updates to any registered DataListeners
     * Parties interested in Data updates should preferentially register as listeners with 
     * the Engine rather than registering directly with the dataitem
     * @param data
     */
    @Override
    public void dataUpdated(Data data) {
        for (DataListener listener:datalisteners) {
            listener.dataUpdated(data);
        }
    }
    
    @Override
    public void dataUpdate(Data olddata, Data newdata) {
       // we do not need to implement this since it is only invoked by the engine anyway
    }


    /**
     * Returns an object which can be used as a lock for synchronization purposes
     * @return
     */
    public Object getExecutionLock() {
       return executionLock;
    }
    
    // *********************************    ADD RESOURCES    ***************************************
    
    /** Imports known operations and registers them with the engine. Also installs standard parsers */
    private void importOperations() {
        logMessage("Importing Operations");
        for (String opname:operations) {
           //System.err.println("Installing operation: "+opname);
           try { 
              Class operationClass=Class.forName("priorseditor.engine.operations.Operation_"+opname);
              Operation operation=(Operation)operationClass.newInstance();
              operation.setEngine(this);
              registerReservedWord(opname);
              String[] words=operation.getReservedWords();
              if (words!=null) {
                  for (String word:words) registerReservedWord(word);
              }
              operationslist.put(opname,operation);
           } catch(ClassNotFoundException cnfe) {
               logMessage("WARNING: No class found for operation: "+opname);
           } catch(InstantiationException ie) {
               logMessage("WARNING: InstantiationException for operation: "+opname);
           } catch(IllegalAccessException iae) {
               logMessage("WARNING: IllegalAccessException for operation: "+opname);
           }                                   
        }        
    }
  
    
    /** Imports known data formats and registers them with the engine
     * (For now, this is hard-coded, but I guess it should not be...)
     */
    private void importDataFormats() {
        logMessage("Importing Data formats");
        
        registerDataFormat(DataFormat_Plain.class, null);
       
        registerDataFormat(DataFormat_FASTA.class, new Class[]{DNASequenceData.class,DNASequenceDataset.class});
        registerDataFormat(DataFormat_GFF.class, new Class[]{RegionSequenceData.class,RegionDataset.class});
        registerDataFormat(DataFormat_GTF.class, null);
        registerDataFormat(DataFormat_BED.class, null);
        registerDataFormat(DataFormat_WIG.class, new Class[]{NumericSequenceData.class,NumericDataset.class});
        
        registerDataFormat(DataFormat_Priority.class, null);
        registerDataFormat(DataFormat_PSP.class, null);
        registerDataFormat(DataFormat_MEME.class, null);
        registerDataFormat(DataFormat_PriorityBackground.class, null);
        registerDataFormat(DataFormat_MEME_Background.class, null);
        registerDataFormat(DataFormat_INCLUSive_Background.class, new Class[]{BackgroundModel.class});

        registerDataFormat(DataFormat_INCLUSive_Motif.class, null);
        registerDataFormat(DataFormat_RawPSSM.class, null);
        registerDataFormat(DataFormat_HTML_MotifTable.class, null);
        registerDataFormat(DataFormat_PriorsEditorMotif.class, new Class[]{MotifCollection.class,Motif.class});
      
        registerDataFormat(DataFormat_Weeder.class, null);
        registerDataFormat(DataFormat_MDscan.class, null);
        registerDataFormat(DataFormat_BioProspector.class, null);

    }

    /** Imports known external programs and registers them with the engine
     *  The list if external programgs is based on XML-files found in the 
     *  'external' subdirectory under the main workdirectory
     */
    private void importExternalPrograms() {
        logMessage("Importing External Programs");  
        File externaldir = new File(getWorkDirectory()+java.io.File.separator+"external");
        if (externaldir.exists()) {
            FilenameFilter filter = new FilenameFilter() {
                @Override
                public boolean accept(File dir, String name) {
                    return (name.endsWith(".xml") || name.endsWith(".XML"));
                }
            };
            File[] programfiles=externaldir.listFiles(filter);
            for (File programfile:programfiles) {
                try {
                    ExternalProgram program=ExternalProgram.initializeExternalProgramFromFile(programfile,false);                    
                    if (program.getServiceType().equals("bundled")) {
                        String classname=program.getLocation();
                        if (classname==null || classname.isEmpty()) throw new SystemError("Missing class specification for bundled program: "+program.getName());
                        program=(ExternalProgram)Class.forName(program.getLocation()).newInstance();
                    }
                    program.setConfigurationFile(programfile.getAbsolutePath());
                    registerExternalProgram(program);                     
                } catch (Exception e) {logMessage(e.getMessage());}
                       
            }            
        } else {
            boolean ok=externaldir.mkdirs();
            if (!ok) errorMessage("Unable to create directory for external programs", 0);
        }
        
 }
    /**
     * Registers a new DataFormat from the given class
     * If a list of data types (classes) are given, the new DataFormat will be registered as default output
     * format for these data types
     * @param dataformatClass
     * @param useasdefaultfordatatypes
     */
    public void registerDataFormat(Class dataformatClass, Class[] useasdefaultfordatatypes) {
        try {
            DataFormat newformat=(DataFormat)dataformatClass.newInstance();
            newformat.setEngine(this);        
            dataformatslist.put(newformat.getName(), newformat); 
            if (useasdefaultfordatatypes!=null) {
                for (Class dataclass:useasdefaultfordatatypes) {
                    setDefaultDataFormat(dataclass,newformat.getName());
                }
            }     
        } catch (Exception e) {
            errorMessage("Unable to register dataformat '"+dataformatClass+"' => "+e.toString(), 0);
        }
    }
    
    
    /** Registers the external program with the engine */
    public void registerExternalProgram(ExternalProgram program) {
        program.setEngine(this);
        externalPrograms.put(program.getName(),program); 
        logMessage("Importing External Program: "+program.getName());
    }    
    
    
    /** Removes the external program with the given names from the internal list
     *  The removed program is returned (but note that this is no longer installed!)
     */
    public ExternalProgram removeExternalProgram(String programName) {
        ExternalProgram p=externalPrograms.remove(programName);
        if (p!=null) p.setEngine(null);
        return p;
    }    
    
    /** Reads a configuration file (actually a serialized object) describing the
     *  predefined MotifCollections known to the system
     */
    @SuppressWarnings("unchecked")
    private void importPredefinedMotifCollections() {
        logMessage("Importing Motif Collections");
        File configfile=new File(getWorkDirectory()+File.separator+"predefinedMotifCollectionsConfig.txt");    
        if (configfile.exists()) {
            try {
               predefinedMotifCollections=loadPredefinedMotifCollectionsConfiguration();
            } catch (Exception e) {errorMessage("Unable to load predefined motif collections: "+e.toString(),0);}
        } else {
            File motifcollectiondir=new File(getWorkDirectory()+java.io.File.separator+"PE_MotifCollections");             
            if (!motifcollectiondir.exists()) {
               boolean ok=motifcollectiondir.mkdirs();
               if (!ok) errorMessage("Unable to create directory for predefined motif collections", 0);  
            }
        }        

    }
    
    /** 
     * Returns the name of the file which stores the data for the given motif collection
     * (in PriorsEditorMotif format)
     */
    public String getFilenameForMotifCollection(String motifCollectionName) {
        Object[] info=predefinedMotifCollections.get(motifCollectionName);
        if (info==null) return null;
        else return getWorkDirectory()+File.separator+"PE_MotifCollections"+File.separator+(String)info[0];
    }
     /** 
     * Returns the size of the predefined motif collection with the given name
     * (in PriorsEditorMotif format)
     */
    public int getSizeForMotifCollection(String motifCollectionName) {
        Object[] info=predefinedMotifCollections.get(motifCollectionName);
        if (info==null) return 0;
        else return (Integer)info[1];
    }
    
    /** Returns a list (or rather a Set) of the names of predefined motif collections known to the engine */
    public Set<String> getPredefinedMotifCollections() {
        return predefinedMotifCollections.keySet();               
    }
    
    /** 
     * Registers the given MotifCollection as part of the set of predefined collections 
     * This will save the collection in the working directory (replacing any previous collections
     * with the same name) and also update the 'registry' of predefined collections
     * @param collection The MotifCollection to be registered
     * @return TRUE if registration was succesful
     */
    public void registerPredefinedMotifCollection(MotifCollection collection) throws Exception {
        String name=collection.getName();
        String filename=name.replace(" ", "_"); // just in case
        filename+=".pmx";
        int collectionsize=collection.size();
        String fullpath=getWorkDirectory()+File.separator+"PE_MotifCollections"+File.separator+filename;
        DataFormat format=getDataFormat("PriorsEditorMotif");
        if (format==null) throw new SystemError("Unknown dataformat: PriorsEditorMotif");
        saveDataToFile(collection, format, fullpath);
        Object[] oldentry=predefinedMotifCollections.get(name);
        predefinedMotifCollections.put(name, new Object[]{filename,new Integer(collectionsize)});
        try {
            savePredefinedMotifCollectionsConfiguration();
        } catch (Exception e) {
           if (oldentry==null) predefinedMotifCollections.remove(name);
           else predefinedMotifCollections.put(name,oldentry);
           throw e;
        } 
     }
     /**
      * Registers a predefined MotifCollection which is read from a stream (in PriorsEditorFormat)
      * The collection is saved locally beneath the work directory
      * @param stream
      * @param filename
      * @throws java.lang.Exception
      */
     public void registerPredefinedMotifCollectionFromStream(InputStream stream, String filename) throws Exception {
        String name=filename;
        String fullpath=getWorkDirectory()+File.separator+"PE_MotifCollections"+File.separator+filename;
        File outputfile=new File(fullpath);
        BufferedReader inputStream=null;
        BufferedWriter outputStream=null;
        int collectionsize=0;
        try {
            inputStream=new BufferedReader(new InputStreamReader(stream));
            outputStream=new BufferedWriter(new FileWriter(outputfile));
            String line;
            while((line=inputStream.readLine())!=null) {
                outputStream.write(line+"\n");
                if (line.startsWith("#ID")) collectionsize++;
                else if (line.startsWith("#Collection")) {
                    String[] parts=line.split("=");
                    name=parts[parts.length-1].trim();                               
                }
            }
        } finally {
            try {
                if (inputStream!=null) inputStream.close();
                if (outputStream!=null) outputStream.close();
            } catch (IOException ioe) {}
        }        
        Object[] oldentry=predefinedMotifCollections.get(name);
        predefinedMotifCollections.put(name, new Object[]{filename,new Integer(collectionsize)});
        try {
            savePredefinedMotifCollectionsConfiguration();
        } catch (Exception e) {
           if (oldentry==null) predefinedMotifCollections.remove(name);
           else predefinedMotifCollections.put(name,oldentry);
           throw e;
        } 
    }
   
    /** Returns true if this is the first time the user starts PriorsEditor */
    public boolean isFirstTimeStartup() {
        File test=new File(getWorkDirectory()+File.separator+"predefinedMotifCollectionsConfig.txt");
        return !test.exists(); // the config-file above is created on first startup. If it does not exists yet it means this is the first time
    }     
     
    public void savePredefinedMotifCollectionsConfiguration() throws Exception {
           File configfile=new File(getWorkDirectory()+File.separator+"predefinedMotifCollectionsConfig.txt");
           BufferedWriter filewriter = new BufferedWriter(new FileWriter(configfile));
           for (String name:predefinedMotifCollections.keySet()) {
               Object[] entry=predefinedMotifCollections.get(name);
               String filename=(String)entry[0];
               int size=(Integer)entry[1];
               filewriter.write(name+"\t"+size+"\t"+filename+"\n");               
           }
           filewriter.close();
           if (!configfile.exists()) throw new ExecutionError("Unable to save config file for predefined motif collections");
    }
     
    public HashMap<String, Object[]> loadPredefinedMotifCollectionsConfiguration() throws Exception {
           File configfile=new File(getWorkDirectory()+File.separator+"predefinedMotifCollectionsConfig.txt");
           if (!configfile.exists()) return null;
           HashMap<String,Object[]> config = new HashMap<String,Object[]>();
           BufferedReader filereader = new BufferedReader(new FileReader(configfile));
           String line;
           while ((line=filereader.readLine())!=null) {
               if (line.trim().isEmpty() || line.startsWith("#")) continue;
               String[] split=line.split("\t");
               if (split.length!=3) throw new SystemError("Unrecognized line format in config file for predefined motif collections. Expected 3 fields but found "+split.length);
               String collectionName=split[0];
               String filename=split[2];
               int size=0;
               try {
                   size=Integer.parseInt(split[1]);
               } catch (NumberFormatException e) {
                  filereader.close();
                  throw new SystemError("Unable to parse expected numerical value in config file for predefined motif collections. Value = '"+split[1]+"'"); 
               }
               config.put(collectionName, new Object[]{filename,new Integer(size)});
           }
           filereader.close();
           return config;   
    }
     
     
     
     
    /** Saves the specified data object to file in the given dataformat */
    public void saveDataToFile(Data data, DataFormat dataformat, String filename) throws Exception {
        OutputData output=new OutputData("output_"+data.getName());
        if (dataformat!=null && !dataformat.canFormatOutput(data)) throw new SystemError("can not format '"+data.getName()+"' ["+data.getTypeDescription()+"] in format '"+dataformat.getName()+"'");
        dataformat.format(data, output, null, null);
        output.saveToFile(filename);        
    }
            
    /** Saves an object to file by serialization */
    public void saveSerializedObject(Object object, String filename) throws Exception {
        ObjectOutputStream stream=null;
        try {
             stream = new ObjectOutputStream(new FileOutputStream(filename));
             stream.writeObject(object);
             stream.close();       
        } catch (Exception e) {
            throw e;
        } finally {try {if (stream!=null) stream.close();} catch (Exception x){}}        
    }
    
    /** Loads a serialized object from file */
    public Object loadSerializedObject(String filename) throws Exception {
        ObjectInputStream stream=null;
        try {
             stream=new ObjectInputStream(new FileInputStream(filename));
             Object value=stream.readObject();
             stream.close();
             return value;  
        } catch (Exception e) {
            return null;
        } finally {try {if (stream!=null) stream.close();} catch (Exception x){}}     
    }
    
   /**
    * This method should be called whenever an unexpected (uncaught) exception
    * is encountered during the execution of the program. At the very least,
    * it should be called by the defaultUncaughtExceptionHandler (if exists).
    * The purpose of this method is to file a report to an appropriate bugtracking
    * system so that the issue can be dealt with by the developers and fixed in
    * future releases of PriorsEditor 
    * @param e The Error or Exception that caused the problem
    */     
   public void reportError(Throwable e) {
       logMessage("Sending bug report!");
       String message=e.getClass().getCanonicalName()+" : "+e.getMessage();
       message+="\n===================================";
       StackTraceElement[] trace=e.getStackTrace();
       for (StackTraceElement el:trace) {
           message+="\n"+el.toString();
       }
       try {
        URL url;
        HttpURLConnection urlConnection;
        DataOutputStream printout;
        BufferedReader reader;
        url = new URL ("http://tare.medisin.ntnu.no/priorseditor/reportbug.php");
        urlConnection = (HttpURLConnection)url.openConnection();
        urlConnection.setDoInput(true);
        urlConnection.setDoOutput(true);
        urlConnection.setUseCaches(false);
        urlConnection.setRequestMethod("POST");
        urlConnection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
        printout = new DataOutputStream (urlConnection.getOutputStream());
        String content = "message=" + URLEncoder.encode(message,"UTF-8");
        printout.writeBytes(content);
        printout.flush();
        printout.close();
        reader = new BufferedReader(new InputStreamReader(urlConnection.getInputStream ()));
        String line;
        while ((line = reader.readLine()) != null) {
            logMessage(line);
        }
        reader.close();
        // logMessage("Bug report successful");
     } catch (Exception ex) {
           logMessage("An error occurred while sending bug report: "+ex.getClass().getCanonicalName()+" : "+ex.getMessage());
     } // nothing to do really   
   }     
        
   
    @Override
    public void uncaughtException(Thread t, Throwable e) {
        if (client!=null && client.handleUncaughtException(e)) {
            // client fixed the problem!
        } else {
            if (client!=null) client.shutdown();
            System.exit(1);
        }
    }        
       
/** Deletes the given file or directory (and subdirectories) */
public boolean deleteTempFile(File tempdir) {
        if (tempdir.isDirectory()) {
            boolean ok=true;
            File[] files=tempdir.listFiles();
            for (File file:files) {
                if (file.isDirectory()) ok = ok && deleteTempFile(file);
                else ok = ok && file.delete();
            }
            ok = ok && tempdir.delete();
            return ok;
        } else return tempdir.delete();
}
    
/** Performs necessary cleanup (like deleting temporary directories) */
public void shutdown() {
    File tempdir=new File(getTempDirectory());
    deleteTempFile(tempdir);
}

@Override
protected void finalize() throws Throwable {
   super.finalize();
   File tempdir=new File(getTempDirectory());
   deleteTempFile(tempdir); // delete tempdir recursively
}

// ------------- just a few utilities below this line. These can be used anywere  ------------------------      
       
    /** Returns the Watson-Crick reverse complementary sequence for the given input sequence */
    public static String reverseSequence(String seq) {
        int length=seq.length();
        StringBuilder buffer=new StringBuilder(seq);
        for (int i=0;i<length;i++) {
            buffer.setCharAt(length-(i+1), reverseBase(seq.charAt(i)) );
        }
        return buffer.toString();
    }
    
    /** Returns the Watson-Crick reverse complementary sequence for the given input sequence */
    public static StringBuilder reverseSequence(StringBuilder seq) {
        if (seq==null) return null;
        int length=seq.length();
        StringBuilder buffer=new StringBuilder(seq);
        for (int i=0;i<length;i++) {
            buffer.setCharAt(length-(i+1), reverseBase(seq.charAt(i)) );
        }
        return buffer;
    }    
    
    /** Returns the Watson-Crick reverse complementary sequence for the given input sequence */
    public static char[] reverseSequence(char[] seq) {
        char[] buffer=new char[seq.length];
        for (int i=0;i<seq.length;i++) {
            switch (seq[i]) {
                case 'A':buffer[buffer.length-(i+1)]='T';break;
                case 'a':buffer[buffer.length-(i+1)]='t';break;
                case 'C':buffer[buffer.length-(i+1)]='G';break;
                case 'c':buffer[buffer.length-(i+1)]='g';break;
                case 'G':buffer[buffer.length-(i+1)]='C';break;
                case 'g':buffer[buffer.length-(i+1)]='c';break;
                case 'T':buffer[buffer.length-(i+1)]='A';break;
                case 't':buffer[buffer.length-(i+1)]='a';break;
                default:buffer[buffer.length-(i+1)]=buffer[i];break;
            }
        }
        return buffer;
    }      
    
    
   private static char reverseBase(char c) {
      switch (c) {
          case 'A':return 'T';
          case 'C':return 'G';
          case 'G':return 'C';
          case 'T':return 'A';      
          case 'a':return 't';
          case 'c':return 'g';
          case 'g':return 'c';
          case 't':return 'a';      
      } 
      return c;  
   }   

    /** Takes a list of Strings and returns a new string which is a concatination of
     *  the elements in the original list separated by by a defined separator string
     *  @param list List of strings
     *  @param separator A string which will be inserted between each pair of string in the list
     */
    public static String splice(String[] list, String separator) {
        StringBuilder string=new StringBuilder();
        for (int i=0;i<list.length;i++) {
            string.append(list[i]);
            if (i<list.length-1) string.append(separator);
        }
        return string.toString();
    }

    /** Takes a list of Strings and returns a new string which is a concatination of
     *  the elements in the original list separated by by a defined separator string
     *  @param list List of strings
     *  @param separator A string which will be inserted between each pair of string in the list
     */
    public static String splice(List<String> list, String separator) {
        StringBuilder string=new StringBuilder();
        int size=list.size();
        for (int i=0;i<size;i++) {
            string.append(list.get(i));
            if (i<size-1) string.append(separator);
        }
        return string.toString();
    }


}
