/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.operations.Operation_new;
import priorseditor.engine.protocol.ParseError;

/**
 *
 * @author kjetikl
 */
public class BackgroundModel extends Data {
    private static char[] bases={'A','C','G','T'};
    private String name="";
    private int order=0;
    private static String typedescription="Background Model";
    private HashMap<String,Float> transitionMatrix; // oligo => probability (e.g. "AGT" => 0.07 or "A" => 0.25) 
    // NOTE: the transitionMatrix is stored "flat" and the key is composed of "prefix+next base". Hence the number of bases in the key is the model order+1 !!
    private float[] oligoFrequencies=null; //  
    private String predefined=null; // name of predefined model
    private int organism=0; // 
    private float[] snf=new float[]{0.25f,0.25f,0.25f,0.25f}; // single nucleotide frequencies
    
    /**
     * Creates a default Markov Model of 0th order with equal base probabilities
     * @param name
     */
    public BackgroundModel(String name) {
        this.name=name;       
        transitionMatrix=createTransitionMatrix(new float[]{0.25f,0.25f,0.25f,0.25f});
        oligoFrequencies=new float[]{0.25f,0.25f,0.25f,0.25f};
        order=0;
    }
    
    /** 
     * Creates a new Markov model based on the given list of values
     * The number of values in the matrixvalues list should be a power of 4 and the order
     * of the model is determined from the number of values (4=0th order, 16=1st order, 64=2nd order etc.)
     * If matrixvalues is NULL the order is determined from the oligovalues parameter
     * If both these are NULL a model of 
     * This constructor is used by the NEW operation. If any of the parameters
     * @param snf an array of 4 values containing the Single Nucleotide Frequencies for A,C,G and T (0th order model)
     * @param oligovalues an array containing the oligo-frequencies
     * @param matrixvalues an array containing the probabilities for the transition matrix (as 1-dimensional flat format)
     */
    public BackgroundModel(String name, float[] snf, float[] oligovalues, float[] matrixvalues) {
        this.name=name;   
        order=-1;
        if (snf!=null) this.snf=snf;
        else {
            if (oligovalues!=null && oligovalues.length==4) this.snf=new float[]{oligovalues[0],oligovalues[1],oligovalues[2],oligovalues[3]};
            else this.snf=new float[]{0.25f,0.25f,0.25f,0.25f};
        }

        if (matrixvalues!=null) {
          transitionMatrix=createTransitionMatrix(matrixvalues);
          order=(int)(Math.log(transitionMatrix.size())/Math.log(4))-1;           
        }        
        if (oligovalues!=null) {
            this.oligoFrequencies=oligovalues;
            if (order<0) order=(oligoFrequencies.length==4)?0:(int)(Math.log(oligoFrequencies.length)/Math.log(4));           
        }
        if (order<0) order=0;
        if (oligovalues==null) {
            if (order==0) {
                this.oligoFrequencies=new float[]{snf[0],snf[1],snf[2],snf[3]};
            } else {
                this.oligoFrequencies=new float[(int)Math.pow(4, order)];
                float val=1.0f/oligoFrequencies.length;
                for (int i=0;i<oligoFrequencies.length;i++) oligoFrequencies[i]=val;
            }

        }
        if (matrixvalues==null) {
            matrixvalues=new float[(int)Math.pow(4,order+1)];
            float val=0.25f;
            for (int i=0;i<matrixvalues.length;i++) matrixvalues[i]=val;
            transitionMatrix=createTransitionMatrix(matrixvalues);
        }

    }
    
    /* 
     * Creates a new Markov model based on the given list of values
     * The first 4 values in the list should specify the Single Nucleotide Frequencies 
     * for A,C,G and T respectively. The remaining values describes the transition
     * matrix and the number of these should be a power of 2 (The order of the model is 
     * determined from the number of values). For 0th order models this matrix
     * description can be omitted.
     * (thus, 0 or 4 values=>0th order, 16 values=>1st order, 64 values=>2nd order etc.)
     * @param values an array containing the 4 SNF probabilitites followed by the 4**n probabilities for the transition matrix (in 1-dimensional flat format)
     *
    public BackgroundModel(String name, float[] values) {
        this.name=name;       
        float[] transitionMatrixValues=null;
        if (values.length==4) {
            snf=new float[]{values[0],values[1],values[2],values[3]};
            transitionMatrixValues=new float[]{values[0],values[1],values[2],values[3]};
            
        } else if (values.length>4) {
            snf=new float[]{values[0],values[1],values[2],values[3]};
            int matrixSize=values.length-4;
            transitionMatrixValues=new float[matrixSize];
            for (int i=4;i<values.length;i++) {
                transitionMatrixValues[i-4]=values[i];
            }
        }         
        transitionMatrix=createTransitionMatrix(transitionMatrixValues);
        order=(int)(Math.log(transitionMatrix.size())/Math.log(4))-1;
    }
    */
    
    
    /** Creates a new model based on the given internal model and Single Nucleotide Frequencies */
    public BackgroundModel(String name, float[] snf, float[] oligofrequencies, HashMap<String,Float> transitionMatrix){
        this.name=name;
        this.transitionMatrix=transitionMatrix;
        this.snf=snf;
        this.oligoFrequencies=oligofrequencies;
        order=(int)(Math.log(transitionMatrix.size())/Math.log(4))-1;
    }
    
      
    /** Creates a transitionMatrix for this Markov Model based on the list of input values. 
     * The values should be first 'ordered by prefix' and then by suffix.
     * Eg. like so: AAA, AAC, AAG, AAT, ACA, ACC, ACG, ACT, AGA, AGC, AGG, AGT, ATA, ATC, ATG, ATT, etc...
     */
    private HashMap<String,Float> createTransitionMatrix(float[] values) {
        int modelOrder=(int)(Math.log(values.length)/Math.log(4))-1;
        //System.err.println("Markov Order="+modelOrder+"    values="+values.length);
        HashMap<String, Float> newmodel=new HashMap<String, Float>();
        for (int i=0;i<values.length;i++) {
            String seq=integerToPattern(i,modelOrder+1);
            newmodel.put(seq, values[i]);
            //System.err.println(seq+"=>"+values[i]);
        }
        return newmodel;
    }
    
    
    
    public static String integerToPattern(int index, int length) {
        String seq="";
        if (length==0) return seq;
        for (int i=1;i<=length;i++) {
            seq=bases[index%4]+seq;
            index=index/4;
        }
        return seq;
    }
    
    public static int patternToInteger(CharSequence pattern) {
        int size=pattern.length();
        int result=0;
        for (int i=0;i<size;i++) {
            char c=pattern.charAt(size-i-1);
            int index=0;
            switch (c) {
               case 'A': index=0; break;
               case 'C': index=1; break;
               case 'G': index=2; break;
               case 'T': index=3; break;
            }
            result+=index*(int)Math.pow(4, i);
        }
        return result;
    }
    
    
    
    
    @Override
    public String getName() {return name;}

    @Override
    public void rename(String newname) {
        this.name=newname;
    }     
    
    
    public int getOrganism() {return organism;}
    
    public void setOrganism(int neworganism) {this.organism=neworganism;}
    
    
    @Override
    public Object getValue() {return this;}
    
    @Override
    public String getValueAsParameterString() { // Returns "fields" separated by semicolons. Each field starts with either SNF:, OLIGOS: or MATRIX: and is followed by a list of comma-separated float-values
        if (predefined!=null) return Operation_new.MODEL_PREFIX+predefined;
        else {
            String parameterstring="SNF:"+snf[0]+","+snf[1]+","+snf[2]+","+snf[3];
            parameterstring+=";OLIGOS:"+getOligoFrequenciesAsString();
            parameterstring+=";MATRIX:"+getTransitionMatrixAsString();
            if (organism!=0) parameterstring+=";ORGANISM:"+organism;
            return parameterstring;
        }
    }
    
    public int getOrder() {return order;}
    
    public void setOrder(int neworder) {this.order=neworder;}
    
    public void setValue(float[] newSNF, float[] newOligoFrequencies, float[] newMatrix) {     
        this.snf=newSNF;
        this.oligoFrequencies=newOligoFrequencies;
        this.transitionMatrix=createTransitionMatrix(newMatrix);
        this.order=(int)(Math.log(transitionMatrix.size())/Math.log(4))-1;
        predefined=null;
        notifyListenersOfDataUpdate();
    }
    
    public void setPredefinedModel(String model) {
        this.predefined=model;
    }
    
    public String getPredefinedModel() {
        return predefined;
    }
    
    /**
     * Given a DNA sequence as input, this function returns a base which according to 
     * the model represents is a suitable continuation of the string provided.
     * The base to return is selected at random according to the model  
     * @param prefix
     * @return
     */
    public char getNextBase(String prefix) throws ExecutionError {
        prefix=preprocessPrefix(prefix);
        if (prefix.length()>order) prefix=prefix.substring(prefix.length()-order, prefix.length()-1);
        else if (prefix.length()<order) prefix=findSuitablePrefix(prefix);
        Float probA=transitionMatrix.get(prefix+"A"); if (probA==null) throw new ExecutionError("Background model '"+name+"' is missing value for prefix="+prefix+"A");
        Float probC=transitionMatrix.get(prefix+"C"); if (probC==null) throw new ExecutionError("Background model '"+name+"' is missing value for prefix="+prefix+"C");
        Float probG=transitionMatrix.get(prefix+"G"); if (probG==null) throw new ExecutionError("Background model '"+name+"' is missing value for prefix="+prefix+"G");
        Float probT=transitionMatrix.get(prefix+"T"); if (probT==null) throw new ExecutionError("Background model '"+name+"' is missing value for prefix="+prefix+"T");
        float cumA=probA.floatValue();
        float cumC=cumA+probC.floatValue();
        float cumG=cumC+probG.floatValue();
        float selected=(float)Math.random();
             if (selected<cumA) return 'A';
        else if (selected<cumC) return 'C';
        else if (selected<cumG) return 'G';
        else return 'T';

    }
    
    /**
     * Given a DNA sequence as input, this function returns a list of base which 
     * according to the model represents is a suitable continuation of the string provided.
     * The list of bases is selected at random according to the model
     * @param prefix A string of DNA letters or null
     * @return
     */
    public char[] getNextBases(String prefix, int length) throws ExecutionError {
        if (prefix.length()>order) prefix=prefix.substring(prefix.length()-order, prefix.length()-1);
        else if (prefix.length()<order) prefix=findSuitablePrefix(prefix);
        char[] buffer=new char[length];
        for (int i=0;i<buffer.length;i++) {
            if (i>=order) prefix=new String(buffer, i-order, order);
            else if (i<order && i>0) { // prefix is made up from part original prefix and part buffer
               prefix=prefix.substring(1); // trim off first letter
               prefix=prefix+buffer[i-1];
            }
            buffer[i]=getNextBase(prefix);
        }
        return buffer;
    }
    
    /**
     * Selects a single base to return based on the single-nucleotide frequency distribution
     * of bases (i.e. a 0th-order model).
     * @return
     */
    public char getBaseFromSNF() {
        float cumA=snf[0];
        float cumC=cumA+snf[1];
        float cumG=cumC+snf[2];
        float selected=(float)Math.random();
             if (selected<cumA) return 'A';
        else if (selected<cumC) return 'C';
        else if (selected<cumG) return 'G';
        else return 'T';        
    }
    
    /**
     * Returns the Single Nucleotide Frequency for the given base according to this model
     * @param base
     * @return
     */
    public float getSNF(char base) {
         switch(base) {       
          case 'A': case 'a': return snf[0];
          case 'C': case 'c':  return snf[1];
          case 'G': case 'g':  return snf[2];
          case 'T': case 't':  return snf[3];
          default: return 0;
        }
    }
    
   /**
     * Returns the Single Nucleotide Frequency all the bases as a float[] (indexes correspond to A,C,G,T)
     * @param base
     * @return a copy of the SNF frequencies for this model
     */
    public float[] getSNF() {
         return new float[]{snf[0],snf[1],snf[2],snf[3]};
    }
    
    
    
    /**
     * Returns the Single Nucleotide Frequency for the given base according to this model
     * @param base
     * @return
     */
    public float getOligoFrequency(String oligo) {
         oligo=oligo.toUpperCase();         
         int index=patternToInteger(oligo);
         if (oligoFrequencies==null || index>=oligoFrequencies.length) return 0; 
         else return oligoFrequencies[index];
    }
    
    
    
    /**
     * Returns the transitionMatrix for this model as a 2-dimensional Float array
     * The first dimension has length 4^order and corresponds to the prefixes while
     * the second dimensions has length 4 corresponding to the 4 bases A,C,G and T
     * @return
     */
    public Float[][] getTransitionMatrix() {
        int size=(int)Math.pow(4, order);
        Float[][] result=new Float[size][4];
        for (int i=0;i<size;i++) {
            String prefix=integerToPattern(i, order);
            result[i][0]=transitionMatrix.get(prefix+"A");
            result[i][1]=transitionMatrix.get(prefix+"C");
            result[i][2]=transitionMatrix.get(prefix+"G");
            result[i][3]=transitionMatrix.get(prefix+"T");
        }
        return result;
    }
    
    /**
     * Returns the transitionMatrix for this model as a large string of comma-separated numbers
     * @return
     */
    public String getTransitionMatrixAsString() {
        String result="";
        int size=(int)Math.pow(4, order);
        for (int i=0;i<size;i++) {
            String prefix=integerToPattern(i, order);
            result+=transitionMatrix.get(prefix+"A")+",";
            result+=transitionMatrix.get(prefix+"C")+",";
            result+=transitionMatrix.get(prefix+"G")+",";
            result+=transitionMatrix.get(prefix+"T")+",";
        }
        result=result.substring(0,result.length()-1);
        return result;
    }
    
    /**
     * Returns the oligoFrequencies for this model as a large string of comma-separated numbers
     * @return
     */
    public float[] getOligoFrequencies() {
        return oligoFrequencies;
    }
    
    /**
     * Returns the oligoFrequencies for this model as a large string of comma-separated numbers
     * @return
     */
    public String getOligoFrequenciesAsString() {
        String result="";
        if (oligoFrequencies==null) return "";
        for (int i=0;i<oligoFrequencies.length-1;i++) {
            result+=oligoFrequencies[i]+",";
        }
        result+=oligoFrequencies[oligoFrequencies.length-1];
        return result;
    }
    
    
    
    /**
     * If the prefix provided is shorter than the model,
     * this method can be used to find a suitable prefix of the correct length 
     */ 
    private String findSuitablePrefix(String prefix) {
        int basesMissing=order-prefix.length();
        if (basesMissing<=0) return prefix;
        // for now we just add random bases to the front of the prefix until it is long enough
        // is there a more "scientific" way to determine a suitable prefix?
        while (prefix.length()<order) {
            prefix=getBaseFromSNF()+prefix;
        }
        return prefix;
    }
    
    
    /** Converts prefix to all uppercase and replaces unknown letters with random bases */
    private String preprocessPrefix(String prefix) {

        prefix=prefix.toUpperCase();
        for (int i=0;i<prefix.length();i++) {
            char atI=prefix.charAt(i);
            if (atI!='A' && atI!='C' && atI!='G' &&atI!='T') {
                prefix.replaceFirst(""+atI,""+getBaseFromSNF());
            }
        }
        return prefix;
    }
    
    @Override
    public void importData(Data source) throws ClassCastException {
        BackgroundModel datasource=(BackgroundModel)source;
        this.name=datasource.name;    
        this.order=datasource.order;
        this.predefined=datasource.predefined;
        this.organism=datasource.organism;
        this.snf=datasource.snf;
        this.oligoFrequencies=datasource.oligoFrequencies;
        this.transitionMatrix=datasource.transitionMatrix;
        //notifyListenersOfDataUpdate(); 
    }
    

    @Override
    public BackgroundModel clone() {
        HashMap<String,Float> newinternalmodel=new HashMap<String,Float>(transitionMatrix.size());
        Set<String> keySet=transitionMatrix.keySet();
        for (String key:keySet) {
            newinternalmodel.put(key,new Float(transitionMatrix.get(key).floatValue()));
        }
        float[] newsnf=new float[snf.length];
        for (int i=0;i<snf.length;i++) newsnf[i]=snf[i];
        float[] newoligos=new float[oligoFrequencies.length];
        for (int i=0;i<oligoFrequencies.length;i++) newoligos[i]=oligoFrequencies[i];
        
        BackgroundModel newmodel=new BackgroundModel(name,newsnf,newoligos,newinternalmodel);
        newmodel.predefined=predefined;
        newmodel.organism=organism;
        return newmodel;
    }
    
    @Override
    public String getTypeDescription() {
        String suffix=null;
        if (order==1) suffix="st";
        else if (order==2) suffix="nd";
        else if (order==3) suffix="rd";
        else suffix="th";
        if (predefined!=null) return typedescription+" : \""+predefined+"\"  "+order+suffix+ " order";
        else return typedescription+" : "+order+suffix+" order";
    }


    public static String getType() {
        return typedescription;
    }    
    
    /** 
     * Returns true if this model has the same Single Nucleotide Frequencies,
     * the same oligo frequency values and the same transition matrix as the given Markov Model
     */
    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof BackgroundModel)) return false;
        BackgroundModel model=(BackgroundModel)other;
        if (getOrder()!=model.getOrder()) return false;
        boolean sameSFN=(snf[0]==model.snf[0] && snf[1]==model.snf[1] && snf[2]==model.snf[2] && snf[3]==model.snf[3]);
        if (!sameSFN) return false;
        if ((this.oligoFrequencies==null && model.oligoFrequencies!=null) || (this.oligoFrequencies!=null && model.oligoFrequencies==null) || (this.oligoFrequencies.length!=model.oligoFrequencies.length)) return false;
        for (int i=0;i<oligoFrequencies.length;i++) {
            if (this.oligoFrequencies[i]!=model.oligoFrequencies[i]) return false;
        }
        Set<String> keySet=transitionMatrix.keySet();
        for (String key:keySet) {
            if (transitionMatrix.get(key).floatValue()!=model.transitionMatrix.get(key).floatValue()) return false;
        }
        return true;
    }
    
    @Override
    public String output() {
        double[] frequencies=convertBackgroundToPlainFormat(this);
        String result="";
        for (int i=0;i<frequencies.length;i++) {
            result+=""+frequencies[i]+"\n";
        }
        return result;
    }
    
    @Override
    public void inputFromPlain(ArrayList<String> input, PriorsEditorEngine engine) throws ParseError {
        Iterator iter=input.iterator();
        while (iter.hasNext()) {
            String next=(String)iter.next();
            if (next.trim().isEmpty()) iter.remove();
        }
        float[] values=new float[input.size()];
        for (int i=0;i<input.size();i++) {
            String string=input.get(i);
            try {
                values[i]=Float.parseFloat(string);
            } catch (NumberFormatException nfe) {throw new ParseError("Unable to parse expected numerical value: "+string);}
        }
        String thisname=getName();
        BackgroundModel model=convertPlainFormatToBackground(values);
        importData(model);
        name=thisname;
    }
    
    /** Converts a background in PriorsEditor format to plain (PRIORITY) format */ 
    public static double[] convertBackgroundToPlainFormat(BackgroundModel background) {
        int order=background.getOrder();
        HashMap<String,float[]> conditionalProbabilities=new HashMap<String,float[]>(); // the string is the suffix and the Float contains probabilities for [A,C,G,T] in order
        int arraySize=0;
        for (int k=1;k<=order+1;k++) arraySize+=(int)Math.pow(4,k);
        double[] result=new double[arraySize];
        //System.err.println("ArraySize="+arraySize);
        float[] oligoFrequencies=background.getOligoFrequencies();
        HashMap<String,Float> temp=new HashMap<String,Float>();
        for (int i=0;i<oligoFrequencies.length;i++) {
            String oligo=BackgroundModel.integerToPattern(i, order);
            temp.put(oligo,oligoFrequencies[i]); // initialize temp-map with stored oligo frequencies
            //System.err.println("Temp["+oligo+"] => "+oligoFrequencies[i]);
        }
        //double sum=0;
        if (order>0) { // process the "raw" oligo frequencies with length==order
            for (int i=0;i<Math.pow(4,order);i++) {
                String oligo=BackgroundModel.integerToPattern(i, order); // the smaller oligos. Size is minimum 1
                String prefix=(oligo.length()>1)?oligo.substring(0,oligo.length()-1):"no_prefix";
                char nextBase=oligo.charAt(oligo.length()-1);
                float value=temp.get(oligo); 
                if (!conditionalProbabilities.containsKey(prefix)) conditionalProbabilities.put(prefix, new float[]{0,0,0,0});
                int index=0;
                if (nextBase=='C') index=1; else if (nextBase=='G') index=2; else if (nextBase=='T') index=3;
                conditionalProbabilities.get(prefix)[index]=value;                    
                //System.err.println("   "+nextBase+"|"+prefix+" => "+value);
                //sum+=value;
            }        
            //System.err.println("   Sum="+sum);
        }
        // now calcuate frequencies for all oligos of smaller lengths
        for (int size=order;size>1;size--) { // 
            //System.err.println("Size="+size);
            for (int i=0;i<Math.pow(4,size);i++) {
                String oligo=BackgroundModel.integerToPattern(i, size);
                // an oligo can only contain 2 words that are 1 bp smaller
                String first =oligo.substring(0,oligo.length()-1);
                String second=oligo.substring(1);
                float oligofrequency=temp.get(oligo);
                if (!temp.containsKey(first)) temp.put(first,0f);
                if (!temp.containsKey(second)) temp.put(second,0f);  
                temp.put(first,  temp.get(first) +oligofrequency);
                temp.put(second, temp.get(second)+oligofrequency); 
                //System.err.println("   suboligos: ["+first+","+second+"] => "+oligofrequency+" each");
            }
            //sum=0;
            for (int i=0;i<Math.pow(4,size-1);i++) {
                String oligo=BackgroundModel.integerToPattern(i, size-1); // the smaller oligos. Size is minimum 1
                String prefix=(oligo.length()>1)?oligo.substring(0,oligo.length()-1):"no_prefix";
                char nextBase=oligo.charAt(oligo.length()-1);
                float value=temp.get(oligo)/2f; // divide by 2 since each suboligo is counted twice (once as the "second" word in and once as the "first" word)
                temp.put(oligo, value);
                if (!conditionalProbabilities.containsKey(prefix)) conditionalProbabilities.put(prefix, new float[]{0,0,0,0});
                int index=0;
                if (nextBase=='C') index=1; else if (nextBase=='G') index=2; else if (nextBase=='T') index=3;
                conditionalProbabilities.get(prefix)[index]=value;                    
                // System.err.println("   "+nextBase+"|"+prefix+" => "+value);
                //sum+=value;
            }        
            //System.err.println("   Sum="+sum);
            
        }
        // normalize and fill in result array
        result[0]=background.getSNF('A');
        result[1]=background.getSNF('C');
        result[2]=background.getSNF('G');
        result[3]=background.getSNF('T');
        int index=0;
        for (int size=1;size<order;size++) {
            index+=Math.pow(4,size);
            for (int i=0;i<Math.pow(4,size);i++) {
                String prefix=BackgroundModel.integerToPattern(i, size); 
                float[] values=conditionalProbabilities.get(prefix);
                if (values==null) values=new float[]{0,0,0,0};
                float total=values[0]+values[1]+values[2]+values[3];
                for (int j=0;j<4;j++) {
                    //System.err.println("["+(index+(i*4)+j)+"] => "+(values[j]/total));
                    result[index+(i*4)+j]=(values[j]/total);
                }
            }
            
        }
        // finally fill in the transitions from the matrix 
        if (order>0) {
            index+=Math.pow(4,order);
            Float[][] matrix=background.getTransitionMatrix();
            for (int i=0;i<Math.pow(4,order);i++) {
                for (int j=0;j<4;j++) {
                    result[index+(i*4)+j]=matrix[i][j];
                    //System.err.println("["+(index+(i*4)+j)+"] => "+matrix[i][j]);
                }
            }              
        }
        return result;
    }    
    
    
      /** Converts a background in PriorsEditor format to oligo-frequency format (used by e.g. MEME)
       * For a background model of order k, the resulting array contains oligo-frequencies for all 
       * word lengths from 1 up to k+1. Thus a zero-order model contains frequencies for "a,c,g,t" while
       * a first-order model contains frequencies for "a,c,g,t" followed by "aa,ac,ag,at,ca,cc,cg,ct,ga,gc,gg,gt,ta,tc,tg,tt"
       * the sum of frequencies for each word-length must sum to 1.
       */ 
    public static double[] convertBackgroundToFrequencyFormat(BackgroundModel background) {
        int order=background.getOrder();
        int arraySize=0;
        for (int k=1;k<=order+1;k++) arraySize+=(int)Math.pow(4,k);
        double[] result=new double[arraySize];
        //System.err.println("ArraySize="+arraySize);
        float[] oligoFrequencies=background.getOligoFrequencies();
        HashMap<String,Float> temp=new HashMap<String,Float>();
        for (int i=0;i<oligoFrequencies.length;i++) {
            String oligo=BackgroundModel.integerToPattern(i, order);
            temp.put(oligo,oligoFrequencies[i]); // initialize temp-map with stored oligo frequencies
            //System.err.println("Temp["+oligo+"] => "+oligoFrequencies[i]);
        }
        // calculate frequencies for all oligos of smaller lengths
        for (int size=order;size>1;size--) { // process oligo size from max to 1.
            // Frequencies for the largest oligo size is already in the "temp" table. Now process smaller sizes based on frequencies of longer words
            for (int i=0;i<Math.pow(4,size);i++) {
                String oligo=BackgroundModel.integerToPattern(i, size);
                // an oligo can only contain 2 words that are 1 bp smaller
                String first =oligo.substring(0,oligo.length()-1);
                String second=oligo.substring(1);
                float oligofrequency=temp.get(oligo); // find frequency for this oligo and add to both oligos 1 bp smaller
                oligofrequency=oligofrequency/2.0f;
                if (!temp.containsKey(first)) temp.put(first,0f);
                if (!temp.containsKey(second)) temp.put(second,0f);  
                temp.put(first,  temp.get(first) +oligofrequency); 
                temp.put(second, temp.get(second)+oligofrequency); 
                //System.err.println("   suboligos: ["+first+","+second+"] => "+oligofrequency+" each");
            }
        }
        int index=0;
        for (int size=1;size<=order;size++) {      
            for (int i=0;i<Math.pow(4,size);i++) {
                String oligo=BackgroundModel.integerToPattern(i, size); 
                float value=temp.get(oligo);
                //System.err.println("Value["+oligo+"] index="+(index)+", i="+(index+i)+" => "+value);
                result[index+i]=value;                
            }     
            index+=Math.pow(4,size);   
        }
        result[0]=background.getSNF('A');
        result[1]=background.getSNF('C');
        result[2]=background.getSNF('G');
        result[3]=background.getSNF('T');        
        // finally fill in the transitions from the matrix 
        if (order>0) {
            Float[][] matrix=background.getTransitionMatrix();
            index=result.length-(matrix.length*4);
            for (int i=0;i<Math.pow(4,order);i++) { // for each oligo prefix
                String oligo=BackgroundModel.integerToPattern(i, order); 
                float oligofrequency=temp.get(oligo);
                for (int j=0;j<4;j++) {
                    result[index+(i*4)+j]=oligofrequency*matrix[i][j];
                    //System.err.println("Index["+(index+(i*4)+j)+"] oligo="+oligo+"   freq="+oligofrequency+"     i,j="+i+","+j+" => "+matrix[i][j]);
                }
            }              
        }
        return result;
    }    
      
    
    
    
    
    /** Converts a background in plain (PRIORITY) format to PriorsEditor format 
     * The float values represent transition probabilities at different orders (oligo lengths)
     * each consecutive four values should sum to 1 and represent a transition probabilities from a prefix (which could be empty)
     * to the bases a,c,g and t. Thus, the first 4 values are transition from "null" to a,c,g,t the next 16 are transitions from "a","c","g" and "t" to another base (4 of each)
     * then follows transitions from "aa" and other oligos of length 2 etc.
     */ 
    public static BackgroundModel convertPlainFormatToBackground(float[] values) throws ParseError {
        int order=-1;
        int expectedsize=0;
        while (expectedsize<values.length) {
           order++; // try higher orders
           expectedsize+=Math.pow(4,order+1);
        }
        if (expectedsize>values.length || values.length<4) throw new ParseError("The model does not contain a valid number of probabilities (found "+values.length+")");
        int matrixsize=(int)Math.pow(4,order+1);
        float[] snf=Arrays.copyOfRange(values, 0, 4); // first 4 values represent single nucleotide frequencies. Note that last coordinate in range is exclusive
        float[] oligofrequencies=Arrays.copyOfRange(values, 0, 4); 
        float[] transitionMatrix=Arrays.copyOfRange(values, values.length-matrixsize, values.length);
        if (order>=2) { // need to calculate oligofrequencies from conditional probabilities
            HashMap<String,Float> map=new HashMap<String,Float>(values.length-matrixsize);
            map.put("A",snf[0]);
            map.put("C",snf[1]);
            map.put("G",snf[2]);
            map.put("T",snf[3]);
            int index=0;
            int currentorder=2;
            while (currentorder<=order) {
                index=index+(int)Math.pow(4,currentorder-1);
                int blocksize=(int)Math.pow(4,currentorder);
                for (int i=0;i<blocksize;i++) {
                    String pattern=integerToPattern(i,currentorder);
                    float conditionalProb=values[index+i];
                    String prefix=pattern.substring(0,pattern.length()-1);
                    //System.err.println("Pattern="+pattern+",  prefix="+prefix);
                    float prefixProb=map.get(prefix);
                    map.put(pattern, conditionalProb*prefixProb); // The pattern represents the prefix oligo followed by the next base, e.g. for P(A|GTC) the pattern will be GTCA                    
                }
                currentorder++;
            }
            int numberofoligos=(int)Math.pow(4,order);
            oligofrequencies=new float[numberofoligos];
            //float totalcheck=0;
            for (int i=0;i<numberofoligos;i++) {
                String pattern=integerToPattern(i,order);
                oligofrequencies[i]=map.get(pattern);
                //totalcheck+=oligofrequencies[i];
            }
            //System.err.println("total check="+totalcheck);
        }
        BackgroundModel bgmodel=new BackgroundModel("temporary", snf, oligofrequencies, transitionMatrix);
        return bgmodel;
    }     
    
    /** Converts an oligo frequency table to PriorsEditor format
     * For models of order k the values-table should contain oligo frequencies for all word-lengths
     * from 1 to k+1. Thus, a second-order background model should start with the oligo frequencies for a,c,g and t
     * then for aa,ac,ag,at,ca,cc,cg,ct,ga,gc,gg,gt,ta,tc,tg,tt and last aaa,aac,aag,aat,aca,acc.. etc.
     * The probabilities for each block (oligos of same length) should sum to 1
     */ 
    public static BackgroundModel convertFrequencyFormatToBackground(float[] values) throws ParseError {
        int order=-1;
        int expectedsize=0;
        while (expectedsize<values.length) {
           order++; // try higher orders
           expectedsize+=Math.pow(4,order+1);
        }
        if (expectedsize>values.length || values.length<4) throw new ParseError("The model does not contain a valid number of probability values (found "+values.length+")");
        int matrixsize=(int)Math.pow(4,order+1);
        float[] snf=Arrays.copyOfRange(values, 0, 4); // first 4 values represent single nucleotide frequencies. Note that last coordinate in range is exclusive
        float[] oligofrequencies=null;
        if (order>=2) {
            int oligotablesize=(int)Math.pow(4,order);
            oligofrequencies=Arrays.copyOfRange(values, values.length-(matrixsize+oligotablesize), values.length-matrixsize); 
        } else {
            oligofrequencies=Arrays.copyOfRange(values, 0, 4); // for order 0 and 1, these are the same as the SNF values
        }        
        float[] transitionMatrix=Arrays.copyOfRange(values, values.length-matrixsize, values.length);
        // now renormalize the transitionMatrix probabilities. Each row should sum to 1
        for (int i=0;i<transitionMatrix.length;i+=4) {
            float total=transitionMatrix[i]+transitionMatrix[i+1]+transitionMatrix[i+2]+transitionMatrix[i+3];
            transitionMatrix[i]=transitionMatrix[i]/total;
            transitionMatrix[i+1]=transitionMatrix[i+1]/total;
            transitionMatrix[i+2]=transitionMatrix[i+2]/total;
            transitionMatrix[i+3]=transitionMatrix[i+3]/total;
        }
        BackgroundModel bgmodel=new BackgroundModel("temporary", snf, oligofrequencies, transitionMatrix);
        return bgmodel;
    }     
    
public static Object[][] predefinedModels={
    {"Uniform",new Integer(0),new Integer(0),"uniform.bg"},
    {"Ensembl_Human_Upstream_2000US_to_200DS",new Integer(3),new Integer(9606),"Ensembl_human_2000-200bp_aroundTSS.bg"},
    {"Ensembl_Human_Upstream_400US",new Integer(3),new Integer(9606),"Ensembl_human_400bp_upstreamTSS.bg"},
    {"Ensembl_Human_Mouse_Syntenic_1000US_to_200DS",new Integer(3),new Integer(9606),"Ensembl_human_mouseSyntenic_1000-200bp_aroundTSS.bg"},
    {"Ensembl_Human_Mouse_Syntenic_10000US_to_200DS",new Integer(3),new Integer(9606),"Ensembl_human_mouseSyntenic_10000-200bp_aroundTSS.bg"},
    {"Ensembl_Human_Mouse_Syntenic_1000US",new Integer(3),new Integer(9606),"Ensembl_human_mouseSyntenic_up1000bp_fromTSS.bg"},
    {"EPD_cow_0",new Integer(0),new Integer(9913),"epd_bos_taurus_499_chromgenes_non_split_0.bg"},
    {"EPD_cow_1",new Integer(1),new Integer(9913),"epd_bos_taurus_499_chromgenes_non_split_1.bg"},
    {"EPD_cow_2",new Integer(2),new Integer(9913),"epd_bos_taurus_499_chromgenes_non_split_2.bg"},
    {"EPD_cow_3",new Integer(3),new Integer(9913),"epd_bos_taurus_499_chromgenes_non_split_3.bg"},
    {"EPD_chicken_0",new Integer(0),new Integer(9031),"epd_gallus_gallus_499_chromgenes_non_split_0.bg"},
    {"EPD_chicken_1",new Integer(1),new Integer(9031),"epd_gallus_gallus_499_chromgenes_non_split_1.bg"},
    {"EPD_chicken_2",new Integer(2),new Integer(9031),"epd_gallus_gallus_499_chromgenes_non_split_2.bg"},
    {"EPD_chicken_3",new Integer(3),new Integer(9031),"epd_gallus_gallus_499_chromgenes_non_split_3.bg"},
    {"EPD_human_0",new Integer(0),new Integer(9606),"epd_homo_sapiens_499_chromgenes_non_split_0.bg"},
    {"EPD_human_1",new Integer(1),new Integer(9606),"epd_homo_sapiens_499_chromgenes_non_split_1.bg"},
    {"EPD_human_2",new Integer(2),new Integer(9606),"epd_homo_sapiens_499_chromgenes_non_split_2.bg"},
    {"EPD_human_3",new Integer(3),new Integer(9606),"epd_homo_sapiens_499_chromgenes_non_split_3.bg"},
    {"EPD_mouse_0",new Integer(0),new Integer(10090),"epd_mus_musculus_499_chromgenes_non_split_0.bg"},
    {"EPD_mouse_1",new Integer(1),new Integer(10090),"epd_mus_musculus_499_chromgenes_non_split_1.bg"},
    {"EPD_mouse_2",new Integer(2),new Integer(10090),"epd_mus_musculus_499_chromgenes_non_split_2.bg"},
    {"EPD_mouse_3",new Integer(3),new Integer(10090),"epd_mus_musculus_499_chromgenes_non_split_3.bg"},
    {"EPD_rat_0",new Integer(0),new Integer(10116),"epd_rattus_norvegicus_499_chromgenes_non_split_0.bg"},
    {"EPD_rat_1",new Integer(1),new Integer(10116),"epd_rattus_norvegicus_499_chromgenes_non_split_1.bg"},
    {"EPD_rat_2",new Integer(2),new Integer(10116),"epd_rattus_norvegicus_499_chromgenes_non_split_2.bg"},
    {"EPD_rat_3",new Integer(3),new Integer(10116),"epd_rattus_norvegicus_499_chromgenes_non_split_3.bg"},
    {"EPD_vertebrate_0",new Integer(0),new Integer(0),"epd_vertebrates_499_chromgenes_non_split_0.bg"},
    {"EPD_vertebrate_1",new Integer(1),new Integer(0),"epd_vertebrates_499_chromgenes_non_split_1.bg"},
    {"EPD_vertebrate_2",new Integer(2),new Integer(0),"epd_vertebrates_499_chromgenes_non_split_2.bg"},
    {"EPD_vertebrate_3",new Integer(3),new Integer(0),"epd_vertebrates_499_chromgenes_non_split_3.bg"},
    {"EPD_frog_0",new Integer(0),new Integer(8364),"epd_xenopus_laevis_499_chromgenes_non_split_0.bg"},
    {"EPD_frog_1",new Integer(1),new Integer(8364),"epd_xenopus_laevis_499_chromgenes_non_split_1.bg"},
    {"EPD_frog_2",new Integer(2),new Integer(8364),"epd_xenopus_laevis_499_chromgenes_non_split_2.bg"},
    {"EPD_frog_3",new Integer(3),new Integer(8364),"epd_xenopus_laevis_499_chromgenes_non_split_3.bg"},
    {"hs200bp5",new Integer(3),new Integer(9606),"hs200bp5.bg"},
    {"Human_distal",new Integer(3),new Integer(9606),"hs_distal_manual_r18_3.bg"},
    {"Ensembl_Human_23299genes",new Integer(3),new Integer(9606),"hs_ensembl16_23299genes_1000bp_upstream.bg"},
    {"Human_Mouse_upstream",new Integer(3),new Integer(9606),"hs_upstreamcr_mm_3.bg"},
    {"human3",new Integer(3),new Integer(9606),"human3.bg"},
    {"Mouse_distal",new Integer(3),new Integer(10090),"mm_distal_manual_r18_3.bg"},
    {"SCerevisiae_0",new Integer(0),new Integer(4932),"scerevisiae_0.bg"},
    {"SCerevisiae_1",new Integer(1),new Integer(4932),"scerevisiae_1.bg"},
    {"SCerevisiae_2",new Integer(2),new Integer(4932),"scerevisiae_2.bg"},
    {"SCerevisiae_3",new Integer(3),new Integer(4932),"scerevisiae_3.bg"},
    {"SCerevisiae_4",new Integer(4),new Integer(4932),"scerevisiae_4.bg"},
    {"SCerevisiae_5",new Integer(5),new Integer(4932),"scerevisiae_5.bg"}     
};

public static String getFilenameForModel(String modelname) {
    for (Object[] entry:predefinedModels) {
        if (modelname.equals((String)entry[0])) return (String)entry[3];
    }
    return null;
}

/** Given the first line of a file, this method returns the name of the 
 *  dataformat believed to be used for the file, or null if the format is
 *  not recognized
 */
public static String determineDataFormatFromHeader(String firstline) {
         if (firstline.startsWith("#INCLUSive Background Model")) return "INCLUSive_Background_Model";
    else if (firstline.split("\\t").length==2) return "MEME_Background";
    else return "PriorityBackground";
}

}
