/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.Arrays;
import priorseditor.engine.ExecutionError;

/**
 * This class contains DNA data for a single sequence from a specific region in a genome
 * 
 * @author kjetikl
 */
public class DNASequenceData extends FeatureSequenceData {
    private char[] sequencedata; // the DNA sequence it self. The data should come from the direct strand so that sequencedata[0] corresponds to the smaller genomic coordinate
    private static String typedescription = "DNA sequence data";
    private transient Integer temporary_organism; // this is only used if sequence data is read from file. Usually, information about the temporary_organism is taken from the corresponding Sequence
    private transient String temporary_build; // this is only used if sequence data is read from file. Usually, information about the temporary_build is taken from the corresponding Sequence
    
    /**
     * Constructs a new DNASequenceData object from the supplied data
     * 
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param TES
     * @param orientation
     * @param sequencedata
     */
    public DNASequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS, Integer TES, int orientation, char[] sequencedata){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             this.sequencedata=sequencedata;
    }
    
    /**
     * Constructs a new DNASequenceData based on the supplied Sequence
     * 
     * @param sequence
     */
    public DNASequenceData(Sequence sequence, char defaultbase){
             setSequenceName(sequence.getName());
             setGeneName(sequence.getGeneName());
             setChromosome(sequence.getChromosome());
             setRegionStart(sequence.getRegionStart());
             setRegionEnd(sequence.getRegionEnd());
             setStrandOrientation(sequence.getStrandOrientation());
             setTSS(sequence.getTSS());
             setTES(sequence.getTES());
             int size=endPosition-startPosition+1;
             this.sequencedata=new char[size];           
             for (int i=0;i<size;i++) sequencedata[i]=defaultbase;
    } 
    /**
     * Constructs a new DNASequenceData based on the supplied Sequence
     * 
     * @param sequence
     */
    public DNASequenceData(Sequence sequence, char[] sequencedata){
             setSequenceName(sequence.getName());
             setGeneName(sequence.getGeneName());
             setChromosome(sequence.getChromosome());
             setRegionStart(sequence.getRegionStart());
             setRegionEnd(sequence.getRegionEnd());
             setStrandOrientation(sequence.getStrandOrientation());
             setTSS(sequence.getTSS());
             setTES(sequence.getTES());
             this.sequencedata=sequencedata;           
    } 
    
    
    /**
     * Constructs a new DNASequenceData with an initial sequence consisting of only the given default base
     * 
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param orientation
     * @param defaultbase 
     */
    public DNASequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS,  Integer TES, int orientation, char defaultbase){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             int size=endPosition-startPosition+1;
             this.sequencedata=new char[size];           
             for (int i=0;i<size;i++) sequencedata[i]=defaultbase;
    }
    
    /**
     * Constructs a new DNASequenceData with an initial sequence constructed from the given Background Model
     * 
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param orientation
     * @param randomgenerator 
     */
    public DNASequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS, Integer TES, int orientation, BackgroundModel randomgenerator){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             int size=endPosition-startPosition+1;         
             try {
                 sequencedata=randomgenerator.getNextBases("", size);
             } catch (ExecutionError e) {
                  this.sequencedata=new char[size];           
                  for (int i=0;i<size;i++) sequencedata[i]='N';
             }
    }
    
    @Override
    public Object getValue() {return sequencedata;}
    
    
    @Override
     public Object getValueInGenomicInterval(int start, int end) {
        if (end<getRegionStart() || start>getRegionEnd() || end<start) return null;
        if (start<getRegionStart()) start=getRegionStart();
        if (end>getRegionEnd()) end=getRegionEnd();
        int length=end-start+1;
        int relativeStart=getRelativePositionFromGenomic(start);
        char[] subset=new char[end-start+1];
        for (int i=0;i<length;i++) {
            subset[i]=sequencedata[relativeStart+i];
        }
        return subset;
    }
    
    /** Returns TRUE if the direct strand subsequence starting at the specified position 
     *  matches the given oligo (case-insensitive)
     */
    public boolean isDirectMatchAtPosition(int pos, String oligo) {
        int length=oligo.length();
        int start=getRegionStart();
        if (pos<start || pos>getRegionEnd()-length+1) return false;
        for (int i=0;i<length;i++) {
            if (Character.toUpperCase(sequencedata[pos-start+i])!=oligo.charAt(i)) return false;
        }
        return true;
    }

    
    /**
     * Returns the DNA base at the specified position relative to the
     * start of the sequence (if position==0 it returns the first base in the 
     * sequence). If the position is outside the boundaries of the sequence it returns
     * null
     * 
     * @param position
     * @return the DNA base at the specified position, or null
     */       
    public Character getValueAtRelativePosition(int position) {
        if (position<0 || position>=sequencedata.length) return null;
        else return new Character(sequencedata[position]);
    }

   /**
     * Returns the DNA base at the specified genomic position.
     * If the position is outside the boundaries of the sequence it returns null.
     * 
     * @param chromosome
     * @param position
     * @return the DNA base at the specified position, or null
     */       
    public Character getValueAtGenomicPosition(String chromosome, int position) {
        if (!chromosome.equalsIgnoreCase(this.chromosome)) return null;
        if (position<this.startPosition || position>this.endPosition) return null;
        return new Character(sequencedata[position-this.startPosition]);
    }

   /**
     * Returns the DNA base at the specified genomic position.
     * If the position is outside the boundaries of the sequence it returns null.
    * 
     * @param position
     * @return the DNA base at the specified position, or null
     */       
    public Character getValueAtGenomicPosition(int position) {
        if (position<this.startPosition || position>this.endPosition) return null;
        return new Character(sequencedata[position-this.startPosition]);
    }

   /**
     * Sets the DNA base at the specified genomic position.
     * 
     * @param position The genomic position
     * @param base The new base
     * @return false if position is outside sequence bounds else true
     */       
    public boolean setValueAtGenomicPosition(int position, char base) {
        if (position<this.startPosition || position>this.endPosition) return false;
        sequencedata[position-this.startPosition]=base;
        return true;
    }
    
   /**
     * Sets the DNA base at the specified position relative to the start of the sequence.
     * 
     * @param position The relative position
     * @param base The new base 
     * @return false if position is outside sequence bounds else true
     */       
    public boolean setValueAtRelativePosition(int position, char base) {
        if (position<0 || position>=sequencedata.length) return false;
        sequencedata[position]=base;
        return true;
    }

   /**
     * Returns the whole sequence as a string (from direct strand)
    * 
     * @param position
     * @return the whole sequence  as a string
     */       
    public String getSequenceAsString() {
        return new String(sequencedata);
    }

     /**
     * Returns a deep copy of this DNASequenceData object
     */
    @Override
    public DNASequenceData clone() {
        char[] newvalues=new char[sequencedata.length];
        for (int i=0;i<sequencedata.length;i++) {
            newvalues[i]=sequencedata[i];
        }        
        DNASequenceData newdata=new DNASequenceData(sequenceName, associatedGeneName, chromosome, startPosition, endPosition, TSS, TES, orientation,newvalues);
        return newdata;
    }      
    
    @Override
    public void importData(Data source) throws ClassCastException {
        super.importData(source);
        this.sequencedata=((DNASequenceData)source).sequencedata;
        //notifyListenersOfDataUpdate(); 
    }

    @Override
    public void setParent(FeatureDataset parent) {
        this.parent=parent;
    } 
    
    public static String getType() {return typedescription;}   
    
    @Override
    public String getTypeDescription() {return typedescription;}   
    
    
    public String getTemporaryBuild() {return temporary_build;}
    public Integer getTemporaryOrganism() {return temporary_organism;}
    public void setTemporaryBuild(String build) {this.temporary_build=build;}
    public void setTemporaryOrganism(Integer organism) {this.temporary_organism=organism;}


    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof DNASequenceData)) return false;
        if (!super.containsSameData(other)) return false;
        return (Arrays.equals(this.sequencedata,((DNASequenceData)other).sequencedata));
    }
}
