/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.io.Serializable;
import java.util.ArrayList;
import priorseditor.engine.ExecutionError;

/**
 * This class represent a portion of a Data Track for a Sequence.
 * When loading data for FeatureSequenceData objects, data can be retrieved 
 * either from disc cache or from external DataSources. Sometimes, some
 * portions of the data are available in cache while others must be obtained anew.
 * This class represents such subsegments (either from cache or external source)
 * that together are copied into complete FeatureSequenceData objects.
 * 
 * @author kjetikl
 */
public class DataSegment implements Serializable, Comparable<DataSegment> {
    private String trackName;
    private int organism;
    private String genomebuild;
    private String chromosome;
    private int segmentStart=0; // genomic coordinates for this segment
    private int segmentEnd=0;   // genomic coordinates for this segment
    private Object data=null;   // this could be an array (float[] or character[]) or an ArrayList<Region> depending on the data stored
    
    /** Creates */
    public DataSegment(String trackName,int organism, String genomebuild, String chromosome, int start, int end, Object data) {
        this.trackName=trackName;
        this.organism=organism;
        this.chromosome=chromosome;
        this.segmentStart=start;
        this.segmentEnd=end;
        this.data=data;
        this.genomebuild=genomebuild;
    }

    public String getChromosome() {
        return chromosome;
    }
    
    public String getGenomeBuild() {
        return genomebuild;
    }

    public Object getData() {
        return data;
    }

    public int getSegmentEnd() {
        return segmentEnd;
    }

    public int getSegmentStart() {
        return segmentStart;
    }
    
    public void setGenomeBuild(String genomebuild) {
        this.genomebuild=genomebuild;
    }

    public void setSegmentEnd(int end) {
        segmentEnd=end;
    }

    public void setSegmentStart(int start) {
        segmentStart=start;
    }
     
    /** Returns the length of this DataSegment */
    public int getSize() {
        return segmentEnd-segmentStart+1;
    }
    
    public void setSegmentData(Object newdata) {
        this.data=newdata;
    }
    
    /** 
     * This method can be used to add single Regions to the segment if the DataSegment represents
     * a portion of a RegionSequenceData track. The Regions added must have coordinates relative
     * to the start of the Segment. 
     */
    @SuppressWarnings("unchecked")
    public void addRegion(Region region) {
        if (this.data==null || !(this.data instanceof ArrayList)) this.data=new ArrayList<Region>();
        ((ArrayList)this.data).add(region);
    }
    
    /** 
     * Returns a String that can be used as a unique filename for this DataSegment object
     * if it should be saved to a disc cache
     */
    public String getFileName() {
        return trackName+"_"+organism+"_"+genomebuild+"_"+chromosome+"_"+segmentStart+"_"+segmentEnd;
    }

    @Override
    public int compareTo(DataSegment other) {
        if (this.segmentStart==other.segmentStart) return 0;
        else if (this.segmentStart<other.segmentStart) return -1;
        else return 1;
    }
    
    /** Returns true if the DataSegment object contains all the data for its segment */
    public boolean containsData() {
        return (data!=null);
    }
    
    /** Returns true if this DataSegment currently has no associated data */
    public boolean isEmpty() {
        return (data==null);
    }
    
    /** 
     * Copies the contents of this data object into the corresponding subsegment of
     * the supplied FeatureSequenceData object. 
     * The methods returns TRUE if the copy operation was completed successfully
     * or FALSE if not (which can happen if this object contains no data or if
     * the target FeatureSequenceData object was of an incompatible type).
     */
    public boolean copyData(FeatureSequenceData target) throws ExecutionError {
        if (!containsData()) return false;
        //System.err.println("Copying data into "+target.getSequenceName()+"["+target.getRegionStart()+"-"+target.getRegionEnd()+"]   segment="+segmentStart+"-"+segmentEnd+"   type="+data.getClass().toString());
        int useStart=segmentStart;
        int useEnd=segmentEnd;
        int offset=0;
        if (segmentStart<target.getRegionStart()) {
            offset=target.getRegionStart()-segmentStart;
            useStart=target.getRegionStart();
        }
        if (segmentEnd>target.getRegionEnd()) {
            useEnd=target.getRegionEnd();
        }
        int length=useEnd-useStart+1;
        if (target instanceof DNASequenceData) {
            if (!(data instanceof char[])) {throw new ExecutionError("SYSTEM ERROR: Incorrect data type. Expected DNA sequence data - got '"+data.getClass().getSimpleName()+"'. Possibly wrong data track type specification?");}
            DNASequenceData targetdata=(DNASequenceData)target;
            char[] datavalues=(char[])data;
            //System.err.println("Copying char[] into DNASequenceData from "+useStart+"-"+useEnd+"   offset="+offset+"  length="+length);
            for (int i=0;i<length;i++) {
                targetdata.setValueAtGenomicPosition(useStart+i, datavalues[offset+i]); // note that some positions can be outside the target sequence
            }
        }
        else if (target instanceof NumericSequenceData) {
            if (data instanceof double[]) { // just to make it compatible with never versions;
                double[] olddata=(double[])data;
                float[] newdata=new float[olddata.length];
                for (int i=0;i<newdata.length;i++) newdata[i]=(float)olddata[i]; // cannot use System.arraycopy() because type-conversion is involved
                data=newdata;
            }
            if (!(data instanceof float[])) {throw new ExecutionError("SYSTEM ERROR: Incorrect data type. Expected numeric data - got '"+data.getClass().getSimpleName()+"'. Possibly wrong data track type specification?");}
            NumericSequenceData targetdata=(NumericSequenceData)target;
            float[] datavalues=(float[])data;
            //System.err.println("Copying float[] into NumericSequenceData from "+useStart+"-"+useEnd+"   offset="+offset+"  length="+length);
            for (int i=0;i<length;i++) {
                targetdata.setValueAtGenomicPosition(useStart+i, datavalues[offset+i]); // note that some positions can be outside the target sequence
            }
        } 
        else if (target instanceof RegionSequenceData) {
            // Note that coordinates for Regions in the segment should initially be relative to the segment start
            // Befire the regions are added to the target, the Region coordinates are updated so that
            // they are relative to the start of the target sequence data object.
            if (!(data instanceof ArrayList)) {throw new ExecutionError("SYSTEM ERROR: Incorrect data type. Expected region data - got '"+data.getClass().getSimpleName()+"'. Possibly wrong data track type specification?");}
            ArrayList list=(ArrayList)data;
            RegionSequenceData targetdata=(RegionSequenceData)target;
            offset=segmentStart-targetdata.getRegionStart(); // used to adjust the relative start of regions from subsegment in this object to full segment in RegionSequenceData
            for (Object element:list) {
                if (!(element instanceof Region)) {System.err.println("SYSTEM ERROR: Incorrect data type in DataSegment.copyData: source="+element.getClass().toString()+", target="+target.getClass()+toString());return false;}
                Region region=(Region)element;
                region.setRelativeStart(region.getRelativeStart()+offset);
                region.setRelativeEnd(region.getRelativeEnd()+offset);
                if (region.getRelativeStart()>=targetdata.getSize() || region.getRelativeEnd()<0) continue; // Region lies outside targetdata segment
                if (!targetdata.containsRegion(region)) {targetdata.addRegion(region);} 
            }            
        }     
        return true;
    }

}
