/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.Arrays;

/**
 * This class represents Transcription Factor Binding Motifs 
 * (usually as PWMs or consensus strings)
 * 
 * @author kjetikl
 */
public class Motif extends Data implements Comparable {    
    public static final int COUNT_MATRIX=0;
    public static final int FREQUENCY_MATRIX=1;
    public static final int LOG_ODDS_MATRIX=2;
    public static final int UNNORMALIZED_FREQUENCY_MATRIX=3;

    private static String typedescription="Motif";   
    
    private String name=null; // usually accession. This is the name or ID used when referencing this Motif
    private String shortname=null; //
    private String longname=null; //
    private String consensus=null;
    private boolean halfsite=false;
    private float IC_content=0;
    private String organisms=null;
    private int quality=6; // 6 is "unknown"
    private String bindingfactors=null;
    private String classification=null;
    private int matrixtype=COUNT_MATRIX;
    private float[][] matrix; // x rows, 4 columns   
    // private float[] columnICs=null; // caches IC values for each column
    
    /**
     * Constructs a new Motif object with the given name
     * 
     * @param name

     */
    public Motif(String name){
         setName(name);
    }  

    @Override
    public String getName() {
        return name;
    }
    
    public String getShortName() {return shortname;}
    public void setShortName(String motifName) {this.shortname=motifName;}
    
    public String getLongName() {return longname;}
    public void setLongName(String motifName) {this.longname=motifName;}
    
    public String getPresentationName() {
        if (shortname==null || shortname.isEmpty()) return name;
        else return name+"-"+shortname;
    }
        
    public String getConsensusMotif() {
        if (consensus==null && matrix!=null) {
            consensus=getConsensusForMatrix(matrix);
        }    
        return consensus;
    }
    public void setConsensusMotif(String consensus) {this.consensus=consensus;}
    
    public int getMatrixType() {return matrixtype;}
    public void setMatrixType(int type) {matrixtype=type;}
    
    /** This will returned the stored matrix 'as is'
     * The values in the matrix could either be raw frequency counts, normalized counts or log-odds values
     */
    public float[][] getMatrix() {return matrix;}
    public void setMatrix(float[][] newmatrix) {
        int type=findProbableMatrixType(newmatrix);     
        matrix=newmatrix;
        float[][] frequencymatrix=matrix; 
        if (type==LOG_ODDS_MATRIX) frequencymatrix=getMatrixAsFrequencyMatrix(); // just in case
        IC_content=calculateInformationContent(frequencymatrix, false); 
        consensus=getConsensusForMatrix(frequencymatrix);
        setMatrixType(type);   
    }
    
    public String getOrganisms() {return organisms;}
    public void setOrganisms(String organisms) {this.organisms=organisms;}
    
    public int getQuality() {return quality;}
    public void setQuality(int quality) {this.quality=quality;}
    
    public float getICcontent() {
        if (IC_content==0 && matrix!=null) {
            IC_content=calculateInformationContent(matrix, false);
        }
        return IC_content;
    }
    public void setICcontent(float ic) {this.IC_content=ic;}
    
    public String getBindingFactors() {return bindingfactors;}
    public void setBindingFactors(String factors) {this.bindingfactors=factors;}
    
    public String getClassification() {return classification;}    
    public void setClassification(String classification) {
        if (classification!=null && classification.matches("[^0-9\\.]")) {System.err.println("Illegal motif class: '"+classification+"' for "+getPresentationName()); return;} // illegal class
        this.classification=classification;
    }
    
    public boolean isHalfsite() {return halfsite;}
    public void setHalfsite(boolean isHalfsite) {halfsite=isHalfsite;}
    
    /** Returns the length of the motif in base pairs */
    public int getLength() {
        if (matrix==null) return 0;
        else return matrix.length;
    }

    @Override
    public Object getValue() {return this;} // should maybe change later
    
    @Override
    public String getValueAsParameterString() {
        String parameter="";
        if (shortname!=null && !shortname.isEmpty()) parameter+="SHORTNAME:"+shortname+";";
        if (longname!=null && !longname.isEmpty()) parameter+="LONGNAME:"+longname+";";
        //if (consensus!=null && !consensus.isEmpty()) parameter+="CONSENSUS:"+consensus+";";
        if (matrix!=null) parameter+=getMatrixAsString();       
        if (classification!=null) parameter+="CLASS:"+classification+";";
        if (halfsite) parameter+="HALFSITE;";
        if (quality<6) parameter+="QUALITY:"+quality+";";
        if (organisms!=null && !organisms.isEmpty()) parameter+="ORGANISMS:"+organisms+";";
        if (bindingfactors!=null && !bindingfactors.isEmpty()) parameter+="FACTORS:"+bindingfactors+";";
        return parameter;
    }

    /** Returns the stored value for the specified base in the given position (starting at 0) 
     * (If the position is outside the matrix bounds or the base is unknown a value of 0 is returned);
     * This will return the value as stored in the matrix itself, which could be an unnormalized base count value,
     * a base frequency or a log-odds value.
     */
    public float getMatrixValue(int pos, char base) {
        if (matrix==null || pos<0 || pos>=matrix.length) return 0;
        switch (base) {
            case 'A': return matrix[pos][0];
            case 'a': return matrix[pos][0];
            case 'C': return matrix[pos][1];
            case 'c': return matrix[pos][1];
            case 'G': return matrix[pos][2];
            case 'g': return matrix[pos][2];
            case 'T': return matrix[pos][3];
            case 't': return matrix[pos][3];
        }
        return 0;
    }
    
    /** Returns the value for the specified base in the given position (starting at 0) 
     * (If the position is outside the matrix bounds or the base is unknown a value of 0 is returned);
     * This will return the value as stored in the matrix itself
     * Note that unknown/masked bases (other than A,C,G,T) will return the value 0
     */
    public float getBaseFrequency(int pos, char base) {
        if (matrix==null || pos<0 || pos>=matrix.length) return 0;
        float value = 0;
        switch (base) {
            case 'A': value = matrix[pos][0]; break;
            case 'a': value = matrix[pos][0]; break;
            case 'C': value = matrix[pos][1]; break;
            case 'c': value = matrix[pos][1]; break;
            case 'G': value = matrix[pos][2]; break;
            case 'g': value = matrix[pos][2]; break;
            case 'T': value = matrix[pos][3]; break;
            case 't': value = matrix[pos][3]; break;
            default : return 0;
        }
        if (matrixtype==FREQUENCY_MATRIX) return value;
        else if (matrixtype==COUNT_MATRIX || matrixtype==UNNORMALIZED_FREQUENCY_MATRIX) {
            float total=matrix[pos][0]+matrix[pos][1]+matrix[pos][2]+matrix[pos][3];
            return value/total;
        } else if (matrixtype==LOG_ODDS_MATRIX) {
            float fraction=(float)Math.exp(value);
            return fraction*0.25f; // using uniform background frequencies
        } else return value;
    }  
    
    /**
     * Sets the name of the motif
     * 
     * @param a name for this motif 
     */
    public void setName(String name) {
        this.name=name;
    }
    
    @Override
    public void rename(String newname) {
        setName(newname);
    } 
   
    private String getMatrixAsString() {
        String result="A:";
        for (int i=0;i<matrix.length;i++) {
            if (i<matrix.length-1) result+=matrix[i][0]+",";
            else result+=matrix[i][0]+";";
        }
        result+="C:";
        for (int i=0;i<matrix.length;i++) {
            if (i<matrix.length-1) result+=matrix[i][1]+",";
            else result+=matrix[i][1]+";";
        }
        result+="G:";
        for (int i=0;i<matrix.length;i++) {
            if (i<matrix.length-1) result+=matrix[i][2]+",";
            else result+=matrix[i][2]+";";
        }
        result+="T:";
        for (int i=0;i<matrix.length;i++) {
            if (i<matrix.length-1) result+=matrix[i][3]+",";
            else result+=matrix[i][3]+";";
        }
        return result;
    }
       
    
//    private float[] getColumnICs() {
//        if (columnICs==null) {
//            columnICs=
//        }
//        return columnICs;
//    }
    
    
    /** Calculates the IC content of a count matrix
     * @param matrix A count matrix where the first index is the position and the second is the base (A=0,C=1,G=2,T=3)
     * @param a boolean flag that can be used to correct the calculation for small sample biases
     * @return the IC (information content) of the matrix
     */
    public static float calculateInformationContent(float[][] matrix, boolean correctForSmallSampleBias) {
        if (matrix==null) return 0;
        float ic=0;
        for (int i=0;i<matrix.length;i++) {
            ic+=calculateColumnIC(matrix[i][0], matrix[i][1], matrix[i][2], matrix[i][3], correctForSmallSampleBias);
        }
        return ic;
    }
   
    
    /** Calculates the information content in one position (for non-log-odds matrices) */
    public static float calculateColumnIC(float a_count, float c_count, float g_count, float t_count,boolean correctForSmallSampleBias) {
           float total= a_count + c_count + g_count + t_count;
           float Pa=a_count/total;
           float Pc=c_count/total;
           float Pg=g_count/total;
           float Pt=t_count/total;
           float errorterm=3f/(2f*(float)Math.log(2)*total); // approximation. Found this in an appendix to a paper by Thomas D. Schneider
           float entropyA=(Pa>0)?Pa*log2(Pa):0;
           float entropyC=(Pc>0)?Pc*log2(Pc):0;
           float entropyG=(Pg>0)?Pg*log2(Pg):0;
           float entropyT=(Pt>0)?Pt*log2(Pt):0;
           float HsL=-(entropyA+entropyC+entropyG+entropyT);
           if (correctForSmallSampleBias) return 2-errorterm-HsL;
           else return 2-HsL;
    }
    
    /** Return the Log2-value of the input*/
    private static float log2(float value) {
        return (float)(Math.log(value)/Math.log(2));
    }
   
    
    /** 
     * Returns a consensus string for the given matrix
     * The consensus is determined according the the rules described in Transfac
     * 1) If a single base occurs in at least 50% of the sequences in a given position, that base is returned for that position
     * 2) A double-degenerate base (W,S,R,Y,M,K) is returned if two bases together occurs in at least 75% of the sequences and rule 1 does not apply
     * 3) A triple-degenerate base (B,D,H,V) is returned if none of the previous rules apply and one base is missing from all sequences in this position
     * 4) An 'N' is returned if no other rules apply
     */
    public static String getConsensusForMatrix(float[][] matrix) {
       if (matrix==null) return null;
       String consensus="";
       for (int i=0;i<matrix.length;i++) { 
          float a=matrix[i][0];
          float c=matrix[i][1];
          float g=matrix[i][2];
          float t=matrix[i][3];
          float total=a+c+g+t;
          String consensusletter="n";
               if (a/total>=0.5 && a>=2*c && a>=2*g && a>=2*t) consensusletter="A"; 
          else if (c/total>=0.5 && c>=2*a && c>=2*g && c>=2*t) consensusletter="C"; 
          else if (g/total>=0.5 && g>=2*a && g>=2*c && g>=2*t) consensusletter="G"; 
          else if (t/total>=0.5 && t>=2*a && t>=2*c && t>=2*g) consensusletter="T"; 
          else if ((a+c)/total>=0.75) consensusletter="m"; 
          else if ((a+g)/total>=0.75) consensusletter="r"; 
          else if ((a+t)/total>=0.75) consensusletter="w"; 
          else if ((c+g)/total>=0.75) consensusletter="s"; 
          else if ((c+t)/total>=0.75) consensusletter="y"; 
          else if ((g+t)/total>=0.75) consensusletter="k"; 
          else if (a==0) consensusletter="b"; 
          else if (c==0) consensusletter="d"; 
          else if (g==0) consensusletter="h"; 
          else if (t==0) consensusletter="v"; 
          consensus+=consensusletter;
       }          
      return consensus;
    }
    
    
    
    @Override
    public void importData(Data source) {
        this.name=((Motif)source).name;  
        this.consensus=((Motif)source).consensus;
        this.shortname=((Motif)source).shortname;
        this.longname=((Motif)source).longname;
        this.matrixtype=((Motif)source).matrixtype;
        this.halfsite=((Motif)source).halfsite;
        this.IC_content=((Motif)source).IC_content;
        this.organisms=((Motif)source).organisms;
        this.quality=((Motif)source).quality;
        this.bindingfactors=((Motif)source).bindingfactors;
        this.classification=((Motif)source).classification;
        float[][] sourcematrix=((Motif)source).matrix;
        this.matrix=(float[][])sourcematrix.clone();
        //notifyListenersOfDataUpdate(); 
    }
    
    @Override
    public Motif clone() {
        Motif motif=new Motif(name);
        motif.consensus=this.consensus;
        motif.shortname=this.shortname;
        motif.longname=this.longname;
        motif.matrixtype=this.matrixtype;
        motif.halfsite=this.halfsite;
        motif.IC_content=this.IC_content;
        motif.quality=this.quality;
        motif.bindingfactors=this.bindingfactors;
        motif.classification=this.classification;
        motif.matrix=(float[][])this.matrix.clone();
        motif.organisms=this.organisms;
        return motif;
    }
    
    /**
     * Returns true if this Motif equals the other given Motif
     * @param other
     * @return
     */
    @Override
    public boolean containsSameData(Data othermotif) {
        if (othermotif==null || !(othermotif instanceof Motif)) return false;
        Motif other=(Motif)othermotif;
        if ((other.shortname==null && this.shortname!=null) || (other.shortname!=null && this.shortname==null) || (other.shortname!=null && this.shortname!=null && !other.shortname.equals(this.shortname))) return false;
        if ((other.longname==null && this.longname!=null) || (other.longname!=null && this.longname==null) ||  (other.longname!=null && this.longname!=null && !other.longname.equals(this.longname))) return false;
        if ((other.consensus==null && this.consensus!=null) || (other.consensus!=null && this.consensus==null) ||  (other.consensus!=null && this.consensus!=null && !other.consensus.equals(this.consensus))) return false;
        if (other.halfsite!=this.halfsite) return false;
        if (other.quality!=this.quality) return false;
        if ((other.bindingfactors==null && this.bindingfactors!=null) || (other.bindingfactors!=null && this.bindingfactors==null) ||  (other.bindingfactors!=null && this.bindingfactors!=null && !other.bindingfactors.equals(this.bindingfactors))) return false;
        if ((other.classification==null && this.classification!=null) || (other.classification!=null && this.classification==null) ||  (other.classification!=null && this.classification!=null && !other.classification.equals(this.classification))) return false;
        if ((other.organisms==null && this.organisms!=null) || (other.organisms!=null && this.organisms==null) ||  (other.organisms!=null && this.organisms!=null && !other.organisms.equals(this.organisms))) return false;
        if (!Arrays.equals(other.matrix, this.matrix)) return false;
        return true;
    }
    
    public static String getType() {return typedescription;}    

    @Override
    public String getTypeDescription() {return typedescription;}

    @Override
    public int compareTo(Object other) {
        if (!(other instanceof Motif)) return -1;
        return getPresentationName().compareTo(((Motif)other).getPresentationName());
    }

    /**
     * Tries to figure out the most probable type of matrix this is based on the values in the matrix
     * The rules are as follows:
     *      If all matrix values are positive integers it is assumed that the matrix is a count matrix.
     *      else, If all numbers are between 0 and 1 (inclusive) it is assumed that the matrix is a frequency matrix
     *      else, if any bases are negative it is assumed that the matrix is a log-odds matrix
     *      else, it is assumed that the matrix is an unnormalized frequency matrix
     * @return a number specifying the assumed type of matrix (or -1 if matrix is NULL)
     */
    public static int findProbableMatrixType(float[][] targetmatrix) {
        if (targetmatrix==null) return -1;
        boolean allIntegers=true;
        for (int i=0;i<targetmatrix.length;i++) { 
            for (int j=0;j<4;j++) {
                float value=targetmatrix[i][j];
                if (value>=0 && (float)((int)value)!=value) {allIntegers=false;break;} 
            }  
        }
        if (allIntegers) return COUNT_MATRIX;
        boolean allBetween0tand1=true;
        for (int i=0;i<targetmatrix.length;i++) { 
            for (int j=0;j<4;j++) {
                float value=targetmatrix[i][j];
                if (value<0 || value>1) {allBetween0tand1=false;break;} 
            }  
        }    
        if (allBetween0tand1) return FREQUENCY_MATRIX;
        for (int i=0;i<targetmatrix.length;i++) { 
            for (int j=0;j<4;j++) {
                float value=targetmatrix[i][j];
                if (value<0) return LOG_ODDS_MATRIX;
            }  
        }            
        return UNNORMALIZED_FREQUENCY_MATRIX;
    }
    
    /** Returns a normalized frequency matrix for this 
     *  The returned matrix must not be modified. If the original matrix for
     *  this motif is FREQUENCY_MATRIX then the original matrix is returned,
     *  if not a new matrix with converted values are returned!
     */
    public float[][] getMatrixAsFrequencyMatrix() {
        if (matrixtype==FREQUENCY_MATRIX) return getMatrix();
        else if (matrixtype==LOG_ODDS_MATRIX) {
            float[][] result=copyMatrix(getMatrix());
            for (int i=0;i<result.length;i++) {
              for (int j=0;j<4;j++) {
                  result[i][j]=(float)Math.exp(result[i][j])*0.25f;
              }
            } 
            return result;
        } else { // just need to normalize
            float[][] result=copyMatrix(getMatrix());
            for (int i=0;i<result.length;i++) {
                float total=result[i][0]+result[i][1]+result[i][2]+result[i][3];
                result[i][0]=result[i][0]/total;
                result[i][1]=result[i][1]/total;
                result[i][2]=result[i][2]/total;
                result[i][3]=result[i][3]/total;
            }
            return result;
        }
    }
    
    
    /** Given a matrix as double array, this method will return the reverse complement of the source matrix*/
    public static float[][] reverseComplementMatrix(float[][] source) {
        float[][] result=new float[source.length][4];
        for (int i=0;i<source.length;i++) {
            int revpos=(source.length-1)-i;
            result[revpos][0]=source[i][3]; //
            result[revpos][1]=source[i][2]; //
            result[revpos][2]=source[i][1]; //
            result[revpos][3]=source[i][0]; //
        }
        return result;
    }
    /** Given a matrix as double array, this method will return the reverse complement of the source matrix*/
    public static float[][] copyMatrix(float[][] source) {
        float[][] result=new float[source.length][4];
        for (int i=0;i<source.length;i++) {
            result[i][0]=source[i][0]; //
            result[i][1]=source[i][1]; //
            result[i][2]=source[i][2]; //
            result[i][3]=source[i][3]; //
        }
        return result;
    }
    

    
}
