/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.operations.Operation_new;


/**
 * MotifCollections can be used to refer to collections of Transcription Factor Binding Motifs
 * Allthough most get/set methods work directly with Motif
 * objects rather than Motif names (Strings), the internal mechanism of the MotifCollection
 * itself revolves around a plain list of names for the Motifs in the collection.
 * The methods that returns Motif objects or lists do this by dynamically obtaining these Motifs
 * from the Engine (on-the-fly) based on the internal list of names
 *
 * @author kjetikl
 */
public class MotifCollection extends Data {
    private static String typedescription="Motif Collection";
    protected String datasetName;
    protected ArrayList<String> storage; // this is a list of Motif names
    private ArrayList<Motif> payload=null; // this is used to temporarily store Motifs that should be registered with the engine
    private String predefined=null; // name of predefined model
    private String fromTrack=null; // a configuration string if this collection is based on a track. Formatexample: "<trackname>,support>=80%"

    /**
     * Constructs a new initially "empty" Motif collection with the given name
     *
     * @param datasetName A name for this dataset
     */
   public MotifCollection(String datasetName) {
       this.datasetName=datasetName;
       storage=new ArrayList<String>(20);
   }




    /**
     * Specifies a new name for this dataset
     *
     * @param name the name for this dataset
     */
    public void setName(String name) {
        this.datasetName=name;
    }

    @Override
    public void rename(String name) {
        setName(name);
    }

   /**
    * Returns the name of this dataset
    *
    * @return dataset name
    */
    @Override
    public String getName() {
        return datasetName;
    }

    @Override
    public Object getValue() {return this;} // should maybe change later

    @Override
    public String getValueAsParameterString() {
        if (predefined!=null) return Operation_new.COLLECTION_PREFIX+predefined;
        else if (fromTrack!=null) return Operation_new.FROMTRACK_PREFIX+fromTrack;
        else {
            String string="";
            for (int i=0;i<storage.size();i++) {
            if (i==storage.size()-1) string+=storage.get(i);
                else string+=storage.get(i)+",";
            }
            return string;
        }
    }



    /**
     * Returns the names of all the Motifs in this dataset
     *
     * @return A list of Motif names
     */
    public ArrayList<String> getAllMotifNames() {
        return storage;
    }

    /**
     * Returns the presentation names of all the Motifs in this dataset
     *
     * @return A list of Motif presentation names
     */
    public ArrayList<String> getAllMotifPresentationNames(PriorsEditorEngine engine) {
        ArrayList<String> list = new ArrayList<String>(storage.size());
        for (String name:storage) {
            Data item=engine.getDataItem(name);
            if (item!=null && item instanceof Motif) list.add( ((Motif)item).getPresentationName() );
        }
        return list;
    }

    /**
     * Returns all the Motif objects in this dataset
     * (if they are currently registered with the engine)
     *
     * @return A list of Motif objects
     */
    public ArrayList<Motif> getAllMotifs(PriorsEditorEngine engine) {
        ArrayList<Motif> list = new ArrayList<Motif>(storage.size());
        for (String name:storage) {
            Data item=engine.getDataItem(name);
            if (item!=null && item instanceof Motif) list.add((Motif)item);
        }
        return list;
    }



    /**
     * Returns the Motif corresponding to the given name.
     *
     * @param name The name of the Motif
     * @return the specified Motif (if found) or null
     */
    public Motif getMotifByName(String name, PriorsEditorEngine engine) {
        for (String motifname : storage) {
            if (motifname.equals(name)) {
                Data item=engine.getDataItem(name);
                if (item!=null && item instanceof Motif) return (Motif)item;
            }
        }
        return null;
    }

    /**
     * Returns the name of the Motif corresponding to the given index.
     * This method could be used if you want to iterate through all Motifs
     * in a dataset.
     *
     * @param index The index of the Motif
     * @return the name of the Motif at the specified index (if exists) or null
     */
    public String getMotifNameByIndex(int index) {
        return storage.get(index);
    }
    /**
     * Returns the Motif corresponding to the given index.
     * This method could be used if you want to iterate through all Motifs
     * in a dataset.
     *
     * @param index The index of the Motif
     * @return the specified Motif (if exists) or null
     */
    public Motif getMotifByIndex(int index, PriorsEditorEngine engine) {
        String name=storage.get(index);
        Motif motif=null;
        if (name!=null) {
            Data item=engine.getDataItem(name);
            if (item!=null && item instanceof Motif) return (Motif)item;
        }
        return motif;
    }

    /**
     * Returns the index (order within the dataset) of the Motif with the given name
     * If no Motif with the given name exists within the dataset the value -1 is returned.
     *
     * @param name The name of the Motif
     * @return index of the Motif (between 0 and size-1) or -1
     */
    public int getIndexForMotif(String name) {
        for (int i=0;i<storage.size();i++) {
            String motifName=storage.get(i);
            if (motifName.equals(name)) return i;
        }
        return -1;
    }

    /**
     * Returns true if the specified Motif is in this collection
     */
    public boolean contains(Motif motif) {
        if (motif==null) return false;
        String motifName=motif.getName();
        return contains(motifName);
    }

    /**
     * Returns true if a Motif with the given name is in this collection
     */
    public boolean contains(String motifName) {
        return storage.contains(motifName);
    }

    /**
     * Returns true if this MotifCollection objects contains the same Motifs
     * as the given collection (the source (eg motif track or predefined)
     * of these motifs are not considered, only if the lists of motifs are the same)
     * @param collection
     * @return
     */
    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof MotifCollection)) return false;
        MotifCollection collection=(MotifCollection)other;
        if (size()!=collection.size()) return false;
        for (String name:storage) {
            if (!collection.contains(name)) return false;
        }
        return true;
    }

   /**
     * Adds a new Motif object to this collection (if it is not already present)
     *
     * @param motif The Motif to be added
     */
    public void addMotif(Motif motif) {
        if (storage.contains(motif.getName())) return;
        storage.add(motif.getName()); // add to local storage
        notifyListenersOfDataAddition(motif);
    }

   /**
     * Removes a Motif object from this collection (if present)
     *
     * @param motif The Motif to be remvoed
     */
    public void removeMotif(Motif motif) {
        boolean hasremoved=storage.remove(motif.getName()); // remove from local storage
        if (hasremoved) notifyListenersOfDataRemoval(motif);
    }


   /**
     * Removes all Motifs from this dataset
     */
    public void clearAll(PriorsEditorEngine engine) {
        String[] list=new String[storage.size()];
        for (int i=0;i<list.length;i++) {list[i]=storage.get(i);}
        for (String name:list) {
            storage.remove(name);
            Data item=null;
            if (engine!=null) item=engine.getDataItem(name);
            if (item!=null && item instanceof Motif) notifyListenersOfDataRemoval((Motif)item);
         }
    }


    /*
     * Reorders the list of motifs in the dataset by moving the motif
     * at the specified current position to the new position.
     *
     * @param oldposition Position of the motif to be moved
     * @param newposition New position to move motif to
     *
    public void reorderMotifs(int currentposition, int newposition) {
        String temp=storage.remove(currentposition);
        storage.add(newposition,temp);
        notifyListenersOfDataUpdate();
    }
    */

    /**
     * Returns the number of Motifs in this collection
     *
     * @return number of motifs
     */
    public int getNumberofMotifs() {
        return storage.size();
    }

    /**
     * Returns the number of motifs in this collection (same as getNumberofMotifs)
     *
     * @return number of motifs
     */
    public int size() {
        return storage.size();
    }

   /**
     * Returns TRUE if this MotifCollection contains no motifs
     */
    public boolean isEmpty() {
        return storage.size()==0;
    }




    @Override
    public String output() {
        String string="";
        for (int i=0;i<storage.size();i++) {
           string+=storage.get(i)+"\n";
        }
        return string;
    }


     /**
     * Initialize this collection based on which motifs (known to the engine)
     * are found in in a selected number of sequences in the motif track (region dataset)
     * provided as argument
     *
     * @param dataset The motif track (Region dataset) used as basis
     * @param quorum The quorum states the number of sequences the motif should be present in
     * @param exactly If this flag is TRUE the motifs must be present in exactly this number of sequences
     *                If false the quorum number is interpreted such that the motif must be present in "at least" that many sequences
     * @param percentage If this flag is TRUE the quorum number should be between 0 and 100 and will be interpreted as a percentage of the sequences
     *                   If the flag is FALSE the quorum number will be interpreted as the specific number of sequences
     */
   public void createCollectionFromTrack(RegionDataset dataset, int quorum, boolean exactly, boolean percentage, PriorsEditorEngine engine) {
       HashMap<String,Integer> counts=new HashMap<String,Integer>();
       ArrayList<Data> allMotifs=engine.getAllDataItemsOfType(Motif.class);
       storage.clear();
       // configure
       setPredefinedCollectionName(null);
       this.fromTrack=dataset.getName()+",support"+((exactly)?"=":">=")+quorum+((percentage)?"%":""); // sets the 'config' string
       if (percentage) {
          int seqsize=engine.getDefaultSequenceCollection().getNumberofSequences();
          quorum=Math.round((float)quorum/100f*seqsize);
       }
       for (Data motif:allMotifs) {
           counts.put(motif.getName(), new Integer(0));
       }

       ArrayList<FeatureSequenceData> sequences=dataset.getAllSequences();
       for (FeatureSequenceData sequence:sequences) {
           ArrayList<Region> allRegions=((RegionSequenceData)sequence).getOriginalRegions();
           HashSet<String> present=new HashSet<String>();
           for (Region region:allRegions) { // make a set of present region types (motifs)
               present.add(region.getType());
           }
           for (String motiftype:present) { // update motif counts by one if present in current sequence
               if (engine.dataExists(motiftype, Motif.class)) {
                   counts.put(motiftype, counts.get(motiftype)+1);
               }
           }
       }
       for (Data motif:allMotifs) {
           int motifquorum=counts.get(motif.getName());
           if ((exactly && motifquorum==quorum) || (!exactly && motifquorum>=quorum)) storage.add(motif.getName());
       }

   }


    /**
     * This method takes all motifs in the specified dataset and adds them
     * to this dataset
     * @param dataset The dataset to be incorporated into this dataset
     */
    public void merge(MotifCollection dataset) {
        int size=dataset.size();
        for (int i=0;i<size;i++) {
            String motifname=dataset.getMotifNameByIndex(i);
            if (!contains(motifname)) storage.add(motifname); // I use this directly instead of calling addMotif() to limit the number of notifications sent
        }
        notifyListenersOfDataAddition(dataset);
    }

    /**
     * This method removes all motifs from this collection that are also present
     * in the argument dataset
     * @param dataset A collection of motifs to remove from this collection
     */
    public void subtract(MotifCollection dataset) {
        int size=dataset.size();
        for (int i=0;i<size;i++) {
            String motifname=dataset.getMotifNameByIndex(i);
            if (contains(motifname)) storage.remove(motifname); // I use this directly instead of calling removeMotif() to limit the number of notifications sent
        }
        notifyListenersOfDataRemoval(dataset);
    }

    /**
     * This method compares the current contents of this collection with the argument collection
     * and removes any motifs from this collection that are not present in both
     * @param dataset A collection of motifs to keep in this collection (if the already exists).
     * Remove all the rest
     */
    public void intersect(MotifCollection dataset) {
        MotifCollection removed=new MotifCollection("removed");
        int size=storage.size();
        for (int i=0;i<size;i++) { // first find which motifs to remove
            String motifname=storage.get(i);
            if (!dataset.contains(motifname)) {
                removed.storage.add(motifname);
            }
        }
        // then remove them
        size=removed.size();
        for (int i=0;i<size;i++) { // first find which motifs to remove
            String motifname=removed.getMotifNameByIndex(i);
            this.storage.remove(motifname);
        }
        notifyListenersOfDataRemoval(removed);
    }






    @Override
    public void importData(Data source) throws ClassCastException {
        if (source==this) return; // no need to import, the source and target are the same
        MotifCollection datasource=(MotifCollection)source;
        this.datasetName=datasource.datasetName;
        this.predefined=datasource.predefined;
        this.fromTrack=datasource.fromTrack;
        storage.clear();
        for (String motifName:datasource.storage) {
            storage.add(motifName);
        }
        //notifyListenersOfDataUpdate();
    }

    @Deprecated
    /**
     * This should only be used for testing
     */
    public void importMotifList(String[] list) {
        storage.clear();
        for (String motifName:list) {
            storage.add(motifName);
        }
        //notifyListenersOfDataUpdate();
    }


    @Override
    public MotifCollection clone() {
        MotifCollection newcollection= new MotifCollection(datasetName);
        int size=size();
        for (int i=0;i<size;i++) {
            String motifName=getMotifNameByIndex(i);
            newcollection.storage.add(motifName); // I use this directly instead of calling addMotif() to limit the number of notifications sent
        }
        newcollection.predefined=predefined;
        newcollection.fromTrack=fromTrack;
        return newcollection;
    }

    public void setPredefinedCollectionName(String model) {
        this.predefined=model;
    }

    public String getPredefinedCollectionName() {
        return predefined;
    }
    /** Returns TRUE if this MotifCollection returns a system predefined collection (like Transfac or JASPAR) */
    public boolean isPredefined() {
        return predefined!=null;
    }


    /** If this collection is based on which motifs are present in a motif track, this method will return a string describing the settings used for initialization */
    public String getFromTrackString() {
        return fromTrack;
    }
    /** Returns TRUE if this MotifCollection is based on which motifs are present in a motif track */
    public boolean isFromTrack() {
        return fromTrack!=null;
    }

    public static String getType() {return typedescription;}

    @Override
    public String getTypeDescription() {return typedescription+" : "+storage.size();}


    /** This method adds a motif to temporary storage within this collection.
     * (These temporarily stored motifs should be registered with the engine afterwards
     * using the initializeFromPayload() method, however, in recent version this
     * is done automatically in PriorsEditorEngine.storeDataItem(data) which checks if
     * the data item is a MotifCollection and calls initializeFromPayload())
     */
    public void addMotifToPayload(Motif motif) {
        if (payload==null) payload=new ArrayList<Motif>();
        payload.add(motif);
    }

    /** This method removes a motif from temporary storage within this collection.
     */
    public void removeMotifFromPayload(Motif motif) {
        if (payload==null) return;
        else payload.remove(motif);
    }

    /** This method removes a motif from temporary storage within this collection.
     */
    public void removeMotifFromPayloadAtIndex(int index) {
        if (payload==null || index<0 || index>=payload.size()) return;
        else payload.remove(index);
    }

    /** This method renames a motif existing in the temporary storage within this collection.
     *  @return returns the old name of the motif at that index
     */
    public String renameMotifInPayload(int index,String newname) {
        if (payload==null || index<0 || index>=payload.size()) return null;
        Motif motif=payload.get(index);
        String oldname=motif.getName();
        motif.rename(newname);
        return oldname;
    }

    /** This method returns the List of Motifs temporarily stored in the payload of this collection
     */
    public List<Motif> getPayload() {
        return payload;
    }

    /** This method returns the number of Motifs temporarily stored in the payload of this collection
     */
    public int getPayloadSize() {
        if (payload==null) return 0;
        else return payload.size();
    }

    /** This registers all Motifs temporarily stored in the payload and deletes the payload storage */
    public void initializeFromPayload(PriorsEditorEngine engine) throws ExecutionError {
        //System.err.println("Initializing MotifCollection from payload. Payload="+(payload!=null));
        if (payload==null) return;
        for (Motif motif:payload) {
            engine.updateDataItem(motif);
            this.addMotif(motif);
        }
        payload=null; // clear payload after initialization!
    }

    public boolean hasPayload() {
        if (payload==null) return false;
        else return !payload.isEmpty();
    }

    public boolean isImportedCollection() {
        return (hasPayload() && predefined==null);
    }


/** Given the first line of a file, this method returns the name of the
 *  dataformat believed to be used for the file, or null if the format is
 *  not recognized
 */
public static String determineDataFormatFromHeader(String firstline) {
         if (firstline.startsWith("#INCLUSive Motif")) return "INCLUSive_Motif_Model";
    else if (firstline.startsWith("#PriorsEditorMotif")) return "PriorsEditorMotif";
    else return null;
}

}

