/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;


/**
 *
 * @author kjetikl
 */
public class NumericDataset extends FeatureDataset implements Cloneable {
    private static String typedescription="Numeric dataset";
    private float min=0;
    private float max=1;
    private float baseline=0;
    
    
    public NumericDataset(String datasetName) {
        super(datasetName);
    }
    
   /**
     * Returns the maximum value allowed for this dataset
     * @return the maximum allowed value
     */       
    public Float getMaxAllowedValue() {
        return max;
    }

   /**
     * Sets the maximum value allowed for this dataset
     * The value must be equal to or greater than the baseline value
     * (if the value is smaller than the baseline, then the max value is set to the baseline value instead!)
     * @param max The maximum value for this dataset
     */       
    public void setMaxAllowedValue(float max) {
        if (max<baseline) max=baseline;
        this.max=max;
    }

   /**
     * Returns the minimum value allowed for this dataset
     * @return the minimum allowed value
     */       
    public Float getMinAllowedValue() {
        return min;
    }

   /**
     * Sets the minimum value allowed for this dataset
     * The value must be equal to or smaller than the baseline value
     * (if the value is greater than the baseline, then the min value is set to the baseline value instead!)
     * @param min The minimum value for this dataset
     */       
    public void setMinAllowedValue(float min) {
        if (min>baseline) min=baseline;
        this.min=min;
    }

   /**
     * Returns the baseline value for this dataset
     * @return the baseline value
     */       
    public Float getBaselineValue() {
        return baseline;
    }

   /**
     * Sets the baseline value for this dataset
     * @param baseline The baseline value for this dataset
     */       
    public void setBaselineValue(float baseline) {
        this.baseline=baseline;
    }
    
    /** Updates the allowed min and max values for this dataset based on the data for the current sequences.
     *  If the min/max-values of the current data are outside the allowed range, the range is extended to 
     *  included these values. If the current min/max-values are within the old range, the old allowed boundaries
     *  are kept. However, if the new range is significantly different smaller than the old range (less than 25%)
     *  the allowed range is updated to fit the new current min/max values
     */
    public void updateAllowedMinMaxValuesFromData() {
        float oldmin=min; float oldmax=max; float oldrange=oldmax-oldmin;
        float newmin=Float.MAX_VALUE; float newmax=-Float.MAX_VALUE;
        //min=0; max=1;
        ArrayList<FeatureSequenceData> sequences=getAllSequences();
        for (FeatureSequenceData seq:sequences) {
            float[] newvalues=((NumericSequenceData)seq).updateAllowedMinMaxValuesFromDataRange();
            if (newvalues[0]<min) min=newvalues[0];
            if (newvalues[1]>max) max=newvalues[1];            
            if (newvalues[0]<newmin) newmin=newvalues[0];
            if (newvalues[1]>newmax) newmax=newvalues[1];            
        }
        float newrange=newmax-newmin;
        if (newrange/oldrange<0.25) {min=newmin;max=newmax;}
    }
    
    /** 
     * Returns the actual min/max-values in this dataset based on a search in the sequences
     * (this is used by the SetRangeDialog)
     */
    public float[] getMinMaxValuesFromData() {
        float newmin=Float.MAX_VALUE; float newmax=-Float.MAX_VALUE;
        //min=0; max=1;
        ArrayList<FeatureSequenceData> sequences=getAllSequences();
        for (FeatureSequenceData seq:sequences) {
            float[] newvalues=((NumericSequenceData)seq).getMinMaxFromData();          
            if (newvalues[0]<newmin) newmin=newvalues[0];
            if (newvalues[1]>newmax) newmax=newvalues[1];            
        }
        return new float[]{newmin,newmax};
    }
    
   /**
     * Sets a new value at the specified genomic position. A call to this method
     * will affect all sequences in this dataset which spans the given position 
     * @param chrom the chromsome
     * @param position The genomic position
     * @param value The new value at this position
     * @return FALSE if position is outside bounds of sequence else true
     */       
    public boolean setValueAtGenomicPosition(String chrom, int position, float value) {
        boolean result=false;
        for (FeatureSequenceData sequence:storage) {
            if (!(sequence.getChromosome().equals(chrom))) continue;
            result=result || ((NumericSequenceData)sequence).setValueAtGenomicPosition(position, value);
        }
        return result;
    }    
    
    /** 
     * Returns a deep copy of this object (but without listeners)
     */
    @Override
    public NumericDataset clone() {
        NumericDataset newdataset=new NumericDataset(datasetName); // creates a new dataset with the same name and empty storage
        newdataset.datasetDescription=this.datasetDescription;
        newdataset.min=this.min;
        newdataset.max=this.max;
        newdataset.baseline=this.baseline;
        int size=getNumberofSequences();
        for (int i=0;i<size;i++) {
            NumericSequenceData seq=(NumericSequenceData)getSequenceByIndex(i);
            newdataset.addSequence((NumericSequenceData)seq.clone());
        }
        return newdataset;
    }      

    @Override
    public void importData(Data source) throws ClassCastException {
        super.importData(source); // this will also import the child sequences
        this.max=((NumericDataset)source).max;
        this.min=((NumericDataset)source).min;
        this.baseline=((NumericDataset)source).baseline;
        //notifyListenersOfDataUpdate(); 
    }
    

    public static String getType() {return typedescription;}

    @Override
    public String getTypeDescription() {return typedescription;}   

    @Override
    public boolean containsSameData(Data data) {
        if (data==null || !(data instanceof NumericDataset)) return false;
        NumericDataset other=(NumericDataset)data;
        if (this.min!=other.min) return false;
        if (this.max!=other.max) return false;
        if (this.baseline!=other.baseline) return false;
        for (FeatureSequenceData sequence:getAllSequences()) {
            if (!sequence.containsSameData(other.getSequenceByName(sequence.getName()))) return false;
        }
        return false;
    }

}
