/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import java.util.HashMap;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.protocol.ParseError;

/**
 * This class represents 'vector' objects that can hold a single and separate 
 * numeric value for each sequence. To get a specific value the user supplies the
 * name of the sequence in question. If the user do not supply a sequence name or
 * if no value is registered for that sequence the value returned will be a preset 
 * 'default' value.
 * @author kjetikl
 */
public class NumericMapVariable extends VariableData {
    private String name; 
    private HashMap<String,Float> values=new HashMap<String,Float>();
    private static String typedescription="Numeric Map Variable";
    private Float defaultvalue=new Float(0);
    
    public NumericMapVariable(String name, float defaultvalue) {
        this.name=name;
        this.defaultvalue=new Float(defaultvalue);
    }
    
    public NumericMapVariable(String name, HashMap<String,Float>values, float defaultvalue) {
        this.name=name;
        this.values=values;
        this.defaultvalue=new Float(defaultvalue);
    }
    
    @Override
    public String getName() {return name;}

    @Override
    public void rename(String newname) {
        this.name=newname;
    }     
    
    public void setDefaultValue(float defaultvalue) {
        this.defaultvalue=new Float(defaultvalue);
    }
    
    
    @Override
    public Float getValue() {return defaultvalue;}
    
    @Override
    public String getValueAsParameterString() {
        String string="";
        for (String seq:values.keySet()) {
            string=","+seq+"="+values.get(seq).toString();
        }
        string=string.substring(1);
        return string;
    }
    
    @Override
    public String output() {
        String string="";
        for (String seq:values.keySet()) {
            string+=seq+"="+values.get(seq).toString()+"\n";
        }
        return string;
    }
    
    /**
     * Sets a scalar numeric value to be associated with the given sequence
     * @param sequencename
     * @param value
     */
    public void setValue(String sequencename, float value) {
        values.put(sequencename, value);
    }    
    
    /** 
     * Returns a scalar numeric value associated with the given sequence
     * or a default value if no value is registered for the sequence
     */
    public Float getValue(String sequencename) {
        Float val=values.get(sequencename);
        if (val==null) return defaultvalue;
        else return val;
    }
    
    /** 
     * Returns true if this NumericMapVariable contains an explicitly set
     * value for the given sequence (that is, when asking for the value for
     * the given sequence with getValue(sequencename) it will not just return
     * the default value
     */
    public boolean hasValueForSequence(String sequencename) {
        Float val=values.get(sequencename);
        return (val!=null);
    }
    
    public HashMap<String,Float> getValues() {return values;}
    
    @Override
    public void importData(Data source) throws ClassCastException {
        NumericMapVariable datasource=(NumericMapVariable)source;
        this.name=datasource.name;
        this.defaultvalue=datasource.defaultvalue;   
        this.values.clear();
        for (String key:datasource.values.keySet()) {
            values.put(key, datasource.values.get(key));
        }
        //notifyListenersOfDataUpdate(); 
    }

    /**
     * Returns true if this NumericMapVariable equals the other given NumericMapVariable
     * @param other
     * @return
     */
    @Override
    public boolean containsSameData(Data data) {
        if (data==null || !(data instanceof NumericMapVariable)) return false;
        NumericMapVariable other=(NumericMapVariable)data;
        if (((float)this.defaultvalue)!=((float)other.defaultvalue)) return false;
        for (String key:values.keySet()) {
            if (!other.hasValueForSequence(key)) return false;
            if (((float)values.get(key))!=((float)other.values.get(key))) return false;
        }       
        return true;
    }    
    
    @Override
    public NumericMapVariable clone() {
        HashMap<String,Float> newvalues=new HashMap<String,Float>();
        for (String key:values.keySet()) {
            newvalues.put(key, values.get(key));
        }        
        NumericMapVariable newdata=new NumericMapVariable(name, newvalues, defaultvalue);
        return newdata;
    }      
    

    public static String getType() {return typedescription;}

    @Override
    public String getTypeDescription() {return typedescription;}

    @Override
    public void inputFromPlain(ArrayList<String> input, PriorsEditorEngine engine) throws ParseError {
        values.clear();
        java.util.regex.Pattern pattern=java.util.regex.Pattern.compile("(\\S+)\\s*=\\s*(\\S+)");
        for (String line:input) {
            line=line.trim();
            if (line.startsWith("#") || line.isEmpty()) continue;
            java.util.regex.Matcher matcher=pattern.matcher(line);
            if (matcher.matches()) {
                String sequenceName=matcher.group(1);
                String value=matcher.group(2);
                Data data=engine.getDataItem(sequenceName);
                if (data==null) throw new ParseError("No such sequence: "+line);
                else if (!(data instanceof Sequence)) throw new ParseError("'"+line+"' is not a Sequence object");
                else {
                      try {
                        Float newvalue=Float.parseFloat(value);
                        setValue(sequenceName, newvalue);
                    } catch (NumberFormatException nfe) {throw new ParseError("Unable to parse expected numerical input in PLAIN format for "+getName()+": "+nfe.getMessage());}
                }
            } // end: matcher.matches()
        } // end: for each input line
    } // end: inputFromPlain

}
