/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.Arrays;

/**
 * This class holds feature data for a single sequence from a specified
 * region in a genome. The type of data contained in these objects are "wiggle data"
 * that specify a different value for each position along the sequence.
 * 
 * The data in this object is stored as an array of floating-point values, with
 * on entry for each position.
 * 
 * @author kjetikl
 */
public class NumericSequenceData extends FeatureSequenceData implements Cloneable {
    private float[] numericdata; //  The data should come from the direct strand so that numericdata[0] corresponds to the smaller genomic coordinate
    private float min=0; // these are only for local use. Query the parent to find dataset min/max/baseline values
    private float max=1;  // these are only for local use. Query the parent to find dataset min/max/baseline values
    private float baseline=0;  // these are only for local use. Query the parent to find dataset min/max/baseline values

    private static String typedescription="Numeric data";    
    
    /**
     * Constructs a new NumericSequenceData object from the supplied data
     * 
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param TES
     * @param orientation
     * @param sequencedata
     */
    public NumericSequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS, Integer TES, int orientation, float[] numerictrack){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             this.numericdata=numerictrack;
             for (int i=0;i<numerictrack.length;i++) {
                 if (numerictrack[i]>max) max=numerictrack[i];
                 else if (numerictrack[i]<min) min=numerictrack[i];
             }
    }
    
    /** Resets the allowed min and max values based on the current data
     * the range will include the baseline
     * @return a float array with two components [0]=>min, [1]=>max
     */
    public float[] updateAllowedMinMaxValuesFromDataRange() {
       min=getBaselineValue();
       max=getBaselineValue();
       for (int i=0;i<numericdata.length;i++) {
         if (numericdata[i]>max) max=numericdata[i];
         else if (numericdata[i]<min) min=numericdata[i];
       }       
       return new float[]{min,max};         
    }
    
    /** Returns min and max values based on the current data
     * @return a float array with two components [0]=>min, [1]=>max
     */
    public float[] getMinMaxFromData() {
       float datamin=Float.MAX_VALUE;
       float datamax=-Float.MAX_VALUE;
       for (int i=0;i<numericdata.length;i++) {
         if (numericdata[i]>datamax) datamax=numericdata[i];
         else if (numericdata[i]<datamin) datamin=numericdata[i];
       }       
       return new float[]{datamin,datamax};         
    }
    
    /**
     * Constructs a new "empty" NumericSequenceData object 
     * 
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param orientation
     */
    public NumericSequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS, Integer TES, int orientation, float defaultvalue){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             this.numericdata=new float[endPosition-startPosition+1];
             for (int i=0;i<numericdata.length;i++) numericdata[i]=defaultvalue;
             //for (int i=0;i<numericdata.length;i++) numericdata[i]=(float)Math.random();
    }
    
    /**
     * Creates a new initially "empty" NumericSequenceData object based on the given Sequence template
     * @param sequence
     */
    public NumericSequenceData(Sequence sequence, float defaultvalue) {
             setSequenceName(sequence.getName());
             setGeneName(sequence.getGeneName());
             setChromosome(sequence.getChromosome());
             setRegionStart(sequence.getRegionStart());
             setRegionEnd(sequence.getRegionEnd());
             setStrandOrientation(sequence.getStrandOrientation());
             setTSS(sequence.getTSS());
             setTES(sequence.getTES());
             this.numericdata=new float[endPosition-startPosition+1];
             for (int i=0;i<numericdata.length;i++) numericdata[i]=defaultvalue;        
    }
    
    
    @Override
    public Object getValue() {return numericdata;}
    
    @Override
     public Object getValueInGenomicInterval(int start, int end) {
        if (end<getRegionStart() || start>getRegionEnd() || end<start) return null;
        if (start<getRegionStart()) start=getRegionStart();
        if (end>getRegionEnd()) end=getRegionEnd();
        int length=end-start+1;
        int relativeStart=getRelativePositionFromGenomic(start);
        float[] subset=new float[end-start+1];
        for (int i=0;i<length;i++) {
            subset[i]=numericdata[relativeStart+i];
        }
        return subset;
    } 
    
    /**
     * Returns the data value at the specified position relative to the
     * start of the sequence (if position==0 it returns the value of the first 
     * position in the sequence.) If the position is outside the boundaries of the 
     * sequence it returns null
     * 
     * @return the data value at the specified position, or null
     */       
    public Float getValueAtRelativePosition(int position) {
        if (position<0 || position>=numericdata.length) return null;
        else return new Float(numericdata[position]);
    }

   /**
     * Returns the DNA base at the specified genomic position.
     * If the position is outside the boundaries of the sequence it returns null.
     * @param chromosome    
     * @param position
     * @return the DNA base at the specified position, or null
     */       
    public Float getValueAtGenomicPosition(String chromosome, int position) {
        if (!chromosome.equalsIgnoreCase(this.chromosome)) return null;
        if (position<this.startPosition || position>this.endPosition) return null;
        return new Float(numericdata[position-this.startPosition]);
    }

   /**
     * Returns the DNA base at the specified genomic position for the chromosome of this gene.
     * If the position is outside the boundaries of the sequence it returns null.
     * @param position
     * @return the DNA base at the specified position, or null
     */       
    @Override
    public Float getValueAtGenomicPosition(int position) {
        if (position<this.startPosition || position>this.endPosition) return null;
        return new Float(numericdata[position-this.startPosition]);
    }

   /**
     * Sets a new value at the specified genomic position
     * @param position The genomic position
     * @param value The new value at this position
     * @return FALSE if position is outside bounds of sequence else true
     */       
    public boolean setValueAtGenomicPosition(int position, float value) {
        if (position<this.startPosition || position>this.endPosition) return false;
        numericdata[position-this.startPosition]=value;
        if (value>max) max=value;
        if (value<min) min=value;
        if (parent!=null) {
            if (max>((NumericDataset)parent).getMaxAllowedValue()) ((NumericDataset)parent).setMaxAllowedValue(max);
            if (min<((NumericDataset)parent).getMinAllowedValue()) ((NumericDataset)parent).setMinAllowedValue(min);
        }
        return true;
    }
    
   /**
     * Sets a new value at the specified genomic position
     * @param chrom the chromsome
     * @param position The genomic position
     * @param value The new value at this position
     * @return FALSE if position is outside bounds of sequence else true
     */       
    public boolean setValueAtGenomicPosition(String chrom, int position, float value) {
        if (chrom==null || chrom.equals(this.chromosome)) return false;
        if (position<this.startPosition || position>this.endPosition) return false;
        numericdata[position-this.startPosition]=value;
        if (value>max) max=value;
        if (value<min) min=value;
        if (parent!=null) {
            if (max>((NumericDataset)parent).getMaxAllowedValue()) ((NumericDataset)parent).setMaxAllowedValue(max);
            if (min<((NumericDataset)parent).getMinAllowedValue()) ((NumericDataset)parent).setMinAllowedValue(min);
        }
        return true;
    }

   /**
     * Sets a new value at the specified position relative to the start of the sequence
     * @param position The genomic position
     * @param value The new value at this position
     * @return FALSE if position is outside bounds of sequence else true
     */       
    public boolean setValueAtRelativePosition(int position, float value) {
        if (position<0 || position>=numericdata.length) return false;
        numericdata[position]=value;
        if (value>max) max=value;
        if (value<min) min=value;
        if (parent!=null) {
            if (max>((NumericDataset)parent).getMaxAllowedValue()) ((NumericDataset)parent).setMaxAllowedValue(max);
            if (min<((NumericDataset)parent).getMinAllowedValue()) ((NumericDataset)parent).setMinAllowedValue(min);
        }
        return true;
    }

   
    @Override
    public void setParent(FeatureDataset parent) {
        this.parent=parent;
        if (parent==null) return;
        if (max>((NumericDataset)parent).getMaxAllowedValue()) ((NumericDataset)parent).setMaxAllowedValue(max);
        if (min<((NumericDataset)parent).getMinAllowedValue()) ((NumericDataset)parent).setMinAllowedValue(min);
    }    
    
   /**
     * Returns the maximum value allowed for this dataset
     * @return the maximum allowed value
     */       
    public Float getMaxAllowedValue() {
        if (parent==null) return max;
        else return ((NumericDataset)parent).getMaxAllowedValue();
    }

   /**
     * Returns the minimum value allowed for this dataset
     * @return the minimum allowed value
     */       
    public Float getMinAllowedValue() {
        if (parent==null) return min;
        else return ((NumericDataset)parent).getMinAllowedValue();
    }


   /**
     * Returns the baseline value for this dataset
     * @return the baseline value
     */       
    public Float getBaselineValue() {
        if (parent==null) return baseline;
        else return ((NumericDataset)parent).getBaselineValue();
    }
        
    public void setData(float[] data) {
        numericdata=data;
        for (int i=0;i<numericdata.length;i++) {
          if (numericdata[i]>max) max=numericdata[i];
          else if (numericdata[i]<min) min=numericdata[i];
        }  
        if (parent!=null) {
            if (max>((NumericDataset)parent).getMaxAllowedValue()) ((NumericDataset)parent).setMaxAllowedValue(max);
            if (min<((NumericDataset)parent).getMinAllowedValue()) ((NumericDataset)parent).setMinAllowedValue(min);
        }
    }
      
    public float[] getData() {
       return numericdata;   
    }
    
    /**
     * Returns the smallest value in this Numeric sequence within the specified interval
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getMinValueInInterval(int start, int end) {
        float minval=Float.MAX_VALUE;
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        for (int i=start;i<=end;i++) {
            if (numericdata[i]<minval) minval=numericdata[i];
        }
        return minval;
    }
    
    /**
     * Returns the largest value in this Numeric sequence within the specified interval
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getMaxValueInInterval(int start, int end) {
        float maxval=-Float.MAX_VALUE; // Note: Float.MIN_VALUE is the smallest POSITIVE number!
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        for (int i=start;i<=end;i++) {
            if (numericdata[i]>maxval) maxval=numericdata[i];
        }
        return maxval;
    }
    
    /**
     * Returns the average value within the specified interval in this Numeric sequence 
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getAverageValueInInterval(int start, int end) {
        float avg=0;
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        float length=end-start+1;
        for (int i=start;i<=end;i++) {
            avg+=numericdata[i];
        }
        avg=avg/length;
        return avg;
    }

    /**
     * Returns the sum of values within the specified interval in this Numeric sequence 
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getSumValueInInterval(int start, int end) {
        float sum=0;
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        for (int i=start;i<=end;i++) {
            sum+=numericdata[i];
        }
        return sum;
    }
    /**
     * Returns the product of values within the specified interval in this Numeric sequence 
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getProductValueInInterval(int start, int end) {
        float product=1.0f;
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        for (int i=start;i<=end;i++) {
            product*=numericdata[i];
        }
        return product;
    }

    /**
     * Returns the median value within the specified interval in this Numeric sequence 
     * @param start The start position of the interval (offset from beginning of sequence, i.e. relative position)
     * @param end The end position of the interval (offset from beginning of sequence, i.e. relative position)
     * @return
     */
    public float getMedianValueInInterval(int start, int end) {
        float median=0;
        if (start<0) start=0;
        if (end>numericdata.length-1) end=numericdata.length-1;
        if (end<start) {int end2=end;end=start;start=end2;} // swap
        int length=end-start+1;
        float[] buffer=new float[length];  
        int j=0;
        for (int i=start;i<=end;i++) {
            buffer[j]=numericdata[i];
            j++;
        }
        Arrays.sort(buffer);
        if (length%2==0) { // even number of values. Note that the length can NOT be 0!
          int index=length/2;
          median=(buffer[index-1]+buffer[index])/2;
        } else { // odd number of values
          median=buffer[(int)length/2];  
        }
        return median;
    }

    
    
    /**
     * Returns a deep copy of this NumericSequenceData object
     */
    @Override
    public NumericSequenceData clone() {
        float[] newvalues=new float[numericdata.length];
        for (int i=0;i<numericdata.length;i++) {
            newvalues[i]=numericdata[i];
        }        
        NumericSequenceData newdata=new NumericSequenceData(sequenceName, associatedGeneName, chromosome, startPosition, endPosition, TSS, TES, orientation,newvalues);
        newdata.max=this.max;
        newdata.min=this.min;
        newdata.baseline=this.baseline;
        newdata.parent=this.parent;
        return newdata;
    }    

    
    @Override
    public void importData(Data source) throws ClassCastException {
        super.importData(source);
        this.max=((NumericSequenceData)source).max;
        this.min=((NumericSequenceData)source).min;
        this.baseline=((NumericSequenceData)source).baseline;
        setData(((NumericSequenceData)source).numericdata);
        //notifyListenersOfDataUpdate(); 
    }
    

    public static String getType() {return typedescription;}

    @Override
    public String getTypeDescription() {return typedescription;}

    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof NumericSequenceData)) return false;
        if (!super.containsSameData(other)) return false;
        if (this.max!=((NumericSequenceData)other).max) return false;
        if (this.min!=((NumericSequenceData)other).min) return false;
        if (this.baseline!=((NumericSequenceData)other).baseline) return false;
        return (Arrays.equals(this.numericdata,((NumericSequenceData)other).numericdata));
    }
}
