/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

/**
 * This class basically just provides a predefined lookup-table and associated methods
 * to convert between taxonomy identifier numbers and names (common and latin) for different
 * organisms. If references to organisms are stored as integer (taxonomy IDs) this class can
 * be used to lookup corresponding names
 * 
 * @author Kjetil Klepper
 */
public class Organism {
    public static int HUMAN=9606;
    public static int MOUSE=10090;
    public static int RAT=10116;
    public static int FRUITFLY=7227;
    public static int ZEBRAFISH=7955;
    public static int ARABIDOPSIS=3702;
    public static int FROG=8364; // Xenopus laevis tropicalis
    public static int MAIZE=4577;
    public static int COW=9913;
    public static int CHICKEN=9031;
    public static int YEAST=4932;
    
    /*
     * the predefined Object array contains info on all organisms currently "supported"
     * Each row in the array contains the following information about a single organism:
     * [0] => NCBI taxonomy ID for that organism
     * [1] => A common english name
     * [2] => The latin name
     * [3] => A String-array listing supported genome builds (or null if no builds are supported) 
     * [4] => A String a string specifying the 'clade': mammal, vertebrate, insect or other) 
     */
    private static Object[][] predefined={
        {new Integer(HUMAN),"Human","Homo sapiens",new String[]{"hg18","hg19"},"mammal"},
        {new Integer(MOUSE),"Mouse","Mus musculus",new String[]{"mm9","mm8","mm7"},"mammal"},
        {new Integer(RAT),"Rat","Rattus norvegicus",new String[]{"rn4"},"mammal"},
        {new Integer(FRUITFLY),"Fruit fly","Drosophila melanogaster",new String[]{"dm3"},"insect"},
        {new Integer(ZEBRAFISH),"Zebrafish","Danio rerio",new String[]{"danRer6","danRer5"},"vertebrate"},
     //   {new Integer(ARABIDOPSIS),"Thale cress","Arabidopsis thaliana",null,"other"},       
        {new Integer(FROG),"Frog","Xenopus tropicalis",new String[]{"xenTro2"},"vertebrate"},       
     //   {new Integer(MAIZE),"Maize","Zea mays",null,"other"},       
        {new Integer(COW),"Cow","Bos taurus",new String[]{"bosTau4"},"mammal"},      
        {new Integer(CHICKEN),"Chicken","Gallus gallus",new String[]{"galGal3"},"vertebrate"},      
        {new Integer(YEAST),"Yeast","Saccharomyces cerevisiae",new String[]{"sacCer2"},"other"}      
    };
       
    
    /**
     * Returns the NCBI taxonomy ID for the organism with the specified name (common or latin)
     * (if it is known to the system, or else 0)
     */
    public static int getTaxonomyID(String name) {
        name=name.toLowerCase();
        for (Object[] values:predefined) { // first try exact matches
            if (((String)values[1]).toLowerCase().equals(name)) return ((Integer)values[0]).intValue();
            if (((String)values[2]).toLowerCase().equals(name)) return ((Integer)values[0]).intValue();
        }
        for (Object[] values:predefined) {// then see if part of the name matches
            if (((String)values[1]).toLowerCase().contains(name)) return ((Integer)values[0]).intValue();
            if (((String)values[2]).toLowerCase().contains(name)) return ((Integer)values[0]).intValue();
        }               
        return 0;
    }
    
    /**
     * Returns the common name for the organism associated with the given taxonomy ID
     * If the taxonomy ID is not known to the system the value "NCBI:[ID]" will be returned instead
     * unless ID==0, in which case the value "UNKNOWN" will be returned
     * @param id
     * @return
     */
    public static String getCommonName(int id) {
        if (id==0) return "UNKNOWN";
        for (Object[] values:predefined) {
            if (((Integer)values[0]).intValue()==id) return (String)values[1];
        }
        return "NCBI:"+id; 
    }
    
    /**
     * Returns the latin name for the organism associated with the given taxonomy ID
     * (if it is known to the system, or null)
     * @param id
     * @return
     */
    public static String getLatinName(int id) {
        for (Object[] values:predefined) {
            if (((Integer)values[0]).intValue()==id) return (String)values[2];
        }
        return "NCBI:"+id;  
    }
    
    /**
     * Returns a string containing both the common and latin name for the organism 
     * associated with the given taxonomy ID (if it is known to the system, or null)
     * The format is "common (latin)", e.g. "Human (Homo sapiens)"
     * 
     * @param id
     * @return
     */
    public static String getCombinedName(int id) {
        for (Object[] values:predefined) {
            if (((Integer)values[0]).intValue()==id) return (String)values[1]+" ("+(String)values[2]+")";
        }
        return "NCBI:"+id; 
    }
    
    /**
     * Returns a String array containing the common names of all organisms known to the engine
     * @return
     */
    public static String[] getSupportedOrganisms() {
        String[] results=new String[predefined.length];
        for (int i=0;i<predefined.length;i++) {
            results[i]=(String)predefined[i][1];   
        }
        return results;
    }
    
    /**
     * Returns an Integer array containing the organsm IDs of all organisms known to the engine
     * @return
     */
    public static Integer[] getSupportedOrganismIDs() {
        Integer[] results=new Integer[predefined.length];
        for (int i=0;i<predefined.length;i++) {
            results[i]=(Integer)predefined[i][0];   
        }
        return results;
    }
    
    /**
     * Returns a String array containing the supported genome builds for the given organism
     * or null if no genome builds are known
     * @return
     */    
    public static String[] getSupportedGenomeBuilds(int organism) {
        for (Object[] values:predefined) { 
            if (((Integer)values[0])==organism) return (String[])values[3];
        } 
        return null;
    }
    

    /**
     * Returns true if the given genome build is supported
     */    
    public static boolean isGenomeBuildSupported(String genomebuild) {
        for (Object[] values:predefined) { 
            if (values[3]==null) continue;
            else {
                String[] builds=(String[])values[3];
                for (String build:builds) {
                    if (build.equals(genomebuild)) return true;
                }
            }
        } 
        return false;
    }   
    
    /**
     * Returns the organism ID corresponding to the genomebuild 
     * if the build is supported by the system. If the build is unknown
     * the value 0 is returned
     */    
    public static int getOrganismForGenomeBuild(String genomebuild) {
        for (Object[] values:predefined) { 
            if (values[3]==null) continue;
            else {
                String[] builds=(String[])values[3];
                for (String build:builds) {
                    if (build.equals(genomebuild)) return (Integer)values[0];
                }
            }
        } 
        return 0;
    }    
    
    /**
     * Returns the Clade for the organism with the specified taxonomy ID
     * Returns the value "Unknown" if either the organism is unknown or 
     * the clade for the organism is unknown
     */    
    public static String getCladeForOrganism(int organism) {
        for (Object[] values:predefined) { 
            if ((Integer)values[0]==organism) {
                 String clade=(String)values[4];
                 if (clade==null) return "Unknown";
                 else return clade;
            }
        }
        return "Unknown";
    }    
    
}
