/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.util.ArrayList;
import javax.swing.text.AbstractDocument;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.PlainDocument;
import priorseditor.engine.PriorsEditorEngine;


/**
 * This Data subclass represents output objects made for instance
 * by the "output" operations.
 * 
 * @author kjetikl
 */
public class OutputData extends Data {
    protected String name="Output";
    protected static final String typedescription="Output";
    protected AbstractDocument document;
    protected String dataformat=null; // a string describing the name of the dataformat used when outputting to this data object. If multiple formats are used the special symbol "*" should be used to denote this
    protected transient String filename=null; // associated file name (full path). Although perhaps 
    protected boolean dirtyflag=false; // this flag should be set to true if the object has been updated since the last save
    protected boolean showAsHTML=false;
    protected ArrayList<String> dependencies=null; // lists the (full path names of) files that are associated with this output object (e.g. external image files). The files are usually temp-files and will be copied and renamed on save
    
    @Override
    public String getName() {
        return name;
    }
    

    public static String getType() {
        return typedescription;
    }
    
    @Override
    public String getTypeDescription() {return typedescription;}   
        
    @Override
    public Object getValue() {
        return document;
    }
    
    @Override
    public String getValueAsParameterString() {return document.toString();}
    
    /** Returns the Document that contains the text for this output object */
    public AbstractDocument getDocument() {
        return document;
    }

    /** 
     * Returns the name of the Data Format used for the text in this output object 
     * If there is no text, the format will be NULL. If the object contains text in 
     * several different formats, the special symbol "*" will be returned to denote this.
     */
    public String getDataFormat() {
        return dataformat;
    }
    
    
    /** updates the DataFormat (or formats) used for the text in this output object */
    private void setDataFormat(String dataFormat) {
        if (dataformat==null) dataformat=dataFormat;
        else if (!dataformat.equals(dataFormat)) dataformat="*";          
    }
    
    /** Sets the name of a file to be associated with this Output object */
    public void setFileName(String filename) {
        this.filename=filename;      
    }
    
    /** Gets the name of the file currently associated with this Output object */
    public String getFileName() {
        return filename;      
    }
    
    /** Set this to true if the output in this object containg HTML tags and should be displayed formatted */
    public void setShowAsHTML(boolean flag) {
        showAsHTML=flag;
    }
    
    /** Returns TRUE if this OutputData object contains HTML tags and should be displayed formatted */
    public boolean isHTMLformatted() {
        return showAsHTML;
    }
    
    /** Returns true if the Output object has been updated/changed since the last saved version */
    public boolean isDirty() {
        return dirtyflag;
    } 
    
    @Override
    @SuppressWarnings("unchecked")
    public void importData(Data source) throws ClassCastException {
        this.name=((OutputData)source).name;
        this.filename=((OutputData)source).filename;
        this.dirtyflag=((OutputData)source).dirtyflag;
        this.dataformat=((OutputData)source).dataformat;
        this.showAsHTML=((OutputData)source).showAsHTML;
        if (((OutputData)source).dependencies!=null) this.dependencies=(ArrayList<String>)((OutputData)source).dependencies.clone();
        else this.dependencies=null;
        try { // copy only the text of the document - no other properties! (such as DocumentListeners)
            document.remove(0, document.getLength()); // clear current content
            Document sourcedocument=((OutputData)source).document;
            String fullText=sourcedocument.getText(0, sourcedocument.getLength());
            document.insertString(0, fullText, null);
        } catch (BadLocationException e) {}
        //notifyListenersOfDataUpdate(); 
    }


    @Override
    public void rename(String newName) {
        this.name=newName;
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public OutputData clone() {  
        PlainDocument doc=new PlainDocument();        
        try {
            String fullText=document.getText(0, document.getLength());
            doc.insertString(0, fullText, null);
        } catch (BadLocationException e) {}
        OutputData newdata=new OutputData(name, doc,dataformat);
        newdata.filename=filename;
        newdata.dirtyflag=dirtyflag;
        newdata.showAsHTML=showAsHTML;
        if (dependencies!=null) newdata.dependencies=(ArrayList<String>)dependencies.clone();
        else newdata.dependencies=null;
        return newdata;
    }   
    
    @Override
    public String output() {
        try {
           return document.getText(0, document.getLength());
        } catch (Exception e) {return "Error!";}
    } 
    
   /**
    * Erases the contents of this output object
    * @return true if the clear operation was successful
    */
    public boolean clear() {
        try {
            document.remove(0, document.getLength());
            dataformat=null;
            dirtyflag=true;
            notifyListenersOfDataUpdate();
            return true;
        } catch (BadLocationException e) {return false;}
    }

    /**
     * Appends the given text to the end of this OutputData object
     * @param text The text to be appended at the end
     * @param dataformat The name of the DataFormat used for this text
     * @return true if the append operation was successful
     */
    public boolean append(String text, String dataFormat) {       
        try {
            document.insertString(document.getLength(), text, null);
            setDataFormat(dataFormat);
            dirtyflag=true;
            notifyListenersOfDataUpdate();
            return true;
        } catch (BadLocationException e) {return false;}
    }
    
     /**
     * Creates a new OutputData object with default name "output"
     * @param name
     * @param document
     */
    public OutputData() {
        this.name="output";
        this.document=new PlainDocument();
        this.dataformat=null;
    }   
    
    /**
     * Creates a new OutputData object with the given name
     * @param name
     * @param document
     */
    public OutputData(String name) {
        this.name=name;
        this.document=new PlainDocument();
        this.dataformat=null;
    }
    
    
    /**
     * Creates a new OutputData object with the given name which wraps
     * the given Document
     * @param name
     * @param document
     */
    public OutputData(String name, PlainDocument document, String dataformat) {
        this.name=name;
        this.document=document;
        this.dataformat=dataformat;
    }
    
    /** Created a dependent file and adds it to this output object. 
     *  Dependency files are files such as e.g. image files that are associated with this object and should
     *  be saved together with the object when saving to file
     *  @param @paramengine
     */
    public File createDependentFile(PriorsEditorEngine engine, String suffix) {
       if (!suffix.startsWith(".")) suffix="."+suffix;
       File file=engine.createTempFile(suffix);
       String depfilename=file.getAbsolutePath();
       if (dependencies==null) dependencies=new ArrayList<String>();
       dependencies.add(depfilename);
       return file;
    }
    
    @Override
    public Data getSerializableContent() {
        return this;
        // return clone(); // return clone if we allow listeners to be added to the document
    }
    
    
    /** Saves the OutputData contents to a temp file created by the engine
     * @return a reference to the new temp-file containing the document
     */
    public File saveToTempFile(PriorsEditorEngine engine) {
        File file=engine.createTempFile();
        PrintWriter outputStream=null;
        try {
            outputStream=new PrintWriter(new FileWriter(file));
            Element root=document.getDefaultRootElement();
            int linecount=root.getElementCount();
            for (int i=0;i<linecount;i++) { 
                  Element e=root.getElement(i);  
                  String line=document.getText(e.getStartOffset(),e.getEndOffset()-e.getStartOffset());
                  line=line.trim();
                  if (i<linecount-1 || !line.isEmpty()) outputStream.println(line);
            }                    
        } catch (Exception bad) { 
            return null;
        } finally {
            try {if (outputStream!=null) outputStream.close();} catch (Exception ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing PrintWriter in OutputData.saveToFile(): "+ioe.getMessage());}
        }
        return file;
    }
    
    
    
   /**
     * Saves the contents of this Output object to a file with the given name
     * (and associates this object with the given filename from now on)
     * @param filename The full pathname of the file to save in
     * Note: this method should preferably NOT be called on the EDT
     */
    public void saveToFile(String filename) throws IOException {
        File file=new File(filename);
        saveToFile(file,true);
    }

    /**
     * Saves the contents of this Output object to the given file
     * @param file The target file (note that depending on the contents of this OutputData object other dependent files can be created as well)
     * @param updateSaveStatus If TRUE this OutputData object will be marked as "saved" (not dirty) and associated with the given file so subsequent calls to getFileName() will return the name of this file
     * Note: this method should preferably NOT be called on the EDT
     */
    public void saveToFile(File file, boolean updateSaveStatus) throws IOException {
        String dependencyPrefix=file.getName();
        int suffixpos=dependencyPrefix.lastIndexOf('.');
        if (suffixpos>=0) {
            dependencyPrefix=dependencyPrefix.substring(0,suffixpos);
        }
        Element root=document.getDefaultRootElement();
        int linecount=root.getElementCount();
        PrintWriter outputStream=null;
        try {
            outputStream=new PrintWriter(new FileWriter(file));
            for (int i=0;i<linecount;i++) {
                  Element e=root.getElement(i);
                  String line=document.getText(e.getStartOffset(),e.getEndOffset()-e.getStartOffset());
                  line=line.trim();
                  if (dependencies!=null) line=replaceDependencies(line, dependencyPrefix);
                  if (i<linecount-1 || !line.isEmpty()) outputStream.println(line);
            }
        } catch (BadLocationException bad) {
            System.err.println("SYSTEM ERROR: BadLocationException in OutputData.saveToFile(): "+bad.getMessage());
            return;
        } finally {
            try {if (outputStream!=null) outputStream.close();} catch (Exception ioe) {System.err.println("SYSTEM ERROR: An error occurred when closing PrintWriter in OutputData.saveToFile(): "+ioe.getMessage());}
        }
        if (dependencies!=null) {
            for (int i=0;i<dependencies.size();i++) {
                String oldfilename=dependencies.get(i);
                File oldfile=new File(oldfilename);
                String newfilename=replaceDependencies(oldfilename, dependencyPrefix);
                File newfile=new File(file.getParent(),newfilename); // same dependency files in the same dir as 'file'
                copyFile(oldfile, newfile);
            }
        }
        if (updateSaveStatus) {
            dirtyflag=false;
            setFileName(filename);
        }
    }
 
    /** Looks for parts of the string that looks like dependent files and replaces the text with new filenames */
    private String replaceDependencies(String line, String prefix) {
         for (int i=0;i<dependencies.size();i++) {
             String oldfilename=dependencies.get(i);
             String suffix="";
             int suffixpos=oldfilename.lastIndexOf('.');
             if (suffixpos>=0) {
                 suffix=oldfilename.substring(suffixpos);
             }
             String newfilename=prefix+"_"+(i+1)+suffix;
             line=line.replace(oldfilename, newfilename);
             line=line.replace("file:///", "");
         }
         return line;
    }
    
    
    private void copyFile(File source, File target) throws IOException {
    	 BufferedInputStream bis = new BufferedInputStream(new FileInputStream(source), 4096);
         BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(target), 4096);
         int theChar;
         while ((theChar = bis.read()) != -1) {
               bos.write(theChar);
         }
         bos.close();
         bis.close();     
    }

    @Override
    public boolean containsSameData(Data data) {
        if (data==null || !(data instanceof OutputData)) return false;
        OutputData other=(OutputData)data;
        if (this.dataformat!=null && other.dataformat==null) return false;
        if (this.dataformat==null && other.dataformat!=null) return false;
        if (this.dataformat!=null && other.dataformat!=null && !this.dataformat.equals(other.dataformat)) return false;
        if (this.filename!=null && other.filename==null) return false;
        if (this.filename==null && other.filename!=null) return false;
        if (this.filename!=null && other.filename!=null && !this.filename.equals(other.filename)) return false;
        if (this.dirtyflag!=other.dirtyflag) return false;
        if (this.showAsHTML!=other.showAsHTML) return false;
        if (this.document!=null && other.document==null) return false;
        if (this.document==null && other.document!=null) return false;
        if (this.document!=null && other.document!=null && !this.document.equals(other.document)) return false;
        if (this.dependencies!=null && other.dependencies==null) return false;
        if (this.dependencies==null && other.dependencies!=null) return false;
        if (this.dependencies!=null && other.dependencies!=null && !this.dependencies.equals(other.dependencies)) return false;
        return true;
    }

}   
    
