/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Set;

/**
 *
 * @author kjetikl
 */
public class Region implements Cloneable,Serializable {
    public static final int DIRECT=1;
    public static final int REVERSE=-1;
    public static final int INDETERMINED=0;
    
    
    private int start=0;
    private int end=0;
    private RegionSequenceData parent=null;
    private String type; // e.g. Alu-repeat, NFkB-binding sites
    private int orientation;
    private float score;
    private boolean ismodule=false;
    private HashMap<String,Object> properties; //useful properties: name, score, source, orientation. All names should be lowercase
    
    /**
     * Constructs a new region located at the specied start and end positions
     * relative to the start of the parent RegionSequenceData object (that is,
     * the start and end positions are offsets).  
     * 
     * @param parent
     * @param start
     * @param end
     */
    public Region(RegionSequenceData parent, int start, int end) {
        this.parent=parent;
        this.start=start;
        this.end=end;   
        this.type="<unknown>";
        this.score=0;
        this.orientation=INDETERMINED;
    }
    /**
     * Constructs a new region located at the specied start and end positions
     * relative to the start of the parent RegionSequenceData object (that is,
     * the start and end positions are offsets).  
     * 
     * @param parent
     * @param start
     * @param end
     * @param type
     * @param score
     * @param orientation
     */
    public Region(RegionSequenceData parent, int start, int end, String type, float score, int orientation) {
        this.parent=parent;
        this.start=start;
        this.end=end;   
        this.type=type;
        this.score=score;
        this.orientation=orientation;
    }
    
    /**
     * Returns the parent data sequence of this region
     * @return RegionSequenceData
     */
    public RegionSequenceData getParent() {
        return parent;
    }   
    
     /**
     * Sets the parent data sequence of this region
     * @param newparent
     */
    public void setParent(RegionSequenceData newparent) {
        parent=newparent;
    }   
    /**
     * Returns the start position of this region relative to the parent sequence
     * @return start position
     */
    public int getRelativeStart() {
        return start;
    }
    
   /**
     * Returns the end position of this region relative to the parent sequence
     * @return end position
     */
    public int getRelativeEnd() {
        return end;
    }
    
    /**
     * Returns the start position of this region in genomic coordinates
     * @return start position
     */
    public int getGenomicStart() {
        return parent.getRegionStart()+start;
    }
    
   /**
     * Returns the end position of this region in genomic coordinates
     * @return end position
     */
    public int getGenomicEnd() {
        return parent.getRegionStart()+end;
    }
 
    
    /**
     * Sets the start position of this region relative to the parent sequence
     * @param start New start position
     */
    public void setRelativeStart(int start) {
        this.start=start;
    }
    
   /**
     * Sets the end position of this region relative to the parent sequence
     * @param end New end position
     */
    public void setRelativeEnd(int end) {
        this.end=end;
    }    
    
   /**
    * Returns the length of the region in base pairs
    * @return
    */
    public int getLength() {
        return end-start+1;
    }
    
    
    public String getType() {
        return type;
    }
    
    public void setType(String type) {
        this.type=type;      
    }
    
    public float getScore() {
        return score;
    }
    
    public void setScore(float score) {
        this.score=score;
    }
    
    public int getOrientation() {
        return orientation;
    }
    
    public void setOrientation(int orientation) {
        this.orientation=orientation;
    }
    
    /**
     * 
     * @param property The name of the property
     * @return The value of the property, if specified, else null
     */
    public Object getProperty(String property) {        
                 if (property.equalsIgnoreCase("type")) return getType();
            else if (property.equalsIgnoreCase("score")) return getScore();
            else if (property.equalsIgnoreCase("orientation")) return getOrientation();
            else {
                if (properties==null) return null;
                else return properties.get(property.toLowerCase());
        }
    }
    
    public void setProperty(String property, Object value) {
             if (property.equalsIgnoreCase("type")) setType((String)value);
        else if (property.equalsIgnoreCase("score")) setScore((Float)value);
        else if (property.equalsIgnoreCase("orientation")) setOrientation((Integer)value);
        else {
            if (properties==null) properties=new HashMap<String, Object>(2);
            properties.put(property.toLowerCase(), value);
        }
    }
    
    /**
     * Returns a list containing the names of all non-standard properties associated with this Region 
     */
    public String[] getAllPropertyNames() {
        if (properties==null) return new String[0];
        Set<String> keySet=properties.keySet();
        String[] result=new String[keySet.size()];
        int i=0;
        for (String key:keySet) {
            result[i]=key;
            i++;
        }
        return result;
    }    
    
    /**
     * Checks whether this region is located prior to another region
     * (ie. the start if this region lies before the other region on the direct strand
     * or this region ends before the other regions if they both start at the same position)
     * @param anotherRegion
     * @return true if this region is located prior to the other region
     */
    public boolean isLocatedPriorTo (Region other) {
        if (start<other.start) return true;
        else if (start==other.start && end<other.end) return true;
        else return false;
    }
    
    /**
     * Checks whether this region is located prior to another region when looking at the sequence in reverse orientation
     * (ie. the start if this region lies before the other region on the reverse strand
     * or this region ends before the other regions if they both start at the same position)
     * @param anotherRegion
     * @return true if this region is located prior to the other region on reverse strand
     */
    public boolean isLocatedPriorToOnReverseStrand (Region other) {
        if (end>other.end) return true;
        else if (end==other.end && start<other.start) return true;
        else return false;
    }
    
    /**
     * Returns a copy of this Region object
     */
    @SuppressWarnings("unchecked")
    @Override
    public Region clone() {
        Region newdata=new Region(parent, start, end, type, score, orientation);
        newdata.ismodule=ismodule;
        if (properties!=null) {
            newdata.properties=new HashMap<String, Object>(properties.size());
            for (String key:properties.keySet()) {
                Object value=properties.get(key);
                if (value instanceof Region) newdata.properties.put(key, ((Region)value).clone());
                else newdata.properties.put(key, value);
            }

        }
        return newdata;
    }
    
    @Override
    public String toString() {
        String desc = "chr"+((parent!=null)?parent.getChromosome():"?")+":"+((parent!=null)?getGenomicStart():"?")+"-"+((parent!=null)?getGenomicEnd():"?")+"    [relative="+getRelativeStart()+"-"+getRelativeEnd()+"]";
        if (type!=null) desc+="  type="+type;
        desc+="   score="+score+"  "+getPropertiesAsString();
        return desc;
    }
    
    /** Returns the registered properties as a string with key=value pairs (excluding type, score and orientation) */
    public String getPropertiesAsString() {
        String result=""; // (type!=null)?("type="+type+", "):"";
        if (properties==null) return result;
        Set<String> keys=properties.keySet();
        for (String key:keys) {
            Object value=properties.get(key);
            if (value instanceof Region) value=((Region)value).getType();
            result+=key+"="+value.toString()+", ";
        }               
        return result.substring(0,result.length()-2);
    }
    
    /** 
     * This method returns TRUE if this region is 'similar to' the region supplied
     * The definition of 'similar to' is that the two regions have the same relative location
     * to their respective parents (if any) and that all the other properties are the same
     * (Note that two regions can be 'similar' even though their parents are not the same
     * (in which case they could also be 'similar' even though they are actually located
     * at different genomic positions (albeit same relative positions))). 
     */
    public boolean isSimilarTo(Region otherregion) {
             if (this.start!=otherregion.start) return false;
        else if (this.end!=otherregion.end) return false;
        else if (!this.type.equals(otherregion.type)) return false;
        else if (this.score!=otherregion.score) return false;
        else if (this.orientation!=otherregion.orientation) return false;
        else if (this.ismodule!=otherregion.ismodule) return false;
        else if ((this.properties!=null && otherregion.properties==null) ||
                 (this.properties==null && otherregion.properties!=null) ||
                 (this.properties!=null && otherregion.properties!=null && !this.properties.equals(otherregion.properties))) return false;
        else return true;   
    }
    
}
