/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * This class holds feature data for a single sequence from a specified
 * region in a genome. The type of data contained in these objects are "region data"
 * which are discrete regions with associated values, for instance the location
 * of transcription factor binding sites or SINE repeats within the region.
 * The data itself is stored as a set of such regions that all fall within the
 * span of the genomic sequence itself. The individual regions and their
 * properties are implemented as objects of type Region
 *
 * @author kjetikl
 */
public class RegionSequenceData extends FeatureSequenceData implements Cloneable {
    private static String typedescription="Region data";

    private ArrayList<Region> regions;  // this list should always be kept sorted according to start position of regions!
                                        // if two elements have the same start position, the shortest is placed first in the list
                                        // so that the end position of the last region in the list marks the right boundary of all regions

    private int sequencesize=0; // keeps track of the length of this sequence.
                                // Unlike for DNA and Numeric Data, this type of dataset has
                                // no array that stores data for each base in a 1-to-1 relationship

    private float maxScore=0; // the maximum region score in this sequence

    /**
     * Constructs a new RegionSequenceData object from the supplied data
     *
     * @param sequenceName
     * @param geneName
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param TSS
     * @param TES
     * @param orientation
     */
    public RegionSequenceData(String sequenceName, String geneName, String chromosome, int startPosition, int endPosition, Integer TSS, Integer TES, int orientation){
             setSequenceName(sequenceName);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             this.regions=new ArrayList<Region>();
             this.sequencesize=endPosition-startPosition+1;
    }

    /**
     * Creates a new initially "empty" RegionSequenceData object based on the given Sequence template
     * @param sequence
     */
    public RegionSequenceData(Sequence sequence) {
             setSequenceName(sequence.getName());
             setGeneName(sequence.getGeneName());
             setChromosome(sequence.getChromosome());
             setRegionStart(sequence.getRegionStart());
             setRegionEnd(sequence.getRegionEnd());
             setStrandOrientation(sequence.getStrandOrientation());
             setTSS(sequence.getTSS());
             setTES(sequence.getTES());
             this.regions=new ArrayList<Region>();
             this.sequencesize=endPosition-startPosition+1;
    }

    @Override
    public Object getValue() {return regions;}

    @Override
     public Object getValueInGenomicInterval(int start, int end) {
        if (end<getRegionStart() || start>getRegionEnd() || end<start) return null;
        if (start<getRegionStart()) start=getRegionStart();
        if (end>getRegionEnd()) end=getRegionEnd();
        ArrayList<Region> list=getRegionsOverlappingGenomicInterval(start, end);
        ArrayList<Region> newList=new ArrayList<Region>();
        for (Region region:list) {
            Region newregion=region.clone();
            newregion.setParent(null);
            newList.add(newregion);
        }
        return newList;
    }

    /**
     * Adds a new Region object to this sequence.
     * The new region is added to the list in sorted order
     *
     * @param region The region to be added to the list
     */
    public void addRegion(Region region) {
        if (region.getParent()==null) region.setParent(this);
        int size=regions.size();
        float score=region.getScore();
        if (score>maxScore) setMaxScoreValue(score);
        if (size==0) {regions.add(region);}
        else if (!region.isLocatedPriorTo(regions.get(size-1))) {regions.add(region);} // add to the end
        else {
          int index=0;
          while (index<size) {
            Region nextregion=regions.get(index);
            if (region.isLocatedPriorTo(nextregion)) {regions.add(index, region);index=size;}
            index++;
          }
        }
    }

    /**
     *  This should be called if any regions have updated the start or end positions
     *  after they were added to a sequence. The regions will then be sorted in their proper order again
     */
    public void updateRegionSortOrder() {
        Collections.sort(regions, new RegionSortOrderComparator());
    }

    /**
     * Removes a region from this sequence object
     *
     * @param region The region to be removed
     */
    public void removeRegion(Region region) {
        regions.remove(region);
    }

    /**
     * Returns the number of regions associated with this sequence
     *
     * @return The number of regions
     */
    public int getNumberOfRegions() {
        return regions.size();
    }

    /**
     * Returns the region corresponding to the specified index
     * This function can be used (in combination with getNumberOfRegions)
     * to iterated through the set of Regions
     *
     * @param i The index of the Region to be retrieved
     */
    public Region getRegionByIndex(int i) {
        return (Region)regions.get(i);
    }

    /** Returns min and max region score values based on the current data
     * @return a float array with two components [0]=>min, [1]=>max
     */
    public float[] getMinMaxFromData() {
       float min=Float.MAX_VALUE;
       float max=-Float.MAX_VALUE;
       for (Region region:regions) {
         float score=region.getScore();
         if (score>max) max=score;
         else if (score<min) min=score;
       }
       return new float[]{min,max};
    }

    /**
     * Returns the data value at the specified position relative to the
     * start of the sequence (if position==0 it returns the value of the first
     * position in the sequence.) If the position is outside the boundaries of the
     * sequence it returns null.
     * The value returned is an ArrayList containing Region objects
     * (since individually specified Regions can potentially overlap)
     *
     * @return An ArrayList containing the set of regions (if any) that overlap with the specified position, or null if the position is outside the boundaries of the sequence
     */
    @Override
    public ArrayList<Region> getValueAtRelativePosition(int position) {
        if (position<0 || position>=sequencesize) return null;
        ArrayList<Region> list=new ArrayList<Region>();
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) list.add(r);
        }
        return list;
    }

    /**
     * Returns the data value at the specified genomic position. relative to the
     * If the position is outside the boundaries of the sequence it returns null.
     * The value returned is an ArrayList containing Region objects
     * (since individually specified Regions can potentially overlap)
     *
     * @param chromosome
     * @param position
     * @return An ArrayList containing the set of regions (if any) that overlap with the specified position, or null if the position is outside the boundaries of the sequence
     */
    @Override
    public ArrayList<Region> getValueAtGenomicPosition(String chromosome, int position) {
        if (!chromosome.equalsIgnoreCase(this.chromosome)) return null;
        if (position<this.startPosition || position>this.endPosition) return null;
        position-=this.startPosition; // convert genomic position to an offset!
        ArrayList<Region> list=new ArrayList<Region>();
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) list.add(r);
        }
        return list;
    }
    /**
     * Returns the data value at the specified genomic position.
     * If the position is outside the boundaries of the sequence it returns null.
     * The value returned is an ArrayList containing Region objects
     * (since individually specified Regions can potentially overlap)
     *
     * @param position
     * @return An ArrayList containing the set of regions (if any) that overlap with the specified position, or null if the position is outside the boundaries of the sequence
     */
    @Override
    public ArrayList<Region> getValueAtGenomicPosition(int position) {
        if (position<this.startPosition || position>this.endPosition) return null;
        position-=this.startPosition; // convert genomic position to an offset!
        ArrayList<Region> list=new ArrayList<Region>();
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) list.add(r);
        }
        return list;
    }

    /**
     * Returns TRUE if the position relative to sequence start lies within a region
     *
     * @param position the relative position within the sequence
     * @return TRUE if the position is within a region
     */
    public boolean isWithinRegion(int position) {
        if (position<0 || position>=sequencesize) return false;
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) return true;
        }
        return false;
    }

    /**
     * Returns the number of regions overlapping with the specified genomic position
     *
     * @param position
     * @return The number of regions at this position
     */
    public int getNumberOfRegionsAtGenomicPosition(int position) {
        int number=0;
        if (position<this.startPosition || position>this.endPosition) return 0;
        position-=this.startPosition; // convert genomic position to an offset!
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) number++;
        }
        return number;
    }

    /**
     * Returns the number of regions overlapping with the specified position relative to sequence start
     *
     * @param position
     * @return The number of regions at this position
     */
    public int getNumberOfRegionsAtRelativePosition(int position) {
        int number=0;
        if (position<0 || position>=sequencesize) return 0;
        for (Region r:regions) {
            if (position>=r.getRelativeStart() && position<=r.getRelativeEnd()) number++;
        }
        return number;
    }

    /**
     * Returns the number of regions overlapping with the specified genomic interval
     *
     * @param start
     * @param end
     * @return The number of regions overlapping this interval
     */
    public int getNumberOfRegionsOverlappingGenomicInterval(int start, int end) {
        int number=0;
        if (end<this.startPosition || start>this.endPosition) return 0;
        for (Region r:regions) {
            if (!(end<r.getGenomicStart() || start>r.getGenomicEnd())) number++;
        }
        return number;
    }

    /**
     * Returns a list containing regions overlapping with the specified genomic interval
     *
     * @param start
     * @param end
     * @return a list of regions
     */
    public ArrayList<Region> getRegionsOverlappingGenomicInterval(int start, int end) {
        ArrayList<Region> list=new ArrayList<Region>();
        if (end<this.startPosition || start>this.endPosition) return list;
        for (Region r:regions) {
            if (!(end<r.getGenomicStart() || start>r.getGenomicEnd())) list.add(r);
        }
        return list;
    }

    /**
     * Returns a list containing regions that lies fully within the specified genomic interval
     *
     * @param start
     * @param end
     * @return a list of regions
     */
    public ArrayList<Region> getRegionsWithinGenomicInterval(int start, int end) {
        ArrayList<Region> list=new ArrayList<Region>();
        if (end<this.startPosition || start>this.endPosition) return list;
        for (Region r:regions) {
            if (r.getGenomicStart()>=start && r.getGenomicEnd()<=end) list.add(r);
        }
        return list;
    }

    /**
     * Returns a list of non-overlapping regions that together span the same genomic segments
     * that the original (possibly overlapping) regions span.
     *
     * @return a list of regions
     */
    public ArrayList<Region> getCollapsedRegions() {
        ArrayList<Region> list=new ArrayList<Region>();
        if (regions.isEmpty()) return list;
        Region current=new Region(this, regions.get(0).getRelativeStart(), regions.get(0).getRelativeEnd());
        for (int i=1;i<regions.size();i++) {
            Region next=regions.get(i);
            if (next.getRelativeStart()>current.getRelativeEnd()+1) {
                list.add(current);
                current=new Region(this, next.getRelativeStart(),next.getRelativeEnd());
            } else { // overlapping or back-to-back
                current.setRelativeEnd(next.getRelativeEnd());
            }
        }
        list.add(current);
        return list;
    }

    /**
     * Returns the total number of bases that are spanned by regions within this sequence
     */
    public int getNumberOfBasesSpannedByRegions() {
        if (regions.isEmpty()) return 0;
        int count=0;
        int currentStart=regions.get(0).getRelativeStart(); int currentEnd=regions.get(0).getRelativeEnd();
        for (int i=1;i<regions.size();i++) {
            Region next=regions.get(i);
            if (next.getRelativeStart()>currentEnd+1) {
                count+=(currentEnd-currentStart+1);
                currentStart=next.getRelativeStart(); currentEnd=next.getRelativeEnd();
            } else { // overlapping or back-to-back
                currentEnd=next.getRelativeEnd();
            }
        }
        count+=(currentEnd-currentStart+1);
        return count;
    }

    /**
     * Returns the closest region lying upstream of the specified genomic position
     * or the most downstream region that overlaps the given position, or NULL if no
     * region overlaps or lies upstream of the specified position
     *
     * @param position A genomic position
     * @return a Region
     */
    public Region getClosestUpstreamRegion(int position) {
        Region candidate=null;
        int relativePos=getRelativePositionFromGenomic(position);
        if (orientation==DIRECT) {
               for (Region r:regions) {
                 if (r.getRelativeStart()>relativePos) break;
                 else candidate=r;
               }
               return candidate;
        } else {
             for (int i=regions.size()-1;i>=0;i--) {
                 Region r=regions.get(i);
                 if (r.getRelativeEnd()<relativePos) break;
                 else candidate=r;
             }
             return candidate;
        }
    }


    /**
     * Returns the closest region lying downstream of the specified genomic position
     * or the most upstream region that overlaps the given position, or NULL if no
     * region overlaps or lies downstream of the specified position
     *
     * @param position A genomic position
     * @return a Region
     */
    public Region getClosestDownstreamRegion(int position) {
        Region candidate=null;
        int relativePos=getRelativePositionFromGenomic(position);
        if (orientation==REVERSE) {
               for (Region r:regions) {
                 if (r.getRelativeStart()>relativePos) break;
                 else candidate=r;
               }
               return candidate;
        } else {
             for (int i=regions.size()-1;i>=0;i--) {
                 Region r=regions.get(i);
                 if (r.getRelativeEnd()<relativePos) break;
                 else candidate=r;
             }
             return candidate;
        }
    }

    /**
     * Returns a list of all Regions associated with this RegionSequenceData object
     * The returned ArrayList is a (shallow) clone of the internal list
     */
    @SuppressWarnings("unchecked")
    public ArrayList<Region> getAllRegions() {
        return (ArrayList<Region>)regions.clone();
    }

    /**
     * Returns list of all Regions associated with this RegionSequenceData object
     * The returned ArrayList is the original list (not a copy) and should not be
     * modified!
     */
    public ArrayList<Region> getOriginalRegions() {
        return regions;
    }

    /**
     * Returns true if this RegionSequenceData object already contains a Region
     * that is 'similar' to the one supplied (both with respect to location and other properties).
     */
    public boolean containsRegion(Region newregion) {
        for (Region region:regions) {
            if (region.isSimilarTo(newregion)) return true;
        }
        return false;
    }

    /**
     * clears all regions associated with this sequence
     *
     */
    public void clearRegions() {
        regions.clear();
    }


    @Override
    public void setParent(FeatureDataset parent) {
        this.parent=parent;
        if (maxScore>((RegionDataset)parent).getMaxScoreValue()) ((RegionDataset)parent).setMaxScoreValue(maxScore);

    }

    /** Returns TRUE if this RegionDataset contains motifs (or transcription factors binding sites),
     *  in which case the typenames of the regions should match the names of motifs registered with the engine
     */
    public boolean isMotifTrack() {
        if (parent==null) return false;
        else return ((RegionDataset)parent).isMotifTrack();
    }

    /** Returns TRUE if this RegionDataset contains modules (or composite transcription factors binding sites),
     *  in which case the typenames of the regions should match the names of modules registered with the engine
     */
    public boolean isModuleTrack() {
        if (parent==null) return false;
        else return ((RegionDataset)parent).isModuleTrack();
    }



   /**
     * Returns the maximum score value for regions in this dataset
     * @return the maximum score value
     */
    public Float getMaxScoreValue() {
        if (parent==null) return maxScore;
        else return ((RegionDataset)parent).getMaxScoreValue();
    }

   /**
     * Sets the maximum score value for this dataset
     * @param max The maximum score value for this dataset
     */
    public void setMaxScoreValue(float max) {
        this.maxScore=max;
        if (parent!=null && (maxScore>((RegionDataset)parent).getMaxScoreValue())) ((RegionDataset)parent).setMaxScoreValue(maxScore);
    }


    /**
     * Returns a deep copy of this RegionSequenceData object
     */
    @Override
    public RegionSequenceData clone() {
        RegionSequenceData newdata=new RegionSequenceData(sequenceName, associatedGeneName, chromosome, startPosition, endPosition, TSS, TES, orientation);
        int size=getNumberOfRegions();
        for (int i=0;i<size;i++) {
            Region reg=getRegionByIndex(i);
            newdata.addRegion((Region)reg.clone());
        }
        newdata.maxScore=this.maxScore;
        return newdata;
    }


    @Override
    public void importData(Data source) throws ClassCastException {
        super.importData(source);
        RegionSequenceData rsource=(RegionSequenceData)source;
        this.regions.clear();
        int size=rsource.getNumberOfRegions();
        for (int i=0;i<size;i++) {
            Region reg=rsource.getRegionByIndex(i);
            addRegion((Region)reg.clone());
        }
        this.maxScore=((RegionSequenceData)source).maxScore;
        //notifyListenersOfDataUpdate();
    }


    public static String getType() {return typedescription;}

    @Override
    public String getTypeDescription() {return typedescription;}

    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof RegionSequenceData)) return false;
        if (!super.containsSameData(other)) return false;
        return (regions.equals(((RegionSequenceData)other).regions));
    }

    private class RegionSortOrderComparator implements Comparator<Region> {
            @Override
            public int compare(Region region1, Region region2) {
                if (region1.getRelativeStart()==region2.getRelativeStart()) return 0;
                else if (region1.isLocatedPriorTo(region2)) return -1;
                else return 1;
            }
    }

}
