/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

/**
 *
 * @author kjetikl
 */
/**
 * This class represents a "template" for a genomic region
 * 
 * @author kjetikl
 */
public class Sequence extends Data {
    public static final int DIRECT  =  1;
    public static final int REVERSE = -1;
    
    protected String sequenceName=null;
    protected String associatedGeneName=null;
    protected String chromosome="?";
    protected int startPosition=0;
    protected int endPosition=0;
    protected Integer TSS=null; // Transcription Start Site
    protected Integer TES=null; // Transcription End Site
    protected int orientation=DIRECT;     
    protected int organism;
    protected String genomeBuild;

    private static String typedescription="Sequence";   
    
    /**
     * Constructs a new Sequence object from the supplied data
     * 
     * @param sequenceName
     * @param organism
     * @param chromosome
     * @param startPosition
     * @param endPosition
     * @param geneName
     * @param TSS
     * @param TES
     * @param orientation

     */
    public Sequence(String sequenceName, int organism, String genomebuild, String chromosome, int startPosition, int endPosition, String geneName, Integer TSS, Integer TES, int orientation){
             setSequenceName(sequenceName);
             setOrganism(organism);
             setGeneName(geneName);
             setChromosome(chromosome);
             setRegionStart(startPosition);
             setRegionEnd(endPosition);
             setStrandOrientation(orientation);
             setTSS(TSS);
             setTES(TES);
             setGenomeBuild(genomebuild);
    }
    
  

    @Override
    public String getName() {return getSequenceName();}

    @Override
    public Object getValue() {return this;} // should maybe change later
    
    @Override
    public String getValueAsParameterString() {
        String val=getName()+"["+getRegionAsString()+"]";
        if (orientation==DIRECT) val+="[+]"; else val+="[-]";
        if (TSS!=null) val+="["+TSS.intValue()+":"; else val+="[null:";
        if (TSS!=null) val+=TSS.intValue()+"]"; else val+="null]";
        val+="["+organism+"]";
        return val; 
    }

    /**
     * Returns the size of this sequence (the number of bases spanned by the region)
     * 
     * @return size of sequence
     */
    public int getSize() {return endPosition-startPosition+1;}
    
    /**
     * Sets the name of the sequence
     * 
     * @param sequenceName a name for this sequence 
     */
    public void setSequenceName(String sequenceName) {
        this.sequenceName=sequenceName;
    }
    
    @Override
    public void rename(String name) {
        setSequenceName(name);
    } 
    
    /**
     * Returns the name of the sequence. If the sequence has no specific name this
     * function will return a name based on the genomic region.
     * 
     * @return a name for this sequence
     */
    public String getSequenceName() {
        if (sequenceName!=null) {
            return sequenceName;
        } else {
          return chromosome+":"+startPosition+"-"+endPosition;   
        }
    }
    
    /**
     * Sets the name of the gene associated with this genomic region
     * 
     * @param geneName name of associated gene for this region
     */
    public void setGeneName(String geneName) {
        this.associatedGeneName=geneName;
    }
    
    /**
     * Returns the name of the gene associated with this genomic region
     * 
     * @return a name for this sequence
     */
    public String getGeneName() {
        return associatedGeneName;      
    }

    /**
     * Sets the genome build for this sequence. All coordinates should be interpreted
     * relative to the specified genome build
     * 
     * @param genomebuild A string specifying the genome build. Eg: 'hg18' or 'mm9'
     */
    public void setGenomeBuild(String genomebuild) {
        this.genomeBuild=genomebuild;
    }
    
    /**
     * Returns the genome build for this sequence. All coordinates should be interpreted
     * relative to this genome build
     * 
     * @return A string specifying the genome build. Eg: 'hg18' or 'mm9'
     */
    public String getGenomeBuild() {
        return genomeBuild;      
    }
        
    
    /**
     * Sets the name of the chromosome for this genomic region
     * 
     * @param chromosome chromosome name for this region
     */
    public void setChromosome(String chromosome) {
        this.chromosome=chromosome;
    }
    
    /**
     * Returns the chromosome name for this genomic region
     * 
     * @return chromosome name
     */
    public String getChromosome() {
        return chromosome;      
    }
    
    /**
     * Sets the start position within the chromosome for this genomic region
     * Regions should always be defined relative to the direct strand, so that
     * the start position of a region is always smaller than the end position. 
     * 
     * @param start the start position for this region
     */
    public void setRegionStart(int start) {
        this.startPosition=start;
    }
    
    /**
     * Returns the start position of this genomic region within the chromosome
     * (represented by the smalled genomic coordinate (regionStart <= regionEnd))
     * 
     * @return start position
     */
    public int getRegionStart() {
        return startPosition;      
    }
    
    /**
     * Sets the end position within the chromosome for this genomic region
     * Regions should always be defined relative to the direct strand, so that
     * the start position of a region is always smaller than the end position. 
     * 
     * @param end the end position for this region
     */
    public void setRegionEnd(int end) {
        this.endPosition=end;
    }
    
    /**
     * Returns the end position of this genomic region within the chromosome
     * (represented by the largest genomic coordinate (regionStart <= regionEnd))
     * 
     * @return end position
     */
    public int getRegionEnd() {
        return endPosition;      
    }
    
    /**
     * Converts a relative position to a genomic position for this sequence
     * @return
     */
    public int getGenomicPositionFromRelative(int relative) {
        return startPosition+relative;
    }
    
    /**
     * Converts a genomic position to a relative position for this sequence
     * @return
     */
    public int getRelativePositionFromGenomic(int genomic) {
        return genomic-startPosition;
    }
    
    /**
     * Sets the position of the transcription start site (TSS) for the gene
     * associated with this genomic region. The position of TSS should be 
     * specified in genomic coordinates
     * 
     * @param position of TSS
     */
    public void setTSS(Integer TSS) {
        this.TSS=TSS;
    }
    
    /**
     * Gets the position of the transcription start site (TSS) for the gene
     * associated with this genomic region. The position of the TSS is given 
     * in genomic coordinates
     * 
     * @return position of TSS if specified, otherwise <code>null</code>
     */
    public Integer getTSS() {
        return TSS;      
    }
    /**
     * Sets the position of the transcription end site (TES) for the gene
     * associated with this genomic region. The position of TES should be 
     * specified in genomic coordinates
     * 
     * @param position of TSS
     */
    
    
    public void setTES(Integer TES) {
        this.TES=TES;
    }
    
    /**
     * Gets the position of the transcription end site (TES) for the gene
     * associated with this genomic region. The position of the TES is given 
     * in genomic coordinates
     * 
     * @return position of TSS if specified, otherwise <code>null</code>
     */
    public Integer getTES() {
        return TES;      
    }
    
    /**
     * Sets the strand orientation of the gene associated with this 
     * genomic region. This class provides two static constants that can
     * be used as values: GenomicSequenceRegion.DIRECT and GenomicSequenceRegion.REVERSE
     * 
     * @param orientation The strand orientation of the associated gene
     */
    public void setStrandOrientation(int orientation) {
        this.orientation=orientation;
    }
    
    /**
     * Gets the strand orientation of the gene associated with this 
     * genomic region. This class provides two static constants that can
     * be used as values for comparison: GenomicSequenceRegion.DIRECT and GenomicSequenceRegion.REVERSE
     * 
     * @return the strand orientation of the associated gene 
     */
    public int getStrandOrientation() {
        return orientation;      
    }
    
    /**
     * Returns true if the gene associated with this genomic region is located on
     * the direct strand
     * 
     * @return true if associated gene is located on direct strand
     */
    public boolean isOnDirectStrand() {
        return (orientation==DIRECT);      
    }
    

    /**
     * Returns a String representation of the genomic region (location)
     * associated with this sequence on the form "chromosome:start-end"
     * 
     * @return a string specifying the region for this sequence
     */
    public String getRegionAsString() {
        return "chr"+chromosome+":"+startPosition+"-"+endPosition;
    }

    /**
     * Sets the originating organism for this sequence
     * 
     * @param organism the NCBI taxonomy ID associated with the organism
     */
    public void setOrganism(int organism) {
        this.organism=organism;
    }
    
   /**
    * Returns the common english name of the originating species for this sequence
    * 
    * @return the english name of the originating species
    */
    public String getSpeciesName() {
        return Organism.getCommonName(organism);
    }
    
    
   /**
    * Returns the latin name of the originating species for this sequence
    * 
    * @return the latin name of the originating species
    */
    public String getSpeciesLatinName() {
        return Organism.getLatinName(organism);
    }
    
   /**
    * Returns the NCBI organism taxonomy ID of the originating organism for this sequence
    * 
    * @return the NCBI organism taxonomy ID
    */
    public int getOrganism() {
        return organism;
    }    
    
    
    @Override
    public void importData(Data source) {
        this.sequenceName=((Sequence)source).sequenceName;
        this.associatedGeneName=((Sequence)source).associatedGeneName;
        this.chromosome=((Sequence)source).chromosome;
        this.startPosition=((Sequence)source).startPosition;
        this.endPosition=((Sequence)source).endPosition;
        this.TSS=((Sequence)source).TSS;
        this.TES=((Sequence)source).TES;
        this.orientation=((Sequence)source).orientation;
        this.organism=((Sequence)source).organism;   
        //notifyListenersOfDataUpdate(); 
    }

    @Override
    public Sequence clone() {
        Sequence seq=new Sequence(sequenceName, organism, genomeBuild, chromosome, startPosition, endPosition, associatedGeneName,TSS, TES, orientation);
        return seq;
    }

    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof Sequence)) return false;
        Sequence source =(Sequence)other;
        if (!this.sequenceName.equals(source.sequenceName)) return false;
        if (!this.associatedGeneName.equals(source.associatedGeneName)) return false;
        if (!this.chromosome.equals(source.chromosome)) return false;
        if (this.startPosition!=source.startPosition) return false;
        if (this.endPosition!=source.endPosition) return false;
        if (this.orientation!=source.orientation) return false;
        if (this.organism!=source.organism) return false;
        if (this.TSS==null && source.TSS!=null) return false;
        if (this.TSS!=null && source.TSS==null) return false;
        if (this.TSS!=null && source.TSS!=null && !this.TSS.equals(source.TSS)) return false;
        if (this.TES==null && source.TES!=null) return false;
        if (this.TES!=null && source.TES==null) return false;
        if (this.TES!=null && source.TES!=null && !this.TES.equals(source.TES)) return false;
        //notifyListenersOfDataUpdate();
        return true;
    }

    public static String getType() {return typedescription;}    

    @Override
    public String getTypeDescription() {return typedescription;}   
}    