/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import java.util.HashSet;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.protocol.ParseError;


/**
 * SequenceCollections can be used to refer to subsets of all the Sequence objects that
 * are registered in the Engine. Allthough most get/set methods work directly with Sequence
 * objects rather than Sequence names (Strings), the internal mechanism of the SequenceCollection 
 * itself revolves around a plain list of names for the Sequences in the collection. 
 * The methods that returns Sequences objects or lists do this by dynamically obtaining these Sequences
 * from the Engine (on-the-fly) based on the internal list of names
 * 
 * @author kjetikl
 */
public class SequenceCollection extends Data implements SequenceList {
    private static String typedescription="Sequence Collection"; 
    protected String datasetName;
    protected ArrayList<String> storage; // this is a list of Sequence names
    
    /**
     * Constructs a new initially "empty" Sequence collection with the given name
     * 
     * @param datasetName A name for this dataset
     */
   public SequenceCollection(String datasetName) {
       this.datasetName=datasetName;
       storage=new ArrayList<String>(20);
   }
   
//    /**
//     * Constructs a new Sequence collection with the given name and sequences
//     * The provided list should contain valid sequence names
//     * 
//     * @param datasetName A name for this dataset
//     */
//   public SequenceCollection(String datasetName, ArrayList<String> sequenceNames) {
//       this.datasetName=datasetName;
//       storage=(ArrayList<String>)sequenceNames.clone();
//   }
//   
   
    /**
     * Specifies a new name for this dataset
     * 
     * @param name the name for this dataset
     */
    public void setName(String name) {
        this.datasetName=name;
    }
    
    @Override
    public void rename(String name) {
        setName(name);
    }   
    
    @Override
    public String getName() {
        return datasetName;
    }
    
    @Override
    public Object getValue() {return this;} // should maybe change later
  
    @Override
    public String getValueAsParameterString() {
        String string="";
        for (int i=0;i<storage.size();i++) {
            if (i==storage.size()-1) string+=storage.get(i);
            else string+=storage.get(i)+",";
        }
        return string;
    }  
    
    /**
     * Returns the names of all the Sequences in this dataset
     * 
     * @return A list of Sequence names (note that the order might be 'random')
     */
    @SuppressWarnings("unchecked")
    public ArrayList<String> getAllSequenceNames() {
        return (ArrayList<String>)storage.clone();
    }    
    
    /**
     * Returns all the Sequence objects in this dataset
     * (if they are currently registered with the engine)
     * 
     * @return A list of Sequence objects (note that the order might be 'random')
     */
    public ArrayList<Sequence> getAllSequences(PriorsEditorEngine engine) {
        ArrayList<Sequence> list = new ArrayList<Sequence>(storage.size());
        for (String name:storage) {
            Data item=engine.getDataItem(name); 
            if (item!=null && item instanceof Sequence) list.add((Sequence)item);
        }
        return list;
    }    
    
    /**
     * Returns all the Sequence objects in this dataset
     * (if they are currently registered with the engine)
     * in the same order as they are listed in the default sequence collection
     * @return A list of Sequence objects
     */
    public ArrayList<Sequence> getAllSequencesInDefaultOrder(PriorsEditorEngine engine) {
        ArrayList<Sequence> list = new ArrayList<Sequence>(storage.size());
        ArrayList<String> allnames=engine.getDefaultSequenceCollection().getAllSequenceNames();
        for (String name:allnames) {
            if (!storage.contains(name)) continue;
            Data item=engine.getDataItem(name); 
            if (item!=null && item instanceof Sequence) list.add((Sequence)item);
        }
        return list;
    }    
    
    
            
    /**
     * Returns the Sequence corresponding to the given name.
     * 
     * @param name The name of the sequence
     * @return the specified sequence (if found) or null
     */
    public Sequence getSequenceByName(String name, PriorsEditorEngine engine) {
        for (String sequencename : storage) {
            if (sequencename.equals(name)) {
                Data item=engine.getDataItem(name); 
                if (item!=null && item instanceof Sequence) return (Sequence)item;
            }
        }
        return null;
    }
    
    /**
     * Returns the name of the Sequence corresponding to the given index.
     * This method could be used if you want to iterate through all sequences
     * in a dataset. 
     * 
     * @param index The index of the sequence
     * @return the name of the Sequence at the specified index (if exists) or null
     */
    public String getSequenceNameByIndex(int index) {
        return storage.get(index);
    }
    /**
     * Returns the Sequence corresponding to the given index.
     * This method could be used if you want to iterate through all sequences
     * in a dataset. 
     * 
     * @param index The index of the sequence
     * @return the specified sequence (if exists) or null
     */
    public Sequence getSequenceByIndex(int index, PriorsEditorEngine engine) {
        String name=storage.get(index);
        Sequence sequence=null;
        if (name!=null) {
            Data item=engine.getDataItem(name); 
            if (item!=null && item instanceof Sequence) return (Sequence)item;            
        }
        return sequence;
    }
    
    /**
     * Returns the index (order within the dataset) of the sequence with the given name 
     * If no sequence with the given name exists within the dataset the value -1 is returned.
     * 
     * @param name The name of the sequence
     * @return index of the sequence (between 0 and size-1) or -1
     */
    public int getIndexForSequence(String name) {
        for (int i=0;i<storage.size();i++) {
            String sequenceName=storage.get(i);
            if (sequenceName.equals(name)) return i;
        }
        return -1;
    }
    
    /** 
     * Returns true if the specified Sequence is in this collection 
     */
    public boolean contains(Sequence sequence) {
        if (sequence==null) return false;
        String sequenceName=sequence.getName();
        return contains(sequenceName);
    }
    
    /** 
     * Returns true if a sequence with the given name is in this collection
     */
    public boolean contains(String sequenceName) {
        for (String name:storage) {
            if (name.equals(sequenceName)) return true;
        }
        return false;
    }
    
    /**
     * Returns true if this SequenceCollection objects contains the same sequences
     * as the given collection
     * @param collection
     * @return
     */
    @Override
    public boolean containsSameData(Data other) {
        if (other==null || !(other instanceof SequenceCollection)) return false;
        SequenceCollection collection =(SequenceCollection)other;    
        if (size()!=collection.size()) return false;
        for (String name:storage) {
            if (!collection.contains(name)) return false;
        }
        return true;
    }
    
   /**
     * Adds a new Sequence object to the dataset.
     * 
     * @param sequence The Sequence to be added
     */
    public void addSequence(Sequence sequence) {
        storage.add(sequence.getName()); // add to local storage
        notifyListenersOfDataAddition(sequence);
    }
    
   /**
     * Removes a Sequence object from the dataset.
     * 
     * @param sequence The Sequence to be remvoed
     */
    public void removeSequence(Sequence sequence) {
        storage.remove(sequence.getName()); // remove from local storage
        notifyListenersOfDataRemoval(sequence);
    }
    
    
   /**
     * Removes all Sequences from this dataset
     */
    public void clearAll(PriorsEditorEngine engine) {
        String[] list=new String[storage.size()];
        for (int i=0;i<list.length;i++) {list[i]=storage.get(i);}
        for (String name:list) {
            storage.remove(name);
            Data item=null;
            if (engine!=null) item=engine.getDataItem(name);
            if (item!=null && item instanceof Sequence) notifyListenersOfDataRemoval((Sequence)item);
         }
    }
    
    
    /**
     * Reorders the list of sequences in the dataset by moving the sequence
     * at the specified current position to the new position. 
     * 
     * @param oldposition Position of the sequence to be moved
     * @param newposition New position to move sequence to
     */
    public void reorderSequences(int currentposition, int newposition) {
        String temp=storage.remove(currentposition);
        storage.add(newposition,temp);
        notifyListenersOfDataUpdate();
    }   
    
    /** Sets the order of sequences to the provided list. 
     *  The list must contain the same sequences as the original collection
     */
    public void setSequenceOrder(String[] order) {
         storage.clear();
         for(String s:order) storage.add(s);
         notifyListenersOfDataUpdate();
    }
    
    /** Sets the order of sequences to the provided list. 
     *  The list must contain the same sequences as the original collection
     */
    public void setSequenceOrder(ArrayList<String> order) {
         storage.clear();
         for(String s:order) storage.add(s);
         notifyListenersOfDataUpdate();
    }
    
    /**
     * Returns the number of sequences in this dataset
     * 
     * @return number of sequences
     */
    public int getNumberofSequences() {
        return storage.size();
    }
    
    /**
     * Returns the number of sequences in this dataset (same as getNumberofSequence)
     * 
     * @return number of sequences
     */
    public int size() {
        return storage.size();
    }
    
    /** Returns TRUE is this Sequence Collection has no sequences*/
    public boolean isEmpty() {
        return storage.size()==0;
    }
    

    
    /**
     * This method takes all sequences in the specified dataset and adds them
     * to this dataset
     * @param dataset The dataset to be incorporated into this dataset
     */
    public void merge(SequenceCollection dataset) {
        int size=dataset.getNumberofSequences();
        for (int i=0;i<size;i++) {
            String data=dataset.getSequenceNameByIndex(i);
            if (!contains(data)) storage.add(data); // I use this directly instead of calling addSequence() to limit the number of notifications sent 
        }              
        notifyListenersOfDataAddition(dataset);
    }
    
    @Override
    public void importData(Data source) throws ClassCastException {
        if (source==this) return; // no need to import, the source and target are the same
        SequenceCollection datasource=(SequenceCollection)source;
        this.datasetName=datasource.datasetName;
        storage.clear();
        for (String seqName:datasource.storage) {
            storage.add(seqName);
        }
        //System.err.println("Imported collection:\n"+output());
        //notifyListenersOfDataUpdate(); 
    }
    
    @Override
    public SequenceCollection clone() {
        SequenceCollection newcollection= new SequenceCollection(datasetName);
        int size=getNumberofSequences();
        for (int i=0;i<size;i++) {
            String seqName=getSequenceNameByIndex(i);
            newcollection.storage.add(seqName); // I use this directly instead of calling addSequence() to limit the number of notifications sent 
        }  
        return newcollection;
    }
       
    

    public static String getType() {return typedescription;}
    
    @Override
    public String getTypeDescription() {return typedescription+" : "+storage.size();}

    
    /**
     * Returns an unsorted list containing taxonomy IDs for all organisms that 
     * are represented in this Sequence Collection
     */
    public int[] getOrganisms(PriorsEditorEngine engine) {
        HashSet<Integer> organisms=new HashSet<Integer>();
        for (String name:storage) {
            Data item=engine.getDataItem(name); 
            if (item!=null && item instanceof Sequence) {
                int id=((Sequence)item).getOrganism();
                organisms.add(new Integer(id));
            }
        }
        int[] results=new int[organisms.size()];
        int i=0;
        for (Integer id:organisms) {
            results[i]=id.intValue();
            i++;
        }
        return results;
    }

    /**
     * Returns an unsorted list containing information about distinct genome builds
     * that are present among the sequences in this Sequence Collection
     */
    public String[] getGenomeBuilds(PriorsEditorEngine engine) {
        HashSet<String> organisms=new HashSet<String>();
        for (String name:storage) {
            Data item=engine.getDataItem(name); 
            if (item!=null && item instanceof Sequence) {
                String build=((Sequence)item).getGenomeBuild();
                organisms.add(build);
            }
        }
        String[] results=new String[organisms.size()];
        int i=0;
        for (String string:organisms) {
            results[i]=string;
            i++;
        }
        return results;
    }

    
    
    @Override
    public String output() {
        String string="";
        for (int i=0;i<storage.size();i++) {
            //string+="\""+storage.get(i)+"\"\n";
            string+=storage.get(i)+"\n";
        }
        return string;        
    }

    @Override
    public void inputFromPlain(ArrayList<String> input, PriorsEditorEngine engine) throws ParseError {
        clearAll(engine);
        for (String line:input) {
            line=line.trim();
            if (line.startsWith("#") || line.isEmpty()) continue;
            Data data=engine.getDataItem(line);
            if (data==null) throw new ParseError("No such sequence: "+line);
            else if (!(data instanceof Sequence)) throw new ParseError("'"+line+"' is not a Sequence object");
            else addSequence((Sequence)data);
        }
    }


}

