/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.data;

import java.util.ArrayList;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.protocol.ParseError;

/**
 * StringCollection can be used to refer to collections of Strings (for instance names)
 * 
 * @author kjetikl
 */
public class StringCollection extends Data {
    private static String typedescription="String Collection"; 
    protected String datasetName;
    protected ArrayList<String> storage; // this is a list of Strings (not StringVariables)
    
    /**
     * Constructs a new initially "empty" StringCollection with the given name
     * 
     * @param datasetName A name for this dataset
     */
   public StringCollection(String datasetName) {
       this.datasetName=datasetName;
       storage=new ArrayList<String>(20);
   }
   
    /**
     * Constructs a new StringCollection with the given name initialized with the given set of Strings
     * Note that if one of the supplied values is invalid it will be removed from the collection
     * without notice. Also, all values in the collection are unique and can not be added more than ones
     * (duplicate values are discarded)
     * @param datasetName A name for this dataset
     */
   public StringCollection(String datasetName, ArrayList<String> values) {
       this.datasetName=datasetName;
       storage=values;
       for (String value:storage) {
           if (!StringVariable.isValidValue(value)) storage.remove(value);
       }
   }
   
    /**
     * Constructs a new StringCollection with the given name initialized with the given set of Strings
     * Note that if one of the supplied values is invalid it will be removed from the collection
     * without notice. Also, all values in the collection are unique and can not be added more than ones
     * (duplicate values are discarded)
     * @param datasetName A name for this dataset
     */
   public StringCollection(String datasetName, String[] values) {
       this.datasetName=datasetName;
       storage=new ArrayList<String>(values.length);
       for (int i=0;i<values.length;i++) {
           if (StringVariable.isValidValue(values[i])) storage.add(values[i]);
       }
   }
   
    /**
     * Specifies a new name for this dataset
     * 
     * @param name the name for this dataset
     */
    public void setName(String name) {
        this.datasetName=name;
    }
    
    @Override
    public void rename(String name) {
        setName(name);
    }   
    
   /**
    * Returns the name of this dataset
    * 
    * @return dataset name
    */
    @Override
    public String getName() {
        return datasetName;
    }
    
    @Override
    public Object getValue() {return this;} // should maybe change later
  
    @Override
    public String getValueAsParameterString() {
        String string="";
        for (int i=0;i<storage.size();i++) {
            if (i==storage.size()-1) string+="\""+storage.get(i)+"\"";
            else string+="\""+storage.get(i)+"\",";
        }
        return string;
    } 
    
    public void setValue(ArrayList<String> newvalue) {
        storage=newvalue;
        for (String value:storage) {
           if (!StringVariable.isValidValue(value)) storage.remove(value);
        }
        notifyListenersOfDataUpdate();
    } 
    
    public void setValue(String[] newvalue) {
        storage.clear();
        for (int i=0;i<newvalue.length;i++) {
           if (StringVariable.isValidValue(newvalue[i])) storage.add(newvalue[i]);
        }
        notifyListenersOfDataUpdate();
    } 
    
    
    
    /**
     * Returns all the Strings in this dataset
     * 
     * @return A list of String names
     */
    public ArrayList<String> getAllStrings() {
        return storage;
    }    
           
    /**
     * Returns the String corresponding to the given index.
     * This method could be used if you want to iterate through all Strings
     * in a dataset. 
     * 
     * @param index The index of the String
     * @return the String at the specified index (if exists) or null
     */
    public String getStringByIndex(int index) {
        return storage.get(index);
    }
    
    /**
     * Returns the index (order within the dataset) of the String with the given name 
     * If no String with the given name exists within the dataset the value -1 is returned.
     * 
     * @param string the String
     * @return index of the String (between 0 and size-1) or -1
     */
    public int getIndexForString(String string) {
        for (int i=0;i<storage.size();i++) {
            String storedString=storage.get(i);
            if (storedString.equals(string)) return i;
        }
        return -1;
    }
    

    /** 
     * Returns true if a String with the given name is in this collection
     */
    public boolean contains(String string) {
        for (String stored:storage) {
            if (stored.equals(string)) return true;
        }
        return false;
    }
    
    /**
     * Returns true if this StringCollection objects contains the same Strings
     * as the given collection
     * @param collection
     * @return
     */
    @Override
    public boolean containsSameData(Data data) {
        if (data==null || !(data instanceof StringCollection)) return false;
        StringCollection collection=(StringCollection)data;
        if (size()!=collection.size()) return false;
        for (String name:storage) {
            if (!collection.contains(name)) return false;
        }
        return true;
    }
    
   /**
     * Adds a new String object to this collection
     * Note that all Strings in the collection must be unique
     * (i.e. the collection is a SET of strings)
     * @param string The String to be added
     * @return TRUE if the String was added (or not added because of duplication)
     * and FALSE if the String could not be added because its value was illegal 
     */
    public boolean addString(String string) {
        if (!StringVariable.isValidValue(string)) return false;
        if (!storage.contains(string)) return storage.add(string); // add to local storage
        else return false;
        //notifyListenersOfDataAddition(string);
    }
    
   /**
     * Removes a String object from this collection
     * 
     * @param string The String to be remvoed
     */
    public void removeString(String string) {
        storage.remove(string); // remove from local storage
        //notifyListenersOfDataRemoval(string);
    }
    
    
   /**
     * Removes all Strings from this dataset
     */
    public void clearAll() {
        storage.clear();
    }
    
    
  
    
    /**
     * Returns the number of Strings in this collection
     * 
     * @return number of strings
     */
    public int getNumberofStrings() {
        return storage.size();
    }
    
    /**
     * Returns the number of Strings in this collection (same as getNumberofStrings)
     * 
     * @return number of Strings
     */
    public int size() {
        return storage.size();
    }
    
    

    
    /**
     * This method takes all Strings in the specified dataset and adds them
     * to this dataset
     * @param dataset The dataset to be incorporated into this dataset
     */
    public void merge(StringCollection dataset) {
        int size=dataset.size();
        for (int i=0;i<size;i++) {
            String data=dataset.getStringByIndex(i);
            if (!contains(data)) storage.add(data); // I use this directly instead of calling addString() to limit the number of notifications sent 
        }              
        notifyListenersOfDataAddition(dataset);
    }
    
    @Override
    public void importData(Data source) throws ClassCastException {
        StringCollection datasource=(StringCollection)source;
        this.datasetName=datasource.datasetName;
        storage.clear();
        for (String string:datasource.storage) {
            storage.add(string);
        }
        //notifyListenersOfDataUpdate(); 
    }
    
    @Override
    public StringCollection clone() {
        StringCollection newcollection= new StringCollection(datasetName);
        int size=size();
        for (int i=0;i<size;i++) {
            String string=getStringByIndex(i);
            newcollection.storage.add(string); // I use this directly instead of calling addString() to limit the number of notifications sent 
        }  
        return newcollection;
    }
       
    

    public static String getType() {return typedescription;}
    
    @Override
    public String getTypeDescription() {return typedescription+" : "+storage.size();}

    @Override
    public String output() {
        String string="";
        for (int i=0;i<storage.size();i++) {
            // string+="\""+storage.get(i)+"\"\n";
            string+=storage.get(i)+"\n";
        }
        return string;
    }
    
    @Override
    public void inputFromPlain(ArrayList<String> input, PriorsEditorEngine engine) throws ParseError {
        String[] result=new String[input.size()];
        setValue(input.toArray(result));
    }   
}