/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.dataformat;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.Parameter;
import priorseditor.engine.ParameterExporter;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.data.DataSegment;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.ProgressListener;

/**
 *
 * @author kjetikl
 */
public abstract class DataFormat implements ParameterExporter {
    protected ArrayList<Parameter> parameterFormats=new ArrayList<Parameter>(); 
    protected PriorsEditorEngine engine=null;
    protected ArrayList<ProgressListener> progressListeners=new ArrayList<ProgressListener>();
    private int progress=0;
    
    /**
     * Returns a name for this output format
     */
    public abstract String getName(); 
    
    /**
     * Sets a reference to the engine for this DataFormat
     */
    public void setEngine(PriorsEditorEngine engine) {this.engine=engine;} 
        
    /**
     * This method returns true if the DataFormat object can format output for Data objects
     * of the given type in the objects format. 
     */
    public abstract boolean canFormatOutput(Data data);

    /**
     * This method returns true if the DataFormat object can format output Data objects
     * of the given class in the objects format. 
     */
    public abstract boolean canFormatOutput(Class dataclass);
    
    
    /**
     * This method returns true if the DataFormat object can parse input for Data objects
     * of the given type in the objects format. 
     */
    public abstract boolean canParseInput(Data data);

    /**
     * This method returns true if the DataFormat object can parse input for Data objects
     * of the given class in the objects format. 
     */
    public abstract boolean canParseInput(Class dataclass);
    
    
    
    
    
    /**
     * This method returns a list containing the Data classes that this output
     * format supports  
     */
    public abstract Class[] getSupportedDataTypes();
    
    /**
     * This method returns the default file-suffix to use for the DataFormat
     */
    public abstract String getSuffix();

    /**
     * This method returns a list of alternative suffixes that are common for this data format
     * The default implementation just returns a list containing the default suffix, but this
     * can be overriden by subclasses that want to return more suffixes
     */
    public String[] getSuffixAlternatives() {return new String[]{getSuffix()};}


    /** This method should return TRUE (default) if output in this format can be appended
     *  to an already existing non-empty output object and FALSE if a new object must be
     *  used as target
     */
    public boolean isAppendable() {
        return true;
    }

    /**
     * This method takes a dataobject as input and outputs its content according to the output format 
     * @param dataobject The input data object which is to be formatted for output
     * @param outputobject An OutputData object to use as destination for the output
     * @param settings An object specifying additional output format parameters
     * @param task An ExecutableTask which can be prompted for interruption or status. Note that this can be NULL
     * @return An OutputData object containing the formatted output of the input Data
     * @throws ExecutionError
     * @throws InterruptedException
     */
    public abstract OutputData format(Data dataobject, OutputData outputobject, ParameterSettings settings, ExecutableTask task) throws ExecutionError, InterruptedException;
   
               
    /**
     * Adds a parameter to the DataFormat. This is used for initialization of new DataFormats and should only be called in a constructor or similar setup method
     */
    protected void addParameter(String parameterName, Object defaultValue, Object[] allowedValues, String description) {
        if (defaultValue==null) throw new NullPointerException("Default parameter value cannot be null"); 
        parameterFormats.add(new Parameter(parameterName,defaultValue.getClass(),defaultValue,allowedValues,description));
    }
    
    /**
     * Adds a parameter to the DataFormat. This is used for initialization of new DataFormats and should only be called in a constructor or similar setup method
     */
    protected void addParameter(String parameterName, Object defaultValue, Object[] allowedValues, String description, boolean required, boolean hidden) {
        if (defaultValue==null) throw new NullPointerException("Default parameter value cannot be null"); 
        parameterFormats.add(new Parameter(parameterName,defaultValue.getClass(),defaultValue,allowedValues,description,required,hidden));
    }
        
    /**
     * Adds a parameter to the DataFormat. This is used for initialization of new DataFormats and should only be called in a constructor or similar setup method
     */
    protected void addOptionalParameter(String parameterName, Object defaultValue, Object[] allowedValues, String description) {
        if (defaultValue==null) throw new NullPointerException("Default parameter value cannot be null"); 
        parameterFormats.add(new Parameter(parameterName,defaultValue.getClass(),defaultValue,allowedValues,description,false,false));
    }
    
    @Override
    public String toString() {return getName();}
    
    @Override
    public Object getDefaultValueForParameter(String parameterName) {
        for (Parameter par:parameterFormats) {
            if (parameterName.equals(par.getName())) return par.getDefaultValue();
        }
        return null; 
    }

    @Override
    public Parameter[] getParameters() {
        Parameter[] list=new Parameter[parameterFormats.size()];
        return parameterFormats.toArray(list);
    }
    
    @Override
    public Parameter getParameterWithName(String parameterName) {
        for (Parameter parameter:parameterFormats) {
            if (parameter.getName().equals(parameterName)) return parameter;
        }
        return null;
    }    
    
    /** Replaces the current output parameters with the new set (use this method with caution) */
    public void replaceParameters(Parameter[] parameters) {
        parameterFormats.clear();
        for (Parameter parameter:parameters) parameterFormats.add(parameter);
    }
    
    
     /** Returns an XML element for this DataFormat that can be included in XML-based configuration files */
    public org.w3c.dom.Element getXMLrepresentation(org.w3c.dom.Document document, ParameterSettings settings) {
        org.w3c.dom.Element element = document.createElement("dataformat");
        element.setAttribute("name", getName());
        if (settings!=null) {
            Parameter[] parameters=getParameters();
            for (Parameter parameter:parameterFormats) {
                String value=settings.getParameterAsString(parameter.getName(), parameters);
                org.w3c.dom.Element setting = document.createElement("setting");
                setting.setAttribute("name", parameter.getName());
                setting.setAttribute("class", parameter.getType().getSimpleName());
                setting.setTextContent(value);
                element.appendChild(setting);
            }  
        }
        return element;
    }     
    /**
     * Parses input from a stream and returns a Data object of the appropriate type
     * @param input The source inputstream
     * @param target the Data object where the results should be stored
     * @param settings optional settings for the DataFormat
     * @param task Optional task object that can be used to send abort signals
     * @return
     */
    public abstract Data parseInput(ArrayList<String> input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException;


    public Data parseInput(InputStream input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException {
        BufferedReader inputStream=null;
        ArrayList<String> inputaslist=new ArrayList<String>();
        long lines=0;
        try {
            inputStream=new BufferedReader(new InputStreamReader(input));
            String line;
            while((line=inputStream.readLine())!=null) {
                lines++;
                if (lines%100==0 && (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED)))) throw new InterruptedException();
                inputaslist.add(line);
            }
        } catch (IOException e) { 
            throw new ParseError(e.getClass().getSimpleName()+":"+e.getMessage());
        } finally {
            try {
                if (inputStream!=null) inputStream.close();
            } catch (IOException ioe) {}
        }
        return parseInput(inputaslist, target, settings, task);
    }

    /**
     * Parses textual input represented as a list of Strings and returns a DataSegment containing the relevant data
     * This method should only be used to retrieve partial segments of Feature data
     * @param input The text to be parsed
     * @param target the DataSegment object where the results should be stored
     * @param settings optional settings for the DataFormat
     * @param task Optional task object that can be used to send abort signals     * 
     * @return
     */
    public DataSegment parseInput(ArrayList<String> input, DataSegment target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException {
        throw new ParseError("Inappropriate use of data format '"+getName()+"' to parse feature data");
    }

     /**
     * Parses input from a stream and returns a DataSegment containing the relevant data
     * This method should only be used to retrieve partial segments of Feature data
     * @param input The source inputstream
     * @param target the DataSegment object where the results should be stored
     * @param settings optional settings for the DataFormat
     * @param task Optional task object that can be used to send abort signals     *
     * @return
     */
    public DataSegment parseInput(InputStream input, DataSegment target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException {
        BufferedReader inputStream=null;
        ArrayList<String> inputaslist=new ArrayList<String>();
        long lines=0;
        try {
            inputStream=new BufferedReader(new InputStreamReader(input));
            String line;
            while((line=inputStream.readLine())!=null) {
                lines++;
                if (lines%100==0 && (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
                inputaslist.add(line);
            }
        } catch (IOException e) {
            throw new ParseError(e.getClass().getSimpleName()+":"+e.getMessage());
        } finally {
            try {
                if (inputStream!=null) inputStream.close();
            } catch (IOException ioe) {}
        }
        return parseInput(inputaslist, target, settings, task);
    }

    
    /** Adds a progressListener to this DataFormat object */
    public void addProgressListener(ProgressListener listener) {
        progressListeners.add(listener);
    }
    
    /** Remove a progressListener from this DataFormat object */
    public void removeProgressListener(ProgressListener listener) {
        progressListeners.remove(listener);
    }
    
    /** 
     * Sets the progress of this DataFormat 
     * Be careful to synchronize properly if using DataFormats as singletons
     */
    public void setProgress(int i) {
        if (i>100) progress=100;
        else if (i<0) progress=0;
        else progress=i;
        for (ProgressListener listener:progressListeners) listener.processProgressEvent(this, progress);
    }   
    
    /** 
     * Sets the progress value for this operation according to the relative number
     * of "subtasks" that have been completed so far
     * Be careful to synchronize properly if using DataFormats as singletons
     * 
     * @param i The number of subtasks currently completed
     * @param n The total number of subtasks that must be completed
     */
    public void setProgress(int i, int n) {
        int newprogress;
        if (i>n) newprogress=100;
        else if (i<0) newprogress=0;
        else newprogress=(int)((i*1.0/n)*100);
        setProgress(newprogress);
    }
    
    /** Returns a progress value from this DataFormat */
    public int getProgress() {
        return progress;
    }
}
