/*
 
 
 */

package priorseditor.engine.dataformat;

import java.util.ArrayList;
import java.util.HashMap;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.RegionSequenceData;
import priorseditor.engine.data.RegionDataset;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.data.Region;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.Parameter;

import priorseditor.engine.OperationTask;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.data.DataSegment;
import priorseditor.engine.data.FeatureSequenceData;
import priorseditor.engine.data.SequenceCollection;
/**
 *
 * @author kjetikl
 */
public class DataFormat_BED extends DataFormat {
    private String name="BED";

    private Class[] supportedTypes=new Class[]{RegionSequenceData.class, RegionDataset.class};

    public DataFormat_BED() {
        addOptionalParameter("Add CHR prefix", "yes", new String[]{"yes","no"},"If set, it will add a 'chr' prefix to chromosomes (if not already present)");
        addOptionalParameter("Start position", "0", new String[]{"0","1"},"Selects the position of the first base in a chromosome. The default is 0 for BED, but it is usually 1 for other formats");
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public boolean canFormatOutput(Data data) {
        return (data instanceof RegionSequenceData || data instanceof RegionDataset);
    }

    @Override
    public boolean canFormatOutput(Class dataclass) {
        return (dataclass.equals(RegionSequenceData.class) || dataclass.equals(RegionDataset.class));
    }

    @Override
    public boolean canParseInput(Data data) {
        return (data instanceof RegionSequenceData || data instanceof RegionDataset);
    }

    @Override
    public boolean canParseInput(Class dataclass) {
        return (dataclass.equals(RegionSequenceData.class) || dataclass.equals(RegionDataset.class));
    }

    @Override
    public Class[] getSupportedDataTypes() {
        return supportedTypes;
    }

    @Override
    public String getSuffix() {
        return "bed";
    }

    @Override
    public OutputData format(Data dataobject, OutputData outputobject, ParameterSettings settings, ExecutableTask task) throws ExecutionError, InterruptedException {
        if (!canFormatOutput(dataobject)) {
            if (task!=null) throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format",task.getLineNumber());
            else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        }
        setProgress(5);
        boolean addPrefix=true;
        int startpos=0;
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             String addPrefixString=(String)settings.getResolvedParameter("Add CHR prefix",defaults,engine); 
             String startposString=(String)settings.getResolvedParameter("Start position",defaults,engine);
             if (addPrefixString!=null && addPrefixString.equalsIgnoreCase("no")) addPrefix=false;
             if (startposString!=null && startposString.equalsIgnoreCase("1")) startpos=1;             
          } catch (ExecutionError e) {
             throw e;
          } catch (Exception ex) {
              throw new ExecutionError("An error occurred during output formatting", ex);
          }
        } else {
           String addPrefixString=(String)getDefaultValueForParameter("Add CHR prefix");
           String startposString=(String)getDefaultValueForParameter("Start position");
           if (addPrefixString!=null && addPrefixString.equalsIgnoreCase("no")) addPrefix=false;
           if (startposString!=null && startposString.equalsIgnoreCase("1")) startpos=1;            
        }
        String outputString="";
        if (dataobject instanceof RegionDataset) {
            SequenceCollection sequenceCollection=null;
            if (task instanceof OperationTask) {
                String subsetName=(String)((OperationTask)task).getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
                if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
                Data seqcol=engine.getDataItem(subsetName);
                if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
                if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
                sequenceCollection=(SequenceCollection)seqcol;
            }
            if (sequenceCollection==null) sequenceCollection=engine.getDefaultSequenceCollection();
            outputString=outputMultipleSequences((RegionDataset)dataobject, sequenceCollection, addPrefix, startpos, task, engine);
        } else if (dataobject instanceof RegionSequenceData){
            StringBuilder builder=new StringBuilder();
            outputSequence((RegionSequenceData)dataobject, addPrefix, startpos, task, builder);
            outputString=builder.toString();
        } else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        outputobject.append(outputString,getName());
        setProgress(100);
        return outputobject;
    }

    /** Outputs multiple sequences (i.e. a whole FeatureDataset)
     *  The sequences are output in the order determined by the DefaultSequenceCollection (as shown in the GUI)
     */
    private String outputMultipleSequences(RegionDataset sourcedata, SequenceCollection collection, boolean addPrefix, int startpos, ExecutableTask task, PriorsEditorEngine engine) throws InterruptedException {
        StringBuilder outputString=new StringBuilder();
        int size=collection.getNumberofSequences();
        int i=0;
        ArrayList<Sequence> sequences=collection.getAllSequencesInDefaultOrder(engine);
        for (Sequence sequence:sequences) { // for each sequence
              String sequenceName=sequence.getName();
              RegionSequenceData sourceSequence=(RegionSequenceData)sourcedata.getSequenceByName(sequenceName);
              outputSequence(sourceSequence, addPrefix, startpos, task, outputString);
              if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
              if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
              if (i%3==0) Thread.yield();
              setProgress(i+1,size);
              i++;
        }
        return outputString.toString();
    }


    /** output-formats a single sequence */
    private void outputSequence(RegionSequenceData sequence, boolean addPrefix, int startpos, ExecutableTask task, StringBuilder outputString) throws InterruptedException {
            ArrayList<Region> regionList=sequence.getAllRegions();
            //String sequenceName=sequence.getName();
            String chromosome=sequence.getChromosome();
            if (addPrefix && !chromosome.startsWith("chr")) chromosome="chr"+chromosome;
            //String featureName=sequence.getParent().getName();
            int count=0;
            for (Region region:regionList) {
               count++;
               if (count%200==0) {
                  if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
                  if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
                  Thread.yield();
               }
               String strand="";
               int start=region.getGenomicStart();
               int end=region.getGenomicEnd();
               if (startpos==0) { // use zero-indexed positions instead of one-indexed
                   start--;
                   end--;
               }
               end++; // add 1 because the END-position is non-inclusive.  E.g the range [0,100] includes bases 0 through 99
               String type=region.getType();
                    if (region.getOrientation()==Region.DIRECT)  strand="+";
               else if (region.getOrientation()==Region.REVERSE) strand="-";
               String line=chromosome+"\t"+start+"\t"+end+"\t"+type+"\t"+region.getScore();
               if (!strand.isEmpty()) line+="\t"+strand;
               outputString.append(line+"\n");
            }
    }



    @Override
    public Data parseInput(ArrayList<String> input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
         return (Data)parseInputToTarget(input, target, settings, task);
    }

    @Override
    public DataSegment parseInput(ArrayList<String> input, DataSegment target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
        return (DataSegment)parseInputToTarget(input, target, settings, task);
    }

    /** The following method is a common substitute for the above 2 methods */
    private Object parseInputToTarget(ArrayList<String> input, Object target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
        if (target==null) target=new RegionDataset("temporary");
        if (target instanceof RegionDataset) { // The RegionDataset might not contain RegionSequenceData objects for all Sequences. Add them if they are missing!
            RegionDataset dataset=(RegionDataset)target;
            ArrayList<Data> sequences=engine.getAllDataItemsOfType(Sequence.class);
            for (Data seq:sequences) {
                if (dataset.getSequenceByName(seq.getName())==null) dataset.addSequence(new RegionSequenceData((Sequence)seq));
            }
        }
        boolean addPrefix=true;
        int startpos=0;
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             String addPrefixString=(String)settings.getResolvedParameter("Add CHR prefix",defaults,engine);
             String startposString=(String)settings.getResolvedParameter("Start position",defaults,engine);
             if (addPrefixString!=null && addPrefixString.equalsIgnoreCase("no")) addPrefix=false;
             if (startposString!=null && startposString.equalsIgnoreCase("1")) startpos=1;
          }
          catch (ExecutionError e) {throw new ParseError("An error occurred while resolving parser parameters: "+e.getMessage());}
          catch (Exception ex) {throw new ParseError("An error occurred while resolving parser parameters: "+ex.getMessage());}
        } else {
           String addPrefixString=(String)getDefaultValueForParameter("Add CHR prefix");
           String startposString=(String)getDefaultValueForParameter("Start position");
           if (addPrefixString!=null && addPrefixString.equalsIgnoreCase("no")) addPrefix=false;
           if (startposString!=null && startposString.equalsIgnoreCase("1")) startpos=1;
        }
        int count=0;
        for (String line:input) { // parsing each line in succession
            count++;
            if (count%200==0) {
               if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
               if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
               Thread.yield();
            }
            if (line.startsWith("#") || line.isEmpty()) continue; // assume comment line
            HashMap<String,Object> map=parseSingleLineInStandardFormat(line,startpos);
            String regionchromosome=(String)map.get("CHROMOSOME");
            int regionstart=(Integer)map.get("START");
            int regionend=(Integer)map.get("END");
            RegionSequenceData targetSequence=null;
            if (target instanceof RegionSequenceData) {
               targetSequence=(RegionSequenceData)target;
               if (!targetSequence.getChromosome().equals(regionchromosome)) continue;
               if (regionstart>targetSequence.getRegionEnd() || regionend<targetSequence.getRegionStart()) continue;
               addRegionToTarget(targetSequence,map);
            } else if (target instanceof RegionDataset) { // add region to all applicable sequences
                ArrayList<FeatureSequenceData> sequences=((RegionDataset)target).getAllSequences();
                for (FeatureSequenceData seq:sequences) {
                    targetSequence=(RegionSequenceData)seq;
                    if (!targetSequence.getChromosome().equals(regionchromosome)) continue;
                    if (regionstart>targetSequence.getRegionEnd() || regionend<targetSequence.getRegionStart()) continue;
                    addRegionToTarget(targetSequence,map);
                }
            } else if (target instanceof DataSegment) {
                addRegionToTarget(target,map);
            } else throw new ParseError("SLOPPY PROGRAMMING ERROR: non-Region data as target for BED dataformat: "+target.getClass().getSimpleName());
        }
        return target;
    }


    private void addRegionToTarget(Object target, HashMap<String,Object> map) throws ParseError {
                int start=0, end=0; // these are offset relative to the start of the parent sequence or segment
                int targetStart=0;
                if (target instanceof RegionSequenceData) {
                    targetStart=((RegionSequenceData)target).getRegionStart();
                } else if (target instanceof DataSegment) {
                    targetStart=((DataSegment)target).getSegmentStart();
                } else throw new ParseError("Target object neither RegionSequenceData nor DataSegment in DataFormat_GTF.addRegionToTarget():"+target.toString());
                Object startValue=map.get("START");
                if (startValue instanceof Integer) start=(Integer)startValue;
                Object endValue=map.get("END");
                if (endValue instanceof Integer) end=(Integer)endValue;
                float score=0;
                Object scoreValue=map.get("SCORE");
                if (scoreValue instanceof Float) score=(Float)scoreValue;
                String type=(String)map.get("TYPE"); if (type==null) type="unknown_type";
                String annotatedOrientation=(String)map.get("STRAND"); // the orientation in the BED file
                int orientation=Sequence.DIRECT;
                start-=targetStart;
                end-=targetStart;
                     if (annotatedOrientation.equals("+")) orientation=Region.DIRECT;
                else if (annotatedOrientation.equals("-")) orientation=Region.REVERSE;
                else orientation=Region.INDETERMINED;

                RegionSequenceData parentSequence=null;
                if (target instanceof RegionSequenceData) parentSequence=(RegionSequenceData)target;
                Region newRegion=new Region(parentSequence, start, end, type, score, orientation);
                for (String property:map.keySet()) {
                    if (property.equalsIgnoreCase("CHROMOSOME") ||
                        property.equalsIgnoreCase("FEATURE") ||
                        property.equalsIgnoreCase("SOURCE") ||
                        property.equalsIgnoreCase("START") ||
                        property.equalsIgnoreCase("END") ||
                        property.equalsIgnoreCase("SCORE") ||
                        property.equalsIgnoreCase("TYPE") ||
                        property.equalsIgnoreCase("STRAND")) continue;
                    else {
                        newRegion.setProperty(property, map.get(property));
                    }
                }
                // System.err.println("Add region: "+newRegion.toString());
                if (target instanceof RegionSequenceData) ((RegionSequenceData)target).addRegion(newRegion);
                else if (target instanceof DataSegment) ((DataSegment)target).addRegion(newRegion);
    }


    /** parses a single line in a BED-file and returns a HashMap with the different properties (with values as strings!) according to the capturing groups in the formatString */
    private HashMap<String,Object> parseSingleLineInStandardFormat(String line, int startpos) throws ParseError {
        HashMap<String,Object> result=new HashMap<String,Object>();
        String[] fields=line.split("\t");
        if (fields.length<3) throw new ParseError("Expected at least 3 fields per line in BED-format. Got "+fields.length);
        String chromosome=fields[0];
        if (chromosome.startsWith("chr")) chromosome=chromosome.substring(3);
        result.put("CHROMOSOME",chromosome);
        try {
            int start=Integer.parseInt(fields[1]);
            if (startpos==0) start++; // if BED is zero-indexed, convert to one-indexing which is used by PriorsEditor
            result.put("START",start);
        } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for START: "+e.getMessage());}
        try {
            int end=Integer.parseInt(fields[2]);
            if (startpos==0) end++; // if BED is zero-indexed, convert to one-indexing which is used by PriorsEditor
            end--; // substract 1 because the end-coordinate in BED-files are non-inclusive
            result.put("END",end);
        } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for END: "+e.getMessage());}
        if (fields.length>=4) {
            result.put("TYPE",fields[3]);
        }
        if (fields.length>=5) {
            try {
                result.put("SCORE",Float.parseFloat(fields[4]));
            } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for SCORE: "+e.getMessage());}
        }
        if (fields.length>=6 && (fields[5].equals("+") || fields[5].equals("-"))) {
            result.put("STRAND",fields[5]);
        } else result.put("STRAND","."); // INDETERMINATE ORIENTATION
        return result;
    }


}





