/*
 
 
 */

package priorseditor.engine.dataformat;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.RegionSequenceData;
import priorseditor.engine.data.RegionDataset;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.data.Region;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.ExecutionError;

import priorseditor.engine.OperationTask;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.Parameter;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.DataSegment;
import priorseditor.engine.data.SequenceCollection;
/**
 *
 * @author kjetikl
 */
public class DataFormat_GFF extends DataFormat {
    private String name="GFF";
    private static final int RELATIVE=0;
    private static final int GENOMIC=1;
    private static final String DIRECT="Direct";
    private static final String REVERSE="Reverse";
    private static final String FROM_GENE="From Gene";
    private static final String OPPOSITE="Opposite";
    
    
    private Class[] supportedTypes=new Class[]{RegionSequenceData.class, RegionDataset.class};
    //private String GFFformatString="{SEQUENCENAME}\t{FEATURE}\tmisc_feature\t{START}\t{END}\t{SCORE}\t{STRAND}\t.\t{ATTRIBUTES}";
    private String GFFformatString="";
    
    public DataFormat_GFF() {
        addParameter("Position", "Genomic", new String[]{"Relative","Genomic"},null);
        addParameter("Relative-offset", new Integer(1), null,"Start relative positions at 0 or 1 (only applicable if Position=relative)");
        addParameter("Orientation", FROM_GENE,new String[]{DIRECT,REVERSE,FROM_GENE,OPPOSITE},"Orientation of relative coordinates (only applicable if Position=relative)");
        addOptionalParameter("Sort1", "Position", new String[]{"Position","Type","Score"},null);
        addOptionalParameter("Sort2", "Type", new String[]{"Position","Type","Score"},null);
        addOptionalParameter("Sort3", "Score", new String[]{"Position","Type","Score"},null);   
        addOptionalParameter("Format", GFFformatString, null,null);   
    }
        
    @Override
    public String getName() {
        return name;
    }
    
    @Override
    public boolean canFormatOutput(Data data) {
        return (data instanceof RegionSequenceData || data instanceof RegionDataset);
    }
    
    @Override
    public boolean canFormatOutput(Class dataclass) {
        return (dataclass.equals(RegionSequenceData.class) || dataclass.equals(RegionDataset.class));
    }
    
    @Override
    public boolean canParseInput(Data data) {
        return (data instanceof RegionSequenceData || data instanceof RegionDataset);
    }
    
    @Override
    public boolean canParseInput(Class dataclass) {
        return (dataclass.equals(RegionSequenceData.class) || dataclass.equals(RegionDataset.class));
    }
    
    
    
    @Override
    public Class[] getSupportedDataTypes() {
        return supportedTypes;
    }
    
    @Override
    public String getSuffix() {
        return "gff";
    }
 
    @Override
    public OutputData format(Data dataobject, OutputData outputobject, ParameterSettings settings, ExecutableTask task) throws ExecutionError, InterruptedException {
        if (!canFormatOutput(dataobject)) {
            if (task!=null) throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format",task.getLineNumber());
            else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        }        
        setProgress(5);
        String sort1,sort2,sort3,offsetString,formatString,orientation;
        int relativeoffset=1;
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             sort1=(String)settings.getResolvedParameter("Sort1",defaults,engine); 
             sort2=(String)settings.getResolvedParameter("Sort2",defaults,engine);
             sort3=(String)settings.getResolvedParameter("Sort3",defaults,engine);
             offsetString=(String)settings.getResolvedParameter("Position",defaults,engine);
             relativeoffset=(Integer)settings.getResolvedParameter("Relative-offset",defaults,engine);
             formatString=(String)settings.getResolvedParameter("Format",defaults,engine);
             orientation=(String)settings.getResolvedParameter("Orientation",defaults,engine);
          } catch (ExecutionError e) {
             throw e;
          } catch (Exception ex) {
              throw new ExecutionError("An error occurred during output formatting", ex);
          }
        } else {
           sort1=(String)getDefaultValueForParameter("Sort1");
           sort2=(String)getDefaultValueForParameter("Sort2");
           sort3=(String)getDefaultValueForParameter("Sort3");
           offsetString=(String)getDefaultValueForParameter("Position");
           relativeoffset=(Integer)getDefaultValueForParameter("Relative-offset");
           formatString=(String)getDefaultValueForParameter("Format");
           orientation=(String)getDefaultValueForParameter("Orientation");
        }
        if (formatString!=null && formatString.trim().isEmpty()) formatString=null;
        int offset=GENOMIC;
        if (offsetString!=null && offsetString.equalsIgnoreCase("Relative")) offset=RELATIVE;
        String outputString="";
        RegionComparator regionComparator=new RegionComparator(sort1, sort2, sort3);
        if (dataobject instanceof RegionDataset) {
            SequenceCollection sequenceCollection=null;
            if (task instanceof OperationTask) {
                String subsetName=(String)((OperationTask)task).getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
                if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
                Data seqcol=engine.getDataItem(subsetName);
                if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
                if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
                sequenceCollection=(SequenceCollection)seqcol;
            }
            if (sequenceCollection==null) sequenceCollection=engine.getDefaultSequenceCollection();                   
            outputString=outputMultipleSequences((RegionDataset)dataobject, sequenceCollection, regionComparator,offset,relativeoffset,orientation,formatString, task, engine);
        } else if (dataobject instanceof RegionSequenceData){
            StringBuilder builder=new StringBuilder();
            outputSequence((RegionSequenceData)dataobject,regionComparator,offset,relativeoffset,orientation,formatString, task, builder);
            outputString=builder.toString();
        } else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        outputobject.append(outputString,getName());
        setProgress(100);
        return outputobject;
    }    

    /** Outputs multiple sequences (i.e. a whole FeatureDataset)
     *  The sequences are output in the order determined by the DefaultSequenceCollection (as shown in the GUI)
     */
    private String outputMultipleSequences(RegionDataset sourcedata, SequenceCollection collection, RegionComparator regionComparator, int offset,int relativeoffset, String orientationString, String formatString, ExecutableTask task, PriorsEditorEngine engine) throws InterruptedException {
        StringBuilder outputString=new StringBuilder();        
        int size=collection.getNumberofSequences();
        int i=0;
        ArrayList<Sequence> sequences=collection.getAllSequencesInDefaultOrder(engine);
        for (Sequence sequence:sequences) { // for each sequence   
              String sequenceName=sequence.getName();
              RegionSequenceData sourceSequence=(RegionSequenceData)sourcedata.getSequenceByName(sequenceName);
              outputSequence(sourceSequence,regionComparator,offset,relativeoffset,orientationString,formatString, task, outputString);
              if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
              if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
              if (i%3==0) Thread.yield();
              setProgress(i+1,size);
              i++;
        }      
        return outputString.toString();
    }    
    
    
    /** outputformats a single sequence */
    private void outputSequence(RegionSequenceData sequence, RegionComparator regionComparator, int offset,int relativeoffset, String orientationString, String formatString, ExecutableTask task, StringBuilder outputString) throws InterruptedException {
            ArrayList<Region> regionList=sequence.getAllRegions();
            int shownStrand=Sequence.DIRECT;
                if (orientationString.equals(DIRECT)) shownStrand=Sequence.DIRECT;
            else if (orientationString.equals(REVERSE)) shownStrand=Sequence.REVERSE;
            else if (orientationString.equals(FROM_GENE)) {
               shownStrand=sequence.getStrandOrientation();
            } else if (orientationString.equals(OPPOSITE)) {
               shownStrand=sequence.getStrandOrientation();
               if (shownStrand==Sequence.DIRECT) shownStrand=Sequence.REVERSE;
               else shownStrand=Sequence.DIRECT;
            }
            regionComparator.setOrientation(shownStrand);
            Collections.sort(regionList, regionComparator);
            String sequenceName=sequence.getName();
            String featureName=sequence.getParent().getName();
            String source=sequence.getDataSource();
            int count=0;
            for (Region region:regionList) {
               count++;
               if (count%200==0) {
                  if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
                  if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
                  Thread.yield();
               }
               int start=-1;
               int end=-1;
               String strand=".";
               if (offset==GENOMIC) {
                   start=region.getGenomicStart();
                   end=region.getGenomicEnd();                    
                        if (region.getOrientation()==Region.INDETERMINED) strand=".";
                   else if (region.getOrientation()==Region.DIRECT) strand="+";
                   else if (region.getOrientation()==Region.REVERSE) strand="-";
               } else { // relative offset
                   if (shownStrand==Sequence.DIRECT) {
                       start=region.getRelativeStart()+relativeoffset;
                       end=region.getRelativeEnd()+relativeoffset;                        
                   } else { // relative offset and reverse orientation
                       start=(sequence.getSize()-1)-region.getRelativeEnd()+relativeoffset;
                       end=(sequence.getSize()-1)-region.getRelativeStart()+relativeoffset;                  
                   }               
                        if (region.getOrientation()==Region.INDETERMINED) strand=".";
                   else if (region.getOrientation()==shownStrand) strand="+";
                   else strand="-";
               }


               String attributes="";
               String[] properties=region.getAllPropertyNames();
               for (int i=0;i<properties.length;i++) {
                   Object val=region.getProperty(properties[i]);
                   if (val==null) attributes+=properties[i]+"=null;";
                   else attributes+=properties[i]+"=\""+val.toString()+"\";";
               }
               String line=formatString;
               if (line!=null) {
                   line=line.replace("{SEQUENCENAME}", sequenceName);
                   line=line.replace("{SOURCE}", source);
                   line=line.replace("{FEATURE}", featureName);
                   line=line.replace("{START}", ""+start);
                   line=line.replace("{END}", ""+end);
                   line=line.replace("{SCORE}", ""+region.getScore());
                   line=line.replace("{STRAND}", strand);
                   line=line.replace("{TYPE}", region.getType());
                   line=line.replace("{ATTRIBUTES}", attributes);
                   //line=line.replace("{XXX}", region.getProperty(XXX)); // check for user-defined properties and replace
               } else { // default format
                   String type=region.getType();
                   if (type==null) type=""; else type="type=\""+type+"\";";
                   line=sequenceName+"\t"+featureName+"\tmisc_feature\t"+start+"\t"+end+"\t"+region.getScore()+"\t"+strand+"\t.\t"+type+attributes;
               }
               outputString.append(line+"\n");
            }
    }
    
           
    
    /**
     * Compares two Regions according to specified sorting order
     */
    private class RegionComparator implements Comparator<priorseditor.engine.data.Region> {
        String sort1;
        String sort2;
        String sort3;
        int orientation=Sequence.DIRECT;
        
        public RegionComparator(String sort1,String sort2,String sort3) {
            this.sort1=sort1;
            this.sort2=sort2;
            this.sort3=sort3;
            orientation=Sequence.DIRECT;
        }
        
        public void setOrientation(int orientation) {
            this.orientation=orientation;
        }
        
        @Override
        public int compare(Region region1, Region region2) {
            int result=0;
            result=compareProperty(region1, region2, sort1);
            if (result!=0) return result;
            result=compareProperty(region1, region2, sort2);
            if (result!=0) return result;
            result=compareProperty(region1, region2, sort3);
            return result;
        }
        
        
        private int compareProperty(Region region1,Region region2, String property) {
            if (property.equals("Position")) {
              if (orientation==Sequence.DIRECT) {
                  if (region1.isLocatedPriorTo(region2)) return -1;
                  else if (region2.isLocatedPriorTo(region1)) return 1;
                  else return 0;                  
              } else {
                  if (region1.isLocatedPriorToOnReverseStrand(region2)) return -1;
                  else if (region2.isLocatedPriorToOnReverseStrand(region1)) return 1;
                  else return 0;                  
              }
            } else if (property.equals("Type")) {
                return region1.getType().compareTo(region2.getType());
            } else if (property.equals("Score")) {
                float score1=region1.getScore();
                float score2=region2.getScore();
                if (score1>score2) return -1;
                else if (score2>score1) return 1;
                else return 0;
            } 
            return 0;
        }
        
    }
    
    @Override
    public Data parseInput(ArrayList<String> input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
         return (Data)parseInputToTarget(input, target, settings, task);         
    }
    
    @Override
    public DataSegment parseInput(ArrayList<String> input, DataSegment target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
        return (DataSegment)parseInputToTarget(input, target, settings, task);
    }
    
    /** The following method is a common substitute for the above 2 methods */
    private Object parseInputToTarget(ArrayList<String> input, Object target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
        String formatString="";
        String offsetString="";
        String orientation="";
        int relativeoffset=0;
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             formatString=(String)settings.getResolvedParameter("Format",defaults,engine); 
             offsetString=(String)settings.getResolvedParameter("Position",defaults,engine);
             relativeoffset=(Integer)settings.getResolvedParameter("Relative-offset",defaults,engine);
             orientation=(String)settings.getResolvedParameter("Orientation",defaults,engine);
          } 
          catch (ExecutionError e) {throw new ParseError("An error occurred while resolving parser parameters: "+e.getMessage());}
          catch (Exception ex) {throw new ParseError("An error occurred while resolving parser parameters: "+ex.getMessage());}
        } else {
           formatString=(String)getDefaultValueForParameter("Format");
           offsetString=(String)getDefaultValueForParameter("Position");
           relativeoffset=(Integer)getDefaultValueForParameter("Relative-offset");
           orientation=(String)getDefaultValueForParameter("Orientation");
        } 
        if (formatString!=null && (formatString.trim().isEmpty())) formatString=null;
        if (target==null) target=new RegionDataset("temporary");
        if (target instanceof RegionDataset) { // The RegionDataset might not contain RegionSequenceData objects for all Sequences. Add them if they are missing!
            RegionDataset dataset=(RegionDataset)target;
            ArrayList<Data> sequences=engine.getAllDataItemsOfType(Sequence.class);
            for (Data seq:sequences) {
                if (dataset.getSequenceByName(seq.getName())==null) dataset.addSequence(new RegionSequenceData((Sequence)seq));     
            }            
        }
        int count=0;
        for (String line:input) { // parsing each line in succession
            count++;
            if (count%200==0) {
               if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
               if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
               Thread.yield();
            }            
            if (line.startsWith("#") || line.isEmpty()) continue; // GFF comment line
            HashMap<String,Object> map=null;
            if (formatString!=null) map=parseSingleLineFromPattern(line,formatString);
            else map=parseSingleLineInStandardFormat(line);
            String sequenceName=(String)map.get("SEQUENCENAME");
            RegionSequenceData targetSequence=null;
            //System.err.println("Parsed line: sequenceName="+sequenceName);
            if (target instanceof RegionSequenceData) {
               targetSequence=(RegionSequenceData)target;
               if (targetSequence==null || !targetSequence.getName().equals(sequenceName)) continue; // the sequence mentioned in the file does not correspond to any known sequence
               addRegionToTarget(targetSequence,map,offsetString,relativeoffset,orientation);
            } else if (target instanceof RegionDataset) {
                targetSequence=(RegionSequenceData)((RegionDataset)target).getSequenceByName(sequenceName);
                if (targetSequence==null) continue; // the sequence mentioned in the file does not correspond to any known sequence
                addRegionToTarget(targetSequence,map,offsetString,relativeoffset,orientation);
            } else if (target instanceof DataSegment) {
                addRegionToTarget(target,map,offsetString,relativeoffset,orientation);
            } else throw new ParseError("SLOPPY PROGRAMMING ERROR: non-Region data as target for GFF dataformat: "+target.getClass().getSimpleName());
        }
        return target;      
    }
    
    
    private void addRegionToTarget(Object target, HashMap<String,Object> map, String offsetString, int relativeoffset, String orientationString) throws ParseError {                
                int start=0, end=0; // these are offset relative to the start of the parent sequence
                int targetOrientation=Sequence.DIRECT;
                int targetStart=0;
                int targetSize=0;
                if (target instanceof RegionSequenceData) {
                    targetOrientation=((RegionSequenceData)target).getStrandOrientation();
                    targetStart=((RegionSequenceData)target).getRegionStart();
                    targetSize=((RegionSequenceData)target).getSize();
                } else if (target instanceof DataSegment) {
                    targetStart=((DataSegment)target).getSegmentStart();
                    targetSize=((DataSegment)target).getSize();
                } else throw new ParseError("Target object neither RegionSequenceData nor DataSegment in DataFormat_GFF.addRegionToTarget():"+target.toString());
                Object startValue=map.get("START");
                if (startValue instanceof Integer) start=(Integer)startValue;
                Object endValue=map.get("END");
                if (endValue instanceof Integer) end=(Integer)endValue;
                float score=0;
                Object scoreValue=map.get("SCORE");
                if (scoreValue instanceof Float) score=(Float)scoreValue;
                String type=(String)map.get("TYPE");
                if (type==null) type="unknown_type";
                String annotatedOrientation=(String)map.get("STRAND"); // the orientation in the GFF file
                int annotatedStrand=Sequence.DIRECT; // the strand corresponding to + in the input
                     if (orientationString.equals(DIRECT)) annotatedStrand=Sequence.DIRECT;
                else if (orientationString.equals(REVERSE)) annotatedStrand=Sequence.REVERSE;
                else if (orientationString.equals(FROM_GENE)) {
                   annotatedStrand=targetOrientation;
                } else if (orientationString.equals(OPPOSITE)) {
                   annotatedStrand=targetOrientation;
                   if (annotatedStrand==Sequence.DIRECT) annotatedStrand=Sequence.REVERSE;
                   else annotatedStrand=Sequence.DIRECT;
                }
                int orientation=Sequence.DIRECT;    
                if (annotatedOrientation==null) annotatedOrientation="+";
                if (offsetString.equals("Genomic")) { // always use "DIRECT" orientation for genomic positions
                    start-=targetStart;
                    end-=targetStart;
                         if (annotatedOrientation.equals("+")) orientation=Region.DIRECT;
                    else if (annotatedOrientation.equals("-")) orientation=Region.REVERSE;
                    else orientation=Region.INDETERMINED;
                } else { // relative offsets. And orientation could be DIRECT,REVERSE,FROM GENE or OPPOSITE
                   if (annotatedStrand==Sequence.DIRECT) {
                       start-=relativeoffset;
                       end-=relativeoffset;
                            if (annotatedOrientation.equals(".")) orientation=Region.INDETERMINED;
                       else if (annotatedOrientation.equals("+")) orientation=Region.DIRECT;
                       else orientation=Region.REVERSE;
                   } else { // relative offset and reverse orientation
                       int annotatedStart=start;
                       int annotatedEnd=end;
                       start=(targetSize-1)-annotatedEnd+relativeoffset;
                       end=(targetSize-1)-annotatedStart+relativeoffset;
                            if (annotatedOrientation.equals(".")) orientation=Region.INDETERMINED;
                       else if (annotatedOrientation.equals("+")) orientation=Region.REVERSE;
                       else orientation=Region.DIRECT;
                   }
                }               
                if (end<0 || start>=targetSize) return; // region is outside sequence
                RegionSequenceData parentSequence=null;
                if (target instanceof RegionSequenceData) parentSequence=(RegionSequenceData)target;                        
                Region newRegion=new Region(parentSequence, start, end, type, score, orientation);
                for (String property:map.keySet()) {
                    if (property.equalsIgnoreCase("SEQUENCENAME") || 
                        property.equalsIgnoreCase("FEATURE") || 
                        property.equalsIgnoreCase("SOURCE") || 
                        property.equalsIgnoreCase("START") || 
                        property.equalsIgnoreCase("END") || 
                        property.equalsIgnoreCase("SCORE") || 
                        property.equalsIgnoreCase("TYPE") ||                         
                        property.equalsIgnoreCase("STRAND")) continue;
                    else {
                        newRegion.setProperty(property, map.get(property));
                    }                    
                }
                //System.err.println("Add to ["+target.toString()+"]  offset="+offsetString+"  relative="+relativeoffset+"  orientionString="+orientationString+":  region="+newRegion.toString());
                if (target instanceof RegionSequenceData) ((RegionSequenceData)target).addRegion(newRegion);   
                else if (target instanceof DataSegment) ((DataSegment)target).addRegion(newRegion);
    }
    
    

    
    /** parses a single line in a GFF-file and returns a HashMap with the different properties (with values as strings!) according to the capturing groups in the formatString */
    private HashMap<String,Object> parseSingleLineInStandardFormat(String line) throws ParseError {
        HashMap<String,Object> result=new HashMap<String,Object>();
        String[] fields=line.split("\t");
        if (fields.length<8) throw new ParseError("Expected at least 8 fields per line in GFF-format. Got "+fields.length+":\n"+line);
        //System.err.println("Parsed standard: "+line+" =>"+fields[0]);
        result.put("SEQUENCENAME",fields[0]);
        result.put("FEATURE",fields[1]);
        result.put("SOURCE",fields[1]); // this is correct
        try {
            result.put("START",Integer.parseInt(fields[3]));        
        } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for START: "+e.getMessage());}
        try {
            result.put("END",Integer.parseInt(fields[4]));        
        } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for END: "+e.getMessage());}
        try {
            result.put("SCORE",Float.parseFloat(fields[5]));        
        } catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for SCORE: "+e.getMessage());}
        
        result.put("STRAND",fields[6]);
        if (fields.length>=9) {
            String[] attributes=fields[8].split(";");
            for (String attribute:attributes) {
                String[] pair=attribute.split("=");
                if (pair.length!=2) throw new ParseError("Attribute not in recognized 'key=value' format: "+attribute); 
                String key=pair[0].trim();
                if (key.equalsIgnoreCase("type")) key="TYPE";
                String value=pair[1].trim();
                if (value.startsWith("\"") && value.endsWith("\"")) value=value.substring(1, value.length()-1);
                result.put(key,value);
            }           
        }
        return result;
    }

    /** parses a single line in a GFF-file and returns a HashMap with the different properties (with values as strings!) according to the capturing groups in the formatString */
    private HashMap<String,Object> parseSingleLineFromPattern(String line, String formatString) throws ParseError {
        HashMap<String,Object> result=new HashMap<String,Object>();
        //System.err.println("Parsed pattern: "+line);
        Pattern pattern=Pattern.compile("\\{([A-Z]+)\\}");
        Matcher matcher=pattern.matcher(formatString);
        ArrayList<String> fields=new ArrayList<String>();
        while (matcher.find()) {  
             String field=matcher.group(1);
             fields.add(field);
        }
        //formatString=formatString.replaceAll("\t", "\\t"); //
        formatString=formatString.replaceAll("\\{([A-Z]+)\\}", "([^\t]+)"); //
        pattern=Pattern.compile(formatString);
        matcher=pattern.matcher(line);
        if (matcher.find()) {
           for (int i=1;i<=matcher.groupCount();i++) {
               String field=fields.get(i-1);
               String matchString=matcher.group(i);
               Object value;
               if (field.equals("START") || field.equals("END")) {
                   try {value=Integer.parseInt(matchString);} catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for "+field+": "+e.getMessage());}
               } else if (field.equals("SCORE")) {
                   try {value=Float.parseFloat(matchString);} catch (NumberFormatException e) {throw new ParseError("Unable to parse expected numerical value for "+field+": "+e.getMessage());}           
               } else value=matchString;
               result.put(field,value);
           } 
        } else throw new ParseError("Unable to parse GFF line: "+line);

        return result;
    }

}

        
       
        
        
