/*
 
 
 */

package priorseditor.engine.dataformat;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import javax.imageio.ImageIO;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.ExecutionError;

import priorseditor.engine.ParameterSettings;
import priorseditor.engine.Parameter;
import priorseditor.engine.protocol.ParseError;
import priorseditor.engine.data.Motif;
import priorseditor.engine.data.MotifClassification;
import priorseditor.engine.data.MotifCollection;
import priorseditor.gui.SequenceLogo;

/**
 *
 * @author kjetikl
 */
public class DataFormat_HTML_MotifTable extends DataFormat {
    private String name="HTML_MotifTable";

    private Class[] supportedTypes=new Class[]{MotifCollection.class, Motif.class};
    private String[] allfields=new String[]{"Motif ID","Short name","Long name","Motif","Size","Information content","Classification","Binding factors","Organisms"};
    
    public DataFormat_HTML_MotifTable() {
        addOptionalParameter("Motif ID",Boolean.TRUE, new Boolean[]{true,false},null);
        addOptionalParameter("Short name",Boolean.TRUE, new Boolean[]{true,false},null);
        addOptionalParameter("Long name",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Motif",Boolean.TRUE, new Boolean[]{true,false},null);
        addOptionalParameter("Size",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Information content",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Classification",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Binding factors",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Organisms",Boolean.FALSE, new Boolean[]{true,false},null);
        addOptionalParameter("Sequence logo height",0, new Integer[]{0,100},"<html>Height of sequence logo.<br>To output without graphics set height=0</html>");
        addOptionalParameter("Sequence logo max width",0, new Integer[]{0,10000},"<html>Maximum width of sequence logos<br>If this is set, logos will be scaled to fit within the limit.<br>A value of 0 means no limit</html>");
        addOptionalParameter("Multiline",Boolean.TRUE, new Boolean[]{true,false},"If selected, fields with many entries will be split over several lines");        
        addOptionalParameter("Sort by","Motif ID", new String[]{"Motif ID","Short name","Long name","Size","Information content","Classification"},null);        
        addOptionalParameter("Sort direction","Ascending", new String[]{"Ascending","Descending"},null);        
    }
        
    @Override
    public String getName() {
        return name;
    }
    
    @Override
    public boolean canFormatOutput(Data data) {
        return (data instanceof MotifCollection || data instanceof Motif);
    }
    
    @Override
    public boolean canFormatOutput(Class dataclass) {
        return (dataclass.equals(MotifCollection.class) || dataclass.equals(Motif.class));
    }
    
    @Override
    public boolean canParseInput(Data data) {
        return false;
    }
    
    @Override
    public boolean canParseInput(Class dataclass) {
        return false;
    }
    
      
    @Override
    public Class[] getSupportedDataTypes() {
        return supportedTypes;
    }
    
    @Override
    public String getSuffix() {
        return "html";
    }

    @Override
    public boolean isAppendable() {
        return false;
    }
    
    @Override
    public OutputData format(Data dataobject, OutputData outputobject, ParameterSettings settings, ExecutableTask task) throws ExecutionError, InterruptedException {
        if (!canFormatOutput(dataobject)) {
            if (task!=null) throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format",task.getLineNumber());
            else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        }        
        int sequenceLogoHeight=0; 
        int sequenceLogoMaxWidth=0; 
        boolean multiline=true;
        String sortby="Motif ID";
        String sortdir="Ascending";
        HashSet<String> fieldsToShow=new HashSet<String>();
        setProgress(5);
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             sequenceLogoHeight=(Integer)settings.getResolvedParameter("Sequence logo height",defaults,engine); 
             sequenceLogoMaxWidth=(Integer)settings.getResolvedParameter("Sequence logo max width",defaults,engine); 
             multiline=(Boolean)settings.getResolvedParameter("Multiline",defaults,engine); 
             sortby=(String)settings.getResolvedParameter("Sort by",defaults,engine); 
             sortdir=(String)settings.getResolvedParameter("Sort direction",defaults,engine); 
             
             for (String field:allfields) {
                 if ((Boolean)settings.getResolvedParameter(field,defaults,engine)) fieldsToShow.add(field);
             }             
          } catch (ExecutionError e) {
             throw e;
          } catch (Exception ex) {
              throw new ExecutionError("An error occurred during output formatting", ex);
          }
        } else {
           sequenceLogoHeight=(Integer)getDefaultValueForParameter("Sequence logo height");
           sequenceLogoMaxWidth=(Integer)getDefaultValueForParameter("Sequence logo max width");
           multiline=(Boolean)getDefaultValueForParameter("multiline");
           sortby=(String)getDefaultValueForParameter("Sort by");
           sortdir=(String)getDefaultValueForParameter("Sort direction");
           for (String field:allfields) {
               if ((Boolean)getDefaultValueForParameter(field)) fieldsToShow.add(field);
           }           
        }        

        Color [] basecolors=new Color[]{Color.GREEN,Color.BLUE,new Color(220,220,0),Color.RED};   
             
        SequenceLogo sequencelogo=new SequenceLogo(basecolors,(int)(sequenceLogoHeight*1.15),(int)(sequenceLogoHeight/4f), Color.WHITE);
        sequencelogo.setMaxWidth(sequenceLogoMaxWidth);
        outputobject.setShowAsHTML(true);
        outputobject.append("<html>" +
                "<head>" +
                "<style TYPE=\"text/css\">table {border-collapse: collapse;} td {text-align:left;border-width: 1px;border-style: solid;border-color: #000000;padding:3px;border-collapse: collapse;} th {text-align:center;border-width: 1px;border-style: solid;border-color: #000000;padding:3px;border-collapse: collapse;}" +
                "</style>" +
                "</head><body><table><tr>","HTML");  
        
        for (String field:allfields) {
           boolean show=fieldsToShow.contains(field);
           if (field.equals("Information content")) field="IC";
           if (show) outputobject.append("<th bgcolor=\"#CCCCCC\">"+field+"</th>","HTML");
        }              
        outputobject.append("</tr>","HTML");
        StringBuilder outputString=new StringBuilder();
        if (dataobject instanceof MotifCollection) {
            ArrayList<Motif> motiflist=((MotifCollection)dataobject).getAllMotifs(engine);
            Collections.sort(motiflist,new SortOrderComparator(sortby,sortdir.equals("Ascending")));
            int size=motiflist.size();
            int i=0;

            for (Motif motif:motiflist) {
                if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
                if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
                outputMotif(motif, outputobject, sequencelogo,sequenceLogoHeight,fieldsToShow,multiline,outputString);
                // task.setStatusMessage("Motif "+(i+1)+" of "+size);
                setProgress(i+1, size);
                i++;
                if (i%5==0) Thread.yield();                                 
            } 
        } else if (dataobject instanceof Motif){
            outputMotif((Motif)dataobject, outputobject, sequencelogo,sequenceLogoHeight,fieldsToShow,multiline,outputString);
        } else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        outputobject.append(outputString.toString(),"HTML");
        outputobject.append("</table></html>","HTML");
        return outputobject;
    }    
    
    
    /** outputformats a single motif */
    private void outputMotif(Motif motif, OutputData outputobject, SequenceLogo sequencelogo, int height, HashSet<String> fieldsToShow, boolean multiline, StringBuilder line) {
        line.append("<tr>");
        for (String field:allfields) {
           if (!fieldsToShow.contains(field)) continue;
           String value="";
           if (field.equals("Motif")) {
                value=motif.getConsensusMotif();
                if (height>0) {
                    File imagefile=outputobject.createDependentFile(engine,"gif");
                    try {
                        sequencelogo.setMotif(motif);
                        saveSequenceLogoImage(imagefile,sequencelogo, height);
                    } catch (IOException e) {
                        engine.errorMessage("An error occurred when creating image file: "+e.toString(),0);
                    }
                    value="<img src=\"file:///"+imagefile.getAbsolutePath()+"\" />";
                }               
           } else value=getPropertyAsString(motif, field, multiline);
           if (value==null) value="";
           line.append("<td>");
           line.append(value);
           line.append("</td>");
        }  
        line.append("</tr>");  
    }

      

    @Override
    public Data parseInput(ArrayList<String> input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException  {
       throw new ParseError("Unable to parse Motif input with data format HTML_MotifTable");
    }
    

    private void saveSequenceLogoImage(File file, SequenceLogo sequencelogo, int motifheight) throws IOException {              
        int width=sequencelogo.getDefaultMotifWidth();
        BufferedImage image=new BufferedImage(width, motifheight, BufferedImage.TYPE_INT_RGB);        
        Graphics2D g=image.createGraphics();    
        g.setColor(Color.WHITE);
        g.fillRect(0, 0, width, motifheight);
        sequencelogo.paintLogo(g);        
        ImageIO.write(image, "gif", file);
        g.dispose();            
    }
    
    
    private String getPropertyAsString(Motif motif, String property, boolean multiline) {
             if (property.equals("Motif ID")) {return motif.getName();}     
        else if (property.equals("Short name")) {return "<nobr>"+motif.getShortName()+"</nobr>";}
        else if (property.equals("Long name")) {return "<nobr>"+motif.getLongName()+"</nobr>";}
        else if (property.equals("Size")) {return ""+motif.getLength();}
        else if (property.equals("Information content")) {return ""+motif.getICcontent();}
        else if (property.equals("Classification")) {
            String classification=MotifClassification.getFullLevelsStringAsHTML(motif.getClassification());
            classification=classification.replace("<html>","");
            classification=classification.replace("</html>","");
            if (!multiline) classification=classification.replace("<br>",", ");
            else classification=classification.replace("<br>","</nobr><br><nobr>");
            return "<nobr>"+classification+"</nobr>"; 
        }
        else if (property.equals("Binding factors")) {
            String bindingfactors=motif.getBindingFactors();
            if (bindingfactors==null) return null;
            if (multiline) bindingfactors=bindingfactors.replace(",", "</nobr><br><nobr>");
            return "<nobr>"+bindingfactors+"</nobr>";
        }
        else if (property.equals("Organisms")) {
            String organisms=motif.getOrganisms();
            if (organisms==null) return null;
            if (multiline) organisms=organisms.replace(",", "</nobr><br><nobr>");
            return "<nobr>"+organisms+"</nobr>";    
        }
        else return "**** UNKNOWN ***";
    }
    
    private class SortOrderComparator implements Comparator<Motif> {
            private String property=null;
            private boolean ascending=true;
            
            public SortOrderComparator(String property, boolean ascending) {
                this.property=property;
                this.ascending=ascending;
            }
        
            @Override
            public int compare(Motif motif1, Motif motif2) { // these are two motifnames
                int result=compareProperty(motif1, motif2, property);
                if (result==0) {
                   result=compareProperty(motif1, motif2, "Motif ID"); // this should not result in 0 
                }
                if (ascending) return result;
                else return (result*(-1));
            }    
            
            public int compareProperty(Motif motif1, Motif motif2, String property) {
                     if (property.equals("Motif ID")) {return motif1.getName().compareTo(motif2.getName());}
                else if (property.equals("Short name")) {return motif1.getShortName().compareToIgnoreCase(motif2.getShortName());}
                else if (property.equals("Long name")) {return motif1.getLongName().compareToIgnoreCase(motif2.getLongName());}
                else if (property.equals("Size")) {return ((Integer)motif1.getLength()).compareTo((Integer)motif2.getLength());}
                else if (property.equals("Information content")) {return ((Float)motif1.getICcontent()).compareTo((Float)motif2.getICcontent());}
                else if (property.equals("Classification")) {return motif1.getClassification().compareTo(motif2.getClassification());}
                else return 0;
            } 
    }      
    
}

        
       
        
        
