/*
 
 
 */

package priorseditor.engine.dataformat;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.OutputData;
import priorseditor.engine.data.NumericDataset;
import priorseditor.engine.data.NumericSequenceData;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.OperationTask;
import priorseditor.engine.Parameter;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.DataSegment;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.data.SequenceCollection;
import priorseditor.engine.protocol.ParseError;

/**
 *
 * @author kjetikl
 */
public class DataFormat_Priority extends DataFormat {
    private String name="PRIORITY";    
    private Class[] supportedTypes=new Class[]{NumericSequenceData.class, NumericDataset.class};

    private static final String DIRECT="Direct";
    private static final String REVERSE="Reverse";
    private static final String FROM_GENE="From Gene";
    private static final String OPPOSITE="Opposite";
    private static final String PARAMETER_STRAND_ORIENTATION="Orientation";
    private static final String PARAMETER_SEPARATOR="Separator";
    private static final String TAB_SEPARATOR="TAB";
    private static final String COMMA_SEPARATOR="Comma";   
    private static final String SPACE_SEPARATOR="Space";     
    
    public DataFormat_Priority() {
        addParameter(PARAMETER_STRAND_ORIENTATION, FROM_GENE, new String[]{DIRECT,REVERSE,FROM_GENE,OPPOSITE},"Sequence orientation");
        addParameter(PARAMETER_SEPARATOR, COMMA_SEPARATOR, new String[]{COMMA_SEPARATOR,TAB_SEPARATOR,SPACE_SEPARATOR},"Character that separates values");
    }
        
    @Override
    public String getName() {
        return name;
    }
    
    @Override
    public boolean canFormatOutput(Data data) {
        return (data instanceof NumericSequenceData || data instanceof NumericDataset);
    }
    
    @Override
    public boolean canFormatOutput(Class dataclass) {
        return (dataclass.equals(NumericSequenceData.class) || dataclass.equals(NumericDataset.class));
    }
    
    @Override
    public boolean canParseInput(Data data) {
        return (data instanceof NumericSequenceData || data instanceof NumericDataset);
    }
    
    @Override
    public boolean canParseInput(Class dataclass) {
        return (dataclass.equals(NumericSequenceData.class) || dataclass.equals(NumericDataset.class));
    }
    
    @Override
    public Class[] getSupportedDataTypes() {
        return supportedTypes;
    }
    
    @Override
    public String getSuffix() {
        return "prior";
    }
 
    @Override
    public OutputData format(Data dataobject, OutputData outputobject, ParameterSettings settings, ExecutableTask task) throws ExecutionError, InterruptedException {
        if (!canFormatOutput(dataobject)) {
            if (task!=null) throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format",task.getLineNumber());
            else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        }        
        setProgress(5);
        String orientation;
        String separator=COMMA_SEPARATOR;
        if (settings!=null) {
          try{
             Parameter[] defaults=getParameters();
             orientation=(String)settings.getResolvedParameter(PARAMETER_STRAND_ORIENTATION,defaults,engine);
             separator=(String)settings.getResolvedParameter(PARAMETER_SEPARATOR,defaults,engine);             
          } catch (ExecutionError e) {
             throw e;
          } catch (Exception ex) {
              throw new ExecutionError("An error occurred during output formatting", ex);
          }
        } else {
           orientation=(String)getDefaultValueForParameter(PARAMETER_STRAND_ORIENTATION);
           separator=(String)getDefaultValueForParameter(PARAMETER_SEPARATOR);
        }
        if (separator.equals(TAB_SEPARATOR)) separator="\t";
        else if (separator.equals(COMMA_SEPARATOR)) separator=",";
        else separator=",";
        String outputString="";
        if (dataobject instanceof NumericDataset) {
            SequenceCollection sequenceCollection=null;
            if (task instanceof OperationTask) {
                String subsetName=(String)((OperationTask)task).getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
                if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
                Data seqcol=engine.getDataItem(subsetName);
                if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
                if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
                sequenceCollection=(SequenceCollection)seqcol;
            }
            if (sequenceCollection==null) sequenceCollection=engine.getDefaultSequenceCollection();                   
            outputString=outputMultipleSequences((NumericDataset)dataobject, sequenceCollection, orientation, separator, task, engine);
        } else if (dataobject instanceof NumericSequenceData){
            StringBuilder builder=new StringBuilder();
            outputSequence((NumericSequenceData)dataobject, orientation, separator, builder);
            outputString=builder.toString();
        } else throw new ExecutionError("Unable to format object '"+dataobject+"' in "+name+" format");
        outputobject.append(outputString,getName());
        setProgress(100);
        return outputobject;
    }    
    
    /** Outputs multiple sequences (i.e. a whole FeatureDataset)
     *  The sequences are output in the order determined by the DefaultSequenceCollection (as shown in the GUI)
     */
    private String outputMultipleSequences(NumericDataset sourcedata, SequenceCollection collection, String orientation, String separator, ExecutableTask task, PriorsEditorEngine engine) throws InterruptedException {
        StringBuilder outputString=new StringBuilder();        
        int size=collection.getNumberofSequences();
        int i=0;
        ArrayList<Sequence> sequences=collection.getAllSequencesInDefaultOrder(engine);
        for (Sequence sequence:sequences) { // for each sequence   
              String sequenceName=sequence.getName();
              NumericSequenceData sourceSequence=(NumericSequenceData)sourcedata.getSequenceByName(sequenceName);
              outputSequence(sourceSequence, orientation, separator, outputString);
              if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
              if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
              if (i%3==0) Thread.yield();
              setProgress(i+1,size);
              i++;
        }      
        return outputString.toString();
    } 
    
    
    /** outputformats a single sequence into StringBuilder*/
    private void outputSequence(NumericSequenceData sequence, String orientationString, String separator, StringBuilder outputString) {
            int shownStrand=Sequence.DIRECT;
                if (orientationString.equals(DIRECT)) shownStrand=Sequence.DIRECT;
            else if (orientationString.equals(REVERSE)) shownStrand=Sequence.REVERSE;
            else if (orientationString.equals(FROM_GENE)) {
               shownStrand=sequence.getStrandOrientation();
            } else if (orientationString.equals(OPPOSITE)) {
               shownStrand=sequence.getStrandOrientation();
               if (shownStrand==Sequence.DIRECT) shownStrand=Sequence.REVERSE;
               else shownStrand=Sequence.DIRECT;
            }
            outputString.append(">");
            outputString.append(sequence.getSequenceName());
            outputString.append("\n");
            int start=sequence.getRegionStart();
            int end=sequence.getRegionEnd();
            if (shownStrand==Sequence.DIRECT) {
                for (int i=start;i<=end;i++) {
                    float value=sequence.getValueAtGenomicPosition(i);
                    outputString.append(value);
                    outputString.append(separator);
                }
            } else {// output reverse strand
                for (int i=end;i>=start;i--) {
                    float value=sequence.getValueAtGenomicPosition(i);
                    outputString.append(value);
                    outputString.append(separator);
                }                
            }
            outputString.setCharAt(outputString.length()-1, '\n');// change last comma to newline
    }

    @Override
    public Data parseInput(ArrayList<String> input, Data target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException {
        if (target==null) {
            target=new NumericDataset("temporary");
        }
        if (target instanceof NumericDataset) setupNumericDataset((NumericDataset)target);
        if (!(target instanceof NumericSequenceData || target instanceof NumericDataset)) throw new ParseError("SLOPPY PROGRAMMING ERROR: non-NumericSequenceData passed as parameter to parseInput in DataFormat_Priority");        
        if (input.size()<1) return target; // throw new ParseError("Empty input");
        String orientation;
        String separator;
        if (settings!=null) {
           try {
              Parameter[] defaults=getParameters();
              orientation=(String)settings.getResolvedParameter(PARAMETER_STRAND_ORIENTATION,defaults,engine);          
              separator=(String)settings.getResolvedParameter(PARAMETER_SEPARATOR,defaults,engine);          
           } catch (Exception ex) {
              throw new ParseError("An error occurred during parsing:"+ex.getMessage());
           }
        } else {
            orientation=(String)getDefaultValueForParameter(PARAMETER_STRAND_ORIENTATION);
            separator=(String)getDefaultValueForParameter(PARAMETER_SEPARATOR);
        }
        if (separator.equals(TAB_SEPARATOR)) separator="\t";
        else if (separator.equals(COMMA_SEPARATOR)) separator=",";
        else separator=",";        
        if (target instanceof NumericSequenceData) target=parseSingleSequenceInput(input, (NumericSequenceData)target, orientation, separator);
        else if (target instanceof NumericDataset) target=parseMultipleSequenceInput(input, (NumericDataset)target, orientation, separator, task);
        else throw new ParseError("SLOPPY PROGRAMMING ERROR: non-DNASequenceData passed as parameter to parseInput in DataFormat_FASTA");
        return target;
    }
    
    

    
    private NumericSequenceData parseSingleSequenceInput(ArrayList<String> input, NumericSequenceData target, String orientationString, String separator) throws ParseError, InterruptedException {
        int useStrand=determineOrientation(target,orientationString);
        String complete="";
        int last=input.size();
        int first=-1;
        int i=0;
        for (String line:input) {
            if (line.startsWith(">")) {
                String targetName=getSequenceNameFromHeader(line);
                if (targetName==null) throw new ParseError("Unable to extract sequence name from header: "+line);
                if (targetName.equals(target.getSequenceName())) first=i;
                else {
                    if (first<0) continue; // sequence not found yet
                    else {last=i;break;} // found sequence after target sequence
                }
            }
            i++;
        }
        if (first<0) return target; // sequence not found in file
        List<String> sub=input.subList(first, last);
        for (String string:sub) complete+=string.trim();   
        target.setData(parseSingleSequence(target.getSize(), complete, useStrand, separator));
        return target;
    }
    
    private int determineOrientation(NumericSequenceData target, String orientationString) {
        int useStrand=Sequence.DIRECT;
            if (orientationString.equals(DIRECT)) useStrand=Sequence.DIRECT;
        else if (orientationString.equals(REVERSE)) useStrand=Sequence.REVERSE;
        else if (orientationString.equals(FROM_GENE)) {
           useStrand=target.getStrandOrientation();
        } else if (orientationString.equals(OPPOSITE)) {
           useStrand=target.getStrandOrientation();
           if (useStrand==Sequence.DIRECT) useStrand=Sequence.REVERSE;
           else useStrand=Sequence.DIRECT;
        }
        return useStrand;
    }
    
    private float[] parseSingleSequence(int expectedsize, String valuesString, int useStrand, String separator) throws ParseError {
        if (separator.equals(",")) separator="\\s*"+separator+"\\s*";
        String[] elements=valuesString.split(separator);
        if (elements.length!=expectedsize) throw new ParseError("Number of values found for sequence ("+elements.length+") does not match expected sequence length ("+expectedsize+")");
        float[] values=new float[elements.length];
        if (useStrand==Sequence.DIRECT) {
            for (int i=0;i<elements.length;i++) {
                try {
                    values[i]=Float.parseFloat(elements[i]);                    
                } catch (NumberFormatException nfe) {throw new ParseError("Unable to parse expected numeric value: "+elements[i]);}
            }
        } else {
            for (int i=0;i<elements.length;i++) {
                try {
                    values[values.length-(i+1)]=Float.parseFloat(elements[i]);                    
                } catch (NumberFormatException nfe) {throw new ParseError("Unable to parse expected numeric value: "+elements[i]);}
            }            
        }
        return values;
    }
    

    private NumericDataset parseMultipleSequenceInput(ArrayList<String> input, NumericDataset target, String orientationString, String separator, ExecutableTask task) throws ParseError, InterruptedException {
        int first=-1; int last=0; int size=input.size();
        ArrayList<int[]> startandstop=new ArrayList<int[]>();
        for (int i=0;i<size;i++) {
            String line=input.get(i).trim();
            if (line.startsWith(">")) {
                if (first<0) {
                    first=i;
                } else {
                    last=i;
                    startandstop.add(new int[]{first,last});
                    first=i;
                    last=0;
                }
            }
        }
        if (first>0) {
            startandstop.add(new int[]{first,size});
        }
        if (first<0) return target; // sequence not found in file
        int count=0;
        for (int[] pair:startandstop) {
            count++;
            if (count%3==0) {
              if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
              if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
              Thread.yield();
            }             
            int start=pair[0]; int end=pair[1];
            if (end<=start+1) continue; // no sequence?
            String sequencename=getSequenceNameFromHeader(input.get(start));
            if (sequencename==null) throw new ParseError("Unable to extract sequence name from header: "+input.get(start));
            NumericSequenceData seq=(NumericSequenceData)target.getSequenceByName(sequencename);
            if (seq==null) continue; // unknown sequence
            String complete="";
            for (int j=start+1;j<end;j++) complete+=input.get(j).trim();
            int useStrand=determineOrientation(seq,orientationString);  
            seq.setData(parseSingleSequence(seq.getSize(), complete, useStrand, separator));
        }
        return target;
    }    
    
    
    
    
    private String getSequenceNameFromHeader(String header) {
        Pattern pattern=Pattern.compile(">\\s*(\\w+)");
        Matcher matcher=pattern.matcher(header);
        if (matcher.find()) { 
            String sequenceName=matcher.group(1);
            return sequenceName;
        } else return null;
    }    
    
    @Override
    public DataSegment parseInput(ArrayList<String> input, DataSegment target, ParameterSettings settings, ExecutableTask task) throws ParseError, InterruptedException {
        String complete="";
        int count=0;
        String separator;
        if (settings!=null) {
           try {
              Parameter[] defaults=getParameters();
               separator=(String)settings.getResolvedParameter(PARAMETER_SEPARATOR,defaults,engine);          
           } catch (Exception ex) {
              throw new ParseError("An error occurred during parsing:"+ex.getMessage());
           }
        } else {
            separator=(String)getDefaultValueForParameter(PARAMETER_SEPARATOR);
        }
        if (separator.equals(TAB_SEPARATOR)) separator="\t";
        else if (separator.equals(COMMA_SEPARATOR)) separator=",";
        else separator=",";          
        for (String line:input) {
            line=line.trim();
            if (line.startsWith(">")) continue;
            else if (line.isEmpty()) continue;
            else if (line.startsWith("minmotiflength")) continue;
            else if (line.startsWith("maxmotiflength")) continue;
            else complete+=line;
            count++;
            if (count%200==0) {
              if (task!=null) task.checkExecutionLock(); // checks to see if this task should suspend execution
              if (Thread.interrupted() || (task!=null && task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();                
              Thread.yield();
            }             
        }
        target.setSegmentData(parseSingleSequence(target.getSize(), complete, Sequence.DIRECT, separator));
        return target;
    }
    
    
    /** adds in missing sequences */
    private void setupNumericDataset(NumericDataset dataset) {
        ArrayList<Data> sequences=engine.getAllDataItemsOfType(Sequence.class);
        for (Data seq:sequences) {
            if (dataset.getSequenceByName(seq.getName())==null) dataset.addSequence(new NumericSequenceData((Sequence)seq,0f));     
        }          
    }
}
