/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.datatracks;


import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.data.DataSegment;

/**
 * A DataSource objects that retrieves data using a simple HTTP "GET" request to obtain a
 * webpage containing the relevant data. The webpage is then parsed by a DataFormat object
 * (according to the specified dataFormat) to construct a FeatureSequenceData object.
 * Information necessary to obtain the data includes the base-URL that points to the web
 * page from where the data can be obtained and a parameterStringTemplate which specifies
 * all additional settings. The parameter string should contain templates for CHROMOSOME,
 * START and END which will be substituted with correct values for each sequence before the
 * page is downloaded. Thus, the final http-request will look something like this: 
 * "http://baseURL?resolvedParameterString" 
 * 
 * @author Kjetil Klepper
 */
public class DataSource_http_GET extends DataSource {

    private String baseURL;
    private String parameterStringTemplate; // template for the "query" part of the URL
 
    /**
     * Creates a new instance of DataSource_http_GET based on the given arguments
     * @param name
     * @param organism
     * @param baseURL
     * @param dataFormatName
     * @param delay
     * @param parameterStringTemplate
     */
    public DataSource_http_GET(DataTrack datatrack, int organism, String genomebuild, String baseURL, String dataFormatName, String parameterStringTemplate) {
        super(datatrack,organism, genomebuild, dataFormatName);
        this.baseURL=baseURL;
        this.parameterStringTemplate=parameterStringTemplate;        
    }
    
    @Override
    public String getServerAddress() {
        try {
            URL url=new URL(baseURL);
            String host=url.getHost();
            return host;
        } catch (MalformedURLException e) {return null;}
    } 
    
    @Override
    public String getProtocol() {return HTTP_GET;}    
    
    public String getParameter() {
        return parameterStringTemplate;
    }
    public void setParameter(String parameter) {
        this.parameterStringTemplate=parameter;
    }
    public String getBaseURL() {
        return baseURL;
    }
    public void setBaseURL(String baseURL) {
        this.baseURL=baseURL;
    }    
    @Override
    public DataSegment loadDataSegment(DataSegment segment, ExecutableTask task) throws Exception {
        int timeout=dataloader.getEngine().getNetworkTimeout();
        String chromosome=segment.getChromosome();
        String start=""+segment.getSegmentStart();
        String end=""+segment.getSegmentEnd();
        String resolvedParameterString=parameterStringTemplate;
        resolvedParameterString=resolvedParameterString.replace(PATTERN_TEMPLATE_CHROMOSOME, chromosome);
        resolvedParameterString=resolvedParameterString.replace(PATTERN_TEMPLATE_START, start);
        resolvedParameterString=resolvedParameterString.replace(PATTERN_TEMPLATE_END, end); 
        if (!baseURL.startsWith("http://")) baseURL="http://"+baseURL;
        URL url=new URL(baseURL+"?"+resolvedParameterString);
        //System.err.println("Get page: "+url.toString());
        resolveDataFormat();       
        int waitperiod=getCurrentWaitingPeriod();
        if (waitperiod>0) {Thread.sleep(waitperiod);} // throws InterruptedException        
        ArrayList<String> page=getPage(url, timeout,task);
        if (dataformat!=null) dataformat.parseInput(page,segment,dataformatSettings, task);
        notifyDataSourceAccess();
        return segment;
    }
    
 
    
    private ArrayList<String> getPage(URL url, int timeout, ExecutableTask task) throws Exception {
        ArrayList<String> document=new ArrayList<String>();
        InputStream inputStream = null;
        BufferedReader dataReader = null;
        URLConnection connection=url.openConnection();
        connection.setConnectTimeout(timeout);
        inputStream=connection.getInputStream();
        dataReader = new BufferedReader(new InputStreamReader(inputStream));
        long count=0;
        String line;
        try {
           while ((line=dataReader.readLine())!=null){
              count++;
              if (count%100==0 && (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED))) throw new InterruptedException();
              document.add(line);
           }
        } catch (Exception e) {throw e;}
        finally {
            if (dataReader!=null) dataReader.close();
            if (inputStream!=null) inputStream.close();
        }    
        notifyDataSourceAccess(); 
        return document;
    }

    @Override
    public DataSource clone() {
        DataSource_http_GET copy=new DataSource_http_GET(dataTrack, organism, genomebuild, baseURL, dataformatName, parameterStringTemplate);
        copy.delay=this.delay;
        copy.dataformat=this.dataformat;
        copy.dataformatSettings=this.dataformatSettings;
        return copy;
    } 

    @Override
    public org.w3c.dom.Element getXMLrepresentation(org.w3c.dom.Document document) {
        org.w3c.dom.Element element = super.getXMLrepresentation(document);
        org.w3c.dom.Element protocol=document.createElement("Protocol");
        protocol.setAttribute("type", "GET");
        org.w3c.dom.Element url=document.createElement("URL");
        url.setTextContent(baseURL);
        org.w3c.dom.Element parameter=document.createElement("Parameter");
        parameter.setTextContent(parameterStringTemplate);
        protocol.appendChild(url);
        protocol.appendChild(parameter);
        element.appendChild(protocol);
        resolveDataFormat(); // try to locate data format based on name just in case
        if (dataformat!=null) {
            org.w3c.dom.Element dataformatelement=dataformat.getXMLrepresentation(document,dataformatSettings);
            element.appendChild(dataformatelement);
        }
        return element;
    }        

}
