/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.datatracks;

import java.util.ArrayList;
import java.util.HashSet;
import priorseditor.engine.data.DNASequenceDataset;
import priorseditor.engine.data.RegionDataset;
import priorseditor.engine.data.NumericDataset;

/**
 * The DataTrack objects make up a set of precollected (feature) datatracks that
 * is available for use in the engine. The DataTrack objects itself provides a 
 * general description of the data, such as a descriptive (and unique) name, 
 * the data type (numeric, region, sequence) etc.
 * 
 * Associated with the each DataTrack object is a list of DataSource objects that specify
 * sources for this track for individual organisms (and/or mirror-sites). 
 * 
 * @author Kjetil Klepper
 */
public class DataTrack implements Cloneable {
    public static final String NUMERIC_DATA="numeric";
    public static final String REGION_DATA="region";
    public static final String SEQUENCE_DATA="sequence";
        
    private String name;
    private Class datatype;
    private String description;
    private String sourceSite;
    private ArrayList<DataSource> datasources; // 
    
    private static int counter=1;
    private int myindex=0;
    
    public DataTrack(String name,Class datatype,String sourceSite,String description) {
        this.name=name;
        this.datatype=datatype;
        this.sourceSite=sourceSite;
        this.description=description;
        datasources=new ArrayList<DataSource>();
        myindex=counter++;
    }
      
    
    
    /**
     * Returns the name for this DataTrack
     * @return
     */
    public String getName() {
        return name;
    }
    
    /**
     * Returns the data type (as Class) for this DataTrack
     * @return
     */
    public Class getDataType() {
        return datatype;
    }
    
    /**
     * Returns a short description for this DataTrack
     * @return
     */
    public String getDescription() {
        return description;
    }
    
    /**
     * Returns a name of the source site for this Data Track (e.g. UCSC, NCBI, Ensembl)
     * @return
     */
    public String getSourceSite() {
        return sourceSite;
    }
    
    /**
     * Sets a new name for this DataTrack
     */
    public void setName(String newname) {
        name=newname;
    }
    
    /**
     * Sets a new data type for this DataTrack
     */
    public void setDataType(Class newclass) {
        datatype=newclass;
    }
    
    /**
     * Sets a short description for this DataTrack
     */
    public void setDescription(String newdescription) {
         description=newdescription;
    }
    
    /**
     * Sets a name of the source site for this Data Track (e.g. UCSC, NCBI, Ensembl)
     */
    public void setSourceSite(String newsource) {
        sourceSite=newsource;
    }
    
    
    @Override
    public String toString() {
        return name;
    }
    
    /**
     * Returns a list of available DataSource objects for the specified organism and genomebuild
     * or null if the organism/build is not supported
     * @param organism
     * @return
     */
    public DataSource[] getDataSources(int organism, String genomebuild) {
        int count=0;
        for (DataSource source:datasources) {
            if (source.getOrganism()==organism && source.getGenomeBuild().equals(genomebuild)) count++;
        }
        if (count==0) return null;
        DataSource[] sources=new DataSource[count];
        int i=0;
        for (DataSource source:datasources) {
            if (source.getOrganism()==organism && source.getGenomeBuild().equals(genomebuild)) {sources[i]=source;i++;}
        }
        return sources;
    }
         
    /** Returns an ordered list of datasources available for this DataTrack */
    public ArrayList<DataSource> getDatasources() {
        return datasources;
    }
    
    /**
     * Adds a new DataSource for this DataTrack
     * @param source
     * @return
     */
    public void addDataSource(DataSource source) {
        datasources.add(source);
    }
        
    /**
     * Replaces a  DataSource for this DataTrack
     * If the oldsource is not present in the list it will not be removed and 
     * the new source will be added to the end of the list
     * @param source
     * @return
     */
    public void replaceDataSource(DataSource oldsource, DataSource newsource) {
        int i=datasources.indexOf(oldsource);
        if (i>=0) datasources.set(i, newsource);
        else datasources.add(newsource);
    }
        
    /**
     * Returns true if this DataTrack is available for the specified Organism and genomebuild
     * @param organism Taxonomy ID for the organism (e.g. human=9606)
     */
    public boolean isSupported(int organism, String genomebuild) {
        for (DataSource source:datasources) {
            String sourcebuild=source.getGenomeBuild();
            if (source.getOrganism()==organism && sourcebuild!=null && sourcebuild.equals(genomebuild)) return true;
        }
        return false;
    }
    
    /**
     * Returns true if this DataTrack is available for the specified genomebuild
     * @param genomebuild A string specifying the genome build (e.g. "hg18" or "mm9")
     */
    public boolean isSupported(String genomebuild) {
        for (DataSource source:datasources) {
            String sourcebuild=source.getGenomeBuild();
            if (sourcebuild!=null && sourcebuild.equals(genomebuild)) return true;
        }
        return false;
    }
    
    
    /**
     * Returns true if this DataTrack is available for all the genome builds in the list
     * If the list is empty the return value is FALSE
     * @param builds List of genome builds for the organisms
     */
    public boolean isSupported(String[] builds) {
        if (builds==null || builds.length==0) return false;
        for (String build:builds) {
            if (!isSupported(build)) return false;
        }
        return true;
    }
    
//    /**
//     * Returns true if this DataTrack is available for all the Organisms in the list
//     * If the list is empty the return value is FALSE
//     * @param organisms List of taxonomy IDs for the organisms
//     */
//    public boolean isSupported(int[] organisms) {
//        if (organisms==null || organisms.length==0) return false;
//        for (int id:organisms) {
//            if (!isSupported(id)) return false;
//        }
//        return true;
//    }  
    
    /**
     * Returns a list of taxonomy IDs for the organisms that are supported for this DataTrack
     * @return
     */
    public int[] getSupportedOrganisms() {
        if (datasources==null) return new int[0];
        int size=datasources.size();
        HashSet<Integer> map=new HashSet<Integer>(3);
        for (DataSource source:datasources) {
            map.add(new Integer(source.getOrganism()));
        }
        int[] result=new int[map.size()];
        int i=0;
        for (Integer integer:map) {
            result[i]=integer.intValue();
            i++;
        }
        return result;
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public DataTrack clone() {
        DataTrack copy=new DataTrack(this.name, this.datatype, this.sourceSite, this.description);
        copy.datasources=new ArrayList<DataSource>(this.datasources.size());
        for (DataSource source:this.datasources) {
            DataSource clonedsource=(DataSource)source.clone();
            clonedsource.dataTrack=copy;
            copy.datasources.add(clonedsource);
        }
        return copy;
    }

     /** Returns an XML element for this DataTrack that can be included in DataConfiguration XML-documents */
    public org.w3c.dom.Element getXMLrepresentation(org.w3c.dom.Document document) {
        org.w3c.dom.Element element = document.createElement("DataTrack");
        String tracktype="";
             if (datatype==DNASequenceDataset.class) tracktype=SEQUENCE_DATA;
        else if (datatype==NumericDataset.class) tracktype=NUMERIC_DATA;
        else if (datatype==RegionDataset.class) tracktype=REGION_DATA;
        element.setAttribute("name", name);
        element.setAttribute("type", tracktype);
        element.setAttribute("source", sourceSite);
        element.setAttribute("description", description);
        for (DataSource source:datasources) {
            org.w3c.dom.Element child=source.getXMLrepresentation(document);
            element.appendChild(child);
        }
        return element;
    }      
    
    
    public void debug() {
        System.err.println("Track["+myindex+"]: "+name+" ["+datatype.toString()+"] {"+sourceSite+"} "+description);
          for (DataSource source:datasources) {
              source.debug();
          }
        
    }
    
}
