/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.operations;

import priorseditor.engine.OperationTask;
import java.util.ArrayList;
import java.util.Iterator;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.data.*;

/**
 * FeatureTransformOperation functions as a superclass for operations that implement simple transforms
 * of FeatureDataset objects. The superclass implements some functionality that is common for these transforms
 * and introduces a "inversion-of-control" paradigm whereby subclasses only need to implement the abstract
 * method {@code transformSequence} which performs the transformation of a single sequence. This method is
 * then called in turn for all applicable sequences in the chosen sequence collection.
 * @author kjetikl
 */
public abstract class FeatureTransformOperation extends Operation {

    
    /**
     * Override this function to implement transform of a single FeatureSequenceData object
     * The transformSequence method should call the boolean function {@code isConditionSatisfied} for each position along the sequence to verify
     * that the position should indeed be transformed according to the 'where-clause' of the operation task.
     * (if the isConditionSatisfied(i) function returns false then that position should be skipped, if true it should be transformed)
     * 
     * @param sourceSequence The original sequence to transformed. <b>NOTE:</b> This sequence should NOT be altered by the transform!!!
     * @param targetSequence The result of the transformation should be stored in this object
     * @return
     */   
    public abstract void transformSequence(final FeatureSequenceData sourceSequence, FeatureSequenceData targetSequence, OperationTask task) throws Exception;   

    /**
     * This method is called automatically before any transformation are executed
     * Subclasses should use this method to parse and resolve their own parameters
     */ 
    public abstract void resolveParameters(OperationTask task) throws Exception;   
       
    
    @Override
    public boolean execute(OperationTask task) throws Exception {
        if (engine.getDefaultSequenceCollection().size()==0) throw new ExecutionError("No sequences are selected");
        String sourceDatasetName=task.getSourceDataName();
        String targetDatasetName=task.getTargetDataName();
        FeatureDataset sourceDataset=(FeatureDataset)engine.getDataItem(sourceDatasetName);
        if (sourceDataset==null) throw new ExecutionError("Unable to locate data object '"+sourceDatasetName+"'",task.getLineNumber());
        if (!canUseAsSource(sourceDataset)) throw new ExecutionError(sourceDatasetName+"("+sourceDataset.getTypeDescription()+") is of a type not supported by the '"+getName()+"' operation",task.getLineNumber());
        FeatureDataset targetDataset=(FeatureDataset)sourceDataset.clone(); // Double-buffer.
        targetDataset.setName(targetDatasetName);
        
        Condition condition=(Condition)task.getParameter("where");
        if (condition!=null) condition.resolve(engine, task);  
        
        Condition_within within=(Condition_within)task.getParameter("within");
        if (within!=null) within.resolve(engine, task);        
        
        String subsetName=(String)task.getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
        if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
        Data seqcol=engine.getDataItem(subsetName);
        if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
        if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
        SequenceCollection sequenceCollection=(SequenceCollection)seqcol;
        task.setParameter(OperationTask.SOURCE, sourceDataset);
        task.setParameter(OperationTask.TARGET, targetDataset);
        task.setParameter(OperationTask.SEQUENCE_COLLECTION, sequenceCollection);
        resolveParameters(task);
        
        int i=0;
        ArrayList<Sequence> sequences=sequenceCollection.getAllSequences(engine);
        if (isSubrangeApplicable() && within!=null) { // remove sequences with no selection windows (if within-condition is used)
            Iterator iter=sequences.iterator();
            while (iter.hasNext()) {
                Sequence seq = (Sequence) iter.next();
                if (!within.existsSelectionWithinSequence(seq.getName(), task)) iter.remove();
            }           
        }
        int size=sequences.size();
        for (Sequence sequence:sequences) { // for each sequence
            task.checkExecutionLock(); // checks to see if this task should suspend execution
            if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
            String sequenceName=sequence.getName();

            FeatureSequenceData sourceSequence=(FeatureSequenceData)sourceDataset.getSequenceByName(sequenceName);
            FeatureSequenceData targetSequence=(FeatureSequenceData)targetDataset.getSequenceByName(sequenceName); // just in case...
            transformSequence(sourceSequence,targetSequence, task);         

            task.setStatusMessage("Executing operation: "+task.getOperationName()+" ("+(i+1)+"/"+size+")");
            task.setProgress(i+1, size);
            i++;
            Thread.yield();
        }   
        if (targetDataset instanceof NumericDataset) ((NumericDataset)targetDataset).updateAllowedMinMaxValuesFromData();
        task.checkExecutionLock(); // checks to see if this task should suspend execution
        if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
        targetDataset.setIsDerived(true);
        try {engine.updateDataItem(targetDataset);} catch (ClassCastException ce) {throw new ExecutionError("Incompatible assignment:"+ce.getMessage(),task.getLineNumber());}
        return true;
    }
       
    
    
    protected boolean positionSatisfiesCondition(String sequencename, int pos, OperationTask task) throws Exception {
        Condition_position condition=(Condition_position)task.getParameter("where");
        Condition_within within=(Condition_within)task.getParameter("within");  
        if (within!=null) {
            if (!within.isConditionSatisfied(sequencename, pos, task)) return false;
        }
        if (condition==null) return true; // no where condition is set, so the transform should be applied everywhere
        else return condition.isConditionSatisfied(sequencename, pos, task);
    }
    
   
    protected boolean regionSatisfiesCondition(String sequencename, Region region, OperationTask task) throws Exception {
        Condition_region condition=(Condition_region)task.getParameter("where");
        Condition_within within=(Condition_within)task.getParameter("within");  
        if (within!=null) {
            if (!within.isConditionSatisfied(sequencename, region.getGenomicStart(),region.getGenomicEnd(), task)) return false;
        }
        if (condition==null) return true; // no where condition is set, so the transform should be applied everywhere
        else return condition.isConditionSatisfied(sequencename, region, task);
    }
    
   
          
            
}

