/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.operations;

import priorseditor.engine.OperationTask;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.Data;
import priorseditor.engine.protocol.*;

/**
 * An operation is a language independent abstraction of 
 * @author kjetikl
 */
public abstract class Operation {
    protected PriorsEditorEngine engine;
    
    public Operation(PriorsEditorEngine engine) {
        this.engine=engine;
    }

    public Operation() { }

    /** 
     * Sets the engine for this operation. This method should always be called 
     * directly after instantiation of new operations with a zero-parameters constructor
     */
    public void setEngine(PriorsEditorEngine engine) {
        this.engine=engine;
    }
    /** 
     * Gets the engine for this operation.
     */
    public PriorsEditorEngine getEngine() {
       return engine;
    }
    
    /**
     * Returns a descriptive name or command for the operation
     * 
     * @return the name of the operation (command)
     */
    public abstract String getName();
    
    /**
     * Returns a short description of the command to be used for Tool Tips or similar
     * @return 
     */
    public abstract String getDescription();
    
     /**
     * Returns an array of classes that this operation can work on
     * @return 
     */  
    public abstract Class[] getDataSourcePreferences();
    
    /**
     * Returns an array of words that this operation needs to 'reserve'.
     * Reserved words can not be used as names for datatracks or variables in protocol scripts
     * (note that names in protocol-scripts are case-sensitive)
     * @return a list of reserved words
     */  
    public String[] getReservedWords() {return null;}
    
    /**
     * This function should return true if the supplied data argument is of
     * a type that the operation can work on, or false if the operation can
     * not be performed on this type of data
     * @param object
     * @return true if this operation can work on data of the specified type
     */
    public final boolean canUseAsSource(Data data) {
        if (data==null) return false;
        Class[] sources=getDataSourcePreferences();
        if (sources==null) return true;
        for (Class source:sources) {
            if (source.isInstance(data)) return true;
        }
        return false;
    }
    
    /**
     * This function should return true if the supplied class argument represents
     * a type that the operation can work on, or false if the operation can
     * not be performed on this class type
     * @param object
     * @return true if this operation can work on data of the specified class
     */
    @SuppressWarnings("unchecked")
    public final boolean canUseAsSource(Class type) {
        if (type==null) return false;
        Class[] sources=getDataSourcePreferences();
        if (sources==null) return true;
        for (Class source:sources) {
            if (source.isAssignableFrom(type)) return true;
        }
        return false;
    }
  
    /** 
     * This method should return TRUE if the operation can be applied to
     * subregions (windows) within a sequence. For instance, transform operations that
     * replace values at each position with a new value can usually be applied
     * to portions of a sequence, while the "convert" operation which converts
     * numeric data to regions data or vice versa must be applied to the whole
     * dataset or not at all
     */
    public boolean isSubrangeApplicable() {
        return false; 
    }
    
    /** 
     * This method should return TRUE if the operation can have multiple datasets
     * as input such as the combine_X operations
     */
    public boolean canHaveMultipleInput() {
        return false; 
    }
    
    
    /**
     * This function executes the operation according to the parameters specified
     * in the supplied object (containing information about source and target datasets,
     * as well as operation specific parameters)
     * 
     * @param parameters 
     * @return true if the execution was successful
     */
    public abstract boolean execute(OperationTask operationtask) throws Exception;
    
    /**
     * This function returns a textual representation of an operation (with arguments)
     * in the format of the specified Protocol
     * 
     * @param parameters 
     * @return A string that can be included in protocol scripts
     */
    public final String getCommandString(OperationTask task, Protocol protocol) {
        return protocol.getCommandString(task);
    }

    
  
}
