/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.operations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import priorseditor.engine.OperationTask;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.Parameter;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.*;
import priorseditor.external.ExternalProgram;
import priorseditor.external.MotifDiscovery;

/**
 *
 * @author kjetikl
 */
public class Operation_motifDiscovery extends Operation {
    private static final String name="motifDiscovery";
    private static final String description="Performs de novo motif discovery with a chosen algorithm";
    public static final String ALGORITHM="algorithm";
    public static final String PARAMETERS="parameters";
    public static final String MOTIFCOLLECTION="MotifCollection";
    public static final String MOTIFPREFIX="Motifs prefix";
    private Class[] datasourcePreferences=new Class[]{DNASequenceDataset.class};



    @Override
    public Class[] getDataSourcePreferences() {
        return datasourcePreferences;
    }

    @Override
    public String getDescription() {
        return description;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public String toString() {return name;}

    public void resolveParameters(OperationTask task, MotifDiscovery program) throws Exception {
        ParameterSettings parameterSettings=(ParameterSettings)task.getParameter(Operation_motifDiscovery.PARAMETERS);
        if (parameterSettings==null) throw new ExecutionError("Missing parameters for motif discovery algorithm");
        Parameter[] arguments=program.getParameters();
        for (int i=0;i<arguments.length;i++) {
            String parameterName=arguments[i].getName();
            Object value=parameterSettings.getResolvedParameter(parameterName, program.getParameters(), engine);
            if (arguments[i].isRequired() && (value==null || value.toString().isEmpty())) throw new ExecutionError("Missing value for required parameter '"+parameterName+"'");
            task.setParameter(parameterName,value);
        }
        String subsetName=(String)task.getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
        if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
        Data seqcol=engine.getDataItem(subsetName);
        if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
        if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
        SequenceCollection sequenceCollection=(SequenceCollection)seqcol;
        task.setParameter(OperationTask.SEQUENCE_COLLECTION, sequenceCollection);
        if (sequenceCollection.size()==0) throw new ExecutionError("Attempting to perform motif discovery on empty Sequence Collection '"+sequenceCollection.getName()+"'",task.getLineNumber());
        else if (sequenceCollection.size()==1) task.getEngine().logMessage("WARNING: Attempting to perform motif discovery on Sequence Collection '"+sequenceCollection.getName()+"' containing only 1 sequence");
    }



    @Override
    public boolean execute(OperationTask task) throws Exception {
        if (engine.getDefaultSequenceCollection().size()==0) throw new ExecutionError("No sequences are selected");
        String algorithmName=(String)task.getParameter(Operation_motifDiscovery.ALGORITHM);
        Object object=engine.getExternalProgram(algorithmName);
        if (object==null) throw new ExecutionError("Unknown algorithm "+algorithmName);
        if (!(object instanceof MotifDiscovery)) throw new ExecutionError(algorithmName+" is not a motif discovery algorithm");
        MotifDiscovery program = (MotifDiscovery)object;
        task.setStatusMessage("Executing operation: "+task.getOperationName());
        task.setProgress(5);
        String sourceName=task.getSourceDataName();
        String targetName=task.getTargetDataName();
        if (targetName==null || targetName.isEmpty()) throw new ExecutionError("Missing name for target data object",task.getLineNumber());
        String motifCollectionName=(String)task.getParameter(Operation_motifDiscovery.MOTIFCOLLECTION);
        if (motifCollectionName==null || motifCollectionName.isEmpty()) throw new ExecutionError("Missing name returned motif collection",task.getLineNumber());

        Data sourceData=engine.getDataItem(sourceName);
        if (sourceData==null) throw new ExecutionError("Unable to locate data object '"+sourceName+"'",task.getLineNumber());
        if (!canUseAsSource(sourceData)) throw new ExecutionError("Operation 'motifDiscovery' can not work on '"+sourceName+"'",task.getLineNumber());
        resolveParameters(task,program);
        RegionDataset targetData = performMotifDiscovery(program, (DNASequenceDataset)sourceData, targetName,motifCollectionName, task);
        targetData.setMotifTrack(true);
        MotifCollection motifsData=(MotifCollection)task.getParameter("Motifs");
        task.setProgress(100);
        task.checkExecutionLock(); // checks to see if this task should suspend execution
        if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
        try {engine.updateDataItem(targetData);} catch (ClassCastException ce) {throw new ExecutionError("Incompatible assignment: "+ce.getMessage(),task.getLineNumber());}
        try {engine.updateDataItem(motifsData);} catch (ClassCastException ce) {throw new ExecutionError("Incompatible assignment: "+ce.getMessage(),task.getLineNumber());}
        return true;
    }

    private RegionDataset performMotifDiscovery(MotifDiscovery program, DNASequenceDataset sourceData, String targetName,String motifCollectionName, OperationTask task) throws Exception {
        task.setParameter(ExternalProgram.SOURCES,new Data[]{sourceData});
        program.execute(task);
        String motifprefix=(String)task.getParameter(MOTIFPREFIX);
        if (motifprefix==null) motifprefix="MM";
        RegionDataset targetData=(RegionDataset)task.getParameter("Result");
        MotifCollection motifsData=(MotifCollection)task.getParameter("Motifs");
        filterDuplicateMotifs(targetData, motifsData);
        int numberofmotifsfound=motifsData.getPayloadSize(); // after filtering duplicates
        String[] newnames=task.getEngine().getNextAvailableDataNames(motifprefix, 5, numberofmotifsfound);
        HashMap<String,String> newnamesmap=new HashMap<String,String>(numberofmotifsfound);
        for (int i=0;i<numberofmotifsfound;i++) {
            String oldname=motifsData.renameMotifInPayload(i, newnames[i]);
            newnamesmap.put(oldname,newnames[i]);
        }
        motifsData.setName(motifCollectionName);
        annotateTFBS(targetData,sourceData,newnamesmap);
        targetData.setName(targetName);
        return targetData;
    }

    /** Adds a 'sequence' property to the Regions and renames the region if necessary*/
    private void annotateTFBS(RegionDataset targetData, DNASequenceDataset sourceData, HashMap<String,String> newnames) {
        ArrayList<FeatureSequenceData> list=targetData.getAllSequences();
        for (FeatureSequenceData sequenceData:list) {
            RegionSequenceData regionSequenceData=(RegionSequenceData)sequenceData;
            ArrayList<Region> regions=regionSequenceData.getAllRegions();
            for (Region region:regions) {
                int start=region.getGenomicStart();
                int end=region.getGenomicEnd();
                int orientation=region.getOrientation();
                String oldtype=region.getType();
                String newtype=newnames.get(oldtype);
                if (newtype==null) newtype=oldtype; // this should not happen
                region.setType(newtype);
                DNASequenceData dnaSeq=(DNASequenceData)sourceData.getSequenceByName(regionSequenceData.getName());
                char[] site=(char[])dnaSeq.getValueInGenomicInterval(start, end);
                if (orientation==Region.DIRECT) region.setProperty("sequence", new String(site));
                else region.setProperty("sequence", new String(PriorsEditorEngine.reverseSequence(site)));
            }
        }
    }


    /** Removes motifs that are duplicates of other motifs in the set */
    private void filterDuplicateMotifs(RegionDataset dataset, MotifCollection motifs) {
        int numberofmotifsfound=motifs.getPayloadSize();
        List<Motif> motiflist=motifs.getPayload();
        boolean[] issimilar=new boolean[numberofmotifsfound]; // flag which marks whether the motif at this index is similar to any previous motifs
        HashMap<String,ArrayList<int[]>> occurrences=separateMotifs(dataset);
        for (int i=1;i<numberofmotifsfound;i++) {
            for (int j=i-1;j>=0;j--) {
                String motifname1=motiflist.get(i).getName();
                String motifname2=motiflist.get(j).getName();
                ArrayList<int[]> list1=occurrences.get(motifname1);
                ArrayList<int[]> list2=occurrences.get(motifname2);
                if (motifOccurrencesEquals(list1, list2)) {issimilar[i]=true;break;}
            }
        }
        for (int i=numberofmotifsfound-1;i>=0;i--) {
            if (issimilar[i]) {
                String motifname1=motiflist.get(i).getName();
                motifs.removeMotifFromPayloadAtIndex(i);
                filterRegions(dataset,motifname1);
            }
        }
    }

    /** Returns TRUE if the two names motifs occupy the same sites in the region dataset */
    private HashMap<String,ArrayList<int[]>> separateMotifs(RegionDataset dataset) {
        HashMap<String,ArrayList<int[]>> result=new HashMap<String,ArrayList<int[]>>();
        ArrayList<FeatureSequenceData> sequences=dataset.getAllSequences();
        for (int i=0;i<sequences.size();i++) {
            RegionSequenceData seq=(RegionSequenceData)sequences.get(i);
            ArrayList<Region> allregions=((RegionSequenceData)seq).getOriginalRegions();
            for (Region region:allregions) {
                 String motiftype=region.getType();
                 if (!result.containsKey(motiftype)) result.put(motiftype, new ArrayList<int[]>());
                  result.get(motiftype).add(new int[]{i,region.getRelativeStart(),region.getRelativeEnd()});
            }
        }
        return result;
    }

    /** Returns true if the two lists contains regions that fully overlap each other
     *  The lists should sorted the same way
     */
    private boolean motifOccurrencesEquals(ArrayList<int[]> list1, ArrayList<int[]> list2) {
        if (list1.size()!=list2.size()) return false;
        for (int i=0;i<list1.size();i++) {
            int[] r1=list1.get(i);
            int[] r2=list2.get(i);
            if (r1[0]!=r2[0] || r1[1]!=r2[1] || r1[2]!=r2[2]) return false;
        }
        return true;
    }

    /** Removes all regions corresponding to the given typename from the dataset */
    private void filterRegions(RegionDataset dataset, String motifname) {
        ArrayList<FeatureSequenceData> sequences=dataset.getAllSequences();
        for (int i=0;i<sequences.size();i++) {
            RegionSequenceData seq=(RegionSequenceData)sequences.get(i);
            ArrayList<Region> allregions=((RegionSequenceData)seq).getOriginalRegions();
            ArrayList<Region> tobefiltered=new ArrayList<Region>();
            for (Region region:allregions) {
                 if (motifname.equals(region.getType())) tobefiltered.add(region);
            }
            for (Region region:tobefiltered) {
                seq.removeRegion(region);
            }
        }
    }
}