/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.operations;

import java.util.ArrayList;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.OperationTask;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.data.*;
import priorseditor.engine.data.FeatureSequenceData;

/**
 *
 * @author kjetikl
 */
public class Operation_score extends FeatureTransformOperation {
    
    public static final String MOTIFNAME="MotifName";
    private static final String MOTIF="Motif";
    
    private static final String name="score";
    private static final String description="Scores a DNA sequence with a single motif model";
    private Class[] datasourcePreferences=new Class[]{DNASequenceDataset.class};



    @Override
    public Class[] getDataSourcePreferences() {
        return datasourcePreferences;
    }

    @Override
    public String getDescription() {
        return description;
    }

    @Override
    public String getName() {
        return name;
    }
    
    @Override
    public String toString() {return name;}
    
    @Override
    public boolean isSubrangeApplicable() {return true;}
    
    @Override
    public void resolveParameters(OperationTask task) throws Exception {
        String motifName=(String)task.getParameter(MOTIFNAME);
        if (motifName==null || motifName.isEmpty()) throw new ExecutionError("Missing motif specification");     
        Object motif=engine.getDataItem(motifName);
        if (motif==null) throw new ExecutionError("No such data object: "+motifName);
        if (!(motif instanceof Motif || motif instanceof MotifCollection)) throw new ExecutionError(motifName+" is not a Motif or Motif Collection");            
        task.setParameter(MOTIF, motif);
      
    }

    
    @Override
    public void transformSequence(FeatureSequenceData sourceSequence, FeatureSequenceData targetSequence, OperationTask task) throws Exception {
        String seqname=sourceSequence.getName();        
        Object data=task.getParameter(MOTIF);
        if (data instanceof Motif) {
            Motif motif = (Motif)data;
            for (int i=sourceSequence.getRegionStart();i<=sourceSequence.getRegionEnd();i++) {
                if (positionSatisfiesCondition(seqname,i,task)) {
                    float score=matrixMatcher((DNASequenceData)sourceSequence,i,motif);
                   ((NumericSequenceData)targetSequence).setValueAtGenomicPosition(i,score);                   
                } // satisfies 'where'-condition
            }            
        } else if (data instanceof MotifCollection) {
            ArrayList<Motif> allMotifs=((MotifCollection)data).getAllMotifs(engine);
            for (int i=0;i<sourceSequence.getSize();i++) {
                ((NumericSequenceData)targetSequence).setValueAtRelativePosition(i, (float)(-Float.MAX_VALUE)); // set to lowest possible value
            }         
            for (Motif motif:allMotifs) { // search with each motif in turn
                for (int i=sourceSequence.getRegionStart();i<=sourceSequence.getRegionEnd();i++) {
                    if (positionSatisfiesCondition(seqname,i,task)) {
                        float score=matrixMatcher((DNASequenceData)sourceSequence,i,motif);
                        float currentscore=((NumericSequenceData)targetSequence).getValueAtGenomicPosition(i);        
                        if (score>currentscore) ((NumericSequenceData)targetSequence).setValueAtGenomicPosition(i,score);                   
                    } // satisfies 'where'-condition
                }
            }
        }
    }


    @Override
    public boolean execute(OperationTask task) throws Exception {
        if (engine.getDefaultSequenceCollection().size()==0) throw new ExecutionError("No sequences are selected");
        String sourceDatasetName=task.getSourceDataName();
        String targetDatasetName=task.getTargetDataName();
        FeatureDataset sourceDataset=(FeatureDataset)engine.getDataItem(sourceDatasetName);
        if (sourceDataset==null) throw new ExecutionError("Unable to locate data object '"+sourceDatasetName+"'",task.getLineNumber());
        if (!canUseAsSource(sourceDataset)) throw new ExecutionError(sourceDatasetName+"("+sourceDataset.getTypeDescription()+") is of a type not supported by the '"+getName()+"' operation",task.getLineNumber());
        FeatureDataset targetDataset=new NumericDataset(targetDatasetName);
        
        Condition condition=(Condition)task.getParameter("where");
        if (condition!=null) condition.resolve(engine, task);        
        resolveParameters(task);
        
        String subsetName=(String)task.getParameter(OperationTask.SEQUENCE_COLLECTION_NAME);
        if (subsetName==null || subsetName.isEmpty()) subsetName=engine.getDefaultSequenceCollectionName();
        Data seqcol=engine.getDataItem(subsetName);
        if (seqcol==null) throw new ExecutionError("No such collection: '"+subsetName+"'",task.getLineNumber());
        if (!(seqcol instanceof SequenceCollection)) throw new ExecutionError(subsetName+" is not a sequence collection",task.getLineNumber());
        SequenceCollection sequenceCollection=(SequenceCollection)seqcol;

        int size=sequenceCollection.getNumberofSequences();
        int i=0;
        ArrayList<Sequence> sequences=sequenceCollection.getAllSequences(engine);
        for (Sequence sequence:sequences) { // for each sequence
            task.checkExecutionLock(); // checks to see if this task should suspend execution
            if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
            String sequenceName=sequence.getName();
            FeatureSequenceData sourceSequence=(FeatureSequenceData)sourceDataset.getSequenceByName(sequenceName);
            FeatureSequenceData targetSequence=new NumericSequenceData(sequence,0);
            transformSequence(sourceSequence,targetSequence, task);             
            targetDataset.addSequence(targetSequence);                    
            task.setStatusMessage("Executing operation: "+task.getOperationName()+" ("+(i+1)+"/"+size+")");
            task.setProgress(i+1, size);
            i++;
            Thread.yield();
        }   
        ((NumericDataset)targetDataset).updateAllowedMinMaxValuesFromData();
        task.checkExecutionLock(); // checks to see if this task should suspend execution
        if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
        targetDataset.setIsDerived(true);
        try {engine.updateDataItem(targetDataset);} catch (ClassCastException ce) {throw new ExecutionError("Incompatible assignment:"+ce.getMessage(),task.getLineNumber());}
        return true;
    }    
    
    
    private float matrixMatcher(DNASequenceData sourceSequence,int genomicPosition, Motif motif) {
         int motifsize=motif.getLength();
         if (genomicPosition>sourceSequence.getRegionEnd()-motifsize+1) return 0; // unable to match near end
         char[] subseq=(char[])sourceSequence.getValueInGenomicInterval(genomicPosition, genomicPosition+motifsize-1);
         float score=0;         
         for (int i=0;i<subseq.length;i++) {
             float logodds=(float)Math.log((motif.getBaseFrequency(i, subseq[i])+0.01)/(0.25+0.04));
             score+=logodds;
         }
         return score;
    }
}
