/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.protocol;

import java.util.ArrayList;
import java.util.HashMap;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.Data;

/**
 * This class represents a look-up table that which maps data object names to their data types (class)
 * Such a table should contain all references to all "known" and (hitherto) "unknown" data objects
 * in the system. The table is populated from data items registered in the engine or by parsing protocol
 * scripts (that have not necessarily been executed yet) looking for potential data types. 
 * DataTypeTables can be used by protocol editors for beautyfication (coloring Data items) or by operation 
 * edit dialogs to correctly populateFromEngine drop-down menus with Data items of different types.
 * 
 * @author kjetikl
 */
public class DataTypeTable {

    private HashMap<String,Class> map;
    private PriorsEditorEngine engine;
    
    public DataTypeTable(PriorsEditorEngine engine) {
        this.engine=engine;
        map=new HashMap<String, Class>();
    }
    
    /**
     * Returns the class for the data object with the given name (or null of unknown)
     * @param dataName
     */
    public Class getClassFor(String dataName) {
        if (map.containsKey(dataName)) return map.get(dataName);
        else return null;
    }
    
    /**
     * Returns true if the table contains the given name
     * @param dataName
     */
    public boolean contains(String dataName) {
        return map.containsKey(dataName);
    }
    
    /**
     * Registers the class for the data object with the given name
     * @param dataName The name of the 
     * @param type
     * @throws ParseError If the entered name is not a valid data object name according to the rules
     */
    public void register(String dataName, Class type) throws ParseError {
        String nameError=engine.checkNameValidity(dataName, false);
        if (nameError!=null) throw new ParseError("\""+dataName+"\" is not a legal name for the data object. "+nameError);
        map.put(dataName, type);
    }
    
    /**
     * Returns a list containing the names of all data items of the given type
     * @param type
     * @return
     */
    @SuppressWarnings("unchecked")
    public ArrayList<String> getAllDataItemsOfType(Class type) {
        ArrayList<String> list=new ArrayList<String>();
        for (String key:map.keySet()) {
            if (type.isAssignableFrom(map.get(key))) list.add(key);
        }
        return list;
    }
    
    /**
     * Returns a list containing the names of all data items of the given types
     * @param type
     * @return
     */
    @SuppressWarnings("unchecked")
    public ArrayList<String> getAllDataItemsOfType(Class[] types) {
        ArrayList<String> list=new ArrayList<String>();
        for (Class type:types) {
          for (String key:map.keySet()) {
            if (type.isAssignableFrom(map.get(key))) list.add(key);
          }
        }
        return list;
    }
    
    /**
     * Populates the list with entries from the engine
     */
    public void populateFromEngine() {
        if (engine==null) return;
        ArrayList<Data> list=engine.getAllDataItemsOfType(Data.class);
        for (Data item:list) {
            map.put(item.getName(),item.getClass());
        }
    }

    /**
     * Returns the number of entries registered in this DataTypeTable
     * @return
     */
    public int size() {
        return map.size();
    }
    
    /**
     * Clears all registered entries from the table
     */
    public void clear() {
        map.clear();
    }
    
    public void debug() {
        System.err.println("====== ( "+this+" ) Size:"+map.size()+"  ======");
        for (String key:map.keySet()) {
            System.err.println("    "+key+" => "+map.get(key).toString());
        }
        System.err.println("----------------------------------------\n");
    }
}
