/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.protocol;

import java.util.HashMap;
import java.util.List;
import javax.swing.text.Document;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.ProtocolTask;
import priorseditor.engine.OperationTask;

/**
 *
 * @author kjetikl
 */
public abstract class Protocol {
    protected PriorsEditorEngine engine;
    private String name="script";
    private HashMap<String,OperationParser> operationParsers;
    private ParametersParser parametersParser=null;
    private DataTypeTable dataTypeLookupTable;

    
    public Protocol(PriorsEditorEngine engine) {
        this.engine=engine;
        operationParsers=new HashMap<String, OperationParser>();
        dataTypeLookupTable=new DataTypeTable(engine);
    }
    
    /**
     * Returns a name for this protocol script. The name can for instance be used as filename
     * (by adding a suffix and a path)
     * @return
     */
    public String getName() {
        return name;
    }
    
    /**
     * Sets a name for this protocol script. The name can for instance be used as filename
     * (by adding a suffix and a path)
     * @param A name to use for this protocol scriopt
     */
    public void setName(String name) {
        this.name=name;
    }
    
    /**
     * Returns a DataTypeTable which can be used to register and query information
     * regarding the types of data objects mentioned on the protocol
     * @return
     */
    public DataTypeTable getDataTypeLookupTable() {
        return dataTypeLookupTable;
    }
    
    /** 
     * Returns a Documents object containing the protocol script formated in some language
     * The type of Document returned may be dependent on the type of Protocol
     */
    public abstract Document getDocument();
    
    /**
     * This function returns a textual representation of a single operation (with arguments)
     * or other executable task in the language of this Protocol
     * 
     * @return A string that can be included in protocol scripts
     */
    public abstract String getCommandString(ExecutableTask task);
        
    
    /**
     * Inserts an operation-command into the protocol at the specified caret position
     * @param task
     * @param pos
     */
    public abstract boolean insertOperationAt(ExecutableTask task, int pos);
    
    /**
     * Appends an operation-command to the end of the protocol 
     * @param task
     */
    public abstract boolean appendOperation(ExecutableTask task);

    /**
     * Registers an operationParser for the specified operation
     * @param operationName
     * @param opparser
     */
    public void registerOperationParser(String operationName, OperationParser opparser) {
        operationParsers.put(operationName, opparser);
    }
    
    /**
     * Registers an Parameters parser to be used for parsing parameters settings 
     * for instance for external programs or data formats
     * @param operationName
     * @param opparser
     */
    public void registerParametersParser(ParametersParser parametersParser) {
        this.parametersParser=parametersParser;
    }
    
    /**
     * Returns the operationParser registered for the specified operation
     * @param operationName
     * @param opparser
     */
    public OperationParser getOperationParser(String operationName) {
        return operationParsers.get(operationName);
    }
    
    /** 
     * Returns the registered ParametersParser
     */
    public ParametersParser getParametersParser() {
        return parametersParser;
    }
    
    /** 
     * Parses the protocol and returns a task which can be executed by the engine
     * If any errors are encountered while parsing, the parsing is aborted and
     * a ParseError is thrown instead
     */
    public abstract ProtocolTask parse() throws ParseError;
    
    /** 
     * Tries to parse the n'th command in the protocol (starting at 1) and returns an OperationTask
     * object containing the results of the parsing. If any errors are encountered 
     * while parsing, the parsing is aborted and a ParseError is thrown instead
     * If the nth command is "empty" or otherwise not possible to execute a NULL value is returned 
     * Note that the "validity" of the n'th line must be viewed in the context of the previous lines.
     * This is important to take into considerations where a data item's type might change in the
     * course of a protocol script execution (for instance through "convert" operations). 
     * Thus, when parsing, not only the n'th line should be parsed, but the entire protocol up to the 
     * n'th line in order to validate correctly the types of all data items.
     * However, if this command is called multiple times for successive lines this constant reparsing
     * of all previous lines is reduntant. If the "clearTypeTable" flag is set to TRUE, the
     * method should not assume that correct registration of data types in previous lines have been
     * performed. Hence, in this case any implementation of this method should start by clearing the
     * current DataTypeTable and reparsing all previous lines. However, if the flag is set to FALSE one
     * can assume that the DataTypeTable is in correct state at the time the method is called and the
     * previous lines need not be parsed. 
     * 
     * @param n The number of the command to be parsed (equals the line number for standard protocols)
     * @param clearTypeTable See above for explaination
     */
    public abstract OperationTask parseCommand(int n, boolean clearTypeTable) throws ParseError;
    
    
    /**
     * This method parses all the commands (or lines) in a protocol script, up to and including the
     * specified commandNumber, and updates the DataTypeTable to reflect the type state at the given command (or line)
     * NOTE: Calling this method will update the internal DataTypeTable of the Protocol (which is returned)
     * @param commandNumber The number of the command to be parsed (equals the line number for standard protocols)
     */
    public abstract DataTypeTable getDataTypeStateAtCommand(int commandNumber);
    
    
    /**
     * This method returns true if the Protocol script has no contents (including comments)
     */
    public abstract boolean isEmpty();
    
}
