/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.protocol;

import priorseditor.engine.OperationTask;
import priorseditor.engine.Parameter;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.NumericVariable;
import priorseditor.engine.dataformat.DataFormat;


/**
 *
 * @author kjetikl
 */
public class StandardParametersParser extends ParametersParser {

    public StandardParametersParser(PriorsEditorEngine engine, Protocol protocol) {
        setEngine(engine);
        setProtocol(protocol);
    }
    
   
    /**
     * Given a text string containing parameter settings (for instance a portion from a protocol script)
     * formatted as as "name=value" pairs separated by commas, the method will update the specified parameters 
     * with the given values. ParameterSettings that are not mentioned in the text will be assigned default values
     * 
     * @param text
     * @return
     */
    @Override
    @SuppressWarnings("unchecked")
    public ParameterSettings parse(String text, Parameter[] parameterFormats) throws ParseError {
        ParameterSettings settings=new ParameterSettings();
        String[] parts=text.split(",");
        for (String part:parts) {
            String[] fields=part.split("=");
            if (fields.length!=2) throw new ParseError("Error in parameters '"+part+"'");
            String parname=fields[0].trim();
            String parvalueString=fields[1].trim();
            if (!parameterExists(parname, parameterFormats)) throw new ParseError("Unrecognized parameter '"+parname+"'");
            Class type=getTypeForParameter(parname, parameterFormats);            
            Object parvalue=null;
            if (type==null) {
                parvalue=parvalueString;
            } else if (type.equals(Float.class)) {
                  try {
                      float value=Float.parseFloat(parvalueString);
                      parvalue=new Float(value);
                  } catch(NumberFormatException e) {
                      Class currentType=protocol.getDataTypeLookupTable().getClassFor(parvalueString);
                      if (currentType==null) throw new ParseError("Unknown data item: "+parvalueString);
                      else if (currentType.equals(NumericVariable.class)) parvalue=parvalueString;
                      else throw new ParseError("Value for parameter '"+parname+"="+parvalueString+"' is not a valid number or Numeric Variable");
                  }
            } else if (type.equals(Integer.class)) {
                  try {
                      int value=Integer.parseInt(parvalueString);
                      parvalue=new Integer(value);
                  } catch(NumberFormatException e) {
                      Class currentType=protocol.getDataTypeLookupTable().getClassFor(parvalueString);
                      if (currentType==null) throw new ParseError("Unknown data item: "+parvalueString);
                      else if (currentType.equals(NumericVariable.class)) parvalue=parvalueString;
                      else throw new ParseError("Value for parameter '"+parname+"="+parvalueString+"' is not a valid integer or Numeric Variable");
                  }
            } else if (type.equals(Boolean.class)) {
                    if (parvalueString.equalsIgnoreCase("true")) parvalue=Boolean.TRUE;
                    else if (parvalueString.equalsIgnoreCase("false")) parvalue=Boolean.FALSE;
                    else throw new ParseError("Value for parameter '"+parname+"="+parvalueString+"' is not a valid boolean value");
            } else if (Data.class.isAssignableFrom(type)) {
                    // perhaps we should do even more checks here...
                    Class currentType=protocol.getDataTypeLookupTable().getClassFor(parvalueString);
                    if (currentType==null) throw new ParseError("Unknown data item: "+parvalueString);
                    Object av=getAllowedValuesForParameter(parname, parameterFormats);
                    if (!(av instanceof Class || av instanceof Class[])) throw new ParseError("SYSTEM ERROR: 'allowed values' for parameter is neither Class nor Class[]");
                    else if (av instanceof Class) {
                        Class allowedType=(Class)av;
                        if (allowedType.isAssignableFrom(currentType)) parvalue=parvalueString;
                        else throw new ParseError("Value for parameter '"+parname+"="+parvalueString+"' is not of valid type");
                    } else {
                        if (testAll(currentType,(Class[])av)) parvalue=parvalueString;
                        else throw new ParseError("Value for parameter '"+parname+"="+parvalueString+"' is not of valid type");
                    } 
            } else {
                parvalue=parvalueString;
            }
            //System.err.println("Parser DataFormat setting: "+parname+"="+parvalue);
            settings.setParameter(parname, parvalue);
            // assign default values to missing parameters
            for(Parameter item:parameterFormats) {
                String pname=item.getName();
                if (settings.getParameterAsString(pname, parameterFormats)==null)  settings.setParameter(pname,item.getDefaultValue());
            }
        }
        return settings;
    }    
    

    
    /**
     * This method returns a string containing a concatination of all the explicitly set parameters and their values. 
     * The string can be included in protocol scripts and should be parsable by the {@code parseParameterSettings(String)} method
     * @param list If this parameter is not null only the parameters included in the list will be output and in the order given
     * 
     * @return
     */
    @Override
    public String getCommandString(Parameter[] list, ParameterSettings settings) {
        String string="";
        if (list==null) {
          String[] keys=settings.getParameters();
          for (int i=0;i<keys.length;i++) {
            String parametername=keys[i];
            String value=settings.getParameterAsString(parametername,list);
            if (value==null) continue;
            string+=parametername+"="+value+",";
          }
        } else {
          for (int i=0;i<list.length;i++) {
              String parametername=list[i].getName();
              String value=settings.getParameterAsString(parametername,list);
              if (value==null) continue;
              string+=parametername+"="+value+",";
          }  
        }
        if (string.endsWith(",")) string=string.substring(0, string.length()-1);
        return string;
    }                
    
    /**
     * Returns an object representing the allowed values for the parameter with the given name,
     * or null if no there is no such parameter
     * @param parameterName
     * @param parameterFormats
     * @return
     */
    private Object getAllowedValuesForParameter(String parameterName, Parameter[] parameterFormats) {
        Object def=null;
        if (parameterFormats==null) return null;
        for(Parameter item:parameterFormats) {
            if (item.getName().equals(parameterName)) {def=item.getAllowedValues();break;}
        }
        return def;
    }
    
    /**
     * Returns a Class object representing the type for the parameter with the given name,
     * or null if no there is no such parameter
     * @param parameterName
     * @param parameterFormats
     * @return
     */
    private Class getTypeForParameter(String parameterName, Parameter[] parameterFormats) {
        Class def=null;
        if (parameterFormats==null) return null;
        for(Parameter item:parameterFormats) {
            if (item.getName().equals(parameterName)) {def=item.getType();break;}
        }
        return def;
    }   
    
    /** Returns true if the list of Parameters contains a parameter corresponding to the given name */
    private boolean parameterExists(String parameterName, Parameter[] parameterFormats) {
        if (parameterFormats==null) return false;
        for(Parameter item:parameterFormats) {
            if (item.getName().equals(parameterName)) return true;
        }
        return false;
    }
    
    @SuppressWarnings("unchecked")
    private boolean testAll(Class type, Class[] list) {
        for (Class c:list) {
            if (c.isAssignableFrom(type)) return true;
        }
        return false;
    }
    
    
    
    /** Parses and sets the Data Format parameter settings from a string with name-value pairs
     *  The ParameterSettings which results from this parse is placed among the parameters
     *  of the task using the name provided with the 'taskParameterName' argument
     *  (using: task.setParameter(taskParameterName,parametersettings);)
     *  
     *  @param taskParameterName The name of the task parameter the settings should be stored as 
     *
     */
    protected void parseFormatParameterSettings(String text, DataFormat formatter, OperationTask task, String taskParameterName) throws ParseError {
        ParametersParser parametersParser=protocol.getParametersParser();
        try {
           ParameterSettings settings=parametersParser.parse(text, formatter.getParameters());
           task.setParameter(taskParameterName, settings);
        } catch (ParseError parseError) {
          parseError.setLineNumber(task.getLineNumber()); 
          throw parseError;
        }       
    } 
    
    protected void parseFormatParameterSettings(String text, Parameter[] parameters, OperationTask task, String taskParameterName) throws ParseError {
        ParametersParser parametersParser=protocol.getParametersParser();
        try {
           ParameterSettings settings=parametersParser.parse(text, parameters);
           task.setParameter(taskParameterName, settings);
        } catch (ParseError parseError) {
          parseError.setLineNumber(task.getLineNumber()); 
          throw parseError;
        }       
    } 
    
    /** Returns a parameterString for the DataFormat settings which are store in the OperationTask
     *  as a parameter under the name 'taskParameterName'
     */
    protected String getParameterSettingsAsString(DataFormat formatter, OperationTask task, String taskParameterName) {
        ParameterSettings settings=(ParameterSettings)task.getParameter(taskParameterName);
        ParametersParser parametersParser=protocol.getParametersParser();
        if (settings!=null) {
            return parametersParser.getCommandString(formatter.getParameters(), settings);
        }
        else return "";        
    }    
    
    protected String getParameterSettingsAsString(Parameter[] parameters, OperationTask task, String taskParameterName) {
        ParameterSettings settings=(ParameterSettings)task.getParameter(taskParameterName);
        ParametersParser parametersParser=protocol.getParametersParser();
        if (settings!=null) {
            return parametersParser.getCommandString(parameters, settings);
        }
        else return "";        
    }    
    
    
}
