/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.engine.protocol;


import java.io.Serializable;
import java.util.ArrayList;
import priorseditor.engine.OperationTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.PlainDocument;
import priorseditor.engine.CompoundTask;
import priorseditor.engine.operations.*;
import priorseditor.engine.ProtocolTask;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.ExecutableTask;


/**
 * This class represents a protocol in the "standard protocol language"
 * 
 * @author kjetikl
 */
public class StandardProtocol extends Protocol {
    private PlainDocument document;
    private String operationParserClassPrefix="priorseditor.engine.protocol.StandardOperationParser_";
    private boolean isDirtyFlag=false;
    private String savedFileName=null;    
 
    /**
     * Creates a new empty instance of StandardProtocol
     * @param engine
     */
    public StandardProtocol(PriorsEditorEngine engine) {
         super(engine);
         document=new PlainDocument();      
         register();
    }
    
    /**
     * Creates a new instance of StandardProtocol
     * @param engine
     * @param text If this is not NULL the text will serve as the contents for the protocol
     */
    public StandardProtocol(PriorsEditorEngine engine, String text) {
         super(engine);
         document=new PlainDocument();
         if (text!=null) try {
             document.insertString(0, text, null);
         } catch (BadLocationException e) {}
         register();
    }
    
   /**
     * Creates a new instance of StandardProtocol based on a serialized object
     * @param engine
     * @param serializedProtocol
     */
    public StandardProtocol(PriorsEditorEngine engine, SerializedStandardProtocol serializedProtocol) {
        super(engine);
        setName(serializedProtocol.name);
        this.isDirtyFlag=serializedProtocol.isDirtyFlag;
        this.savedFileName=serializedProtocol.savedFileName;
        document=new PlainDocument();
        if (serializedProtocol.text!=null) try {
             document.insertString(0, serializedProtocol.text, null);
        } catch (BadLocationException e) {}
        register();     
    }
    
   /**
     * Creates a new instance of StandardProtocol based on a serialized object and a PlainDocument
     * The PlainDocument should contain the protocol text 
     * @param engine
     * @param serializedProtocol
     */
    public StandardProtocol(PriorsEditorEngine engine, SerializedStandardProtocol serializedProtocol, PlainDocument usedocument) {
        super(engine);
        setName(serializedProtocol.name);
        this.isDirtyFlag=serializedProtocol.isDirtyFlag;
        this.savedFileName=serializedProtocol.savedFileName;
        document=usedocument;
        register();     
    }
    
   
    
    
    private void register() {
         registerParametersParser(new StandardParametersParser(engine,this));
         for (Operation operation:engine.getAllOperations()) {
             String operationname=operation.getName();
             String classname=operationParserClassPrefix+operationname;
             try { 
              Class parserClass=Class.forName(classname);
              OperationParser parser=(OperationParser)parserClass.newInstance();
              parser.setEngine(engine);
              parser.setProtocol(this);
              registerOperationParser(operationname, parser);
              //operation.setParser(parser);
           } catch(ClassNotFoundException cnfe) {
               System.err.println("SYSTEM ERROR: No StandardOperationParser found for operation: "+operationname);
           } catch(InstantiationException ie) {
               System.err.println("SYSTEM ERROR: InstantiationException for StandardOperationParser: "+operationname);
           } catch(IllegalAccessException iae) {
               System.err.println("SYSTEM ERROR: IllegalAccessException for StandardOperationParser: "+operationname);
           }             
         }            
    }
    
    
    @Override
    public Document getDocument() {
        return document;
    }    
    
    @Override
    public ProtocolTask parse() throws ParseError {
        getDataTypeLookupTable().clear(); 
        getDataTypeLookupTable().populateFromEngine();
        ProtocolTask scriptTask=new ProtocolTask(getName());
        Element root=document.getDefaultRootElement();
        int linecount=root.getElementCount();
        //System.err.println("Protocol has "+linecount+" lines");
        for (int i=0;i<linecount;i++) {
           Element element=root.getElement(i);           
           String line;
           try {
              line=document.getText(element.getStartOffset(),element.getEndOffset()-element.getStartOffset());
           } catch(BadLocationException ble) {throw new ParseError(ble.getMessage());}
           line=preprocess(line);   
           if (line.isEmpty()) continue;
           String operationName=null;
           Pattern pattern=Pattern.compile("^(([a-zA-Z_0-9-]+?|\\[([a-zA-Z_0-9-,\\s]+?)\\])\\s*=\\s*)?(\\w+)");
           Matcher matcher=pattern.matcher(line);
           if (matcher.find()) {
               operationName=matcher.group(4);
           } else throw new ParseError("Unable to parse line "+(i+1)+":\n"+line);
           Operation operation=engine.getOperation(operationName);
           if (operation==null) throw new ParseError("Unrecognized operation '"+operationName+"'",(i+1));
           OperationParser opparser=getOperationParser(operationName);
           if (opparser==null) throw new ParseError("SYSTEM ERROR: Missing parser for '"+operationName+"'",(i+1));
           try {
               ExecutableTask optask=opparser.parse(line);
               optask.setLineNumber(i+1);
               scriptTask.addTask(optask);
           } catch (ParseError operror) {
               throw new ParseError(operror.getMessage(),(i+1));
           }
        }
        return scriptTask;
    }
    
    @Override
    public OperationTask parseCommand(int linenumber, boolean clearTypeTable) throws ParseError {
        Element root=document.getDefaultRootElement();
        int linecount=root.getElementCount();
        if (linenumber>linecount) return null;
        if (clearTypeTable) {
            getDataTypeLookupTable().clear(); 
            getDataTypeLookupTable().populateFromEngine();
            for (int i=0;i<linenumber-1;i++) { // this will populateFromEngine the DataTypeTable correctly up the the line in question
              Element e=root.getElement(i);  
              parseCommandAtNthLine(e);
            }
        }
        Element element=root.getElement(linenumber-1);           
        String line;
        try {
           line=document.getText(element.getStartOffset(),element.getEndOffset()-element.getStartOffset());
        } catch(BadLocationException ble) {throw new ParseError(ble.getMessage());}
        //System.err.println("PARSING line: "+line);
        line=preprocess(line);   
        if (line.isEmpty()) return null;
        String operationName=null;
        String[] splitOnWithinRange=line.split(" within\\s?\\[");
        if (splitOnWithinRange.length>2) throw new ParseError("Multiple within-clauses",linenumber);
        String cmdline=splitOnWithinRange[0];
        Pattern pattern=Pattern.compile("^(([a-zA-Z_0-9-]+?|\\[([a-zA-Z_0-9-,\\s]+?)\\])\\s*=\\s*)?(\\w+)");
        Matcher matcher=pattern.matcher(cmdline);
        if (matcher.find()) {          
               operationName=matcher.group(4);
        } else throw new ParseError("Unable to parse line "+(linenumber)+":\n"+line);
        Operation operation=engine.getOperation(operationName);
        if (operation==null) throw new ParseError("Unrecognized operation '"+operationName+"'",linenumber);
        OperationParser opparser=getOperationParser(operationName);
        if (opparser==null) throw new ParseError("SYSTEM ERROR: Missing parser for '"+operationName+"'",linenumber);
        OperationTask optask=null;
        optask=opparser.parse(cmdline); // this call can throw a ParseError which will be propagated
        if (splitOnWithinRange.length>1) {
            if (!splitOnWithinRange[1].endsWith("]")) throw new ParseError("Within-clause missing end bracket",linenumber);
            if (splitOnWithinRange[1].length()<=1) throw new ParseError("Empty within-clause",linenumber);
            String withinclause=splitOnWithinRange[1].substring(0, splitOnWithinRange[1].length()-1);            
            if (!operation.isSubrangeApplicable()) throw new ParseError("The '"+operationName+"' operation can not be applied to window selections",linenumber);
            ((StandardOperationParser)opparser).parseWithinCondition(withinclause,optask); // this will set the within condition in the task
        }
        return optask;
    }
    
    /** 
     * This method is called by parseCommand(N) in order to parse previous lines (from 1 to N-1) 
     * in the script and populateFromEngine the DataTypeTable so that the type-"state" of all variables in 
     * DataTypeTable corresponds to the environment at line N
     * The method does not produce any OperationTasks
     */
    private void parseCommandAtNthLine(Element element) {          
        String line;
        try {
           line=document.getText(element.getStartOffset(),element.getEndOffset()-element.getStartOffset());
        } catch(BadLocationException ble) {return;}
        line=preprocess(line);   
        if (line.isEmpty()) return;
        String operationName=null;
        Pattern pattern=Pattern.compile("^(([a-zA-Z_0-9-]+?|\\[([a-zA-Z_0-9-,\\s]+?)\\])\\s*=\\s*)?(\\w+)");
        Matcher matcher=pattern.matcher(line);
        if (matcher.find()) {
               operationName=matcher.group(4);
        } else return;
        Operation operation=engine.getOperation(operationName);
        if (operation==null) return;
        OperationParser opparser=getOperationParser(operationName);
        if (opparser==null) return;
        try {opparser.parse(line);} catch (ParseError e) {} // this call can throw a ParseError        
    }

    
    @Override
    public DataTypeTable getDataTypeStateAtCommand(int commandNumber) {
        DataTypeTable typetable=getDataTypeLookupTable();
        typetable.clear(); 
        typetable.populateFromEngine();
        Element root=document.getDefaultRootElement();
        int linecount=root.getElementCount();
        if (commandNumber>linecount) commandNumber=linecount;
        for (int i=0;i<commandNumber;i++) { 
              Element e=root.getElement(i);  
              parseCommandAtNthLine(e);
        }   
        return typetable;
    }
  
    
    
    @Override
    public boolean appendOperation(ExecutableTask task) {
        try {
            String cmd=getCommandString(task)+"\n";
            document.insertString(document.getLength(), cmd, null);
            return true;
        } catch(Exception e) {return false;}
    }



    @Override
    public boolean insertOperationAt(ExecutableTask task, int pos) {
        try {
            String cmd=getCommandString(task)+"\n";
            document.insertString(pos, cmd, null);
            return true;
        } catch(Exception e) {return false;}
    }

    
    
    @Override
    public String getCommandString(ExecutableTask task) {
        if (task instanceof CompoundTask) {
            CompoundTask block = (CompoundTask)task;
            String protocolstring="";
            ArrayList<ExecutableTask> list=block.getTaskList();
            for (ExecutableTask ex:list) {
                protocolstring+=ex.getCommandString(this)+"\n"; // no block delimiters at this time...
            }
            return protocolstring.trim();
        } else if (task instanceof OperationTask) {
            Operation op=((OperationTask)task).getOperation();
            OperationParser opparser=getOperationParser(op.getName());
            String command=opparser.getCommandString(((OperationTask)task));
            Condition_within within=(Condition_within)((OperationTask)task).getParameter("within");
            if (within!=null) {
                command+=((StandardOperationParser)opparser).getCommandString_condition(within);
            }
            return command;
        } else return null;
    }
    
    
    /**
     * Performs preprocessing of a line by removing comments and excessive whitespace 
     * @param line The part of the text to preprocess
     * @return The prep
     */
    private String preprocess(String line) {
        line=line.trim();
        line=line.replaceAll("\\s+", " "); // collapse runs of whitespace into single spaces
        int commentstart=line.indexOf("#"); // this is hacky!!!   This does not distinguish between comment signs and "protected" hashes within quotes
        if (commentstart==-1) return line;
        else {
            return line.substring(0,commentstart);
        }
    }
    
    
    /**
     * Sets the status of the "dirty" flag for this Protocol.
     * Protocol scripts should be marked as dirty if the contents of the Protocol
     * has been changed compared to the last saved version.
     */ 
    public void setDirtyFlag(boolean dirty) {
        isDirtyFlag=dirty;
    }
    
    /**
     * Returns true if this Protocol is marked as "dirty"
     */ 
    public boolean isDirty() {
        return isDirtyFlag;
    }
    
    
    @Override
    public boolean isEmpty() {
        if (document.getLength()==0) return true; 
        String text="";
        try {
            text=document.getText(0, document.getLength());
        } catch (BadLocationException e) {System.err.println("SYSTEM ERROR: BadLocationException in method Protocol.isEmpty()");}
        return text.trim().isEmpty();
    }
    
    /**
     * Returns the filename of the file currently associated with this Protocol script
     * @return
     */
    public String getFileName() {
        return savedFileName;
    }
    
    /**
     * Sets a filename to be associated with this Protocol script.
     * The Protocol script should already have been saved to a file with the given name
     * @return
     */
    public void setFileName(String filename) {
         savedFileName=filename;
    }
            
    /** Exports a serialized version of this protocol, which can be loaded into a protocol through import() */ 
    public SerializedStandardProtocol getSerializedProtocol() {
        return new SerializedStandardProtocol(this);
    }
    
    

}
