/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.external;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.util.ArrayList;
import priorseditor.engine.Parameter;
import priorseditor.engine.ParameterExporter;
import priorseditor.engine.PriorsEditorEngine;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamResult;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import priorseditor.engine.ExecutableTask;
import priorseditor.engine.ExecutionError;
import priorseditor.engine.OperationTask;
import priorseditor.engine.ParameterSettings;
import priorseditor.engine.SystemError;
import priorseditor.engine.data.*;
import priorseditor.engine.dataformat.DataFormat;

/**
 * This superclass is the top-level class for wrapping external programs such
 * as motif discovery or motif scanning tools
 * @author kjetikl
 */
public class ExternalProgram implements ParameterExporter {
    public static final String SOURCES="SOURCES";
    
    protected ArrayList<Parameter> regularParameterFormats=new ArrayList<Parameter>(); 
    protected ArrayList<Parameter> sourceParameterFormats=new ArrayList<Parameter>(); 
    protected ArrayList<Parameter> resultsParameterFormats=new ArrayList<Parameter>(); 
    protected PriorsEditorEngine engine;
    protected String name;     
    protected String programclass;
    protected String serviceType;
    protected String location;
    protected String command;
    private HashMap<String,Object> properties=new HashMap<String,Object>();
    private ArrayList<String> allArguments=new ArrayList<String>();
    private HashMap<String,String> argumentSwitches=new HashMap<String,String>();
    private HashMap<String,String> argumentTypes=new HashMap<String,String>();
    private HashMap<String,String> dataFormatNames=new HashMap<String,String>();
    private HashMap<String,ParameterSettings> dataFormatSettings=new HashMap<String,ParameterSettings>();
    private ArrayList<String[]> executableSources; // each source is described by an array of five elements {OS, version, URL, compression-type and relative path of target executable within ZIP (when applicable)}
    private HashMap<String,String> implicitArguments=new HashMap<String,String>(); // contains 'values' of implicit arguments
    private HashMap<String,String> redirectedArguments=new HashMap<String,String>(); // contains STDIN or STDOUT flags for arguments with redirection
    private ArrayList<String[]> referenceLinks=new ArrayList<String[]>(); // contains information about crossreferences that must be resolved
    private HashMap<String,HashMap<String,String>> optionValueMap=null; // keeps information about String parameters that have different 'display values' and 'actual values'
    private ArrayList<String> temporaryfilenames=new ArrayList<String>(); // keeps track of files that are explicitly registered as being temporary (in the XML-config)
    private String configurationfile=null;
    private boolean cygwin=false; // set true if the program runs in Windows under Cygwin
    private boolean surroundFilenamesWithQuotes=true;
    
    public String getName() {
        return name;
    }
   
    public String getDescription() {
        return (String)properties.get("description");
    }
    
    public String getProgramClass() {
        return programclass;
    }
    
    /** Returns a value for the given property (not parameter)
     *  or NULL if no such property exists. Common properties for ExternalPrograms
     *  are e.g. 'author','citation' ...
     *  @return the property value should be either a single String or an ArrayList<String>
     */
    public Object getProperty(String key) {
        return properties.get(key);
    }
    
    /** Sets a value for the given property (not parameter)
     *  Common properties for ExternalPrograms are e.g. 'author','citation' ...
     *  the property value should be either a single String or an ArrayList<String>
     *  if a property with the given name already exists the new value will be added
     *  to the existing list unless overwrite is set to TRUE
     */
    @SuppressWarnings("unchecked")
    public void setProperty(String key, Object value, boolean overwrite) {
        if (overwrite) properties.put(key,value);
        else {
            Object current=getProperty(key);
            if (current==null) {
                properties.put(key,value);
            } else if (current instanceof ArrayList) {
                ((ArrayList)current).add(value);
            } else { // the property is currently a single value, replace with a list containing both the old and new values
                ArrayList list=new ArrayList();
                list.add(current);
                list.add(value);
                properties.put(key,list);
            }
        }
    }
 
    /** Removes all registered properties for this program */
    public void clearProperties() {
        properties.clear();
    }
    
    /** Returns true if the program was made for Unix/Linux and must run in Cygwin to work on Windows*/
    public boolean isRunInCygwin() {
        return cygwin;
    }
    
    /** Specifies if the program must be run in Cygwin in order to work on Windows */
    public void setRunInCygwin(boolean runInCygwin) {
        cygwin=runInCygwin;
    }
    
    
    
    /** Returns the type of service this ExternalProgram represents
     *  e.g. "bundled" for (Java) programs that are included with PriorsEditor 
     *  or   "local" for locally installed executable programs 
     *  or   "soap" for web services accessed via SOAP protocol  
     */
    public String getServiceType() {
        return serviceType;
    }
    
    /** Returns the location of this ExternalProgram
     *  Depending on the service type this could be the full pathname of a file
     *  on the local filesystem, a JAVA class (for bundled software) or an internet address
     */
    public String getLocation() {
        return location;
    }
    
    /** Sets the location of this external program
     *  Depending on the service type this could be the full pathname of a file
     *  on the local filesystem, a JAVA class (for bundled software) or an internet address     
     */
    public void setLocation(String location) {
        this.location=location;
    }
    
    /** Returns the command to use in order to execute this program
     */
    public String getCommand() {
        return command;
    }
    
    /** Sets the command to use to execute this program
     */
    public void setCommand(String command) {
        this.command=command;
    }
    
    /** Sets a reference to the configuration file for this program */
    public void setConfigurationFile(String filename) {
        configurationfile=filename;
    }
    
    /** Returns a reference to the configuration file for this program */
    public String getConfigurationFile() {
        return configurationfile;
    }
    
    
    /**
     * Returns the directory associated with program (as long as the program is installed locally)
     * If getServiceType() is not "local" the method will return an empty string;
     * @return
     */
    public String getApplicationDirectory() {
        if (!getServiceType().equals("local")) return "";
        File appfile=new File(getLocation());
        String dir=appfile.getParent();
        return (dir!=null)?dir:"";
    }
    
    /** Returns a list of String arrays describing available sources
     *  from which the executable program can be downloaded. 
     *  Each String[] contains 
     *      [0] the Operating system 
     *      [1] program version
     *      [2] an URL (as String)
     *      [3] A String specifiying the compression format (e.g. "ZIP") or null if the file was not compressed
     *      [4] An (optional) String specifying the relative path to the target executable file if this is located somewhere inside a ZIP-file
     */
    public ArrayList<String[]> getExecutableSources() {
        return executableSources;
    }
    
    
    /** Returns true if the parameter with the given name represents a 'source' parameter */
    public boolean isSourceParameter(String parameterName) {
        for (Parameter param:sourceParameterFormats) {
            if (param.getName().equals(parameterName)) return true;
        }
        return false;
    }
    /** Returns true if the parameter with the given name represents a 'results' parameter */
    public boolean isResultsParameter(String parameterName) {
        for (Parameter param:resultsParameterFormats) {
            if (param.getName().equals(parameterName)) return true;
        }
        return false;        
    }

    /** Returns the number of data objects produced by this ExternalProgram */
    public int getNumberOfResultParameters() {
        return resultsParameterFormats.size();
    }

    /** Returns the class type for the result parameter with the given index number (starting at 0) */
    public Class getTypeForResultParameter(int resultParameterNumber) {
        if (resultParameterNumber<0 || resultParameterNumber>=resultsParameterFormats.size()) return null;
        return resultsParameterFormats.get(resultParameterNumber).getType();
    }

   /** Returns the name of the result parameter with the given index number (starting at 0) */
    public String getNameForResultParameter(int resultParameterNumber) {
        if (resultParameterNumber<0 || resultParameterNumber>=resultsParameterFormats.size()) return null;
        return resultsParameterFormats.get(resultParameterNumber).getName();
    }
    
    /**
     * Adds a regular parameter to the External Program. This is used for initialization of External Program objects and should only be called in a constructor or similar setup method
     */
    protected void addParameter(String parameterName, Class type, Object defaultValue, Object[] allowedValues, String description, boolean required, boolean hidden) {
        regularParameterFormats.add(new Parameter(parameterName,type,defaultValue,allowedValues,description,required,hidden));
    }
    /**
     * Adds a source parameter to the External Program. This is used for initialization of External Program objects and should only be called in a constructor or similar setup method
     */
    protected void addSourceParameter(String parameterName, Class type, Object defaultValue, Object[] allowedValues, String description) {
        sourceParameterFormats.add(new Parameter(parameterName,type,defaultValue,allowedValues,description));
    }
    /**
     * Adds a parameter to the External Program. This is used for initialization of External Program objects and should only be called in a constructor or similar setup method
     */
    protected void addResultParameter(String parameterName, Class type, Object defaultValue, Object[] allowedValues, String description) {
        resultsParameterFormats.add(new Parameter(parameterName,type,defaultValue,allowedValues,description));
    }
        
    @Override
    public String toString() {return getName();}

    @Override
    public Object getDefaultValueForParameter(String parameterName) {
        for (Parameter par:regularParameterFormats) {
            if (parameterName.equals(par.getName())) return par.getDefaultValue();
        }
        return null;
    }

    @Override
    public Parameter[] getParameters() {
        Parameter[] list=new Parameter[regularParameterFormats.size()];
        return regularParameterFormats.toArray(list);
    }
    
    @Override
    public Parameter getParameterWithName(String parameterName) {
        for (Parameter parameter:regularParameterFormats) {
            if (parameter.getName().equals(parameterName)) return parameter;
        }
        return null;
    }
    

    private Parameter getParameterWithName(String parameterName, ArrayList<Parameter> list) {
        for (Parameter parameter:list) {
            if (parameter.getName().equals(parameterName)) return parameter;
        }
        return null;
    }
    
    /**
     * Returns a list of all arguments (names) for this program in the order in which they 
     * shall be listed as arguments for command line execution.
     * This list will typically be superset of the exported parameters, i.e. this list will
     * include the names of all exported parameters and in addition the names of any source
     * and target arguments.
     * @return
     */
    public ArrayList<String> getAllArguments() {
      return allArguments;   
    }
    
    /** Returns the type designation for this argument (e.g. "flag|valued option|implicit")
     */
    public String getArgumentType(String argumentName) {
        return argumentTypes.get(argumentName);
    }
    
    private void setArgumentType(String argumentName, String type) {
        argumentTypes.put(argumentName,type);
    }
    
    /** Returns the (system dependent) switch for the argument with the given name
     * for instance: for a file argument that is specified on the command line as
     * "-f <file>" the method will return "-f"
     */
    public String getArgumentSwitch(String argumentName) {
        return argumentSwitches.get(argumentName);
    }
    
    private void setArgumentSwitch(String argumentName, String switchString) {
        argumentSwitches.put(argumentName,switchString);
    }
    
    /** Returns a value (usually a filename) to use for an implicit argument
     *  Implicit arguments are settings that are usually hardcoded in the external program
     *  and can not (and should not) be specified at the comman line.
     *  An example can be a program that always outputs results to a file called "results.txt"
     *  Then the name of the resultsfile should not be specified as a regular argument but as an implicit argument with filename "results.txt"
     */
    public String getImplicitArgument(String argumentName) {
        return implicitArguments.get(argumentName);
    }
    
    private void setImplicitArgument(String argumentName, String filename) {
        implicitArguments.put(argumentName,filename);
    }
    
    public boolean isImplicitArgument(String argumentName) {
        return implicitArguments.containsKey(argumentName);
    }


    /** Returns either STDIN or STDOUT if the argument represents
     *  data which should be redirected to or from the program
     *  or NULL if the argument does not require redirection
     */
    public String getRedirectionForArgument(String argumentName) {
        return redirectedArguments.get(argumentName);
    }

    /** Sets a redirection destination for the argument
     *  @param argumentName
     *  @param destination (should be either STDIN or STDOUT)
     */
    public void setRedirectionForArgument(String argumentName, String destination) {
        redirectedArguments.put(argumentName,destination);
    }

    /** Returns TRUE if this argument is (directly) associated with
     *  a redirection to/from STDIN or STDOUT
     */
    public boolean isRedirectionArgument(String argumentName) {
        return redirectedArguments.containsKey(argumentName);
    }

    /** Returns TRUE if this argument is directly or indirectly (linked)
     *  associated with a redirection to/from STDIN or STDOUT
     */
    public boolean isLinkedToRedirectionArgument(String argumentName) {
        if (redirectedArguments.containsKey(argumentName)) return true;
        String linked=getParameterReferenceLink(argumentName, null);
        if (linked!=null) {
             return redirectedArguments.containsKey(linked);
        } else return false;
    }

    /** Returns TRUE if the external program requires any form of redirection
     *  of data from STDIN or STDOUT
     */
    public boolean programRequiresRedirection() {
        return !redirectedArguments.isEmpty();
    }
    
    /** Returns the Data Format required by the argument with the given name
     * If no explicit Data Format is specified for the argument or if the format
     * is unknown, a NULL value is returned
     */
    public DataFormat getArgumentDataFormat(String argumentName) {
        String dataFormatName=dataFormatNames.get(argumentName);
        if (dataFormatName==null) return null;
        return engine.getDataFormat(dataFormatName);
    }

    private void setArgumentDataFormat(String argumentName, String dataformatName) {
        dataFormatNames.put(argumentName, dataformatName);
    }    
    
    /** Returns the Data Format Parameter settings required by the argument with the given name
     * If no explicit Parameter settings is specified for the argument, a NULL value is returned
     */
    public ParameterSettings getArgumentDataFormatSettings(String argumentName) {
        return dataFormatSettings.get(argumentName);
    }

    private void setArgumentDataFormatSettings(String argumentName, ParameterSettings settings) {
        dataFormatSettings.put(argumentName,settings);
    }
    
    public void setEngine(PriorsEditorEngine engine) {
        this.engine=engine;
    }
    
    /**
     * Saves the configuration of this external program in XML format to the given file
     */
    public void saveConfigurationToFile(File configurationfile) throws Exception {
        Document document=getXMLrepresentation();
        TransformerFactory factory=TransformerFactory.newInstance();
        factory.setAttribute("indent-number", new Integer(3));
        Transformer transformer=factory.newTransformer();
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        DOMSource source=new DOMSource(document); 
        StreamResult result=new StreamResult(new OutputStreamWriter(new FileOutputStream(configurationfile),"UTF-8"));
        transformer.transform(source, result);
    }

    /** Returns an XML representation of the configuration of the ExternalProgram */
    public Document getXMLrepresentation() throws ParserConfigurationException {
         DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
         Document document = builder.newDocument();
         Element program=document.createElement("program");
         program.setAttribute("name", getName());
         program.setAttribute("class", getProgramClass());
         if (isRunInCygwin()) program.setAttribute("cygwin", "yes");
         Element propertiesElement=document.createElement("properties");
         for (String key:properties.keySet()) {
             Object propertyvalue=properties.get(key);
             if (propertyvalue instanceof ArrayList) {
                 for (Object val:(ArrayList)propertyvalue) {
                     Element propertyElement = document.createElement(key);
                     propertyElement.setTextContent(val.toString());
                     propertiesElement.appendChild(propertyElement);  
                 }
             } else if (propertyvalue!=null) {
                 Element propertyElement = document.createElement(key);
                 propertyElement.setTextContent(propertyvalue.toString());
                 propertiesElement.appendChild(propertyElement);                 
             }
         }
         program.appendChild(propertiesElement); 

         Element serviceElement=document.createElement("service");
         serviceElement.setAttribute("type", serviceType);
         serviceElement.setAttribute("location", location);
         program.appendChild(serviceElement);   
         
         String commandline=getCommand();
         if (commandline!=null && !commandline.isEmpty()) {
             Element commandElement=document.createElement("command");
             commandElement.setTextContent(commandline);
             program.appendChild(commandElement); 
         }
         
         for (String parameterName:allArguments) {
             Element parameterElement=document.createElement("parameter");
             String parameterType=null;
             Parameter parameter;
             if (isSourceParameter(parameterName)) {
                 parameterType="source";
                 parameter=getParameterWithName(parameterName,sourceParameterFormats); 
             } else if (isResultsParameter(parameterName)) {
                 parameterType="result";
                 parameter=getParameterWithName(parameterName,resultsParameterFormats); 
             } else {
                parameterType="regular";
                parameter=getParameterWithName(parameterName,regularParameterFormats);      
             }
             parameterElement.setAttribute("name", parameterName);
             parameterElement.setAttribute("class", parameter.getType().getSimpleName());
             parameterElement.setAttribute("type", parameterType);
             parameterElement.setAttribute("required",(parameter.isRequired())?"yes":"no");  
             if (parameter.isHidden()) parameterElement.setAttribute("hidden","yes"); 
             String paramlink=getParameterReferenceLink(parameterName,null);
             if (paramlink!=null) parameterElement.setAttribute("link", paramlink); 
             
             if (parameter.getType()==Integer.class || parameter.getType()==Float.class) {
                  Object values = parameter.getAllowedValues();
                  if (values instanceof Object[]) {
                         Element minElement=document.createElement("min");
                         Element maxElement=document.createElement("max");
                         minElement.setTextContent(((Object[])values)[0].toString());
                         maxElement.setTextContent(((Object[])values)[1].toString());
                         parameterElement.appendChild(minElement);                       
                         parameterElement.appendChild(maxElement);                       
                  }
             } else if (parameter.getType()==String.class) {
                  Object values = parameter.getAllowedValues();
                  if (values instanceof Object[]) {
                      for (Object value:(Object[])values) {
                         String valueAsString=value.toString();
                         Element optionElement=document.createElement("option");
                         if (optionValueMap!=null && optionValueMap.containsKey(parameterName)) {
                             HashMap<String,String>map=optionValueMap.get(parameterName);                             
                             String usevalue=map.get(valueAsString);
                             if (usevalue!=null) optionElement.setAttribute("value", usevalue);                             
                         }
                         optionElement.setTextContent(valueAsString);
                         parameterElement.appendChild(optionElement);                            
                      }
                  } else if (values!=null) {
                     Element optionElement=document.createElement("option");
                     optionElement.setTextContent(values.toString());
                     parameterElement.appendChild(optionElement);                      
                  }
             }
             Object def = parameter.getDefaultValue();
             if (def!=null) {
                 Element defaultValueElement=document.createElement("default");
                 defaultValueElement.setTextContent(def.toString());
                 parameterElement.appendChild(defaultValueElement);
             }  
             String descriptionString = parameter.getDescription();
             if (descriptionString!=null) {
                 Element descriptionElement=document.createElement("description");
                 descriptionElement.setTextContent(descriptionString);
                 parameterElement.appendChild(descriptionElement);
             }            
             DataFormat dataformat=getArgumentDataFormat(parameterName);
             if (dataformat!=null) {
                 Element dataformatElement=document.createElement("dataformat");
                 dataformatElement.setAttribute("name", dataformat.getName());
                 ParameterSettings dataformatsettings=getArgumentDataFormatSettings(parameterName);
                 if (dataformatsettings!=null) {
                     String[] settings=dataformatsettings.getParameters();
                     for (String setting:settings) { 
                         Element settingsElement=document.createElement("setting");
                         settingsElement.setAttribute("name", setting);
                         settingsElement.setAttribute("class", dataformatsettings.getTypeForParameter(setting, dataformat.getParameters()).getSimpleName());
                         String link=getParameterReferenceLink(parameterName,setting);
                         if (link!=null) settingsElement.setAttribute("link", link); 
                         else {
                             Object defaultValue=dataformatsettings.getParameterAsString(setting, dataformat.getParameters());
                             if (defaultValue!=null) settingsElement.setTextContent(defaultValue.toString());
                         }
                         dataformatElement.appendChild(settingsElement);                         
                     }                     
                 }
                 parameterElement.appendChild(dataformatElement);
             }           
             Element argumentElement=document.createElement("argument");
             String argumentType=getArgumentType(parameterName);
             argumentElement.setAttribute("type", argumentType);
             String implicitfilename=getImplicitArgument(parameterName);
             if (implicitfilename!=null && !implicitfilename.isEmpty()) argumentElement.setAttribute("filename", implicitfilename);
             String argumentSwitch=getArgumentSwitch(parameterName);
             if (argumentSwitch!=null) argumentElement.setAttribute("switch", argumentSwitch);
             parameterElement.appendChild(argumentElement);
             program.appendChild(parameterElement); 
         }        
         for (String temp:temporaryfilenames) {
             Element tempdirElement=document.createElement("temporary");
             tempdirElement.setAttribute("filename", temp);
             program.appendChild(tempdirElement); 
         }
         document.appendChild(program);
         return document;
    }

    /** Initializes an ExternalProgram based on an XML-configuration file read from a file
     * @param configurationfile The XML-formatted configuration file of the program
     * @param keepAllProperties If this is TRUE all non-essential properties of the program (such as program description, license text, author contact information and citation information)
     *        is retained in the ExternalProgram object (possibly hogging memory). If the flag is false, only essential info is kept
     */    
    public static ExternalProgram initializeExternalProgramFromFile(File configurationfile, boolean keepAllProperties) throws SystemError {
        try {
            FileInputStream inputstream=new FileInputStream(configurationfile);
            return initializeExternalProgramFromStream(inputstream, configurationfile.getName(),keepAllProperties); 
        }
        catch (FileNotFoundException fnf) {throw new SystemError("File not found error: "+configurationfile.getAbsolutePath());}        
    }    
    
    /** Initializes an ExternalProgram based on an XML-configuration file read from a stream
     * 
     *  Note: this method is used to initialize (on startup) "installed programmes" whose XML-representations 
     *  are stored "internally" in the local PriorsEditor work-directory. However, it is also used to parse
     *  new config-files that are in process of being installed. These new XML-files are first parsed by
     *  this method, then a local XML-"copy" is stored by generating a new XML-representation with the
     *  getXMLrepresentation() method. The stored copy is subsequently used (and parsed by this method)
     *  whenever PriorsEditor starts up.
     *  @param inputstream The InputStream from which to read the XML-formatted configuration file of the program
     *  @param sourceName the name (e.g. a filename) that describes the original source for the inputstream. This is used when reporting errors
     *  @param keepAllProperties If this is TRUE all non-essential properties of the program (such as program description, license text, author contact information and citation information)
     *         is retained in the ExternalProgram object (possibly hogging memory). If the flag is false, only essential info is kept
     * 
     */  
    public static ExternalProgram initializeExternalProgramFromStream(InputStream inputstream, String sourcename, boolean keepAllProperties) throws SystemError {
        ExternalProgram externalProgram=null;
        try {
         DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
         Document doc = builder.parse(inputstream);
         NodeList programnodes = doc.getElementsByTagName("program");
         if (programnodes.getLength()==0) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n No 'program' node found");
         Element programNode = (Element) programnodes.item(0);
         String programName=programNode.getAttribute("name");
         if (programName==null || programName.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'name' attribute for program");
         String programCygwin=programNode.getAttribute("cygwin");
         String programclass=programNode.getAttribute("class");
         if (programclass==null || programclass.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'class' attribute for program");
              if (programclass.equalsIgnoreCase("MotifDiscovery")) externalProgram=new MotifDiscovery();
         else if (programclass.equalsIgnoreCase("MotifScanning")) externalProgram=new MotifScanning();
         else externalProgram=new ExternalProgram();
         externalProgram.name=programName;
         externalProgram.programclass=programclass;
         externalProgram.cygwin=(programCygwin!=null && (programCygwin.equalsIgnoreCase("yes") || programCygwin.equalsIgnoreCase("true"))); // this is a boolean property

         // --- Read properties ---
         NodeList propertiesnodes = programNode.getElementsByTagName("properties");
         if (propertiesnodes.getLength()>0 && keepAllProperties) {
             Element propertiesNode = (Element) propertiesnodes.item(0);
             NodeList propertiesList=propertiesNode.getChildNodes();
             for (int i=0;i<propertiesList.getLength();i++) {
                 Node property = (Node)propertiesList.item(i);
                 String propertyName=property.getNodeName();
                 if (propertyName.equals("#text")) continue; // a hack !
                 String propertyValue=property.getTextContent();
                 propertyValue=propertyValue.replace("&lt;", "<"); // unescape escaped html
                 propertyValue=propertyValue.replace("&gt;", ">"); // unescape escaped html
                 externalProgram.setProperty(propertyName, propertyValue, false);
                 // System.err.println("Adding Property: "+propertyName+" => "+propertyValue);
             }
         }

         // --- Read service ---
         NodeList servicenodes = programNode.getElementsByTagName("service");
         if (servicenodes.getLength()==0) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n No 'service' node found");
         Element serviceNode = (Element) servicenodes.item(0);
         externalProgram.serviceType=serviceNode.getAttribute("type");
         if (externalProgram.serviceType==null || externalProgram.serviceType.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'type' attribute for service");
         externalProgram.location=serviceNode.getAttribute("location");
         NodeList executableSourcesList = serviceNode.getElementsByTagName("source");
         if (executableSourcesList.getLength()>0) {
             externalProgram.executableSources=new ArrayList<String[]>(executableSourcesList.getLength());
             for (int i=0;i<executableSourcesList.getLength();i++) {
                 Element source = (Element)executableSourcesList.item(i);
                 String os=source.getAttribute("os");
                 String version=source.getAttribute("version");
                 String url=source.getAttribute("url");
                 String compression=source.getAttribute("compression");
                 if (compression!=null && (compression.equalsIgnoreCase("none") || compression.isEmpty())) compression=null;
                 String relativeZIPpath=source.getAttribute("targetInZIP");
                 if (compression!=null && (relativeZIPpath==null || relativeZIPpath.isEmpty())) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Location of executable file within the archive must be specified!");
                 externalProgram.executableSources.add(new String[]{os,version,url,compression,relativeZIPpath});
             }
         }

         // --- Read command line ---
         NodeList commandnodes = programNode.getElementsByTagName("command");
         if (commandnodes.getLength()==0) externalProgram.command=null; // throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n No 'command' node found");
         else {
             Element commandNode = (Element) commandnodes.item(0);
             externalProgram.command=commandNode.getTextContent();
         }

         //if (externalProgram.location==null || externalProgram.location.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]: Missing 'location' attribute for service");
         // --- Read parameter tags ---
         NodeList parameternodes = programNode.getElementsByTagName("parameter");
         for (int i=0;i<parameternodes.getLength();i++) {
             boolean required=true;
             boolean hidden=false;
             Element parameter = (Element)parameternodes.item(i);
             String parameterName=parameter.getAttribute("name");
             if (parameterName==null || parameterName.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'name' attribute for parameter #"+(i+1));
             String parameterType=parameter.getAttribute("type");
             if (parameterType==null || parameterType.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'type' attribute for parameter '"+parameterName+"'");
             String parameterRequired=parameter.getAttribute("required");
             if (parameterRequired!=null && parameterRequired.equalsIgnoreCase("no")) required=false;
             String parameterHidden=parameter.getAttribute("hidden");
             if (parameterHidden!=null && parameterHidden.equalsIgnoreCase("yes")) hidden=true;
             String parameterClass=parameter.getAttribute("class");
             if (parameterClass==null || parameterClass.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'class' attribute for parameter '"+parameterName+"'");
             Class parameterclass=null;
             try {
                 parameterclass=getClassForName(parameterClass);
             } catch (ClassNotFoundException cnf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unrecognized class '"+parameterClass+"'");}
             String parameterLink=parameter.getAttribute("link");
             if (parameterLink!=null && !parameterLink.isEmpty()) {
                 externalProgram.referenceLinks.add(new String[]{parameterName,null,parameterClass,parameterLink});
             }

             String parameterDescription=null;
             NodeList parameterDescriptionList = parameter.getElementsByTagName("description");
             if (parameterDescriptionList.getLength()>0) {
                parameterDescription = ((Element)parameterDescriptionList.item(0)).getTextContent();
                if (parameterDescription.isEmpty()) parameterDescription=null;
             }
             String defaultValueString=null;
             NodeList parameterDefaultValueList = parameter.getElementsByTagName("default");
             if (parameterDefaultValueList.getLength()>0) {
                 defaultValueString = ((Element)parameterDefaultValueList.item(0)).getTextContent();
             }
             Object parameterDefaultValue=null;
             Object[] allowedValues=null;
             if (parameterClass.equals("Integer")) {
                 if (defaultValueString==null || defaultValueString.isEmpty()) parameterDefaultValue=null;
                 else try {parameterDefaultValue=new Integer(Integer.parseInt(defaultValueString));} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric default value for parameter '"+parameterName+"' => "+defaultValueString);}
                 String minValue=getTextContentFromChildElement(parameter,"min");
                 String maxValue=getTextContentFromChildElement(parameter,"max");
                 int min=0;
                 int max=1;
                 if (minValue!=null && !minValue.isEmpty()) try {min=Integer.parseInt(minValue);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric minimum value for parameter '"+parameterName+"' => "+minValue);}
                 if (maxValue!=null && !maxValue.isEmpty()) try {max=Integer.parseInt(maxValue);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric maximum value for parameter '"+parameterName+"' => "+maxValue);}
                 allowedValues=new Integer[]{min,max};
             } else if (parameterClass.equals("Double") || parameterClass.equals("Float")) {
                 if (defaultValueString==null || defaultValueString.isEmpty()) parameterDefaultValue=null;
                 else try {parameterDefaultValue=new Double(Double.parseDouble(defaultValueString));} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric default value for parameter '"+parameterName+"' => "+defaultValueString);}
                 String minValue=getTextContentFromChildElement(parameter,"min");
                 String maxValue=getTextContentFromChildElement(parameter,"max");
                 double min=0;
                 double max=1;
                 if (minValue!=null && !minValue.isEmpty()) try {min=Double.parseDouble(minValue);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric minimum value for parameter '"+parameterName+"' => "+minValue);}
                 if (maxValue!=null && !maxValue.isEmpty()) try {max=Double.parseDouble(maxValue);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected numeric maximum value for parameter '"+parameterName+"' => "+maxValue);}
                 allowedValues=new Double[]{min,max};
             } else if (parameterClass.equals("String")) {
                 if (defaultValueString==null || defaultValueString.isEmpty()) parameterDefaultValue=null;
                 else parameterDefaultValue=defaultValueString;
                 allowedValues=getStringOptions(parameter, parameterName, externalProgram);
             } else if (parameterClass.equals("Boolean")) {
                 if (defaultValueString==null || defaultValueString.isEmpty()) parameterDefaultValue=null;
                 else try {parameterDefaultValue=Boolean.parseBoolean(defaultValueString);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Unable to parse expected boolean default value for parameter '"+parameterName+"' => "+defaultValueString);}
                 allowedValues=new Boolean[]{Boolean.TRUE,Boolean.FALSE};
             } else if (Data.class.isAssignableFrom(parameterclass)) {
                  allowedValues=new Class[]{parameterclass};
                  parameterDefaultValue=null;
             } else {throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n parameter-class '"+parameterClass+"' for parameter '"+parameterName+"' is neither Data nor allowed basic type (Integer|Float|Boolean|String)");}
             // Read argument switch
             NodeList argumentlist = parameter.getElementsByTagName("argument");
             if (argumentlist.getLength()>0) { // Note that only one argument element is allowed per parameter
                String type=((Element)argumentlist.item(0)).getAttribute("type");
                externalProgram.setArgumentType(parameterName, type);
                if (type.equalsIgnoreCase("STDIN") || type.equalsIgnoreCase("STDOUT")) externalProgram.setRedirectionForArgument(parameterName, type.toUpperCase());
                String implicitfilename=((Element)argumentlist.item(0)).getAttribute("filename");
                if (implicitfilename!=null && !implicitfilename.isEmpty()) externalProgram.setImplicitArgument(parameterName,implicitfilename);
                String switchString=((Element)argumentlist.item(0)).getAttribute("switch");
                if (switchString!=null && !switchString.isEmpty()) externalProgram.setArgumentSwitch(parameterName,switchString);
             }
             NodeList dataformatlist = parameter.getElementsByTagName("dataformat");
             if (dataformatlist.getLength()>0) {
                Element dataformatElement=((Element)dataformatlist.item(0));
                setDataFormatAndSettings(parameterName, dataformatElement, externalProgram);
             }
                  if (parameterType.equals("regular")) externalProgram.addParameter(parameterName,parameterclass,parameterDefaultValue,allowedValues,parameterDescription,required,hidden);
             else if (parameterType.equals("source")) externalProgram.addSourceParameter(parameterName,parameterclass,parameterDefaultValue,allowedValues,parameterDescription);
             else if (parameterType.equals("result")) externalProgram.addResultParameter(parameterName,parameterclass,parameterDefaultValue,allowedValues,parameterDescription);
             externalProgram.allArguments.add(parameterName);
         }
         NodeList tempdirnodes = programNode.getElementsByTagName("temporary");
         for (int i=0;i<tempdirnodes.getLength();i++) {
             Element tempdir = (Element)tempdirnodes.item(i);
             String tempdirName=tempdir.getAttribute("filename");
             if (tempdirName==null || tempdirName.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+sourcename+"]:\n\n Missing 'filename' attribute for temporary-element #"+(i+1));
             externalProgram.temporaryfilenames.add(tempdirName);
         }
       } // -- end try
       catch (SystemError se) {
           throw se;
       }catch (Exception e) {
           //e.printStackTrace(System.err);
           throw new SystemError("An error occurred while loading External Program:\n\n ["+e.getClass().getSimpleName()+"]: "+e.getMessage());
       }
       return externalProgram;
}
    
/** Given a reference to a parent node in an XML-document, and the name of a child-element
 * (assuming only one child element with the given tag name exists) this method will return 
 * the text content of the child (or NULL if no such child exists).
 */    
private static String getTextContentFromChildElement(Element parent, String childName) {
     NodeList list = parent.getElementsByTagName(childName);
     if (list.getLength()>0) {
        return ((Element)list.item(0)).getTextContent();
     } else return null;   
}    
/** Given a reference to a parent node in an XML-document, and the name of child-elements
 * (assuming multiple children with the given tag name can exist) this method will return 
 * a String array containing the text content of all such child tags 
 * (or NULL if no children with the given tag name exists)
 */    
private static String[] getTextContentFromChildElements(Element parent, String childName) {
     NodeList list = parent.getElementsByTagName(childName);
     if (list.getLength()==0) return null;
     String[] result=new String[list.getLength()];
     for (int i=0;i<list.getLength();i++) {         
        result[i]=((Element)list.item(i)).getTextContent();
     }
     return result;
}    

/** 
 * Returns a list of 'option' fields for parameters of the String type
 * Also populates the optionValueMap lookup-table if the options have alternative values
 */    
private static String[] getStringOptions(Element parent, String parameterName, ExternalProgram program) {
     NodeList list = parent.getElementsByTagName("option");
     if (list.getLength()==0) return null;
     String[] result=new String[list.getLength()];
     for (int i=0;i<list.getLength();i++) {         
        result[i]=((Element)list.item(i)).getTextContent();
        org.w3c.dom.Attr valueattr=((Element)list.item(i)).getAttributeNode("value");
        if (valueattr!=null) {
            if (program.optionValueMap==null) program.optionValueMap=new HashMap<String,HashMap<String,String>>();
            HashMap<String,String>map=program.optionValueMap.get(parameterName);
            if (map==null) {map=new HashMap<String,String>();program.optionValueMap.put(parameterName, map);}
            map.put(result[i], valueattr.getValue());            
        }
     }
     return result;
}    




/** 
 * Reads a part of a tree that corresponds to a DataFormat element and sets 
 * the relevant properties in the external program configuration 
 */
private static void setDataFormatAndSettings(String parameterName, Element dataformatElement, ExternalProgram externalProgram) throws SystemError {
    //System.err.println("Reading dataformat for "+parameterName+"  "+dataformatElement.getTagName()+"  "+dataformatElement.getAttribute("name"));
     String formatName=dataformatElement.getAttribute("name");
     if (formatName==null || formatName.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\nMissing 'name' attribute for dataformat");    
     // if (engine.getDataFormat(formatName)==null) throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\nUnknown dataformat '"+formatName+"'");    
     externalProgram.setArgumentDataFormat(parameterName, formatName);
     NodeList parameterlist = dataformatElement.getElementsByTagName("setting");
     if (parameterlist.getLength()==0) return; // no more to do here
     ParameterSettings settings=new ParameterSettings();
     for (int i=0;i<parameterlist.getLength();i++) {         
         Element dataformatParameter=(Element)parameterlist.item(i);
         String dataformatParameterName=dataformatParameter.getAttribute("name");
         if (dataformatParameterName==null || dataformatParameterName.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Missing 'name' attribute for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");
         String dataformatParameterClass=dataformatParameter.getAttribute("class");
         if (dataformatParameterClass==null || dataformatParameterClass.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Missing 'type' attribute for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");
         String dataformatParameterLink=dataformatParameter.getAttribute("link");
         if (dataformatParameterLink!=null && !dataformatParameterLink.isEmpty()) { // the value of this setting links a different parameter
             externalProgram.referenceLinks.add(new String[]{parameterName,dataformatParameterName,dataformatParameterClass,dataformatParameterLink});
             settings.setParameter(dataformatParameterName, null);
             continue; // the reference links will be resolved later!
         }
         String dataformatParameterValueString=dataformatParameter.getTextContent();
         if (dataformatParameterValueString==null || dataformatParameterValueString.isEmpty()) throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Missing 'value' attribute for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");
         Object dataFormatParameterValue=null;
         if (dataformatParameterClass.equals("Integer")) {
             try {dataFormatParameterValue=new Integer(Integer.parseInt(dataformatParameterValueString));} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Unable to parse expected numeric value for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");}
         } else if (dataformatParameterClass.equals("Float")) {
             try {dataFormatParameterValue=new Float(Float.parseFloat(dataformatParameterValueString));} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Unable to parse expected numeric value for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");}
         } else if (dataformatParameterClass.equals("String")) {
             dataFormatParameterValue=dataformatParameterValueString;
         } else if (dataformatParameterClass.equals("Boolean")) {
             try {dataFormatParameterValue=Boolean.parseBoolean(dataformatParameterValueString);} catch (NumberFormatException nf) {throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Unable to parse expected boolean value for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");}
         } else {throw new SystemError("An error occurred while loading External Program ["+externalProgram.getName()+"]:\n\n Illegal type for parameter in dataformat["+formatName+"] for program-parameter '"+parameterName+"'");}
         settings.setParameter(dataformatParameterName, dataFormatParameterValue);
     } 
     externalProgram.setArgumentDataFormatSettings(parameterName, settings);
     //System.err.println("Settings for "+parameterName+" => ");
}

/**
 * This method resolves any bindings between DataFormat parameter settings and
 * values of other parameters. (It is possible to link or bind the value of 
 * a DataFormat parameter to values of normal ExternalProgram parameters, so that
 * the first will use the value of the latter). 
 * 
 * @param program A link to the external program
 * @param argument The name of the parameter whose DataFormat settings might need to be resolved
 * @param settings The settings for the DataFormat used by the given parameter
 * @throws java.lang.Exception
 */
public void resolveDataFormatLinks(ExternalProgram program, String argument, ParameterSettings settings, OperationTask task) throws Exception {    
    for (String[] vector:referenceLinks) {
        if (!vector[0].equals(argument) || vector[1]==null) continue; //
        String dataformatParameterName=vector[1];
        String link=vector[3];            
        Parameter externalProgramParameter=program.getParameterWithName(link);
        if (externalProgramParameter==null) throw new ExecutionError("DataFormat setting '"+dataformatParameterName+"' used by parameter '"+argument+"' links to unknown parameter '"+link+"'");
        Object value=task.getParameter(link);
        settings.setParameter(dataformatParameterName, value);
    }
}

/** Returns the name of a parameter that is referenced by the argument parameter  (or null if argument parameter has no references)
 *  This method is used to resolve parameter links in both normal parameters and dataformat settings.
 *  When used for dataformat settings the name of 'parent' parameter (normal parameter) is the first argument
 *  and the name of the dataformat-setting parameter is the second argument.
 *  When used to retrieve links for normal parameters, the second argument should be null  
 */
public String getParameterReferenceLink(String parameterName, String dataformatParameter) {  
    for (String[] vector:referenceLinks) {
        if (!vector[0].equals(parameterName)) continue; //
        if (dataformatParameter==null && vector[1]==null) return vector[3]; // vector[3] holds the link
        if (dataformatParameter!=null && vector[1]!=null && vector[1].equals(dataformatParameter)) return vector[3]; //
    }
    return null;
}



/**
 * This method executes the given program external program
 * (the superclass executes locally installed programs only!)
 * All needed resources (like parameter settings) should be available in the
 * OperationTask object before execution and on successful completion the results
 * will be made available in the OperationTask object
 */
public void execute(OperationTask task) throws Exception {
        int sourcecount=-1;
        int resultscount=-1;
        int expectedSourceCount=sourceParameterFormats.size();
        int expectedResultsCount=resultsParameterFormats.size();
        Data[] datasources=(Data[])task.getParameter(SOURCES);
        if ((datasources!=null && datasources.length<expectedSourceCount) || (expectedSourceCount>0 && datasources==null)) throw new ExecutionError("SYSTEM ERROR: Not enough sources added to task for execution of external program "+getName()+" (Expected "+expectedSourceCount+")");
        File[] resultFiles=new File[expectedResultsCount];
        task.setStatus(ExecutableTask.WAITING); // indeterminate progress
        String commandLine=getCommand();
        boolean hasCommandLine=false; // is the command line explicitly stated in the XML config-file?
        if (commandLine==null || commandLine.isEmpty()) commandLine="%PROGRAM";
        else hasCommandLine=true;

        Object[] redirarg=getRedirectedArgumentNames((hasCommandLine)?commandLine:null);

        String[] STDIN_arguments=(String[])redirarg[0]; // the lengths of these arrays equals the number of commands to be executed in succession
        String[] STDOUT_arguments=(String[])redirarg[1]; // and if the String[x]!=null it means that the given argument should be used as STDIN/OUT

        // build a command line with applicable arguments and their values (as specified in the OperationTask)
        HashMap<String,Object> resolvedArguments=new HashMap<String,Object>(allArguments.size()); // keeps the values for arguments that have been resolved so far
        for (String argument:allArguments) {
            String currentargumentCommandLineString=""; //this is output on the command line (either appended or replaced for references)
            Parameter parameter=getParameterWithName(argument);
            if (parameter!=null) { // the argument is a regular parameter!
                Object dataobject=task.getParameter(argument);
                if (getParameterReferenceLink(argument, null)!=null) { // the parameter references another parameter
                   Object linkedValue=resolvedArguments.get(getParameterReferenceLink(argument, null));
                   resolvedArguments.put(argument, linkedValue);
                   currentargumentCommandLineString=getCommandStringForParameter(argument,linkedValue);
                }
                else if (dataobject==null || dataobject.toString().isEmpty()) {}  // this probably means that an optional parameter has been omitted. Do not mention it on the command line
                else if (isComplexType(dataobject)) {
                   File tempFile;
                   if (dataobject instanceof FeatureDataset) {
                       DataFormat dataobjectFormat=getArgumentDataFormat(argument);
                       if (dataobjectFormat==null) throw new ExecutionError("SYSTEM ERROR: Unspecified (or unrecognized) data format for "+getName()+" argument "+argument);
                       ParameterSettings dataobjectFormatSettings=getArgumentDataFormatSettings(argument);
                       resolveDataFormatLinks(this,argument,dataobjectFormatSettings,task);
                       if (isImplicitArgument(argument)) tempFile=new File(engine.getTempDirectory(),resolveGlobalDirectoryReferences(getImplicitArgument(argument)));
                       else tempFile=engine.createTempFile();
                       outputSourceSequences((FeatureDataset)dataobject,dataobjectFormat,dataobjectFormatSettings,task,tempFile);
                   }
                   else tempFile=createTempFile(this, argument, task);
                   resolvedArguments.put(argument, tempFile);
                   currentargumentCommandLineString=getCommandStringForParameter(argument,tempFile);
                } else { // argument is non-complex type: String, Number or Boolean
                   Object value=dataobject;
                   if (value instanceof String && optionValueMap!=null && optionValueMap.containsKey(argument)) {
                       HashMap<String,String> map=optionValueMap.get(argument);
                       if (map.containsKey((String)value)) {
                           String valuestring=map.get((String)value); // lookup
                           if (valuestring==null || valuestring.isEmpty()) value=null;
                           else value=valuestring;
                       }
                   }
                   resolvedArguments.put(argument, value);
                   currentargumentCommandLineString=getCommandStringForParameter(argument,value);                }
            } else if (isSourceParameter(argument)) {
                DataFormat sourceFormat=getArgumentDataFormat(argument);
                if (sourceFormat==null) throw new ExecutionError("SYSTEM ERROR: Unspecified (or unrecognized) data format for "+getName()+" source file");
                ParameterSettings sourceFormatSettings=getArgumentDataFormatSettings(argument);
                try {
                    sourcecount++;
                    Data datasource=datasources[sourcecount];
                    File sourceFile;
                    if (getParameterReferenceLink(argument, null)!=null) { // the parameter references another parameter
                         Object linkedValue=resolvedArguments.get(getParameterReferenceLink(argument, null));
                         if (linkedValue==null) throw new ExecutionError("Source parameter links to unresolved parameter (or target parameter has value NULL)");
                         if (!(linkedValue instanceof File)) throw new ExecutionError("Source parameters can only reference 'complex' data types");
                         sourceFile=(File)linkedValue;
                    }
                    else if (datasource instanceof FeatureDataset) {
                        if (isImplicitArgument(argument)) sourceFile=resolveImplicitFileForArgument(argument);
                        else sourceFile=engine.createTempFile();
                        resolveDataFormatLinks(this,argument,sourceFormatSettings,task);
                        outputSourceSequences((FeatureDataset)datasource,sourceFormat,sourceFormatSettings,task,sourceFile);
                    }
                    else sourceFile=createTempFile(this, argument, task); // this outputs the data to a temp-file (dataformat and settings are found in the task object)
                    resolvedArguments.put(argument, sourceFile);
                    currentargumentCommandLineString=getCommandStringForParameter(argument,sourceFile);
                } catch (IOException io) {
                    throw new ExecutionError("Unable to create needed temporary file for source ("+io.getClass().getSimpleName()+")\n\n"+io.getMessage(),io);
                }
            } else if (isResultsParameter(argument)) {
                resultscount++;
                if (getParameterReferenceLink(argument, null)!=null) { // the parameter references another parameter
                     Object linkedValue=resolvedArguments.get(getParameterReferenceLink(argument, null));
                     if (linkedValue==null) throw new ExecutionError("Results parameter links to unresolved parameter (or target parameter has value NULL)");
                     if (!(linkedValue instanceof File)) throw new ExecutionError("Results parameters can only reference 'complex' data types");
                     resultFiles[resultscount]=(File)linkedValue;
                }
                else if (isImplicitArgument(argument)) {
                    resultFiles[resultscount]=resolveImplicitFileForArgument(argument);
                    resultFiles[resultscount].deleteOnExit();
                }
                else resultFiles[resultscount]=engine.createTempFile(); // these temp-files are deleted on exit
                if (resultFiles[resultscount]==null) throw new ExecutionError("Unable to create needed temporary file for results");
                resolvedArguments.put(argument, resultFiles[resultscount]);
                currentargumentCommandLineString=getCommandStringForParameter(argument,resultFiles[resultscount]);
            } else throw new ExecutionError("SYSTEM ERROR: Unknown parameter: "+argument);
            String argumentReferenceName="{"+argument+"}"; // use the name in braces to refer to it on the command line
            if (hasCommandLine) {
                //System.err.println("["+argument+"]  refName='"+argumentReferenceName+"'  value='"+currentargumentCommandLineString+"'");
                if (argumentReferenceName!=null && !argumentReferenceName.isEmpty()) commandLine=commandLine.replace(argumentReferenceName, currentargumentCommandLineString); // replace references to the argument in the command line with the arguments resolved value
            } else { // no explicit command line is given. Just add arguments to the end of the line
                if (!currentargumentCommandLineString.isEmpty() && !isImplicitArgument(argument)) {
                    commandLine+=" "+currentargumentCommandLineString;
                } // not reference in the command line to this parameter. Add it to the end?
            }
        } // end of foreach argument

        // resolve additional references in the command line
        commandLine=resolveGlobalDirectoryReferences(commandLine);
        String programLocation=getLocation();
        if (programLocation.contains(" ")) programLocation="\""+programLocation+"\""; // escape filenames containing spaces!
        commandLine=commandLine.replace("%PROGRAM", programLocation); //
        // replace path-references: ${path}$  and $'{path}'$
        HashMap<String,String> tobereplaced=new HashMap<String,String>();
        Pattern pattern = Pattern.compile("\\$\\{(.+?)\\}\\$");
        Matcher matcher=pattern.matcher(commandLine);
        while (matcher.find()) {
            String path=convertFilePath(matcher.group(1));
            tobereplaced.put(matcher.group(0),path);
        }
        for (String key:tobereplaced.keySet()) {
            String replacement=tobereplaced.get(key);
            commandLine=commandLine.replace(key, replacement);
        }
        tobereplaced.clear();
        pattern = Pattern.compile("\\$'\\{(.+?)\\}'\\$");
        matcher=pattern.matcher(commandLine);
        while (matcher.find()) {
            String path=quoteFilePathIfNecessary(matcher.group(1));
            tobereplaced.put(matcher.group(0),path);
        }
        for (String key:tobereplaced.keySet()) {
            String replacement=tobereplaced.get(key);
            commandLine=commandLine.replace(key, replacement);
        }
        // System.err.println(commandLine);
        // There might be several commands in one commandLine (separated by semicolons).
        // Execute each command in turn
        String[] commands=commandLine.split("\\s*;\\s*");
        for (int commandNumber=0;commandNumber<commands.length;commandNumber++) { // for each command
           String commandline=commands[commandNumber];
           String[] argumentslist=commandline.split("\\s+");
           engine.logMessage(commandline);
           String STDIN_argument=STDIN_arguments[commandNumber];
           String STDOUT_argument=STDOUT_arguments[commandNumber];

           // -- Create a process and run it. Also create an OutputMonitor to catch STDOUT and STDERR activity
           ProcessBuilder pb = new ProcessBuilder(Arrays.asList(argumentslist));

           pb.redirectErrorStream(STDOUT_argument==null);
           File appdir=null;
           if (getApplicationDirectory()!=null) appdir=new File(getApplicationDirectory());
           pb.directory(appdir);
           //pb.directory(new File(engine.getTempDirectory())); // use the directory of this PriorsEditor session as work-dir for the process

           Process process = pb.start();

           OutputMonitor monitor=null;

           InputFeeder STDINfeeder=null;
           if (STDIN_argument!=null) {
               Object file=resolvedArguments.get(STDIN_argument);
               if (file==null || !(file instanceof File)) throw new ExecutionError("SYSTEM ERROR: Argument '"+STDIN_argument+"' redirected to STDIN has not been assigned a file");
               else STDINfeeder=new InputFeeder(task, process, Thread.currentThread(), (File)file);
               //System.err.println("STDIN ["+STDIN_argument+"] => "+((File)file).getAbsolutePath());
           }
           OutputMonitor STDOUTreceiver=null;
           if (STDOUT_argument!=null) {
               Object file=resolvedArguments.get(STDOUT_argument);
               if (file==null || !(file instanceof File)) throw new ExecutionError("SYSTEM ERROR: Argument '"+STDOUT_argument+"' redirected to STDOUT has not been assigned a file");
               else STDOUTreceiver=new OutputMonitor(task, process, Thread.currentThread(), (File)file, "STDOUT");
               //System.err.println("STDOUT ["+STDOUT_argument+"] => "+((File)file).getAbsolutePath());
               monitor=new OutputMonitor(task, process, Thread.currentThread(), null,"STDERR");
           } else monitor=new OutputMonitor(task, process, Thread.currentThread(),null, "MERGED");

           int flag=0;
           try {
               if (STDOUTreceiver!=null) STDOUTreceiver.start(); // this thread stores the STDOUT output in a file
               monitor.start();
               if (STDINfeeder!=null) STDINfeeder.start(); // start feeding data to the process' STDIN
               flag=process.waitFor();
               if (flag!=0) engine.logMessage(getName()+" finished with error code: "+flag);
           } catch (Exception e) { // this could be an Interrupted exception or else
               // destroying the process causes closing of STDIN/STDOUT/STDERR streams
               // which will cause the auxilary threads to finish
               process.destroy();
               // Wait for the auxilary threads to finish. I am not sure if this is necessary, but just in case
               if (STDINfeeder!=null) try {STDINfeeder.join();} catch (InterruptedException ie) {}
               if (STDOUTreceiver!=null) try {STDOUTreceiver.join();} catch (InterruptedException ie) {}
               if (monitor!=null) try {monitor.join();} catch (InterruptedException ie) {}

               deleteFiles(temporaryfilenames);

               throw e;
           }
       } // end for each command

       // -- parse result files and set results parameters in the task. Also delete temporary files
       for (int i=0;i<expectedResultsCount;i++) {
           Parameter resultsParameter=resultsParameterFormats.get(i);
           String resultParameterName=resultsParameter.getName();
           ArrayList<String> result=readFile(resultFiles[i]);
           DataFormat resultsFormat=getArgumentDataFormat(resultParameterName);
           if (resultsFormat==null) throw new ExecutionError("SYSTEM ERROR: Unspecified (or unrecognized) data format for "+getName()+" results ("+resultParameterName+")");
           ParameterSettings resultsFormatSettings=getArgumentDataFormatSettings(resultParameterName);
           resolveDataFormatLinks(this,resultParameterName,resultsFormatSettings,task);
           Class resultsDataType=resultsParameter.getType();
           Data resultsData=createResultsData(resultsDataType,resultParameterName);
           resultsFormat.parseInput(result, resultsData, resultsFormatSettings, task);
           task.setParameter(resultParameterName, resultsData);
       }
       task.setStatus(ExecutableTask.RUNNING);
       deleteFiles(temporaryfilenames);
}


/**
 * This method returns an array with 2 elements both of which are String arrays (String[])
 * The two String[] arrays contains names of arguments that are linked to redirection wrt
 * STDIN or STOUD (the first array is for STDIN the second for STDOUT). The length of the
 * String[] arrays is the same as the number of commands that should be executed in succesion
 * when this external program is run. E.g. lets say two commands should be run and the values
 * of the arrays are STDIN[0]==arg1, STDIN[1]==null, STDOUT[0]==null, STDOUT[1]==arg2.
 * This means that for the first command there is an argument (arg1) which should be feed to the
 * process through STDIN but no data needs to be read back from STDOUT. For the second command
 * no data needs to be feed to STDIN but this command produces output to STDOUT which must be
 * read and associated with the argument 'arg2'.
 *
 * @param commandLine
 * @return
 */
private Object[] getRedirectedArgumentNames(String commandLine) throws ExecutionError {
    String[] STDIN;
    String[] STDOUT;
    if (commandLine==null || !commandLine.contains(";")) { // single command
        STDIN=new String[1];
        STDOUT=new String[1];
        for (String key:redirectedArguments.keySet()) {
            String destination=redirectedArguments.get(key);
            if (destination.equalsIgnoreCase("STDIN")) {
               if (STDIN[0]!=null) throw new ExecutionError("Only one argument can be redirected to STDIN for each command");
               else STDIN[0]=key;
            } else if (destination.equalsIgnoreCase("STDOUT")) {
                if (STDOUT[0]!=null) throw new ExecutionError("Only one argument can be redirected to STDOUT for each command");
                else STDOUT[0]=key;
            } else throw new ExecutionError("Unknown redirection destination '"+destination+"' (Allowed values are STDIN and STDOUT)");
        }
    } else { // the commandLine consists of multiple commands which should be executed in order
        String[] commands=commandLine.split("\\s*;\\s*");
        STDIN=new String[commands.length];
        STDOUT=new String[commands.length];
        Pattern pattern = Pattern.compile("\\{(.+?)\\}");

        for (int i=0;i<commands.length;i++) {
            String subcommand=commands[i];
            Matcher matcher = pattern.matcher(subcommand);
            while (matcher.find()) {
                String argumentname=matcher.group(1);
                if (!redirectedArguments.containsKey(argumentname)) continue;
                String destination=redirectedArguments.get(argumentname);
                //System.err.println("Redirecting '"+argumentname+"' to "+destination+" for command="+i);
                if (destination.equalsIgnoreCase("STDIN")) {
                   if (STDIN[i]!=null) throw new ExecutionError("Only one argument can be redirected to STDIN for each command");
                   else STDIN[i]=argumentname;
                } else if (destination.equalsIgnoreCase("STDOUT")) {
                    if (STDOUT[i]!=null) throw new ExecutionError("Only one argument can be redirected to STDOUT for each command");
                    else STDOUT[i]=argumentname;
                } else throw new ExecutionError("Unknown redirection destination '"+destination+"' (Allowed values are STDIN and STDOUT)");
            }
        }

    }

    return new Object[]{STDIN,STDOUT};
}



/** Given an argument name and the resolved value of this argument (which could be null),
 *  this method returns a string that could be inserted in a command line to represent this argument.
 *  Depending on the argument type, the returned string could be a switch followed by a value,
 *  just the switch (for Boolean flags), just the value or simply and empty string.
 */
private String getCommandStringForParameter(String argumentname, Object value) {
   if (value==null || value.toString().isEmpty()) return "";
   if (isLinkedToRedirectionArgument(argumentname)) return ""; // redirection argument should not be written to command line. They are implicit!
   String argumentswitch=getArgumentSwitch(argumentname);
   String cmdString=(argumentswitch!=null)?argumentswitch+" ":"";
   if (value instanceof File) cmdString+=convertFilePath(((File)value).getAbsolutePath());
   else if (!(value instanceof Boolean)) cmdString+=(value.toString());
   else if (value instanceof Boolean && ((Boolean)value).booleanValue()==false) cmdString=""; // do not output anything for boolean flags that are false
   return cmdString;
}

private void deleteFiles(ArrayList<String> files) {
   for (String filename:files) {
       filename=resolveGlobalDirectoryReferences(filename); 
       File tempfile=new File(filename);          
       if (!engine.deleteTempFile(tempfile)) engine.logMessage("WARNING: Unable to delete temporary file (or directory): "+tempfile.getAbsolutePath());
   }    
} 

private File resolveImplicitFileForArgument(String argument) {
    String impArg=getImplicitArgument(argument);
    String resolvedArg=resolveGlobalDirectoryReferences(impArg);
    if (impArg.equals(resolvedArg)) return new File(engine.getTempDirectory(),impArg); // the reference is local, so put it under temp-dir
    else return new File(resolvedArg);
}

private String resolveGlobalDirectoryReferences(String command) {
    command=command.replace("%WORKDIR/", engine.getTempDirectory()+File.separator); // 
    command=command.replace("%WORKDIR\\", engine.getTempDirectory()+File.separator); // 
    command=command.replace("%APPDIR/", getApplicationDirectory()+File.separator); // 
    command=command.replace("%APPDIR\\", getApplicationDirectory()+File.separator); //    
    return command;
}

/** 
 * Does extra processing of a filename path 
 * Converts pathnames to 'Cygwin format' if the program is run under cygwin 
 * and adds quotes around the filename if necessary
 */
private String convertFilePath(String pathname) {
   if (isRunInCygwin() && System.getProperty("os.name").startsWith("Windows")) {
       if (pathname.length()>=3 && pathname.charAt(1)==':' && pathname.charAt(2)=='\\') {
           char driveLetter=Character.toLowerCase(pathname.charAt(0));
           pathname="/cygdrive/"+driveLetter+"/"+pathname.substring(3);
       }   
       pathname=pathname.replace("\\", "/");
       // if (pathname.contains(" ")) pathname="\""+pathname+"\"";
   }
   if (System.getProperty("os.name").startsWith("Windows") && pathname.contains(" ")) pathname="\""+pathname+"\"";
   return pathname;
}
/**
 * Adds quotes around a filename (path) if the OS is Windows and the pathname contains spaces
 */
private String quoteFilePathIfNecessary(String pathname) {
   if (System.getProperty("os.name").startsWith("Windows") && pathname.contains(" ")) pathname="\""+pathname+"\"";
   return pathname;
}


/** This might be a little overkill since data will mostly be of FeatureDataset type (?) */ 
private Data createResultsData(Class type, String tempName) throws ExecutionError {
         if (type==RegionDataset.class) return new RegionDataset(tempName);
    else if (type==NumericDataset.class)  return new NumericDataset(tempName);
    else if (type==DNASequenceDataset.class)  return new DNASequenceDataset(tempName);
    else if (type==MotifCollection.class) return new MotifCollection(tempName);
    else if (type==Motif.class) return new Motif(tempName);
    else if (type==StringVariable.class) return new StringVariable(tempName,"empty");
    else if (type==StringCollection.class) return new StringCollection(tempName);
    else if (type==NumericVariable.class) return new NumericVariable(tempName,0);
    else if (type==NumericMapVariable.class) return new NumericMapVariable(tempName,0);
    else if (type==BackgroundModel.class) return new BackgroundModel(tempName);
    else if (type==SequenceCollection.class) return new SequenceCollection(tempName);
    throw new ExecutionError("Results data object not created (type="+type.getSimpleName()+")");
}

private static Class getClassForName(String classname) throws ClassNotFoundException {
         if (classname.equalsIgnoreCase("RegionDataset")) return RegionDataset.class;
    else if (classname.equalsIgnoreCase("NumericDataset"))  return NumericDataset.class;
    else if (classname.equalsIgnoreCase("DNASequenceDataset"))  return DNASequenceDataset.class;
    else if (classname.equalsIgnoreCase("MotifCollection")) return MotifCollection.class;
    else if (classname.equalsIgnoreCase("Motif")) return Motif.class;
    else if (classname.equalsIgnoreCase("StringVariable")) return StringVariable.class;
    else if (classname.equalsIgnoreCase("StringCollection")) return StringCollection.class;
    else if (classname.equalsIgnoreCase("NumericVariable")) return NumericVariable.class;
    else if (classname.equalsIgnoreCase("NumericMapVariable")) return NumericMapVariable.class;
    else if (classname.equalsIgnoreCase("BackgroundModel")) return BackgroundModel.class;      
    else if (classname.equalsIgnoreCase("String")) return String.class;      
    else if (classname.equalsIgnoreCase("Integer")) return Integer.class;      
    else if (classname.equalsIgnoreCase("Float")) return Float.class;   
    else if (classname.equalsIgnoreCase("Double")) return Float.class; // for forward compatibility :P
    else if (classname.equalsIgnoreCase("Boolean")) return Boolean.class;
    else if (classname.equalsIgnoreCase("SequenceCollection")) return SequenceCollection.class;
    throw new ClassNotFoundException(classname);
}



private void outputSourceSequences(FeatureDataset sourceData, DataFormat outputFormat, ParameterSettings outputSettings, OperationTask task, File sequencesFile) throws Exception {
        SequenceCollection sequenceCollection=(SequenceCollection)task.getParameter(OperationTask.SEQUENCE_COLLECTION);
        ArrayList<Sequence> sequences=sequenceCollection.getAllSequences(engine);
        OutputData sequenceOutput=new OutputData(sourceData+"_sourceTemp");
        for (Sequence sequence:sequences) { // for each sequence
            task.checkExecutionLock(); // checks to see if this task should suspend execution
            if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) throw new InterruptedException();
            String sequenceName=sequence.getName();
            FeatureSequenceData sourceSequence=(FeatureSequenceData)sourceData.getSequenceByName(sequenceName);         
            try {                 
                 outputFormat.format(sourceSequence, sequenceOutput, outputSettings, task);
           } catch (Exception e) {throw e;}              
        }                   
        sequenceOutput.saveToFile(sequencesFile.getAbsolutePath());  
    }
    


/** Creates a temporary file holding the data for the given argment 
 *  The argument data is output in a format (with settings) as specified in the OperationTask object
 */        
private File createTempFile(ExternalProgram program, String argument, OperationTask task) throws Exception {
    File outputfile=engine.createTempFile();
    if (outputfile==null) throw new ExecutionError("Unable to create necessary temporary file");
    DataFormat dataformat=program.getArgumentDataFormat(argument);
    if (dataformat==null) throw new ExecutionError("SYSTEM ERROR: Unspecified (or unrecognized) data format for "+program.getName()+" argument "+argument);
    ParameterSettings formatsettings=program.getArgumentDataFormatSettings(argument);
    resolveDataFormatLinks(program,argument,formatsettings,task);
    OutputData output=new OutputData(argument+"Temp");
    Data dataobject=(Data)task.getParameter(argument);
    dataformat.format(dataobject, output, formatsettings, task);
    output.saveToFile(outputfile.getAbsolutePath());
    return outputfile;
}
    
/** */
private ArrayList<String> readFile(File file) throws Exception {
    BufferedReader inputStream = null;
    ArrayList<String> result=new ArrayList<String>();
    try {
        inputStream =  new BufferedReader(new FileReader(file));
        String l;
        while ((l = inputStream.readLine()) != null) {
            result.add(l);
        }
    } catch (Exception e) {
        throw e;
    } finally {
        if (inputStream != null) {try {inputStream.close();} catch (Exception c) {}}
    }
    return result;
}    
    
private boolean isComplexType(Object object) {
    if (object instanceof FeatureDataset) return true;
    if (object instanceof FeatureSequenceData) return true;
    if (object instanceof Motif) return true;
    if (object instanceof MotifCollection) return true;
    if (object instanceof StringCollection) return true;
    if (object instanceof Sequence) return true;
    if (object instanceof SequenceCollection) return true;
    if (object instanceof BackgroundModel) return true;
    if (object instanceof NumericMapVariable) return true;
    return false;
}
    

    

/**
 * This class monitors activity on STDOUT and or STDERR made by the external process
 * and conveys this information to the task as a status message
 * @param outputfile If this is not NULL the data read will be written to this file
 * @param type should be STDOUT, STDERR or MERGED
 *
*/
private class OutputMonitor extends Thread {
        ExecutableTask task;
        Process process;
        Thread taskThread;
        File outputfile=null;
        String type;
        BufferedReader input=null;
        BufferedWriter output=null;

        public OutputMonitor(ExecutableTask task, Process process, Thread taskThread, File outputfile, String type) {
            this.task=task;
            this.process=process;
            this.taskThread=taskThread;
            this.outputfile=outputfile;
            this.type=type;
        }


        @Override
        public void run() {
            String line;
            try {
               if (type.equalsIgnoreCase("STDERR")) input=new BufferedReader(new InputStreamReader(process.getErrorStream())); //
               else input=new BufferedReader(new InputStreamReader(process.getInputStream())); // STDERR and STDOUT may have been combined into one stream (getInputStream)

               if (outputfile!=null) output = new BufferedWriter( new OutputStreamWriter( new FileOutputStream(outputfile)));
               //System.err.println("Monitor running");
               while ((line = input.readLine()) != null) {
                   //engine.logMessage(line); // for debugging only!!
                   if (outputfile!=null) {
                       output.write(line);
                       output.newLine();
                       //output.flush();
                   } else {
                       task.setStatusMessage(line);
                   }
                   // System.err.println(line);
                   task.checkExecutionLock(); // checks to see if this task should suspend execution
                   if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) {taskThread.interrupt();break;}
               }
               input.close();
               if (output!=null) {
                   output.flush();
                   output.close();
               }
               //System.err.println("Monitor finished");
            }
            catch (IOException ioe) {engine.errorMessage("FILE ERROR: "+ioe.getMessage(), 0);}
            catch (Exception e) {}
            finally {
                try {
                    if (input!=null) input.close();
                    if (output!=null) {output.close();}
                } catch (Exception ie) {}
            }
        }

} // end OutputMonitor

/**
 * This class can be used to feed the contents of a file to the STDIN of the given process
 */
private class InputFeeder extends Thread {
        ExecutableTask task;
        Process process;
        Thread taskThread;
        File file;
        BufferedWriter output=null;
        BufferedReader input = null;

        public InputFeeder(ExecutableTask task, Process process, Thread taskThread, File file) {
            this.task=task;
            this.process=process;
            this.taskThread=taskThread;
            this.file=file;
        }


        @Override
        public void run() {
            String line;
            try {
              output = new BufferedWriter( new OutputStreamWriter( process.getOutputStream()));
              input=new BufferedReader(new InputStreamReader(new FileInputStream(file)));
              while ((line = input.readLine()) != null) {
                    output.write(line);
                    output.newLine();
                    if (Thread.interrupted() || task.getStatus().equals(ExecutableTask.ABORTED)) {taskThread.interrupt();break;}
              }
              input.close();
              output.close();
            }
            catch (IOException ioe) {engine.errorMessage("ERROR while feeding data to process STDIN: "+ioe.getMessage(),0);}
            catch (Exception e) {}
            finally {
                try {
                   if (input!=null) input.close();
                   if (output!=null) output.close();
                } catch (Exception ie) {}
            }
        }

} // end InputFeeder

}    

