/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.gui;
import java.io.*;
import java.lang.management.ManagementFactory;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.*;
import javax.swing.event.UndoableEditEvent;
import java.util.HashSet;
import java.util.HashMap;
import org.jdesktop.application.ApplicationContext;
import org.jdesktop.application.LocalStorage;
import priorseditor.engine.PriorsEditorEngine;

/**
 *
 * @author kjetikl
 */
public class GUIUndoManager extends UndoManager implements UndoableEditListener {
    private ExtendedUndoableEditSupport support;
    private Object source;
    private HashSet<String> usedCacheKeys;
    private HashMap<String,Object> virtualMemoryStorage; // to be used only as last resort
    private ApplicationContext context;
    private PriorsEditorGUI gui;
    private static int counter=0;
    private UndoableEditListener forwardListener=null;
    private String cacheDirectoryName=null; // the cacheDirectory is unique for this PriorsEditor process
    private String cacheDirectoryFullPath=null; // the cacheDirectory is unique for this PriorsEditor process
    
    /** Creates a new UndoManager to be used with the GUI */
    public GUIUndoManager(ApplicationContext context, PriorsEditorGUI gui) {
        super();
        this.context=context;
        this.gui=gui;
        cacheDirectoryName="UndoCache_"+ManagementFactory.getRuntimeMXBean().getName();
        try { 
           cacheDirectoryFullPath=context.getLocalStorage().getDirectory().getCanonicalPath()+File.separator+cacheDirectoryName;
           File dir=new File(cacheDirectoryFullPath);
           if (!dir.mkdirs()) throw new IOException("Unable to create cache directory on local disc => "+cacheDirectoryFullPath);
           dir.deleteOnExit(); // tells the VM to delete the cache directory automatically on exit
        } catch (Exception e) {gui.getEngine().logMessage("UNDO cache error: "+e.getMessage());}
        support=new ExtendedUndoableEditSupport();
    }
    
    @Override
    protected void finalize() throws Throwable {
       super.finalize();
       File cachedir=new File(cacheDirectoryFullPath);
       deleteTempFile(cachedir); // delete cachedir recursively  
    }
    
    /**
     * If the parameter recipient is set to something different than NULL,
     * incoming undoableEditEvents will not be added to the undo queue but
     * forwarded to the specified recipient instead. If recipient is NULL, 
     * undoEvents will be added to the queue as normal
     * This can be used to intercept events coming in to the GUIUndoManager.
     * 
     * @param recipient 
     */
    public void forwardUndoEvents(UndoableEditListener recipient) {
        forwardListener=recipient;
    }
    
    /** A convenience method to get to the engine */
    public PriorsEditorEngine getEngine() {
       return gui.getEngine();
    }
    
    @Override
    public synchronized boolean addEdit(UndoableEdit edit) {
        if (forwardListener!=null) {
            forwardListener.undoableEditHappened(new UndoableEditEvent(this,edit));
            return true;
        }
        boolean ok=super.addEdit(edit);
        if (ok) support.postEdit(edit);
        //gui.debugMessage("added Edit to undoManager => "+edit.toString()+"   Queue:"+edits.size()+"  ok="+ok);
        return ok;
    }
        
    @Override
    public synchronized void undoableEditHappened(UndoableEditEvent e) {
        if (forwardListener!=null) {
            forwardListener.undoableEditHappened(e);
            return;
        }
        UndoableEdit edit=e.getEdit();
        source=e.getSource();
        addEdit(edit);       
    }
    
    public synchronized void addUndoableEditListener(UndoableEditListener listener) {
        support.addUndoableEditListener(listener);    
    }
    
    public synchronized void removeUndoableEditListener(UndoableEditListener listener) {
        support.removeUndoableEditListener(listener);    
    }
    
    /**
     * Returns a unique ID that can be used to identify an UndoableEdit
     * @return
     */
    public String getUniqueUndoID() {
        counter++;
        return "Undo_"+counter;
    }
    
    
    public void debug() {
        gui.debugMessage("Undoables left="+edits.size());
        for (UndoableEdit edit:edits) {
            gui.debugMessage("  "+edit.getPresentationName()+"["+edit.getClass().toString()+"]  undo=>"+edit.canUndo()+"  redo=>"+edit.canRedo()+"    Manager:undo="+canUndo()+" redo="+canRedo()+" both="+canUndoOrRedo());
        }
    }
    
    @Override
    public void discardAllEdits(){
        super.discardAllEdits();
        gui.updateUndoRedoStates();
    }
    
    /**
     * Temporarily stores an object so that it can be retrieved later  
     * (usually on local disc but if that does not work then it stores it directly in virtual memory).
     * This is an easy way of implementing undos (by storing previous version of the data)
     * @param key A unique name to associate with the object. This name is used to retrieve it later
     * @param value The object to be stored
     * @return True if the object could be stored on disc, false if it had to be stored in memory
     */
    public boolean storeObjectInCache(String key,Object value) {
        if (usedCacheKeys==null) usedCacheKeys=new HashSet<String>();
            String filename=cacheDirectoryFullPath+File.separator+key+".ser";  
             //gui.debugMessage("Storing object of class "+value.getClass().toString()+" to '"+filename+"'");
             ObjectOutputStream stream=null;
        try {
             stream = new ObjectOutputStream(new FileOutputStream(filename));
             stream.writeObject(value);
             stream.close();
             usedCacheKeys.add(key);         
        } catch (Exception e) {
            gui.getEngine().logMessage("Note: Unable to cache '"+value+"' to local disc. Using virtual memory for storage instead (Exception message:"+e.toString()+")");
            if (virtualMemoryStorage==null) virtualMemoryStorage=new HashMap<String, Object>();
            virtualMemoryStorage.put(key, value);
            return false;
        } finally {try {if (stream!=null) stream.close();} catch (Exception x){}}
        return true;
    }
    
    /**
     * Fetches an object that has been temporarily stored
     * This is an easy way of implementing undos (by retrieving previous versions of the data that has been stored)
     * @param key The unique name of the object
     */
    public Object getObjectFromCache(String key) {
        try {
             String filename=cacheDirectoryFullPath+File.separator+key+".ser";  
             ObjectInputStream stream=new ObjectInputStream(new FileInputStream(filename));
             Object value=stream.readObject();
             stream.close();
             return value;        
        } catch (Exception e) {
            if (virtualMemoryStorage.containsKey(key)) return virtualMemoryStorage.get(key);
            else {
                gui.getEngine().logMessage("Undo Error: Unable to load cached object from local disc: "+e.toString());
                return null;
            }
        }

    }
    
    /**
     * Clears all data that has been stored in local cache (as well as the cache itself)
     * This method should only be called on system shutdown to remove any 
     * temporary files that have been created
     * 
     * This method has been deprecated since it seems to fail for Web Start applications
     */
    @Deprecated
    public void clearCacheOld() throws IOException {
        LocalStorage localstorage=context.getLocalStorage();
        if (usedCacheKeys==null) return;
        for (String key:usedCacheKeys) {
            String filename=cacheDirectoryName+File.separator+key+".ser";
            if (!localstorage.deleteFile(filename)) System.err.println("Unable to delete UNDO file '"+filename+"'");
        }
        usedCacheKeys.clear();
        File dir=new File(cacheDirectoryFullPath);
        if (!dir.delete()) System.err.println("WARNING: unable to delete UNDO directory!");
        // the temporary cache directory is already registered for automatic deletion when the VM terminates
    }
    
    /**
     * Clears all data that has been stored in local cache (as well as the cache itself)
     * This method should only be called on system shutdown to remove any 
     * temporary files that have been created
     */    
    public void clearCache() {
        File dir=new File(cacheDirectoryFullPath);
        File[] files=dir.listFiles();
        for (File file:files) {
           if (!file.delete()) System.err.println("WARNING: unable to delete UNDO file: "+file.getAbsolutePath());  
        }
        if (!dir.delete()) System.err.println("WARNING: unable to delete UNDO directory!");
        if (usedCacheKeys!=null) usedCacheKeys.clear();
     }
    
    
    /**
     * Clears a cached object associated with the given key
     */
    public void clearCachedObject(String key) {
        //gui.debugMessage("Clearing cached object: "+key);
        if (usedCacheKeys==null) return;
        if (virtualMemoryStorage!=null && virtualMemoryStorage.containsKey(key)) virtualMemoryStorage.remove(key);
        else {
           //try {
//              LocalStorage localstorage=context.getLocalStorage();
//              String filename=cacheDirectoryName+File.separator+key+".ser";
//              if (!localstorage.deleteFile(filename)) gui.debugMessage("Unable to delete cached file '"+filename+"'");
                // the above strategy fails for web start applications
               File file=new File(cacheDirectoryFullPath+File.separator+key+".ser");
               if (!file.delete()) gui.debugMessage("Unable to delete cached file '"+file.getAbsolutePath()+"'");
           //} catch (IOException e) {gui.debugMessage("ERROR in 'undo.clearCachedObject("+key+")': "+e.toString());}
        }
        usedCacheKeys.remove(key);
    }
    
    /** Returns the next significant edit to be undone if undo is invoked. It is not removed from the queue */
    public UndoableEdit peek() {
        return editToBeUndone();
    }
    
    /** Removes the last UndoableEdit that was added to the queue and returns it */
    public UndoableEdit pop() {
        int size=edits.size();
        UndoableEdit lastedit=edits.elementAt(size-1);
        edits.removeElementAt(size-1);
        gui.updateUndoRedoStates();
        gui.getFrame().repaint(); // This seems necessary but it shouldn't be
        return lastedit;
    }
    
    @Override
    public synchronized void undo() throws CannotUndoException {
        super.undo();
        gui.updateUndoRedoStates();
        gui.getFrame().repaint(); // This seems necessary but it shouldn't be
    }
    
    @Override
    public synchronized void redo() throws CannotUndoException {
        super.redo();
        gui.updateUndoRedoStates();
        gui.getFrame().repaint(); // This seems necessary but it shouldn't be
    }
    
    /** Returns an UndoMonitor that can be used to track changes 
     *  The monitor has already been registered with this UndoManager
     */
    public UndoMonitor getUndoMonitor(String presentationName) {
        return new UndoMonitor(this, presentationName);
    }
    
    /**
     * Returns true if the specified edit is already added to the undo-queue
     * @param edit
     * @return
     */
    public boolean isInUndoQueue(UndoableEdit edit) {
        return edits.contains(edit); // "edits" is the undo-queue
    }
    
    
    /** */
    private class ExtendedUndoableEditSupport extends UndoableEditSupport { // I honestly am not sure what this class does...
        public synchronized void postEdit(UndoableEdit edit) {
            if (source!=null) realSource=source;
            super.postEdit(edit);
        }
    }
    
    
    /** Deletes the given file or directory (and subdirectories) */
    private boolean deleteTempFile(File tempdir) {
        if (tempdir.isDirectory()) {
            boolean ok=true;
            File[] files=tempdir.listFiles();
            for (File file:files) {
                if (file.isDirectory()) ok = ok && deleteTempFile(file);
                else ok = ok && file.delete();
            }
            ok = ok && tempdir.delete();
            return ok;
        } else return tempdir.delete();
    }    
    
}
