/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import javax.swing.JTable;
import priorseditor.engine.data.Region;
import javax.swing.table.DefaultTableCellRenderer;
import priorseditor.engine.data.Motif;

/**
 * This class implements functionality to draw sequence logos of motifs.
 * It implements the TableCellRenderer interface in order to directly incorporate
 * the Sequence logo into JTables.
 * If the class is used as a TableCellRenderer, the motif to paint is taken from
 * the table (which must contain instances of Motif objects). Also, if the table
 * contains a column describing the motif orientation it is possible to link to
 * this column and paint the motif in the orientation described.
 * If the class is used outside tables to paint single Sequence logos, the motif
 * (and orientation to use) must be set directly.
 * @author kjetikl
 */
public class SequenceLogo extends DefaultTableCellRenderer {
    private Font logofont;
    private char[] bases=new char[]{'A','C','G','T'};
    private int fontheight=25; // if you change this size you must also change the ascentCorrection below accordingly (but I am not sure about the relationship...)
    private int ascentCorrection=6; // this is a needed hack because the ascent returned by the FontMetrics is not really the same height as that of a capital letter (even though that was stated in the documentation)
    private Color[] basecolors;
    private Motif motif=null;
    private Color background=Color.WHITE;
    private Color selectedbackground=Color.WHITE;
    private boolean antialias=true;
    private JTable table=null;
    private int strandcolumn=-1; // use this column to decide the strand orientation if motifs are within a JTable
    private int strand=Region.DIRECT; // use this column for strand orientation if painting single motif
    private int currentrow=-1;
    private boolean rowselected=false;
    private int maxwidth=0;


    public SequenceLogo(Color[] basecolors, int fontheight, int ascentcorrection, Color background) {
        this.fontheight=fontheight;
        this.ascentCorrection=ascentcorrection;
        logofont=new Font(Font.SANS_SERIF,Font.BOLD,fontheight);
        this.basecolors=basecolors;
        this.background=background;
    }

    public SequenceLogo(Color[] basecolors) {
        logofont=new Font(Font.SANS_SERIF,Font.BOLD,fontheight);
        this.basecolors=basecolors;
    }


    public void setFontHeight(int fontheight, int ascentcorrection) {
        this.fontheight=fontheight;
        this.ascentCorrection=ascentcorrection;
        logofont=new Font(Font.SANS_SERIF,Font.BOLD,fontheight);
    }

    public void setStrandColumn(JTable table, int column) {
        this.table=table;
        this.strandcolumn=column;
    }

    /** */
    public void setStrand(int strand) {
        this.strand=strand;
    }

    public void setMotif(Motif motif) {
        this.motif=motif;
    }

    /** Sets a maximum width for the logo (this only applies outside)
     *  Set to 0 or negative number to allow infinite width
     */
    public void setMaxWidth(int max) {
        this.maxwidth=max;
    }



    /** Specifies whether or not the base letters should be drawn in antialias mode
     *  (antialiasing gives nicer graphics but might be slower)
     */
    public void setUseAntialias(boolean useantialias) {
        antialias=useantialias;
    }

   /** Specifies whether or not the base letters should be drawn in antialias mode
     *  (antialiasing gives nicer graphics but might be slower)
     */
    public void setSelectedBackgroundColor(Color color) {
        selectedbackground=color;
    }


    @Override
    protected void paintComponent(Graphics graphics) {
        int xoffset=0; int yoffset=0;
        if (motif==null) return;
        float[][] matrix=motif.getMatrix();
        if (motif.getMatrixType()==Motif.LOG_ODDS_MATRIX) matrix=motif.getMatrixAsFrequencyMatrix();
        if (table!=null && strandcolumn>=0) {
            Integer usestrand=(Integer)table.getValueAt(currentrow, strandcolumn);
            if (usestrand==Region.REVERSE) matrix=Motif.reverseComplementMatrix(matrix);
        } else {
            if (strand==Region.REVERSE) matrix=Motif.reverseComplementMatrix(matrix);
        }
        float[] letterXoffset=new float[4];
        if (matrix==null) return;
        Graphics2D g = (Graphics2D)graphics;
        FontMetrics metrics=g.getFontMetrics(logofont);
        int ascent=metrics.getAscent()-ascentCorrection;
        int widthA=metrics.charWidth('A');
        int widthC=metrics.charWidth('C');
        int widthG=metrics.charWidth('G');
        int widthT=metrics.charWidth('T');
        letterXoffset[0]=(widthG-widthA)/2;
        letterXoffset[1]=(widthG-widthC)/2;
        letterXoffset[3]=(widthG-widthT)/2;
        int logowidth=matrix.length*widthG;
        if (antialias) g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g.setFont(logofont);
        g.setColor(java.awt.Color.RED);
        AffineTransform restore=g.getTransform();
        AffineTransform save=g.getTransform();
        float width=getWidth();
        if (maxwidth>0 && (width>maxwidth || width==0)) width=maxwidth;
        float scaleX=(logowidth>width && width>0)?(width/(float)logowidth):1f;
        save.scale(scaleX, 1); // scale X-direction so that logo fits irrespective of size
        AffineTransform scaleY = new AffineTransform();
        if (rowselected) {
            if (selectedbackground!=null) {
               g.setColor(selectedbackground);
               g.fillRect(xoffset, yoffset, logowidth, ascent);
            }
            else if (background!=null) {
               g.setColor(background);
               g.fillRect(xoffset, yoffset, logowidth, ascent);
            }
        } else {
            if (background!=null) {
               g.setColor(background);
               g.fillRect(xoffset, yoffset, logowidth, ascent);
            }
        }
        float xpos=xoffset;
        for (int i=0;i<matrix.length;i++) {
            float[] counts=new float[]{matrix[i][0],matrix[i][1],matrix[i][2],matrix[i][3]};
            float total=counts[0]+counts[1]+counts[2]+counts[3];
            float ic=Motif.calculateColumnIC(counts[0], counts[1], counts[2], counts[3], false);
            int[] sorted=new int[4]; // sorted in ascending order. Values are base-indices (i.e. 0=>A, 1=>C, 2=>G, 3=>T)
            int indexA=0, indexC=0, indexG=0, indexT=0;
            if (counts[0]>=counts[1]) indexA++;
            if (counts[0]>=counts[2]) indexA++;
            if (counts[0]>=counts[3]) indexA++;
            if (counts[1]>counts[0]) indexC++;
            if (counts[1]>=counts[2]) indexC++;
            if (counts[1]>=counts[3]) indexC++;
            if (counts[2]>counts[0]) indexG++;
            if (counts[2]>counts[1]) indexG++;
            if (counts[2]>=counts[3]) indexG++;
            if (counts[3]>counts[0]) indexT++;
            if (counts[3]>counts[1]) indexT++;
            if (counts[3]>counts[2]) indexT++;
            sorted[indexA]=0;
            sorted[indexC]=1;
            sorted[indexG]=2;
            sorted[indexT]=3;
            float currentYoffset=yoffset+ascent-(ascent*ic/2f);
            for (int j=3;j>=0;j--) { // draws letters from top to bottom (most frequent first)
                int base=sorted[j];
                float fraction=counts[base]/total;
                scaleY.setTransform(save);
                scaleY.translate(letterXoffset[base],currentYoffset); // translated in 1-to-1 scale
                scaleY.scale(1,ic/2f); // scale by IC-content in position
                scaleY.scale(1, fraction);
                g.setColor(basecolors[base]);
                g.setTransform(scaleY);
                g.drawString(""+bases[base], xpos, ascent); // draw all letters at same position and use transform to place them correctly
                currentYoffset+=ascent*fraction*(ic/2f);
            } // end for each base
            xpos+=widthG; //
        } // end for each position
        g.setColor(Color.BLACK);
        g.setTransform(restore);
        g.drawRect(xoffset,yoffset,(int)(scaleX*logowidth)-1,ascent);
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
    }


    @Override
    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
        currentrow=row;
        rowselected=isSelected;
        if (value!=null && value instanceof Motif) motif=(Motif)value;
        else motif=null;
        return this;
    }

    /** Returns the expected width for this motif */
    public int getDefaultMotifWidth() {
        FontMetrics metrics=getFontMetrics(logofont);
        int widthG=metrics.charWidth('G');
        if (motif==null) return 0;
        int logowidth=motif.getLength()*widthG;
        if (maxwidth>0 && logowidth>maxwidth) return maxwidth;
        else return logowidth;
    }

    /** Paints the sequence logo into the given Graphics object */
    public void paintLogo(Graphics g) {
         paintComponent(g);
    }
}
