/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package priorseditor.gui;

import java.util.ArrayList;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import priorseditor.engine.PriorsEditorEngine;
import priorseditor.engine.data.Data;
import java.util.HashMap;
import java.util.Set;
import javax.swing.undo.UndoableEdit;
import priorseditor.engine.DataListener;
import priorseditor.engine.data.MotifCollection;
import priorseditor.engine.data.SequenceCollection;

/**
 * UndoMonitors are used by the GUI to monitor executions of operations and scripts.
 * They register as DataListeners for the duration of a task execution and keep
 * track of which data items that change and how (if they are updated or created anew)
 * For data items that update during execution the UndoMonitor is responsible for caching 
 * the original data as it was before the execution started so that all the changes that has
 * been done to it can later be undone (since only the original state is stored there is no
 * need for step-wise rollback to undo multiple operations that have affected the same data).
 * The UndoMonitors should also cache the resulting data after successful execution to facilitate 
 * efficient REDO functionality after UNDO (rather than executing the whole script again).
 * 
 * The UNDO/REDO information is stored in two HashMaps. Any object that is affected during
 * execution should be registered by name in these maps (at least in the originalData map).
 * 
 * 
 * @author kjetikl
 */
public class UndoMonitor implements DataListener, UndoableEdit {
    private GUIUndoManager undoManager=null; // provides functionality for caching UNDO-data
    
    private PriorsEditorEngine engine=null;
    private HashMap<String,UndoInformation> originalData=new HashMap<String, UndoInformation>(); // stores UNDO information
    private HashMap<String,UndoInformation> finalData=new HashMap<String, UndoInformation>();  // stores REDO information
    private boolean redoInformationStored=false;
    private boolean isDone=false;
    private String presentationName;

    
    /**
     * Creates a new UndoManager
     */        
    public UndoMonitor(GUIUndoManager undoManager, String presentationName) {
        this.undoManager=undoManager;
        this.presentationName=presentationName;
        this.engine=undoManager.getEngine();
    }
    
    /**
     * Registers the UndoMonitor with the engine. Once the monitor is registered
     * it will listen for data updates and track information needed to undo 
     * all changes until it is deregistered.
     */
    public void register() {
        //engine.logMessage("UndoMonitor Registered => "+presentationName);
        cleanup();        
        engine.addDataListener(this);
    }
    
    /**
     * Deregisters the UndoMonitor with the engine. Once the monitor is deregistered,
     * it will no longer listen for data updates. If the execution of the operation/script
     * was a success, the boolean argument should be set to <tt>true</tt> to signal that
     * the final state resulting from the operation should be stored to facilitate efficient
     * REDO functionality if the user ever chooses to UNDO and then REDO (caching this data
     * will perhaps be more efficient than executing the whole script anew). 
     * However, if the execution was aborted by the user or otherwise stopped because of 
     * unforseen errors there will be an automatic rollback. In this case there will be
     * no reason to REDO the same execution again; the final state needs not to be 
     * stored and the argument should be set to <tt>false</tt>
     * @param storeFinalState set this to <tt>true</tt> if the execution of the script was
     * successful and the final state should be cached to facilitate speedy REDO functionality later.
     */
    public void deregister(boolean storeFinalState) {
        //System.err.println("UndoMonitor DEregistered => "+presentationName+"  EDT="+SwingUtilities.isEventDispatchThread());
        engine.removeDataListener(this);
        isDone=true;
        if (storeFinalState) {
            // go through the originalData HashMap to see which data items have been updated, then store the final results of these
            Set<String> updatedDataitems = originalData.keySet();
            for (String name:updatedDataitems) {
               //engine.logMessage("Deregister: Storing REDO info about:" +name);
               Data data=engine.getDataItem(name);
               if (data==null) { // data has been removed?
                   finalData.put(name, null);     
               } else {
                   String undoID=undoManager.getUniqueUndoID();
                   if (!undoManager.storeObjectInCache(undoID, data.getSerializableContent())) engine.logMessage("Problems occurred while caching data for " +name);;
                   finalData.put(name, new UndoInformation(data, undoID));  
                   //engine.logMessage("Final data cached to disc for " +name+" undoID="+undoID);
               }
            }
            redoInformationStored=true;
        }
    }
      
    /**
     * Undos all the data changes traced by this UndoMonitor while it was registered with the engine
     * This method is used by the undo() method to reset changes and should only be called directly 
     * to immediately do rollback after abnormal termination of unsuccessful tasks (for instance due 
     * to errors/exceptions during execution or abortion by the user) in which case the normal undo/redo
     * functions are disabled
     */
    public void rollback() {
        //System.err.println("- Rollback -");
        Set<String> updatedDataitems = originalData.keySet();
        ArrayList<Data> delayedRecovery = new ArrayList<Data>();        
        for (String name:updatedDataitems) {
            //engine.logMessage("RB: Updating " +name);
            UndoInformation undoInfo=originalData.get(name);
             if (undoInfo==null) { // the data item was created during execution. Just remove it and all will be as it was before
                 //engine.logMessage("RB[" +name+"] UndoInfo == null. Data item was added");
                 engine.removeDataItem(name); 
             } else {
                 Data updatedItem=engine.getDataItem(name);
                 //engine.logMessage("RB[" + name + "] UndoInfo != null. Data item was updated");
                 Data storedItem=null;
                 String undoID=undoInfo.undoTicket;
                 if (undoID==null) storedItem=undoInfo.dataitem;
                 else storedItem=(Data)undoManager.getObjectFromCache(undoID);  
                 //engine.logMessage("rollback data item:"+storedItem.toString());
                 if (storedItem instanceof MotifCollection || storedItem instanceof SequenceCollection) delayedRecovery.add(storedItem);
                 else try {engine.updateDataItem(storedItem);} catch (Exception e) {} 
             }
        }        
        for (Data item:delayedRecovery) try {engine.updateDataItem(item);} catch (Exception e) {}
    }
 
   /**
     * Redos changes that has been undone by reestablishing the state of data as it
     * was at the point when the monitor was deregistered
     */
    private void rollforward() {
        Set<String> updatedDataitems = finalData.keySet();
        ArrayList<Data> delayedRecovery = new ArrayList<Data>();
        for (String name:updatedDataitems) {
            //engine.logMessage("RF: Updating " +name);
            UndoInformation undoInfo=finalData.get(name);
             if (undoInfo==null) { // the data item was removed during execution. So remove it again for redo
                 //engine.logMessage("RF[" +name+"] Undo info == null");
                 engine.removeDataItem(name); 
             } else {
                 //engine.logMessage("RF[" +name+"] Undo info != null");
                 Data storedItem=null;
                 String undoID=undoInfo.undoTicket;
                 //engine.logMessage("undoID="+undoID);
                 if (undoID==null) storedItem=undoInfo.dataitem;
                 else storedItem=(Data)undoManager.getObjectFromCache(undoID);  
                 //engine.logMessage("storedItem="+storedItem);
                 //engine.logMessage("rollforward data item:"+storedItem.toString());
                 if (storedItem instanceof MotifCollection || storedItem instanceof SequenceCollection) delayedRecovery.add(storedItem);
                 else try {engine.updateDataItem(storedItem);} catch (Exception e) {}  
             }
        }
        for (Data item:delayedRecovery) try {engine.updateDataItem(item);} catch (Exception e) {} 
    }
 
    
    
    
    
    @Override
    public void dataAdded(Data data) {
        // System.err.println("dataAdded!! data="+data.getName());
        if (!originalData.containsKey(data.getName())) originalData.put(data.getName(), null); // no need to store the original data, so we just put in NULL to signal this
    }

    @Override
    public void dataRemoved(Data data) {
        //System.err.println("dataRemoved!! data="+data.getName());
        if (originalData.containsKey(data.getName())) return; // the original data is already stored for this object. Further notificiations of updates can be ignored                                                                
        else {
           //System.err.println("   dataRemoved!! stored data!");
           String undoID=undoManager.getUniqueUndoID();
           undoManager.storeObjectInCache(undoID, data.getSerializableContent());
           originalData.put(data.getName(), new UndoInformation(data, undoID));
        }
    }    

    // this method is invoked by the engine just before a data value is updated/overwritten in memory (or before new data items are added)
    @Override
    public void dataUpdate(Data oldvalue, Data newvalue) {
//        System.err.println("dataUpdate!! data="+oldvalue.getName());
//        for (String key:originalData.keySet()) {System.err.println("  ->"+key);}
        if (originalData.containsKey(oldvalue.getName())) return; // the original data is already stored for this object. Further notificiations of updates can be ignored                                                                
        else {
           // engine.logMessage("data update detected for "+oldvalue.getName());
           String undoID=undoManager.getUniqueUndoID();
           undoManager.storeObjectInCache(undoID, oldvalue.getSerializableContent());
           originalData.put(oldvalue.getName(), new UndoInformation(oldvalue, undoID));
        }
    }
    
    
    @Override
    public void dataUpdated(Data data) { } // Not used. We catch updates with dataUpdate(old,new) instead to get access to the original value
    @Override
    public void dataAddedToSet(Data parentDataset, Data child) { }
    @Override
    public void dataRemovedFromSet(Data parentDataset, Data child) { }

    @Override
    public boolean addEdit(UndoableEdit anEdit) { return false; }
    @Override
    public boolean replaceEdit(UndoableEdit anEdit) {return false;}
       
    @Override
    public boolean canRedo() {
        return redoInformationStored && !isDone;
    }
    
    @Override
    public boolean canUndo() {
        // return isDone;
        return true;
    }
    
    @Override
    public void die() { // performs cleanup. This can be useful to do to recycle resources if objects have been cached in memory rather than local disc
        //engine.logMessage("Undo DIE: "+presentationName);
        cleanup();
    }
    
    /**
     * Clears all undo/redo information registered by this UndoMonitor 
     */
    private void cleanup() {
        Set<String> updatedDataitems = originalData.keySet();
        for (String name:updatedDataitems) {
            UndoInformation undoInfo=originalData.get(name);
            if (undoInfo==null) continue;
            String undoID=undoInfo.undoTicket;
            if (undoID!=null) undoManager.clearCachedObject(undoID);
        }
        originalData.clear();
        updatedDataitems = finalData.keySet();
        for (String name:updatedDataitems) {
            UndoInformation undoInfo=finalData.get(name);
            if (undoInfo==null) continue;
            String undoID=undoInfo.undoTicket;
            if (undoID!=null) undoManager.clearCachedObject(undoID);
        }
        finalData.clear();        
    }
    
    
    @Override
    public String getPresentationName() {
        return presentationName;
    }
    
    @Override
    public String getRedoPresentationName() {
        return "Redo "+presentationName;
    }
    
    @Override
    public String getUndoPresentationName() {
        return "Undo "+presentationName;
    }

    @Override
    public boolean isSignificant() { return true; }
    

    
 
    @Override
    public void undo() throws CannotUndoException {
        //if (!undoInformationStored || !isDone) throw new CannotUndoException();
        //else {
            //engine.logMessage("Performing rollback");
            rollback();
            isDone=false;
       // }
    }
    
    @Override
    public void redo() throws CannotRedoException {
      if (!redoInformationStored || isDone)  throw new CannotRedoException();
        else {
            //engine.logMessage("Performing rollforward");
            rollforward();
            isDone=true;
        }   
    }    
    
    
    
    
    /**
     * This private class is just an information wrapper that is inserted into the HashMaps of the outer class
     * If the undoTicket is NULL the whole original dataobject will be stored in this wrapper, else
     * the caching facilities in GUIUndoManager will be used to cache the data objects to disc (in which
     * case the Class of the original dataitem is stored)
     */     
    private class UndoInformation {
        Class thisclass=null;
        String undoTicket=null;
        Data dataitem=null;
        
        public UndoInformation(Data dataitem, String undoTicket) {
            this.undoTicket=undoTicket;
            if (undoTicket==null) {this.dataitem=dataitem;}
            else {thisclass=dataitem.getClass();}
        }
    }
}
