/*
 * VisualizationPanel.java
 *
 * Created on 7. oktober 2008, 14:43
 */

package priorseditor.gui;



import java.awt.Graphics;
import java.awt.print.PageFormat;
import javax.swing.BorderFactory;
import javax.swing.JPanel;
import java.awt.Dimension;
import javax.swing.BoxLayout;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import javax.swing.JLabel;
import javax.swing.Action;
import javax.swing.JOptionPane;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.KeyStroke;
import javax.swing.InputMap;
import javax.swing.ActionMap;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.print.Printable;
import java.awt.print.PrinterAbortException;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import javax.print.PrintService;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;
import javax.swing.RepaintManager;
import priorseditor.engine.data.SequenceCollection;
import priorseditor.engine.data.Sequence;
import priorseditor.engine.data.Data;
import priorseditor.engine.data.FeatureDataset;
import sun.swing.PrintingStatus;
import sun.swing.SwingUtilities2;

/**
 * This class represents the main panel shown in the "Visualization" tab in the GUI.
 * It is responsible for visualizing all the sequences and datatracks (through division of labour
 * to classes such as SequenceVisualizer which draws all chosen tracks for a single sequence).
 * 
 * @author  kjetikl
 */
public class VisualizationPanel extends javax.swing.JScrollPane implements VisualizationSettingsListener, Printable {
    private Action zoomInAction;
    private Action zoomOutAction;
    private Action zoomResetAction;
    private Action zoomToFitAction;
    private Action zoomToCustomAction;
    private Action moveLeftAction;
    private Action moveToLeftEndAction;
    private Action moveRightAction;
    private Action moveToRightEndAction;
    private Action flipOrientationAction;
    private Action directOrientationAction;
    private Action reverseOrientationAction;
    private Action fromGeneOrientationAction;
    private Action alignRightAction;
    private Action alignLeftAction;
    private Action alignNoneAction;
    private Action alignTSSHereAction;
    private Action alignTSSRightAction;
    private Action constrainOnAction;
    private Action constrainOffAction;
    private Action scrollLockOnAction;
    private Action scrollLockOffAction;

    public static final String ACTION_SCROLL_LOCK_ON = "Turn On Scroll Lock";
    public static final String ACTION_SCROLL_LOCK_OFF = "Turn Off Scroll Lock";
    public static final String ACTION_CONSTRAIN_ON = "Turn On Movement Constraint";
    public static final String ACTION_CONSTRAIN_OFF = "Turn Off Movement Constraint";
    
    
    private VisualizationSettings settings;
    private JPanel mainVisualizationPanel; // an inner panel that will be placed inside a JScrollPanel. This is where all the stuff goes!
    private ArrayList<SequenceVisualizer> sequenceVisualizers;   
     
    /** 
     * Creates new VisualizationPanel object
     * @param settings A reference to the singleton VisualizationSettings object
     */
    public VisualizationPanel(VisualizationSettings settings) {
        this.settings=settings;
        
        mainVisualizationPanel=new JPanel();
        mainVisualizationPanel.setLayout(new BoxLayout(mainVisualizationPanel, BoxLayout.Y_AXIS));
        //mainVisualizationPanel.setMaximumSize(new Dimension(Integer.MAX_VALUE,Integer.MAX_VALUE)); 
        mainVisualizationPanel.setBackground(settings.getVisualizationPanelBackgroundColor());
        mainVisualizationPanel.setOpaque(true);    
        settings.addVisualizationSettingsListener(this);
        this.setViewportView(mainVisualizationPanel);
        sequenceVisualizers=new ArrayList<SequenceVisualizer>(0);
        setupMainVisualizationPanel(); 
        setupActions();
        this.setFocusCycleRoot(true);
    }
    
    public JPanel getMainVisualizationPanel() {
        return mainVisualizationPanel;
    }

    
    private void setupActions() {
        zoomInAction=new AbstractAction(SequenceVisualizer.ACTION_ZOOM_IN, new InfoPanelIcon(InfoPanelIcon.ZOOM_IN_ICON)) { public void actionPerformed(ActionEvent e) {zoomInOnAllSequences();} };
        zoomOutAction=new AbstractAction(SequenceVisualizer.ACTION_ZOOM_OUT, new InfoPanelIcon(InfoPanelIcon.ZOOM_OUT_ICON)) { public void actionPerformed(ActionEvent e) {zoomOutOnAllSequences();} };
        zoomResetAction=new AbstractAction(SequenceVisualizer.ACTION_ZOOM_RESET, new InfoPanelIcon(InfoPanelIcon.ZOOM_RESET_ICON)) { public void actionPerformed(ActionEvent e) {zoomResetOnAllSequences();} };
        zoomToFitAction=new AbstractAction(SequenceVisualizer.ACTION_ZOOM_TO_FIT, new InfoPanelIcon(InfoPanelIcon.ZOOM_TO_FIT_ICON)) { public void actionPerformed(ActionEvent e) {zoomToFitOnAllSequences();} };
        zoomToCustomAction=new AbstractAction(SequenceVisualizer.ACTION_ZOOM_TO_CUSTOM_LEVEL, new InfoPanelIcon(InfoPanelIcon.ZOOM_ICON)) { public void actionPerformed(ActionEvent e) {zoomToCustomOnAllSequences();} };
        
        moveLeftAction=new AbstractAction(SequenceVisualizer.ACTION_MOVE_LEFT, new InfoPanelIcon(InfoPanelIcon.LEFT_ICON)) { public void actionPerformed(ActionEvent e) {moveAllSequencesLeft(SequenceVisualizer.moveIncrement);} };
        moveRightAction=new AbstractAction(SequenceVisualizer.ACTION_MOVE_RIGHT, new InfoPanelIcon(InfoPanelIcon.RIGHT_ICON)) { public void actionPerformed(ActionEvent e) {moveAllSequencesRight(SequenceVisualizer.moveIncrement);} };
        moveToLeftEndAction=new AbstractAction(SequenceVisualizer.ACTION_MOVE_TO_LEFT_END, new InfoPanelIcon(InfoPanelIcon.LEFT_END_ICON)) { public void actionPerformed(ActionEvent e) {moveAllSequencesToLeftEnd();} };
        moveToRightEndAction=new AbstractAction(SequenceVisualizer.ACTION_MOVE_TO_RIGHT_END, new InfoPanelIcon(InfoPanelIcon.RIGHT_END_ICON)) { public void actionPerformed(ActionEvent e) {moveAllSequencesToRightEnd();} };
        
        flipOrientationAction=new AbstractAction(SequenceVisualizer.ACTION_ORIENTATION_FLIP, new InfoPanelIcon(InfoPanelIcon.FLIP_ICON)) { public void actionPerformed(ActionEvent e) {flipOrientationOnAllSequences();} };
        directOrientationAction=new AbstractAction(SequenceVisualizer.ACTION_ORIENTATION_DIRECT, null) { public void actionPerformed(ActionEvent e) {showDirectStrandOnAllSequences();} };
        reverseOrientationAction=new AbstractAction(SequenceVisualizer.ACTION_ORIENTATION_REVERSE, null) { public void actionPerformed(ActionEvent e) {showReverseStrandOnAllSequences();} };
        fromGeneOrientationAction=new AbstractAction(SequenceVisualizer.ACTION_ORIENTATION_FROM_GENE, null) { public void actionPerformed(ActionEvent e) {showStrandFromGeneOnAllSequences();} };

        alignLeftAction=new AbstractAction(SequenceVisualizer.ACTION_ALIGN_LEFT, new InfoPanelIcon(InfoPanelIcon.ALIGN_LEFT_ICON)) { public void actionPerformed(ActionEvent e) {alignAllSequencesLeft();} };
        alignRightAction=new AbstractAction(SequenceVisualizer.ACTION_ALIGN_RIGHT, new InfoPanelIcon(InfoPanelIcon.ALIGN_RIGHT_ICON)) { public void actionPerformed(ActionEvent e) {alignAllSequencesRight();} };
        alignTSSRightAction=new AbstractAction(SequenceVisualizer.ACTION_ALIGN_TSS_RIGHT, new InfoPanelIcon(InfoPanelIcon.ALIGN_TSS_ICON)) { public void actionPerformed(ActionEvent e) {alignAllSequencesToTSSRight();} };
        alignNoneAction=new AbstractAction(SequenceVisualizer.ACTION_ALIGN_NONE, new InfoPanelIcon(InfoPanelIcon.ALIGN_NONE_ICON)) { public void actionPerformed(ActionEvent e) {alignAllSequencesNone();} };
        
        
        constrainOnAction=new AbstractAction(ACTION_CONSTRAIN_ON, new InfoPanelIcon(InfoPanelIcon.CONSTRAIN_ICON)) { 
            public void actionPerformed(ActionEvent e) { setConstrainOnAllSequences(true);} 
        };
        constrainOffAction=new AbstractAction(ACTION_CONSTRAIN_OFF, new InfoPanelIcon(InfoPanelIcon.CONSTRAIN_ICON)) { 
            public void actionPerformed(ActionEvent e) { setConstrainOnAllSequences(false);} 
        };
        scrollLockOnAction=new AbstractAction(ACTION_SCROLL_LOCK_ON, new InfoPanelIcon(InfoPanelIcon.LOCK_ICON)) { 
            public void actionPerformed(ActionEvent e) { setScrollLockOnAllSequences(true);} 
        };
        scrollLockOffAction=new AbstractAction(ACTION_SCROLL_LOCK_OFF, new InfoPanelIcon(InfoPanelIcon.LOCK_ICON)) { 
            public void actionPerformed(ActionEvent e) { setScrollLockOnAllSequences(false);} 
        };
       

        
        
        InputMap inputmap=this.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        ActionMap actionmap=this.getActionMap();
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PLUS, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_IN);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ADD, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_IN);
        actionmap.put(SequenceVisualizer.ACTION_ZOOM_IN, zoomInAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_MINUS, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_OUT);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_SUBTRACT, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_OUT);
        actionmap.put(SequenceVisualizer.ACTION_ZOOM_OUT, zoomOutAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_RESET);
        actionmap.put(SequenceVisualizer.ACTION_ZOOM_RESET, zoomResetAction);
          
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_0, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_TO_FIT);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD0, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_TO_FIT);
        actionmap.put(SequenceVisualizer.ACTION_ZOOM_TO_FIT, zoomToFitAction);
                        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_Z, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ZOOM_TO_CUSTOM_LEVEL);
        actionmap.put(SequenceVisualizer.ACTION_ZOOM_TO_CUSTOM_LEVEL, zoomToCustomAction);
        
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_MOVE_LEFT);
        actionmap.put(SequenceVisualizer.ACTION_MOVE_LEFT, moveLeftAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_MOVE_RIGHT);
        actionmap.put(SequenceVisualizer.ACTION_MOVE_RIGHT, moveRightAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, InputEvent.ALT_DOWN_MASK+InputEvent.SHIFT_DOWN_MASK), SequenceVisualizer.ACTION_MOVE_TO_LEFT_END);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_L, 0), SequenceVisualizer.ACTION_MOVE_TO_LEFT_END);
        actionmap.put(SequenceVisualizer.ACTION_MOVE_TO_LEFT_END, moveToLeftEndAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, InputEvent.ALT_DOWN_MASK+InputEvent.SHIFT_DOWN_MASK), SequenceVisualizer.ACTION_MOVE_TO_RIGHT_END);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R, 0), SequenceVisualizer.ACTION_MOVE_TO_RIGHT_END);
        actionmap.put(SequenceVisualizer.ACTION_MOVE_TO_RIGHT_END, moveToRightEndAction);
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_SLASH, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_FLIP);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_DIVIDE, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_FLIP);
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_FLIP);
        actionmap.put(SequenceVisualizer.ACTION_ORIENTATION_FLIP, flipOrientationAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_DIRECT);
        actionmap.put(SequenceVisualizer.ACTION_ORIENTATION_DIRECT, directOrientationAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_REVERSE);
        actionmap.put(SequenceVisualizer.ACTION_ORIENTATION_REVERSE, reverseOrientationAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME, InputEvent.ALT_DOWN_MASK), SequenceVisualizer.ACTION_ORIENTATION_FROM_GENE);
        actionmap.put(SequenceVisualizer.ACTION_ORIENTATION_FROM_GENE, fromGeneOrientationAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_C,InputEvent.ALT_DOWN_MASK), ACTION_CONSTRAIN_ON);
        actionmap.put(ACTION_CONSTRAIN_ON, constrainOnAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_C,InputEvent.ALT_DOWN_MASK+InputEvent.SHIFT_DOWN_MASK), ACTION_CONSTRAIN_OFF);
        actionmap.put(ACTION_CONSTRAIN_OFF, constrainOffAction);

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.ALT_DOWN_MASK), ACTION_SCROLL_LOCK_ON);
        actionmap.put(ACTION_SCROLL_LOCK_ON, scrollLockOnAction); 
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.ALT_DOWN_MASK+InputEvent.SHIFT_DOWN_MASK), ACTION_SCROLL_LOCK_OFF);
        actionmap.put(ACTION_SCROLL_LOCK_OFF, scrollLockOffAction);  
        
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_L, 0), SequenceVisualizer.ACTION_ALIGN_LEFT);        
        actionmap.put(SequenceVisualizer.ACTION_ALIGN_LEFT, alignLeftAction);  
        
        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R, 0), SequenceVisualizer.ACTION_ALIGN_RIGHT); 
        actionmap.put(SequenceVisualizer.ACTION_ALIGN_RIGHT, alignRightAction);  

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_N, 0), SequenceVisualizer.ACTION_ALIGN_NONE); 
        actionmap.put(SequenceVisualizer.ACTION_ALIGN_NONE, alignNoneAction);  

        inputmap.put(KeyStroke.getKeyStroke(KeyEvent.VK_T, 0), SequenceVisualizer.ACTION_ALIGN_TSS_RIGHT);         
        actionmap.put(SequenceVisualizer.ACTION_ALIGN_TSS_RIGHT, alignTSSRightAction);  
        
        actionmap.put(SequenceVisualizer.ACTION_ALIGN_TSS_HERE, alignTSSHereAction);          
        
    }  
    
    /**
     * Sets up the main visualizationpanel. 
     * This should be recalled every time a new DNASequenceDataset (or single sequence) 
     * is added, removed or reordered in order to update the visualization
     */
    @SuppressWarnings("unchecked")
    private void setupMainVisualizationPanel() {
        ArrayList<SequenceVisualizer> tempList=(ArrayList<SequenceVisualizer>)sequenceVisualizers.clone();
        sequenceVisualizers=new ArrayList<SequenceVisualizer>();
        mainVisualizationPanel.removeAll();
        int maxlabelwidth=0;
        SequenceCollection dataset=settings.getEngine().getDefaultSequenceCollection();
        int size=dataset.getNumberofSequences();
        //settings.getGUI().logMessage("setupMainVisualizationPanel(). Size="+size);
        for (int i=0;i<size;i++) {
                //settings.getGUI().logMessage("loop i="+i+", size="+size);
                Sequence sequence=(Sequence)dataset.getSequenceByIndex(i,settings.getEngine());
                String sequenceName=sequence.getName();
                SequenceVisualizer viz=getSequenceVisualizerFromList(sequenceName,tempList);
                if (viz==null) {
                    viz=new SequenceVisualizer(sequenceName,settings);                
                    viz.alignSequence(settings.getDefaultSequenceAlignment(),false);
                } 
                int labelwidth=viz.getRenderedLabelWidth();
                if (labelwidth>maxlabelwidth) maxlabelwidth=labelwidth;
                sequenceVisualizers.add(viz);
                if (settings.isSequenceVisible(sequenceName)) mainVisualizationPanel.add(viz);
                //settings.getGUI().logMessage("Adding "+sequenceName+" to panel");
        }
        settings.setSequenceLabelMaxWidth(maxlabelwidth);
        mainVisualizationPanel.setBorder(BorderFactory.createEmptyBorder(20, 0, 20, 0));
        //settings.getGUI().logMessage("sequenceVisualizers size="+sequenceVisualizers.size());
    }



    /** 
     * Searches a given list of SequenceVisualizers for one that corresponds 
     * to the given name
     * If found this SequenceVisualizer is returned else NULL is returned
     */
    public SequenceVisualizer getSequenceVisualizerFromList(String sequenceName, ArrayList<SequenceVisualizer> list) {
        for (SequenceVisualizer viz:list) {
            if (viz.getSequenceName().equals(sequenceName)) return viz;
        }
        return null;
    }
    
    /** Returns a list of all SequenceVisualizers currently in use */
    public ArrayList<SequenceVisualizer> getSequenceVisualizers() {
        return sequenceVisualizers;
    }
    
    
   /**
    * This callback method is called by the VisualizationSettings object whenever
    * the datatrack order is rearranged
    */
    public void trackReorderEvent(int type, Data affected) {
        // this event is not important here
    }
    
   /**
    * This callback method is called by the VisualizationSettings object whenever
    * the name of a dataitem is changed (by the user)
    */
    public void dataRenamedEvent(String oldname,String newname) {
        // this event is not important here
    }
    
   /**
    * This callback method is called by the VisualizationSettings object whenever
    * the size of the sequence window is changed
    */    
    public void sequenceWindowSizeChangedEvent(int newsize) {
        // this event is not important here
    }
    
   /**
    * This callback method is called by the VisualizationSettings object whenever
    * the size of the sequence margin is changed
    */  
    public void sequenceMarginSizeChangedEvent(int newsize) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.setSequenceMarginWithoutRepaint(newsize);
        }
        mainVisualizationPanel.repaint();
    }
    
    
   /**
    * This callback method is called by the VisualizationSettings object whenever
    * the order of sequences or sequence sets is rearranged. 
    */
    public void sequencesReorderEvent(int type, Data affected) {
        if (SwingUtilities.isEventDispatchThread()) {
               setupMainVisualizationPanel();
               validate();
               mainVisualizationPanel.repaint();               
       } else {
           Runnable runner = new Runnable() {
               public void run() {
                   setupMainVisualizationPanel();
                   validate();
                   mainVisualizationPanel.repaint();
               }
           };
           SwingUtilities.invokeLater(runner);
      }        
    }

    public void condensedModeChanged(boolean newmode) {
       if (SwingUtilities.isEventDispatchThread()) {
               setupMainVisualizationPanel();
               validate();
               mainVisualizationPanel.repaint();               
       } else {
           Runnable runner = new Runnable() {
               public void run() {
                   setupMainVisualizationPanel();
                   validate();
                   mainVisualizationPanel.repaint();
               }
           };
           SwingUtilities.invokeLater(runner);
      }          
    }
           
    
    /** 
     * Moves all sequences to the left. 
     */
    public void moveAllSequencesLeft(double fraction) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.moveSequenceLeft(fraction,false);
        }
        repaint();
    }   
    
    /** 
     * Moves all sequences to the right. 
     */
    public void moveAllSequencesRight(double fraction) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.moveSequenceRight(fraction,false);
        }
        repaint();
    }     
        
    /** 
     * Moves all sequences to the left end. 
     */
    public void moveAllSequencesToLeftEnd() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.moveSequenceToLeftEnd(false);
        }
        repaint();        

    }   
    
    /** 
     * Moves all sequences to the right end. 
     */
    public void moveAllSequencesToRightEnd() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.moveSequenceToRightEnd(false);
        }
        repaint();
    }
    
    /** 
     * Zooms in on all sequences while conforming to current alignment settings
     */
    public void zoomInOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.zoomInOnSequence(false);
        }
        repaint();
    }  
    
    /** 
     * Zooms out on all sequences while conforming to current alignment settings
     */
    public void zoomOutOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.zoomOutOnSequence(false);
        }
        repaint();
    }  
    
    /** 
     * Resets zoom level on all sequences while conforming to current alignment settings
     */
    public void zoomResetOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.zoomResetOnSequence(false);
        }
        repaint();
    }
    
    /** 
     * Sets individual zoomlevels on all sequences so that all fit within their respective sequence windows
     */
    public void zoomToFitOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.zoomToFitSequence(false);
        }
        repaint();
    }
    
   /**
    * Calling this method triggers a popup dialog that asks the user to enter a new zoom level
    * If the user enters a valid number, the zoom level of all sequences is set to that value
    * (However, if this number is greater than the largest possible zoom, it will be rounded down
    * to the largest possible zoom-level);
    */  
    public void zoomToCustomOnAllSequences() {
         String newZoomString=(String)JOptionPane.showInputDialog(settings.getGUI().getFrame(),"Enter new zoom level","Zoom to custom level",JOptionPane.PLAIN_MESSAGE,null,null,100.0);
         try {
             double zoomlevel=Double.parseDouble(newZoomString);
             if (zoomlevel>settings.getSequenceWindowSize()*100) zoomlevel=settings.getSequenceWindowSize()*100;
             for (SequenceVisualizer viz:sequenceVisualizers) {
               viz.zoomToLevel(zoomlevel,false);
             }
         } catch (Exception e) {} 
        repaint();
    }
   
    /** 
     * Sets sets zoom level on all sequences to specified value (without asking the user to specify it)
     */
    public void zoomToLevelOnAllSequences(double zoomlevel) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.zoomToLevel(zoomlevel,false);
        }
        repaint();
    }    
    
    /** 
     * Flips the current orientation of all sequences
     */
    public void flipOrientationOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.flipSequenceStrand(false);
        }
        repaint();
    }
    
    /** 
     * Sets the orientation of all sequences to display direct strand
     */
    public void showDirectStrandOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.showSequenceDirectStrand(false);
        }
        repaint();
    }
    
    /** 
     * Sets the orientation of all sequences to display reverse strand
     */
    public void showReverseStrandOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.showSequenceReverseStrand(false);
        }
        repaint();
    }
    
    /** 
     * Sets the orientation of all sequences so they correspond the the respective genes
     */
    public void showStrandFromGeneOnAllSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.showSequenceStrandFromGene(false);
        }
        repaint();
    }
    
    
    /** 
     * Sets the alignment of all sequences to LEFT and moves them to the left end. 
     */
    public void alignAllSequencesLeft() {
        for (SequenceVisualizer viz:sequenceVisualizers) {           
            viz.alignSequenceLeft(false);
        }
        settings.setDefaultSequenceAlignment(VisualizationSettings.ALIGNMENT_LEFT); // this will ensure that new sequences added later also will have the same alignment
        repaint();
    }


    /** 
     * Sets the alignment of all sequences to RIGHT and moves them to the right end. 
     */
    public void alignAllSequencesRight() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.alignSequenceRight(false);
        }
        settings.setDefaultSequenceAlignment(VisualizationSettings.ALIGNMENT_RIGHT);
        repaint();
    }
    
    /** 
     * Sets the alignment of all sequences to NONE
     */
    public void alignAllSequencesNone() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.alignSequenceNone(false);
        }
        settings.setDefaultSequenceAlignment(VisualizationSettings.ALIGNMENT_NONE);                
        repaint();
    }
    
    /** 
     * Sets the alignment of all sequences to TSS and moves them so that the TSS of
     * all sequences (or upstream end) is placed at the specified screen coordinate
     * @param x The screen x-coordinate where TSS should be placed
     */
    public void alignAllSequencesToTSSHere(int x) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.alignSequenceTSSatPoint(x,false);
        }
        settings.setDefaultSequenceAlignment(VisualizationSettings.ALIGNMENT_TSS); 
        repaint();
    }    

    /** 
     * Sets the alignment of all sequences to TSS and moves them so that the TSS of
     * all sequences (or upstream end) is placed at the right end of the sequence window
     */
    public void alignAllSequencesToTSSRight() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.alignSequenceTSSRight(false);
        }
        settings.setDefaultSequenceAlignment(VisualizationSettings.ALIGNMENT_TSS);        
        repaint();
    }    

    /** 
     * realigns the sequences using the currently selected alignment for each sequence
     */
    public void realignSequences() {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            String name=viz.getName();
            String currentAlignment=settings.getSequenceAlignment(name);
                 if (currentAlignment.equals(VisualizationSettings.ALIGNMENT_LEFT)) viz.alignSequenceLeft(false);
            else if (currentAlignment.equals(VisualizationSettings.ALIGNMENT_RIGHT)) viz.alignSequenceRight(false);
            else if (currentAlignment.equals(VisualizationSettings.ALIGNMENT_TSS)) viz.alignSequenceTSSRight(false);
        }      
        repaint();
    }    

    
     /** 
     * Sets the movement contraint on all sequences to the specified value
     * @param constrain TRUE if movement should be constrained
     */
    public void setConstrainOnAllSequences(boolean contrain) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.constrainSequence(contrain,false);
        }
        repaint();
    }
    
    /** 
     * Sets the scroll lock setting on all sequences
     * @param lock TRUE if sequences should not be allowed to autoscroll
     */
    public void setScrollLockOnAllSequences(boolean lock) {
        for (SequenceVisualizer viz:sequenceVisualizers) {
            viz.scrollLockSequence(lock,false);
        }
        repaint();
    }
   
    
    /** This method has been stolen from Swing */
    public boolean print(final MessageFormat headerFormat, final MessageFormat footerFormat, final boolean showPrintDialog,final PrintService service,final PrintRequestAttributeSet attributes, final boolean interactive) throws PrinterException {
        final PrinterJob job = PrinterJob.getPrinterJob();
        final Printable printable;
        final PrintingStatus printingStatus;
        final boolean isHeadless = GraphicsEnvironment.isHeadless();
        final boolean isEventDispatchThread = SwingUtilities.isEventDispatchThread();
        if (interactive && ! isHeadless) {
            printingStatus = PrintingStatus.createPrintingStatus(this, job);
            printable = this;
        } else {
            printingStatus = null; 
            printable = this;
        }

        if (service != null) {
            job.setPrintService(service);
        }

        job.setPrintable(printable);

        final PrintRequestAttributeSet attr = (attributes == null) 
            ? new HashPrintRequestAttributeSet() 
            : attributes;        

        if (showPrintDialog && ! isHeadless && ! job.printDialog(attr)) {
            return false;
        }

        /*
         * there are three cases for printing:
         * 1. print non interactively (! interactive || isHeadless)
         * 2. print interactively off EDT
         * 3. print interactively on EDT
         * 
         * 1 and 2 prints on the current thread (3 prints on another thread)
         * 2 and 3 deal with PrintingStatusDialog
         */
        final Callable<Object> doPrint = 
            new Callable<Object>() {
                public Object call() throws Exception {
                    try {
                        job.print(attr);
                    } finally {
                        if (printingStatus != null) {
                            printingStatus.dispose();
                        }
                    }
                    return null;
                }
            };

        final FutureTask<Object> futurePrinting = 
            new FutureTask<Object>(doPrint);

        final Runnable runnablePrinting = 
            new Runnable() {
                public void run() {
                    //disable component
                    boolean wasEnabled = false;
                    if (isEventDispatchThread) {
                        if (isEnabled()) {
                            wasEnabled = true;
                            setEnabled(false);
                        }
                    } else {
                        try {
                            wasEnabled = SwingUtilities2.submit(
                                new Callable<Boolean>() {
                                    public Boolean call() throws Exception {
                                        boolean rv = isEnabled();
                                        if (rv) {
                                            setEnabled(false);
                                        } 
                                        return rv;
                                    }
                                }).get();
                        } catch (InterruptedException e) {
                            throw new RuntimeException(e);
                        } catch (ExecutionException e) {
                            Throwable cause = e.getCause();
                            if (cause instanceof Error) {
                                throw (Error) cause;
                            } 
                            if (cause instanceof RuntimeException) {
                                throw (RuntimeException) cause;
                            } 
                            throw new AssertionError(cause);
                        }
                    }

                    futurePrinting.run();

                    //enable component
                    if (wasEnabled) {
                        if (isEventDispatchThread) {
                            setEnabled(true);
                        } else {
                            try {
                                SwingUtilities2.submit(
                                    new Runnable() {
                                        public void run() {
                                            setEnabled(true);
                                        }
                                    }, null).get();
                            } catch (InterruptedException e) {
                                throw new RuntimeException(e);
                            } catch (ExecutionException e) {
                                Throwable cause = e.getCause();
                                if (cause instanceof Error) {
                                    throw (Error) cause;
                                } 
                                if (cause instanceof RuntimeException) {
                                    throw (RuntimeException) cause;
                                } 
                                throw new AssertionError(cause);
                            }
                        }
                    }
                }
            };
        
        if (! interactive || isHeadless) {
            runnablePrinting.run();
        } else {
            if (isEventDispatchThread) {
                (new Thread(runnablePrinting)).start();
                printingStatus.showModal(true);
            } else {
                printingStatus.showModal(false);
                runnablePrinting.run();
            }
        }
        
        //the printing is done successfully or otherwise. 
        //dialog is hidden if needed.
        try {
            futurePrinting.get();
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        } catch (ExecutionException e) {
            Throwable cause = e.getCause();
            if (cause instanceof PrinterAbortException) {
                if (printingStatus != null
                    && printingStatus.isAborted()) {
                    return false;
                } else {
                    throw (PrinterAbortException) cause;
                }
            } else if (cause instanceof PrinterException) {
                throw (PrinterException) cause;
            } else if (cause instanceof RuntimeException) {
                throw (RuntimeException) cause;
            } else if (cause instanceof Error) {
                throw (Error) cause;
            } else {
                throw new AssertionError(cause);
            }
        }
        return true;
    }

    @Override
    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
        if (pageIndex > 0) {
          return(NO_SUCH_PAGE);
        } else {
          Graphics2D g2d = (Graphics2D)graphics;
          RepaintManager currentManager = RepaintManager.currentManager(mainVisualizationPanel);
          currentManager.setDoubleBufferingEnabled(false);          
          g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
          int height=mainVisualizationPanel.getHeight();
          int width=getImageWidth();
            // scale to fill the page
          double dw = pageFormat.getImageableWidth();
          double dh = pageFormat.getImageableHeight();
          double xScale = dw/(double)width;
          double yScale = dh/(double)height;
          double scale = Math.min(xScale,yScale);
          g2d.scale(scale, scale);        
          mainVisualizationPanel.printAll(g2d);
          currentManager.setDoubleBufferingEnabled(true);
          return(PAGE_EXISTS);
        }
    }
    
    private int getImageWidth() {
        int width=settings.getSequenceLabelWidth()+settings.getSequenceWindowSize()+2*VisualizationSettings.DATATRACK_BORDER_SIZE;
        Font trackLabelsFont=SequenceVisualizer.getTrackLabelsFont();
        int maxtracklabelwidth=0;
        ArrayList<FeatureDataset> trackorder=settings.getDatatrackOrder();
        for (FeatureDataset track:trackorder) {                 
           String trackname=track.getName();
           if (!settings.isVisible(trackname)) continue;
           int tracklabelwidth=mainVisualizationPanel.getFontMetrics(trackLabelsFont).stringWidth(trackname);
           if (tracklabelwidth>maxtracklabelwidth) maxtracklabelwidth=tracklabelwidth;
        }
        width+=maxtracklabelwidth+16; // 16 is just a padding value
        return width;    
    }
    
}    
    
    